#!/bin/bash
#
# Running the test on a virtual machine:
#   $ sudo apt-get update
#   $ sudo apt-get install cifs-utils
#   $ mkdir testcnts
#   $ sudo mount -t cifs //umbrella/Public testcnts/ --verbose -o guest
#   $ cd testcnts/Jetico/containers_for_test
#   $ bash mount_and_check_all.sh | tee ~/bestcrypt_2.0.0_results.txt
#
# Here are pre-requirements for the test:
#  1. Password must be qqqqqqqqq         ( 'q' * 9 )
#  2. Password for hidden part must be qqqqqqqq   ( 'q' * 8 )
#  3. Root directory must have 'md5sums.txt' file
#
# Creation new containers for the test:
#    $ find -type f -a ! -name 'md5sums.txt' | cut -b3- | while read a ; do md5sum "$a" ; done | tee md5sums.txt
#    $ md5sum -c < md5sums.txt
#

set -e


# current container number
n=1

# results statistics
total=0
success=0


get_mountpoint()
{
  echo "$HOME/mp_`echo $1| tr '/.' '_'`"
}


check_9q_main()
{
  local cnt="$1"
  local MP=`get_mountpoint $cnt`

  total=$(( $total + 1 ))

  # try to mount real part of container ('echo' will enter empty password for cancel)
  mkdir -p "$MP"
  echo | env BESTCRYPT_PASSWORD=qqqqqqqqq bctool mount -r "$cnt" "$MP" || return 1
  ( cd "$MP" ; cat md5sums.txt | md5sum -c --quiet && echo "  main   part is OK" ) || return 1
  bctool umount "$MP"
}

check_8q_hidden()
{
  local cnt="$1"
  local MP=`get_mountpoint $cnt`

  total=$(( $total + 1 ))

  # try to mount hidden part
  echo | env BESTCRYPT_PASSWORD=qqqqqqqq bctool mount -r "$cnt" "$MP" || return 1
  ( cd "$MP" ; cat md5sums.txt | md5sum -c --quiet && echo "  hidden part is OK" ) || return 1
  bctool umount "$MP"

  rmdir "$MP"
}


REPORT=`mktemp /tmp/containers_test_report_XXXXXX`

echo >>$REPORT
echo >>$REPORT
echo "\"BestCrypt for Linux\" container compatibility test report:" >>$REPORT
echo >>$REPORT

printf '%4s  %-45s  %10s  %10s\n' "No."  "Container filename" "main" "hidden" >>$REPORT
echo "--------------------------------------------------------------" >>$REPORT


try_main_and_hidden()
{
  local dir="$1"
  local CONTAINERS=`find "$dir" -name '*.jbc' | sort`

for a in $CONTAINERS ; do
  m="  OK  "
  h="  OK  "
    echo "Testing main $a..."
  check_9q_main "$a"   && success=$(( $success + 1 )) || m="failed"
    echo "Testing hidden $a..."
  check_8q_hidden "$a" && success=$(( $success + 1 )) || h="failed"
    printf '%4s  %-45s  %10s  %10s\n' "$n" `basename $a` "$m" "$h" >>$REPORT
  n=$(( $n + 1 ))
done
}


try_main_and_hidden "./created_by_1.8.4_lin/"

try_main_and_hidden "./created_by_1.9_lin/"

try_main_and_hidden "./created_by_2.0.5_lin/"

try_main_and_hidden "./created_by_8_win/"

try_main_and_hidden "./created_by_9_win/"

try_main_and_hidden "created_by_mac_1.2.888"


perc=$(( $success * 100 / $total ))

echo >>$REPORT
echo  "Total score: $perc% ($success of $total)" >>$REPORT

cat $REPORT
