#!/bin/bash
#
# Put deb/rpm package into APT/YUM repository
#
# RUNTIME_DEPENDS: reprepro

set -e
set -u

if [ -n "${DEBUG:-}" ] ; then
  set -x
fi


USAGE="
  This script puts binary packages into APT/YUM repository.

  Usage:

   $0 -d <dist> -c <codename> -b <basedir> pkg1 pkg2 pkg3 ...

      <distribution>       - distribution: debian, ubuntu
      <codename>      - distribution codename (ex.: squeeze, wheezy, lucid, precise, trusty)
      <basedir>   - root directory of package repository (ex. /home/alex/Public/packages/ )

"


GPG_KEY_ID="4823E840"

if ! gpg --list-secret-key 4823E840 ; then
  echo "Error: unable to find GPG key for repository signing"
  echo "Hint: download from RedMine wiki and import this key:"
  echo "    Open http://nsk.jetico.com/redmine/projects/best-crypt-for-linux/wiki"
  echo "    $ su - uploader"
  echo "    $ gpg --import packages_jetico.com.gpgkey"
  exit 2
fi

# parse options
dist=
while getopts "d:b:c:" name
do
  case $name in
  d)
       distribution="$OPTARG"
       ;;
  c)
       codename="$OPTARG"
       ;;
  b)
       basedir="$OPTARG"
       ;;
  ?)   echo "Error: invalid usage" 2>&1
       echo "$USAGE"
       exit 2
       ;;
  esac
done

if [ -z "${distribution:-}" ]; then
  printf 'Option -d is required\n'
  exit 2
fi
if [ -z "${codename:-}" ]; then
  printf 'Option -c is required\n'
  exit 2
fi
if [ -z "${basedir:-}" ]; then
  printf 'Option -b is required\n'
  exit 2
fi

# the rest is list of *.deb files
shift $(($OPTIND - 1))
PKGS="$*"


DEBIAN_APT_REPO_DIR="$basedir/debian"
UBUNTU_APT_REPO_DIR="$basedir/ubuntu"


# Create 'reprepro' config for specified distribution
#
make_config()
{
  local cnf="$1/conf/distributions"
  test -e "$cnf" || { mkdir -p "$1/conf" ; touch "$cnf" ; }
  if ! cat $cnf | grep -q "^Codename: $codename" ; then
cat << EOF >> "$cnf"

Origin: packages.jetico.com
Label: apt repository
Codename: $codename
Architectures: i386 amd64 source
Components: non-free
Description: Jetico Inc. Oy products
SignWith: $GPG_KEY_ID
Pull: $codename

EOF
  fi
}

case $distribution in
  debian|old_debian)
    # debian family
    REPO_TYPE="deb"
    REPO_DIR="$DEBIAN_APT_REPO_DIR"
    SECTION="non-free"
    ;;
  ubuntu|old_ubuntu)
    # ubuntu family
    REPO_TYPE="deb"
    REPO_DIR="$UBUNTU_APT_REPO_DIR"
    SECTION="multiverse"
    ;;
  *)
    echo "Error: unknown distribution name: $distribution"
    exit 1
esac


case $REPO_TYPE in
  deb)
    make_config "$REPO_DIR" "$codename" "$SECTION"
    reprepro --verbose --basedir "${REPO_DIR}" --section "$SECTION" includedeb "$codename" ${PKGS}
    ;;
  *)
    echo "Repository type for upload is undefined"
esac
