/*
 *	Copyright (c) 2014 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	cClusterCache class.
 *	Saves contiguous chains of clusters in table. The table size is passed at cClusterCache constructor.
 *	External functions:
 *
 *		1.	BYTE *allocateChain( DWORD startCLuster, DWORD clusterQuantity )
 *				returs allocated BYTE buffer for ready for writing.
 *			Note, if it is not enough space in table, it finds the element was not used too long ( see findOldChain ),
 *			frees it and allocate new at this place.
 *			
 *		2.	BYTE *getSector( DWORD diskSector )
 *				walks chains table and return pointer to corresponded volume sector ( the volumes geometry parameters are passed to cClusterCache constructor. )
 */

#pragma once

#include "integer.h"	/* Basic integer types */

static inline void _memset( BYTE *bf, BYTE b, size_t size )
{
	for( int i = size; i--; )
		bf[i] = b;
}

class cClusterChain
{
public:
	DWORD m_start;	// Start cluster number
	DWORD m_length;	// Length of the chain in clusters

	BYTE *m_byteBuffer;	// Allocated byte buffer for entire cluster chain ( see allocateChain ).
	clock_t *m_accessT; // Access time table for all clusters, if at least one cluster was not use too long, so all chain will be detected as old and changed by new update.
	
	cClusterChain()
	{
		_memset( (BYTE*)this, 0, sizeof( *this ) );
	}

	clock_t getAccessTime()
	{
		if ( m_accessT == NULL )
			return 0;

		clock_t accessTime = m_accessT[0];

		for( DWORD i = 1; i < m_length; i++ )
		{
			if ( accessTime < m_accessT[i] )
				accessTime = m_accessT[i];
		}

		return accessTime;
	}

	BYTE *allocateChain( DWORD start, DWORD length, DWORD clusterSize )
	{
		free( clusterSize );

		m_accessT = new clock_t [ length ];
		if ( m_accessT == NULL )
			return NULL;

		//
		//	Marks chain as used last.
		//
		clock_t time = clock();

		for( DWORD i = length; i--; )
			m_accessT[i] = time;

		m_byteBuffer = new BYTE[ length * clusterSize ];
		if ( m_byteBuffer == NULL )
			return NULL;

		m_length	= length;
		m_start		= start;

		return m_byteBuffer;
	}

	void free( DWORD clusterSize )
	{
		LIBLOG_INFO("Begin buffer %x length %d * %d", (unsigned int)m_byteBuffer, (int)m_length, (int)clusterSize );

		if ( m_byteBuffer != NULL )
		{
			_memset( m_byteBuffer, 0, clusterSize * m_length );
			delete m_byteBuffer;
			m_byteBuffer = NULL;
			m_length = 0;
		}

		if ( m_accessT != NULL )
		{
			delete m_accessT;
			m_accessT = NULL;
		}

		_memset( (BYTE*)this, 0, sizeof( this ) );

		LIBLOG_INFO("End" );

	}
};

class cClusterCache
{
	int m_chainQ;
	cClusterChain *m_chainT;

	DWORD m_rootDirFirst;
	DWORD m_rootDirLast;
	BYTE *m_rootDirBuffer;

	DWORD m_bytesPerSector;
	DWORD m_clustersOffset;

public:
	DWORD m_sectorsPerCluster;

	cClusterCache( int chainsQ, DWORD clustersOffset, DWORD bytesPerSector, DWORD sectorsPerCluster )
	{
		m_bytesPerSector	= bytesPerSector;
		m_sectorsPerCluster	= sectorsPerCluster;
		m_clustersOffset	= clustersOffset;

		m_chainQ = chainsQ;
		m_chainT = new cClusterChain [ m_chainQ ];
		if ( m_chainT == NULL )
			m_chainQ = 0;
		
		m_rootDirBuffer = 0;
		m_rootDirFirst = 0;
		m_rootDirLast = 0;
	}
	
	~cClusterCache()
	{
		LIBLOG_INFO("Begin T %x Q %d", (unsigned int)m_chainT, (int)m_chainQ );

		for( int i = m_chainQ; i--; )
			m_chainT[i].free( m_bytesPerSector * m_sectorsPerCluster );
		
		delete [] m_chainT;
		m_chainT = NULL;

		freeRootDir();

		LIBLOG_INFO("End" );
	}

	BYTE *allocateChain( DWORD startCLuster, DWORD clusterQuantity )
	{
		int i = findOldChain();

		if ( i == -1 )
			return NULL;

		return m_chainT[i].allocateChain( startCLuster, clusterQuantity, m_bytesPerSector * m_sectorsPerCluster );
	}

	void freeChain( int i )
	{
		m_chainT[i].free( m_bytesPerSector * m_sectorsPerCluster );
	}

	BYTE *getRootDirSector( DWORD diskSector )
	{
		if ( m_rootDirFirst == 0 ) // dj->fs->dirbase
			return NULL;

		if ( diskSector < m_rootDirFirst ) 
			return NULL;
		
		if ( diskSector > m_rootDirLast )
			return NULL;

		if ( m_rootDirBuffer == NULL )
			return NULL;

		return m_rootDirBuffer + ( diskSector - m_rootDirFirst ) * m_bytesPerSector;
	}

	BYTE *allocateRootDirSectors( DWORD sectorFirst, DWORD sectorQ )
	{
		if ( m_bytesPerSector == 0 )
			return NULL;

		m_rootDirBuffer = new BYTE[ m_bytesPerSector * sectorQ ];

		if ( m_rootDirBuffer == NULL )
			return NULL;

		m_rootDirFirst = sectorFirst;
		m_rootDirLast = m_rootDirFirst + sectorQ - 1;

		return m_rootDirBuffer;
	}

	void freeRootDir()
	{
		LIBLOG_INFO("Begin %x last %x  fist %x", (unsigned int)m_rootDirBuffer, (unsigned int)m_rootDirLast, (unsigned int)m_rootDirFirst );

		if ( m_rootDirBuffer == NULL )
			return;

		DWORD Q = m_rootDirLast - m_rootDirFirst;
		if ( Q )
		{
			LIBLOG_INFO("memset Q %x,  m_bytesPerSector  %x", (unsigned int)Q, (unsigned int)m_bytesPerSector );
			_memset( m_rootDirBuffer, 0, m_bytesPerSector * Q );
		}

		delete m_rootDirBuffer;

		m_rootDirBuffer = NULL;
		m_rootDirLast = 0;
		m_rootDirFirst = 0;

		LIBLOG_INFO("End" );
	}

	BYTE *getSector( DWORD diskSector )
	{
		if ( diskSector < m_clustersOffset )
			return getRootDirSector( diskSector );
		
		DWORD sector = diskSector - m_clustersOffset;

		DWORD cluster = sector / m_sectorsPerCluster;
		sector = sector % m_sectorsPerCluster;

		BYTE *buffer = getCluster( cluster + 2 ); // see clust2sect

		if ( buffer == NULL )
			return NULL;

		return buffer + sector * m_bytesPerSector;
	}

	DWORD getPhysSector( DWORD clusterN )
	{
		clusterN -= 2;	// see clust2sect

		return m_clustersOffset + clusterN * m_sectorsPerCluster;
	}

	DWORD getBufferLength( DWORD clusterQ )
	{
		return clusterQ * m_sectorsPerCluster * m_bytesPerSector;
	}

	BYTE *getCluster( DWORD cluster )
	{
		for( int i = m_chainQ; i--; )
		{
			if ( m_chainT[i].m_start > cluster )
				continue;

			if ( m_chainT[i].m_start + m_chainT[i].m_length <= cluster )
				continue;

			DWORD c = cluster - m_chainT[i].m_start;
			m_chainT[i].m_accessT[c] = clock();

			return m_chainT[i].m_byteBuffer + c * m_bytesPerSector * m_sectorsPerCluster;
		}

		return NULL;
	}

	int findChain( DWORD cluster, DWORD length, bool &bIncluded )
	{
		for( int i = m_chainQ; i--; )
		{
			//
			//	Empty chain
			//
			if ( m_chainT[i].m_length == 0 )
				continue;

			DWORD last = cluster + length - 1;
			
			//
			//	Check if new chain is completely included by cyrrent cache item
			//
			if ( m_chainT[i].m_start > last )
				continue;
			
			DWORD lastInCache = m_chainT[i].m_start + m_chainT[i].m_length - 1;

			if ( cluster > lastInCache )
				continue;
			
			bIncluded = ( last <= lastInCache ) && ( cluster >= m_chainT[i].m_start );

			return i;
		}

		return -1;
	}
	
	int findOldChain()
	{
		clock_t firstT = 0;
		int firstN = -1;
		for( int i = m_chainQ; i--; )
		{
			if ( m_chainT[i].m_length == 0 )
				return i;

			clock_t cur = m_chainT[i].getAccessTime();

			if ( ( firstN == -1 ) || ( firstT > cur ) )
			{
				firstT = cur;
				firstN = i;
			}
		}

		return firstN;
	}
};
