#include <string.h>
#include <errno.h>
#include <sys\stat.h>
#include "../include/ntfs-3g/device.h"
#include "../include/ntfs-3g/misc.h"
#include "../../iVolume.h"
#include "../../cADB.h"

/**
 * struct BCContainer_fd -
 */
typedef struct {
	iVolume *volume;
	s64 pos;		/* Logical current position on the volume. */
	s64 part_start;
	s64 part_length;
	int part_hidden_sectors;
	s64 geo_size, geo_cylinders;
	s32 geo_sector_size;
	s64 volume_size;
	u32 geo_sectors, geo_heads;
} BCContainer_fd;

#define EINVAL          22
#define EROFS           30
#define EIO             5

#define	LD_bc_u16(ptr)		(bc_u16)(((bc_u16)*((bc_u8*)(ptr)+1)<<8)|(bc_u16)*(bc_u8*)(ptr))
#define	LD_bc_u32(ptr)		(bc_u32)(((bc_u32)*((bc_u8*)(ptr)+3)<<24)|((bc_u32)*((bc_u8*)(ptr)+2)<<16)|((bc_u16)*((bc_u8*)(ptr)+1)<<8)|*(bc_u8*)(ptr))

static bool checkNTFSSig( bc_u8 *br )
{
	bc_u32 sig = LD_bc_u32(br + 82);	/* File system type (1) */
	if ( (sig & 0xFFFFFF) == 0x544146 ) // FAT
		return false;

	sig = LD_bc_u32(br + 3);	/* File system type (1) */
	if ( (sig & 0xFFFFFFFF) == 0x5346544E ) // NTFS
		return true;

	return false;
}

static int check_geo( BCContainer_fd *fd )
{
	cADB < bc_u8 > br(fd->volume->getSectorSize());
	if ( br.m_body == NULL )
	{
		ntfs_log_debug("Assertion! At check_geo" );
		return -1;
	}

	bc_u64 bootSect = 0;
	bc_u64 partSize = 0;

	if ( fd->volume->readSectors( bootSect, br.m_body, 1 ) != iVolume::iv_Success )
	{
		ntfs_log_debug("Assertion! At check_geo" );
		return -1;
	}

	bc_u16 sig = LD_bc_u16( br.m_body + 510 );
	if ( sig == 0xAA55 )		/* Check record signature (always placed at offset 510 even if the sector size is >512) */
	{
		if ( !checkNTFSSig( br.m_body ) )
		{
			bc_u8 *tbl = br.m_body + 446; /* MBR: Partition table offset */
			if (tbl[4]) /* Is the partition existing? */
			{
				bootSect = LD_bc_u32(&tbl[8]);		/* Partition offset in LBA */
				partSize = LD_bc_u32(&tbl[12]);
				if ( fd->volume->readSectors( bootSect, br.m_body, 1 ) != iVolume::iv_Success )
				{
					ntfs_log_debug("Assertion! At check_geo" );
					return -1;
				}

				sig = LD_bc_u16( br.m_body + 510 );
			}
		}
	}

	if ( ( sig != 0xAA55 ) || !checkNTFSSig( br.m_body ) )
	{
		ntfs_log_debug("Assertion! Disk is not formatted." );
		return -1;
	}

	fd->geo_sector_size = *(bc_u16*)(br.m_body+11); //	11 - /* Sector size [byte] (2) */ fd->volume->getSectorSize()
	fd->part_start = bootSect * fd->geo_sector_size; // 0x100000;
	if ( partSize == 0 )
		partSize = fd->volume->getSectorCount();

	fd->part_length =  partSize * fd->geo_sector_size;

	return 0;
	
	///* Search first partition on the drive. */
	//fmt = check_fs(fs, bsect = 0);		/* Load sector 0 and check if it is an FAT-VBR (in SFD) */
	//if (LD2PT(vol) && !fmt) fmt = 1;	/* Force non-SFD if the volume is forced partition */
	//if (fmt == 1) {						/* Not an FAT-VBR, the physical drive can be partitioned */
	//	/* Check the partition listed in the partition table */
	//	pi = LD2PT(vol);
	//	if (pi) pi--;
	//	tbl = &fs->win[MBR_Table + pi * SZ_PTE];/* Partition table */
	//	if (tbl[4]) {						/* Is the partition existing? */
	//		bsect = LD_DWORD(&tbl[8]);		/* Partition offset in LBA */
	//		fmt = check_fs(fs, bsect);		/* Check the partition */
	//	}
	//}
	//if (fmt == 3) LEAVE_FF( fs, FR_DISK_ERR ); //  return FR_DISK_ERR;
	//if (fmt) LEAVE_FF( fs, FR_NO_FILESYSTEM ); // return FR_NO_FILESYSTEM;		/* No FAT volume is found */

	///* An FAT volume is found. Following code initializes the file system object */

	//if (LD_WORD(fs->win+BPB_BytsPerSec) != SS(fs))		/* (BPB_BytsPerSec must be equal to the physical sector size) */
	//	LEAVE_FF( fs, FR_NO_FILESYSTEM ); // return FR_NO_FILESYSTEM;
}

/**
 * ntfs_device_BCContainer_open - open a device
 * @dev:	a pointer to the NTFS_DEVICE to open
 * @flags:	unix open status flags
 *
 * @dev->d_name must hold the device name, the rest is ignored.
 * Supported flags are O_RDONLY, O_WRONLY and O_RDWR.
 *
 * If name is in format "(hd[0-9],[0-9])" then open a partition.
 * If name is in format "(hd[0-9])" then open a volume.
 * Otherwise open a file.
 */

#ifndef _O_RDWR
#define _O_RDWR         0x0002  /* open for reading and writing */
#endif

static int ntfs_device_BCContainer_open(struct ntfs_device *dev, int flags)
{
	//
	//	If volume is closed the dev->d_private is pointer to iVolume interface
	//	If volume is opened the dev->d_private is pointer to BCContainer_fd structure
	//		and pointer to iVolume is BCContainer_fd->volume
	//	See ntfs_device_BCContainer_close
	//
	iVolume *iv = (iVolume*)dev->d_private;

	BCContainer_fd *fd =  new BCContainer_fd;
	memset( fd, 0, sizeof( BCContainer_fd ) );

//#ifdef _DEBUG
//	fd->part_start = 0x100000;
//	fd->geo_sector_size = 512;
//	fd->part_length = iv->getSectorCount() * iv->getSectorSize();
//	memset( fd, 0, sizeof( BCContainer_fd ) );
//#endif
	fd->volume = iv;
	dev->d_private = fd;

	if ( check_geo( fd ) != 0 )
		return -1;

	ntfs_log_debug("BCContainer_open(%s) -> %p, offset 0x%llx.\n", dev->d_name, dev, fd->part_start);

	/* Setup our read-only flag. */
	if ((flags & _O_RDWR) != _O_RDWR)
		NDevSetReadOnly(dev);

	NDevSetOpen(dev);
	NDevClearDirty(dev);
	return 0;

	//int drive_id = 0, numparams;
	//unsigned int part = 0;
	//char drive_char;
	//BCContainer_fd fd;
	//int err;

	//if (NDevOpen(dev)) {
	//	errno = EBUSY;
	//	return -1;
	//}
	//ntfs_device_BCContainer_init_imports();
	//numparams = sscanf(dev->d_name, "/dev/hd%c%u", &drive_char, &part);
	//if (!numparams
	//    && (dev->d_name[1] == ':')
	//    && (dev->d_name[2] == '\0')) {
	//	drive_char = dev->d_name[0];
	//	numparams = 3;
	//	drive_id = toupper(drive_char) - 'A';
	//}
	//switch (numparams) {
	//case 0:
	//	ntfs_log_debug("BCContainer_open(%s) -> file.\n", dev->d_name);
	//	err = ntfs_device_BCContainer_open_file(dev->d_name, &fd, flags);
	//	break;
	//case 1:
	//	ntfs_log_debug("BCContainer_open(%s) -> drive %d.\n", dev->d_name,
	//			drive_id);
	//	err = ntfs_device_BCContainer_open_drive(drive_id, &fd, flags);
	//	break;
	//case 2:
	//	ntfs_log_debug("BCContainer_open(%s) -> drive %d, part %u.\n",
	//			dev->d_name, drive_id, part);
	//	err = ntfs_device_BCContainer_open_partition(drive_id, part, &fd,
	//			flags);
	//	break;
	//case 3:
	//	ntfs_log_debug("BCContainer_open(%s) -> drive %c:\n",
	//			dev->d_name, drive_char);
	//	err = ntfs_device_BCContainer_open_lowlevel(drive_id, &fd,
	//			flags);
	//	break;
	//default:
	//	ntfs_log_debug("BCContainer_open(%s) -> unknwon file format.\n",
	//			dev->d_name);
	//	err = -1;
	//}
	//if (err)
	//	return err;
	//ntfs_log_debug("BCContainer_open(%s) -> %p, offset 0x%llx.\n", dev->d_name,
	//		dev, fd.part_start);
	///* Setup our read-only flag. */
	//if ((flags & O_RDWR) != O_RDWR)
	//	NDevSetReadOnly(dev);
	//dev->d_private = (BCContainer_fd*)ntfs_malloc(sizeof(BCContainer_fd));
	//memcpy(dev->d_private, &fd, sizeof(BCContainer_fd));
	//NDevSetOpen(dev);
	//NDevClearDirty(dev);
	//return 0;
}


/**
 * ntfs_device_BCContainer_close - close an open ntfs deivce
 * @dev:	ntfs device obtained via ->open
 *
 * Return 0 if o.k.
 *	 -1 if not, and errno set.  Note if error fd->vol_handle is trashed.
 */
static int ntfs_device_BCContainer_close(struct ntfs_device *dev)
{
	//
	//	If volume is closed the dev->d_private is pointer to iVolume interface
	//	If volume is opened the dev->d_private is pointer to BCContainer_fd structure
	//		and pointer to iVolume is BCContainer_fd->volume
	//	See ntfs_device_BCContainer_open
	//
	BCContainer_fd *fd = (BCContainer_fd *)dev->d_private;

	ntfs_log_trace("Closing device %p.\n", dev);

	if (!NDevOpen(dev)) {
		errno = 9; //EBADF;
		return -1;
	}

	iVolume *iv = (iVolume*)fd->volume;
	memset( fd, 0, sizeof( BCContainer_fd ) );
	delete fd;

	dev->d_private = (void*)iv;
	return 0;

	//if (fd->vol_handle != INVALID_HANDLE_VALUE) {
	//	if (!NDevReadOnly(dev)) {
	//		ntfs_device_BCContainer_dismount(fd->vol_handle);
	//		ntfs_device_BCContainer_unlock(fd->vol_handle);
	//	}
	//	if (!CloseHandle(fd->vol_handle))
	//		ntfs_log_trace("CloseHandle() failed for volume.\n");
	//}
	//if (fd->ntdll) {
	//	ntfs_device_BCContainer_setlock(fd->handle,FSCTL_UNLOCK_VOLUME);
	//	rvl = NtClose(fd->handle) == STATUS_SUCCESS;
	//} else
	//	rvl = CloseHandle(fd->handle);
	//free(fd);
	//if (!rvl) {
	//	errno = ntfs_w32error_to_errno(GetLastError());
	//	if (fd->ntdll)
	//		ntfs_log_trace("NtClose() failed.\n");
	//	else
	//		ntfs_log_trace("CloseHandle() failed.\n");
	//	return -1;
	//}
	//return 0;
}


/**
 * ntfs_device_BCContainer_seek - change current logical file position
 * @dev:	ntfs device obtained via ->open
 * @offset:	required offset from the whence anchor
 * @whence:	whence anchor specifying what @offset is relative to
 *
 * Return the new position on the volume on success and -1 on error with errno
 * set to the error code.
 *
 * @whence may be one of the following:
 *	SEEK_SET - Offset is relative to file start.
 *	SEEK_CUR - Offset is relative to current position.
 *	SEEK_END - Offset is relative to end of file.
 */
static s64 ntfs_device_BCContainer_seek(struct ntfs_device *dev, s64 offset,
		int whence)
{
	s64 abs_ofs;
	BCContainer_fd *fd = (BCContainer_fd *)dev->d_private;

	ntfs_log_trace("seek offset = 0x%llx, whence = %d.\n", offset, whence);
	switch (whence) {
	case SEEK_SET:
		abs_ofs = offset;
		break;
	case SEEK_CUR:
		abs_ofs = fd->pos + offset;
		break;
	case SEEK_END:
		/* End of partition != end of disk. */
		if (fd->part_length == -1) {
			ntfs_log_trace("Position relative to end of disk not "
					"implemented.\n");
			errno = EOPNOTSUPP;
			return -1;
		}
		abs_ofs = fd->part_length + offset;
		break;
	default:
		ntfs_log_trace("Wrong mode %d.\n", whence);
		errno = EINVAL;
		return -1;
	}
	if ((abs_ofs < 0)
	    || (abs_ofs > fd->part_length))
	{
		ntfs_log_trace("Seeking outsize seekable area.\n");
		errno = EINVAL;
		return -1;
	}
	fd->pos = abs_ofs;
	return abs_ofs;
}


/**
 * ntfs_device_win32_pio - positioned low level i/o
 * @fd:		win32 device descriptor obtained via ->open
 * @pos:	at which position to do i/o from/to
 * @count:	how many bytes should be transfered
 * @b:		source/destination buffer
 * @write:	TRUE if write transfer and FALSE if read transfer
 *
 * On success returns the number of bytes transfered (can be < @count) and on
 * error returns -1 and errno set.  Transfer starts from position @pos on @fd.
 *
 * Notes:
 *	- @pos, @buf, and @count must be aligned to geo_sector_size
 *	- When dealing with volumes, a single call must not span both volume
 *	  and disk extents.
 *	- Does not use/set @fd->pos.
 */
static s64 ntfs_device_BCContainer_pio( BCContainer_fd *fd, const s64 pos,
		const s64 count, void *rbuf, const void *wbuf)
{
	int res;
	
	if ( fd->geo_sector_size == 0 )
	{
		ntfs_log_trace("Assertion! Sector size is ZERO.\n" );
		return -1;
	}

	if ( ( pos | count ) & ( fd->geo_sector_size - 1 ) )
	{
		ntfs_log_trace("Assertion! pos = 0x%llx, count = 0x%llx, sector size = 0x%llx.\n",
			(long long)pos, (long long)count, (long long)fd->geo_sector_size );
		return -1;
	}

	static bool bUseCache = true;
	static bool bCheckWriting = false;

#ifndef _DebugCache
	bCheckWriting = false;
#endif

	ntfs_log_trace("%s, %s, pos = 0x%llx, count = 0x%llx, direction = %s.\n",
			(bUseCache ? "Use cache!" : "No cache!" ),
			(bCheckWriting ? "Check writing!" : "No check!" ),
			(long long)pos, (long long)count, wbuf != 0 ? "write" :
			"read");

	s64 startSector	= ( pos + fd->part_start ) / fd->geo_sector_size;
	s64 sectorCount	= count / fd->geo_sector_size;

	if ( bUseCache )
	{
		if ( wbuf )
			res = fd->volume->writeSectorsCache( startSector, (const u8*)wbuf, sectorCount );
		else
			res = fd->volume->readSectorsCache( startSector, (u8*)rbuf, sectorCount );
	}
	else
	{
		if ( wbuf )
			res = fd->volume->writeSectors( startSector, (const u8*)wbuf, sectorCount );
		else
			res = fd->volume->readSectors( startSector, (u8*)rbuf, sectorCount );
	}

#ifdef _DebugCache
	if ( wbuf && bCheckWriting )
	{
		u8 *bf = new u8[count];
		if ( bf == NULL )
			ntfs_log_trace("Assertion! Error in allocating! count = 0x%llx, sector size = 0x%llx.\n", (long long)count, (long long)fd->geo_sector_size );
		else
		{
			int lres = fd->volume->readSectors( startSector, (u8*)bf, sectorCount );
			if ( lres != iVolume::iv_Success )
				ntfs_log_trace("%sFile() failed.\n", wbuf != 0 ? "Write" : "Read");
			else
			if ( memcmp( bf, wbuf, count ) != 0 )
				ntfs_log_trace("Assertion! Compare error! startSector = 0x%llx(%lld), count = 0x%llx, sector size = 0x%llx.\n",
						(long long)startSector, (long long)startSector, (long long)count, (long long)fd->geo_sector_size );

			delete bf;
		}
	}

#endif // _DebugCache

	if ( res != iVolume::iv_Success ) {
		errno = res;
		ntfs_log_trace("%sFile() failed.\n", wbuf != 0 ?
						"Write" : "Read");
		return -1;
	}

	if (rbuf && !pos && sectorCount )
	{
		/* get the sector size from the boot sector */
		char *boot = (char*)rbuf;
		
		s32 geo_sector_size = (boot[11] & 255) + ((boot[12] & 255) << 8);

		if ( geo_sector_size != fd->geo_sector_size )
			fd->geo_sector_size = geo_sector_size;
	}

	return sectorCount * fd->geo_sector_size;
}

/**
 * ntfs_device_BCContainer_pread_simple - positioned simple read
 * @fd:		win32 device descriptor obtained via ->open
 * @pos:	at which position to read from
 * @count:	how many bytes should be read
 * @b:		a pointer to where to put the contents
 *
 * On success returns the number of bytes read (can be < @count) and on error
 * returns -1 and errno set.  Read starts from position @pos.
 *
 * Notes:
 *	- @pos, @buf, and @count must be aligned to geo_sector_size.
 *	- When dealing with volumes, a single call must not span both volume
 *	  and disk extents.
 *	- Does not use/set @fd->pos.
 */
static inline s64 ntfs_device_BCContainer_pread_simple(BCContainer_fd *fd, const s64 pos,
		const s64 count, void *b)
{
	return ntfs_device_BCContainer_pio(fd, pos, count, b, (void*)NULL);
}

/**
 * ntfs_device_BCContainer_pwrite_simple - positioned simple write
 * @fd:		win32 device descriptor obtained via ->open
 * @pos:	at which position to write to
 * @count:	how many bytes should be written
 * @b:		a pointer to the data to write
 *
 * On success returns the number of bytes written and on error returns -1 and
 * errno set.  Write starts from position @pos.
 *
 * Notes:
 *	- @pos, @buf, and @count must be aligned to geo_sector_size.
 *	- When dealing with volumes, a single call must not span both volume
 *	  and disk extents.
 *	- Does not use/set @fd->pos.
 */
static inline s64 ntfs_device_BCContainer_pwrite_simple(BCContainer_fd *fd, const s64 pos,
		const s64 count, const void *b)
{
	return ntfs_device_BCContainer_pio(fd, pos, count, (void*)NULL, b);
}

/**
 * ntfs_device_BCContainer_read - read bytes from an ntfs device
 * @dev:	ntfs device obtained via ->open
 * @b:		pointer to where to put the contents
 * @count:	how many bytes should be read
 *
 * On success returns the number of bytes actually read (can be < @count).
 * On error returns -1 with errno set.
 */
static s64 ntfs_device_BCContainer_read(struct ntfs_device *dev, void *b, s64 count)
{
	s64 old_pos, to_read, i, br = 0;
	BCContainer_fd *fd = (BCContainer_fd *)dev->d_private;
	u8 *alignedbuffer;
	int old_ofs, ofs;

	old_pos = fd->pos;
	old_ofs = ofs = old_pos & (fd->geo_sector_size - 1);
	to_read = (ofs + count + fd->geo_sector_size - 1) &
			~(s64)(fd->geo_sector_size - 1);
	/* Impose maximum of 2GB to be on the safe side. */
	if (to_read > 0x80000000) {
		s64 delta = to_read - count;
		to_read = 0x80000000;
		count = to_read - delta;
	}
	ntfs_log_trace("fd = %p, b = %p, count = 0x%llx, pos = 0x%llx, "
			"ofs = %i, to_read = 0x%llx.\n", fd, b,
			(long long)count, (long long)old_pos, ofs,
			(long long)to_read);
	if (!((unsigned long)b & (fd->geo_sector_size - 1)) && !old_ofs &&
			!(count & (fd->geo_sector_size - 1)))
		alignedbuffer = (u8 *)b;
	else {
		alignedbuffer = (u8 *)ntfs_calloc( (size_t)to_read );
		if (!alignedbuffer) {
			errno = -2;
			ntfs_log_trace("VirtualAlloc failed for read.\n");
			return -1;
		}
	}
	if ( old_pos < fd->geo_size )
	{
		s64 vol_to_read = fd->geo_size - old_pos;
		if (count > vol_to_read) {
			br = ntfs_device_BCContainer_pread_simple(fd,
					old_pos & ~(s64)(fd->geo_sector_size - 1),
					ofs + vol_to_read, alignedbuffer);
			if (br == -1)
				goto read_error;
			to_read -= br;
			if (br < ofs) {
				br = 0;
				goto read_partial;
			}
			br -= ofs;
			fd->pos += br;
			ofs = fd->pos & (fd->geo_sector_size - 1);
			if (br != vol_to_read)
				goto read_partial;
		}
	}
	i = ntfs_device_BCContainer_pread_simple(fd,
			fd->pos & ~(s64)(fd->geo_sector_size - 1), to_read,
			alignedbuffer + br);
	if (i == -1) {
		if (br)
			goto read_partial;
		goto read_error;
	}
	if (i < ofs)
		goto read_partial;
	i -= ofs;
	br += i;
	if (br > count)
		br = count;
	fd->pos = old_pos + br;
read_partial:
	if (alignedbuffer != b) {
		memcpy((void*)b, alignedbuffer + old_ofs, (size_t)br);
		ntfs_free( alignedbuffer );
	}
	return br;
read_error:
	if (alignedbuffer != b)
		ntfs_free( alignedbuffer );
	return -1;
}

/**
 * ntfs_device_BCContainer_write - write bytes to an ntfs device
 * @dev:	ntfs device obtained via ->open
 * @b:		pointer to the data to write
 * @count:	how many bytes should be written
 *
 * On success returns the number of bytes actually written.
 * On error returns -1 with errno set.
 */
static s64 ntfs_device_BCContainer_write(struct ntfs_device *dev, const void *b,
		s64 count)
{
	s64 old_pos, to_write, i, bw = 0;
	BCContainer_fd *fd = (BCContainer_fd *)dev->d_private;
	u8 *alignedbuffer;
	int old_ofs, ofs;

	old_pos = fd->pos;
	old_ofs = ofs = old_pos & (fd->geo_sector_size - 1);
	to_write = (ofs + count + fd->geo_sector_size - 1) &
			~(s64)(fd->geo_sector_size - 1);
	/* Impose maximum of 2GB to be on the safe side. */
	if (to_write > 0x80000000) {
		s64 delta = to_write - count;
		to_write = 0x80000000;
		count = to_write - delta;
	}
	ntfs_log_trace("fd = %p, b = %p, count = 0x%llx, pos = 0x%llx, "
			"ofs = %i, to_write = 0x%llx.\n", fd, b,
			(long long)count, (long long)old_pos, ofs,
			(long long)to_write);
	if (NDevReadOnly(dev)) {
		ntfs_log_trace("Can't write on a R/O device.\n");
		errno = EROFS;
		return -1;
	}
	if (!count)
		return 0;
	NDevSetDirty(dev);
	if (!((unsigned long)b & (fd->geo_sector_size - 1)) && !old_ofs &&
			!(count & (fd->geo_sector_size - 1)))
		alignedbuffer = (u8 *)b;
	else {
		s64 end;

		alignedbuffer = (u8 *)ntfs_calloc( (size_t)to_write );
		if (!alignedbuffer) {
			errno = -2;
			ntfs_log_trace("VirtualAlloc failed for write.\n");
			return -1;
		}
		/* Read first sector if start of write not sector aligned. */
		if (ofs) {
			i = ntfs_device_BCContainer_pread_simple(fd,
					old_pos & ~(s64)(fd->geo_sector_size - 1),
					fd->geo_sector_size, alignedbuffer);
			if (i != fd->geo_sector_size) {
				if (i >= 0)
					errno = EIO;
				goto write_error;
			}
		}
		/*
		 * Read last sector if end of write not sector aligned and last
		 * sector is either not the same as the first sector or it is
		 * the same as the first sector but this has not been read in
		 * yet, i.e. the start of the write is sector aligned.
		 */
		end = old_pos + count;
		if ((end & (fd->geo_sector_size - 1)) &&
				((to_write > fd->geo_sector_size) || !ofs)) {
			i = ntfs_device_BCContainer_pread_simple(fd,
					end & ~(s64)(fd->geo_sector_size - 1),
					fd->geo_sector_size, alignedbuffer +
					to_write - fd->geo_sector_size);
			if (i != fd->geo_sector_size) {
				if (i >= 0)
					errno = EIO;
				goto write_error;
			}
		}
		/* Copy the data to be written into @alignedbuffer. */
		memcpy(alignedbuffer + ofs, b, (size_t)count);
	}
	if (old_pos < fd->geo_size)
	{
		s64 vol_to_write = fd->geo_size - old_pos;
		if (count > vol_to_write) {
			bw = ntfs_device_BCContainer_pwrite_simple(fd,
					old_pos & ~(s64)(fd->geo_sector_size - 1),
					ofs + vol_to_write, alignedbuffer);
			if (bw == -1)
				goto write_error;
			to_write -= bw;
			if (bw < ofs) {
				bw = 0;
				goto write_partial;
			}
			bw -= ofs;
			fd->pos += bw;
			ofs = fd->pos & (fd->geo_sector_size - 1);
			if (bw != vol_to_write)
				goto write_partial;
		}
	}
	i = ntfs_device_BCContainer_pwrite_simple(fd,
			fd->pos & ~(s64)(fd->geo_sector_size - 1), to_write,
			alignedbuffer + bw);
	if (i == -1) {
		if (bw)
			goto write_partial;
		goto write_error;
	}
	if (i < ofs)
		goto write_partial;
	i -= ofs;
	bw += i;
	if (bw > count)
		bw = count;
	fd->pos = old_pos + bw;
write_partial:
	if (alignedbuffer != b)
		ntfs_free( alignedbuffer );
	return bw;
write_error:
	bw = -1;
	goto write_partial;
	return -1;
}

/**
 * ntfs_device_BCContainer_pread - positioned low level i/o
 * @fd:		win32 device descriptor obtained via ->open
 * @pos:	at which position to do i/o from/to
 * @count:	how many bytes should be transfered
 * @b:		source/destination buffer
 * @write:	TRUE if write transfer and FALSE if read transfer
 *
 * On success returns the number of bytes transfered (can be < @count) and on
 * error returns -1 and errno set.  Transfer starts from position @pos on @fd.
 *
 * Notes:
 *	- @pos, @buf, and @count must be aligned to geo_sector_size
 *	- When dealing with volumes, a single call must not span both volume
 *	  and disk extents.
 *	- Does not use/set @fd->pos.
 */
static s64 ntfs_device_BCContainer_pread(struct ntfs_device *dev, void *b, s64 count, s64 offset)
{
	s64 got;
	BCContainer_fd *fd;

		/* read the fast way if sector aligned */
	fd = (BCContainer_fd*)dev->d_private;
	if (!((count | offset) & (fd->geo_sector_size - 1))) {
		got = ntfs_device_BCContainer_pio(fd, offset, count, b, (void*)NULL);
	} else {
		if (ntfs_device_BCContainer_seek(dev, offset, 0) == -1)
			got = 0;
		else
			got = ntfs_device_BCContainer_read(dev, b, count);
	}

	return (got);
}

static s64 ntfs_device_BCContainer_pwrite(struct ntfs_device *dev, const void *b, s64 count, s64 offset)
{
	s64 put;
	BCContainer_fd *fd;

		/* write the fast way if sector aligned */
	fd = (BCContainer_fd*)dev->d_private;
	if (!((count | offset) & (fd->geo_sector_size - 1))) {
		put = ntfs_device_BCContainer_pio(fd, offset, count, (void*)NULL, b);
	} else {
		if (ntfs_device_BCContainer_seek(dev, offset, 0) == -1)
			put = 0;
		else
			put = ntfs_device_BCContainer_write(dev, b, count);
	}
	return (put);
}

/**
 * ntfs_device_BCContainer_sync - flush write buffers to disk
 * @dev:	ntfs device obtained via ->open
 *
 * Return 0 if o.k.
 *	 -1 if not, and errno set.
 *
 * Note: Volume syncing works differently in windows.
 *	 Disk cannot be synced in windows.
 */
static int ntfs_device_BCContainer_sync(struct ntfs_device *dev)
{
	int err = 0;

	if (!NDevReadOnly(dev) && NDevDirty(dev)) {
		BCContainer_fd *fd = (BCContainer_fd *)dev->d_private;

		err = fd->volume->sync();

		if ( err != iVolume::iv_Success ) {
			ntfs_log_trace("Could not sync.\n");
			errno = err;
			return -1;
		}

		NDevClearDirty(dev);
	}
	return 0;
}

/**
 * ntfs_device_BCContainer_stat - get a unix-like stat structure for an ntfs device
 * @dev:	ntfs device obtained via ->open
 * @buf:	pointer to the stat structure to fill
 *
 * Note: Only st_mode, st_size, and st_blocks are filled.
 *
 * Return 0 if o.k.
 *	 -1 if not and errno set. in this case handle is trashed.
 */
static int ntfs_device_BCContainer_stat(struct ntfs_device *dev, struct stat *buf)
{
	BCContainer_fd *fd = (BCContainer_fd *)dev->d_private;
	mode_t st_mode;

	st_mode = S_IFBLK;

	memset(buf, 0, sizeof(struct stat));
	buf->st_mode = st_mode;
	buf->st_size = (long long)fd->part_length;
	//if (buf->st_size != -1)
	//	buf->st_blocks = buf->st_size >> 9;
	//else
	//	buf->st_size = 0;
	return 0;
}

#pragma pack(1)
struct sCachPriopity
{
	int m_size;
	int m_priority;
	void *m_pVolume;

	sCachPriopity( int priority = 0, void *pVolume = NULL )
	{
		memset( this, 0, sizeof(*this) );
		m_size = sizeof(*this);
		m_priority = priority;
		m_pVolume = pVolume;
	}
};
#pragma pack()

static int ntfs_device_BCContainer_ioctl(struct ntfs_device *dev, int request, void *argp)
{
#if defined(HDIO_READCACHEPRIORITY)

	if ( request == HDIO_READCACHEPRIORITY )
	{
		sCachPriopity *cp = (sCachPriopity *)argp;

		if ( cp->m_size != sizeof(sCachPriopity) )
		{
			ntfs_log_debug("Assertion! Invalid structure size %d != %d - ioctl %d.\n", cp->m_size, sizeof(sCachPriopity), request);
			return -1;
		}

		iVolume *iv = (iVolume*)(cp->m_pVolume);
		iv->setCurrentReadCachePriority( cp->m_priority );

		return 0;
	}
#endif

	ntfs_log_debug("unimplemented ioctl %d.\n", request);
	errno = EOPNOTSUPP;
	return -1;

//#if defined(BLKGETSIZE) | defined(BLKGETSIZE64)
//	BCContainer_fd *fd = (BCContainer_fd *)dev->d_private;
//#endif
//
//	ntfs_log_trace("BCContainer_ioctl(%d) called.\n", request);
//	switch (request) {
//#if defined(BLKGETSIZE)
//	case BLKGETSIZE:
//		ntfs_log_debug("BLKGETSIZE detected.\n");
//		if (fd->part_length >= 0) {
//			*(int *)argp = (int)(fd->part_length / 512);
//			return 0;
//		}
//		errno = EOPNOTSUPP;
//		return -1;
//#endif
//#if defined(BLKGETSIZE64)
//	case BLKGETSIZE64:
//		ntfs_log_debug("BLKGETSIZE64 detected.\n");
//		if (fd->part_length >= 0) {
//			*(s64 *)argp = fd->part_length;
//			return 0;
//		}
//		errno = EOPNOTSUPP;
//		return -1;
//#endif
//#ifdef HDIO_GETGEO
//	case HDIO_GETGEO:
//		ntfs_log_debug("HDIO_GETGEO detected.\n");
//		return ntfs_BCContainer_hdio_getgeo(dev, (struct hd_geometry *)argp);
//#endif
//#ifdef BLKSSZGET
//	case BLKSSZGET:
//		ntfs_log_debug("BLKSSZGET detected.\n");
//		return ntfs_BCContainer_blksszget(dev, (int *)argp);
//#endif
//#ifdef BLKBSZSET
//	case BLKBSZSET:
//		ntfs_log_debug("BLKBSZSET detected.\n");
//		/* Nothing to do on Windows. */
//		return 0;
//#endif
//	default:
//		ntfs_log_debug("unimplemented ioctl %d.\n", request);
//		errno = EOPNOTSUPP;
//		return -1;
//	}
}

struct ntfs_device_operations ntfs_device_BCContainer_io_ops = {
	ntfs_device_BCContainer_open,
	ntfs_device_BCContainer_close,
	ntfs_device_BCContainer_seek,
	ntfs_device_BCContainer_read,
	ntfs_device_BCContainer_write,
	ntfs_device_BCContainer_pread,
	ntfs_device_BCContainer_pwrite,
	ntfs_device_BCContainer_sync,
	ntfs_device_BCContainer_stat,
	ntfs_device_BCContainer_ioctl
};
