/*
 *	Copyright (c) 2015 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Read cache class.
 */

#pragma once

#include <memory.h>
#include <time.h>
#include "../multi-lib-enc/container/intsize.h"
#include "log.h"

//
//	Enum read cache prioprity.
//		The flag is used at secrtor reading.
//		eRCPNo(0) do not place sectors to cache it saves cache memory if f.e. file bodies were never added to cache.
//		eRCP_1 first chached level is removed first.
//		eRCP_2 does never removed.
//
enum eReadCachePriority  {
	eRCPNo = 0,
	eRCP_1,
	eRCP_2
};


class cSectorChain
{
public:
	bc_u64 m_start;			// Nunber of start sector
	bc_u64 m_length;		// Length of the chain in sectors

	bc_u8 *m_byteBuffer;	// Allocated byte buffer for entire cluster chain ( see allocateChain ).
	clock_t m_accessTime;  // Access time to at leas one sector

	int m_priority;

	cSectorChain()
	{
		memset( (bc_u8*)this, 0, sizeof( *this ) );
	}

	clock_t getAccessTime()
	{
		return m_accessTime;
	}

	bc_u8 *allocateChain( bc_u64 start, bc_u64 length, bc_u32 sectorSize )
	{
		free( sectorSize );

		m_accessTime = clock();

		m_byteBuffer = new bc_u8[ (size_t)(length * sectorSize) ];
		if ( m_byteBuffer == NULL )
			return NULL;

		m_length	= length;
		m_start		= start;

		return m_byteBuffer;
	}

	void free( bc_u32 sectorSize )
	{
		LIBLOG_INFO("Begin buffer %x length %d * %d", (unsigned int)m_byteBuffer, (int)m_length, (int)sectorSize );

		if ( m_byteBuffer != NULL )
		{
			memset( m_byteBuffer, 0, (size_t)(sectorSize * m_length) );
			delete m_byteBuffer;
			m_byteBuffer = NULL;
			m_length = 0;
		}

		memset( (bc_u8*)this, 0, sizeof( this ) );

		LIBLOG_INFO("End" );
	}
};

//
// Class cReadCache
//	It does not cache write operation, it just updates cached sectors at writting operation.
//	Main purpouse is to save clouds reading operation.
//	So it is recommended to preculculate crytical places like MFT, Root directory and read them by one call.
//
class cReadCache
{
	int m_chainQ;
	cSectorChain *m_chainT;

public:

	bc_u32 m_bytesPerSector;

	const cSectorChain *getChain( int i )
	{
		if ( ( i < 0 ) || ( i >= m_chainQ ) )
			return NULL;

		return m_chainT + i;
	}

	cReadCache( int chainsQ = 0, bc_u32 bytesPerSector = 0 )
	{
		m_chainT = NULL;
		m_chainQ = 0;
		m_bytesPerSector = bytesPerSector;
		allocateChainTable( chainsQ, bytesPerSector );
	}

	void freeChainTable()
	{
		if ( m_chainT == NULL )
			return;

		for( int i = m_chainQ; i--; )
			m_chainT[i].free( m_bytesPerSector );
		
		delete [] m_chainT;
		m_chainT = NULL;
		m_chainQ = 0;
	}
	
	void allocateChainTable( int chainsQ, bc_u32 bytesPerSector )
	{
		if ( chainsQ > m_chainQ )
		{
			freeChainTable();
			m_chainQ = chainsQ;
			m_chainT = new cSectorChain[m_chainQ];
			if ( m_chainT == NULL )
				m_chainQ = 0;
		}
		
		m_bytesPerSector = bytesPerSector;
	}

	~cReadCache()
	{
		LIBLOG_INFO( "Begin T %x Q %d", (unsigned int)m_chainT, (int)m_chainQ );
		freeChainTable();
		LIBLOG_INFO( "End" );
	}

	bc_u8 *allocateChain( bc_u64 startSector, bc_u64 sectorQ, int priority, int *pPos = NULL )
	{
		int i = findOldChain();

		if ( pPos != NULL )
			*pPos = i;

		if ( i == -1 )
			return NULL;
		
		m_chainT[i].m_priority = priority;
		m_chainT[i].m_accessTime = clock();

		return m_chainT[i].allocateChain( startSector, sectorQ, m_bytesPerSector );
	}

	void freeChain( int i )
	{
		if ( ( i < 0 ) || ( i >= m_chainQ ) )
			return;

		m_chainT[i].free( m_bytesPerSector );
	}

	bc_u8 *getSectors( bc_u64 sectorN, bc_u64 sectorQ )
	{
		bool bIncluded = false;
		int i = findChain( sectorN, sectorQ, bIncluded );

		if ( i == -1 )
			return NULL;

		if ( !bIncluded )
		{
			freeChain( i );
			return NULL;
		}

		bc_u64 s = sectorN - m_chainT[i].m_start;
		m_chainT[i].m_accessTime = clock();

		return m_chainT[i].m_byteBuffer + s * m_bytesPerSector;
	}

	int findChain( bc_u64 sectorN, bc_u64 sectorQ, bool &bIncluded, bc_u8 **pData = NULL )
	{
		for( int i = m_chainQ; i--; )
		{
			//
			//	Empty chain
			//
			if ( m_chainT[i].m_length == 0 )
				continue;

			bc_u64 last = sectorN + sectorQ - 1;
			
			//
			//	Check if new chain is completely included by cyrrent cache item
			//
			if ( m_chainT[i].m_start > last )
				continue;
			
			bc_u64 lastInCache = m_chainT[i].m_start + m_chainT[i].m_length - 1;

			if ( sectorN > lastInCache )
				continue;
			
			bIncluded = ( last <= lastInCache ) && ( sectorN >= m_chainT[i].m_start );

			if ( bIncluded )
			{
				m_chainT[i].m_accessTime = clock();

				if ( pData != NULL )
				{
					bc_u64 offs = sectorN - m_chainT[i].m_start;
					*pData = m_chainT[i].m_byteBuffer + offs * m_bytesPerSector;
				}
			}

			return i;
		}

		return -1;
	}
	
	//
	//	findOldChain
	//		Finds an old chain with lowest priority to be removed
	//
	int findOldChain()
	{
		clock_t firstT = 0;
		int firstN = -1;
		int firstP = -1;

		for( int i = m_chainQ; i--; )
		{
			if ( m_chainT[i].m_length == 0 )
				return i;

			clock_t lTime = m_chainT[i].m_accessTime;
			int lPriority = m_chainT[i].m_priority;

			if ( ( firstN == -1 ) || ( firstP > lPriority ) || ( firstT > lTime ) )
			{
				firstT = lTime;
				firstN = i;
				firstP = lPriority;
			}
		}

		return firstN;
	}
};
