/*
 *	Copyright (c) 2012-2015 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Interface class for volume.
 */

#pragma once

#include "cReadCache.h"
#include "log.h"

const int maxChainsQ = 20;
const bc_u32 minReadAhead = 256 * 1024; // 256K

//#define _DebugCache

class iVolume
{
	bool m_bWriteProtected;
	int m_currentReadCachePriority; // See eReadCachePriority,  eRCPNo disable caching.
	cReadCache m_cache;

	//
	//	readFromCache
	//		If entire chain is found so fills data and return true.
	//		Otherwise return false.
	//
	bool readFromCache( bc_u64 start_sector, bc_u8 *data, bc_u64 number_of_sectors )
	{
		bool bIncluded = false;
		bc_u8 *chachedData = NULL;

		int i = m_cache.findChain( start_sector, number_of_sectors, bIncluded, &chachedData );
		
		if ( ( i == -1 ) || !bIncluded )
			return false;

		if ( chachedData == NULL )
			return false;

		memcpy( data, chachedData, (size_t)(number_of_sectors * m_cache.m_bytesPerSector) );

		return true;
	}

	//
	//	writeToCache
	//		Finds existing cache sectors
	//			Updates them
	//		If existing cache sectors were not found and cachePriority is more then 0 so create new chain.
	//
	bool writeToCache( bc_u64 start_sector, const bc_u8 *data, bc_u64 number_of_sectors, int cachePriority )
	{
		bool bIncluded = false;
		bc_u8 *chachedData = NULL;

		// Update existing
		while( true )
		{
			int i = m_cache.findChain( start_sector, number_of_sectors, bIncluded, &chachedData );
			
			if ( i == -1 )
				break;
			
			if ( !bIncluded )
			{
				// Removes all partial chains. Rather the object we chached was changed.
				m_cache.freeChain( i );
				continue;
			}
			else
			{
				if ( chachedData == NULL )
					return false;

				// Was updated
				memcpy( chachedData, data, (size_t)(number_of_sectors * m_cache.m_bytesPerSector) );
				return true;
			}
		}

		if ( cachePriority == eRCPNo )
			return true;
		
		// Adds new
		m_cache.allocateChainTable( maxChainsQ, getSectorSize() );

		chachedData = m_cache.allocateChain( start_sector, number_of_sectors, cachePriority );

		if ( chachedData == NULL )
			return false;

		// Was updated
		memcpy( chachedData, data, (size_t)(number_of_sectors * m_cache.m_bytesPerSector) );
		return true;
	}

	//
	//	readAhead
	//		Read once minReadAhead. For cloud reading it almost does not mate if we read 1K or 1M.
	//		But if we read several sectors it significantly saves time.
	//		Note, this function does not work if cache priority is zero and if number_of_sectors to large.
	//
	bool readAhead( bc_u64 start_sector, bc_u8 *data, bc_u64 number_of_sectors, int cachePriority )
	{
		if ( cachePriority <= 0 )
			return false;

		bc_u32 sectorSize = getSectorSize();

		if ( sectorSize == 0 )
			return false;

		bc_u64 readAheadSectorsQ = minReadAhead / sectorSize;

		if ( number_of_sectors > readAheadSectorsQ ) // Odinary way will be used for reading
			return false;

		// Adds new
		m_cache.allocateChainTable( maxChainsQ, sectorSize );

		int pos = 0;

		for( int c = 0; ; c++ )
		{
			if ( c > maxChainsQ ) // Check loop...
				return false;
			bool bIncluded;

			pos = m_cache.findChain( start_sector, readAheadSectorsQ, bIncluded );

			if ( pos == -1 )
				break;

			if ( bIncluded ) // Just check including. Otherwise it we get rproblem with duplicate cache!
				return false;
			// Another case read ahead size is too large.
			// we just reduce it.

			const cSectorChain *sc = m_cache.getChain( pos );

			if ( sc == NULL )
				return false;

			if ( sc->m_start < start_sector ) // Is included by tail
				return false;
			
			if ( ( sc->m_start - start_sector ) > readAheadSectorsQ ) // Noncence
				return false;
			
			readAheadSectorsQ = sc->m_start - start_sector;
		}

		bc_u8 *chachedData = m_cache.allocateChain( start_sector, readAheadSectorsQ, cachePriority, &pos );

		if ( chachedData == NULL )
			return false;

		if ( iv_Success != readSectors( start_sector, chachedData, readAheadSectorsQ ) )
		{
			m_cache.freeChain( pos );
			return false;
		}

		bool bRes = readFromCache( start_sector, data, number_of_sectors );

		LIBLOG_INFO( "(%lld) sector 0x%llx(%lld), size 0x%llx(%lld) = SUCCESS!",
				(long long) readAheadSectorsQ, (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );

		return bRes;
	}

public:

	enum eResult{
		iv_Success,
		iv_IOError
	};

	//
	//	readSectorsCache
	//		read sectors in cache first.
	//
	eResult readSectorsCache( bc_u64 start_sector, bc_u8 *data, bc_u64 number_of_sectors  )
	{
		if ( readFromCache( start_sector, data, number_of_sectors ) )
		{
			LIBLOG_INFO( "readFromCache sector 0x%llx(%lld), size 0x%llx(%lld) = SUCCESS!",  (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
#ifdef _DebugCache
			return checkReadResult( start_sector, data, number_of_sectors );
#endif
			return iv_Success;
		}

		if ( readAhead( start_sector, data, number_of_sectors, m_currentReadCachePriority ) )
		{
#ifdef _DebugCache
			return checkReadResult( start_sector, data, number_of_sectors );
#endif
			return iv_Success;
		}

		eResult res = readSectors( start_sector, data, number_of_sectors );

		if ( res == iv_Success )
			writeToCache( start_sector, data, number_of_sectors, m_currentReadCachePriority );

		return res;
	}

	eResult writeSectorsCache( bc_u64 start_sector, const bc_u8 *data, bc_u64 number_of_sectors  )
	{
		eResult res = writeSectors( start_sector, data, number_of_sectors );

		if ( res == iv_Success )
			writeToCache( start_sector, data, number_of_sectors, m_currentReadCachePriority );

#ifdef _DebugCache
		if ( res == iv_Success )
			res = checkWriteResult( start_sector, data, number_of_sectors );
#endif
		return res;
	}

#ifdef _DebugCache

	bc_u8 *allocateBuffer( bc_u64 number_of_sectors, size_t &bfSize )
	{
		if ( number_of_sectors == 0 )
		{
			LIBLOG_INFO( "Zero number of sectors", __FUNCTION__, __LINE__ );
			return NULL;
		}

		bfSize = (size_t)(number_of_sectors * getSectorSize() );

		if ( bfSize == 0 )
		{
			LIBLOG_INFO( "Zero buffer size", __FUNCTION__, __LINE__ );
			return NULL;
		}

		bc_u8 *bf = new bc_u8[bfSize];

		if ( bf == NULL )
			LIBLOG_INFO( "Error in memory allocation, buffer size %X",  bfSize );

		return bf;	
	}

	eResult checkReadResult( bc_u64 start_sector, bc_u8 *data, bc_u64 number_of_sectors  )
	{
		size_t bfSize = 0;
		bc_u8 *bf = allocateBuffer( number_of_sectors, bfSize );
		if ( bf == NULL )
			return iv_IOError;

		if ( readSectors( start_sector, bf, number_of_sectors ) != iv_Success )
		{
			LIBLOG_INFO( "Assertion! Error in reading data! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
				 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
			delete bf;
			return iv_IOError;
		}
		else
		{
			if ( memcmp( data, bf, bfSize ) != 0 )
			{
				LIBLOG_INFO( "Assertion! Read data is invalid! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
					 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
				delete bf;
				return iv_IOError;
			}
			else
			{
				LIBLOG_INFO( "Read data is OK! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
					 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
			}
		}

		delete bf;
		return iv_Success;

	}

	eResult checkWriteResult( bc_u64 start_sector, const bc_u8 *data, bc_u64 number_of_sectors  )
	{
		size_t bfSize = 0;
		bc_u8 *bf = allocateBuffer( number_of_sectors, bfSize );
		if ( bf == NULL )
			return iv_IOError;

		//
		//	1. Check cached content.
		//
		if ( readFromCache( start_sector, bf, number_of_sectors ) )
		{
			if ( memcmp( data, bf, bfSize ) != 0 )
			{
				LIBLOG_INFO( "Assertion! Cached data is invalid! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
					 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
				delete bf;
				return iv_IOError;
			}
			else
			{
				LIBLOG_INFO( "Cached data is OK! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
					 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
			}
		}

		//
		//	2. Check read ahead.
		//
		if ( readAhead( start_sector, bf, number_of_sectors, m_currentReadCachePriority ) )
		{
			if ( memcmp( data, bf, bfSize ) != 0 )
			{
				LIBLOG_INFO( "Assertion! Read ahead data is invalid! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
					 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
				delete bf;
				return iv_IOError;
			}
			else
			{
				LIBLOG_INFO( "Read ahead data is OK! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
					 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
			}
		}

		//
		//	3. Check direct data.
		//
		if ( readSectors( start_sector, bf, number_of_sectors ) != iv_Success )
		{
			LIBLOG_INFO( "Assertion! Error in reading data! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
				 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
			delete bf;
			return iv_IOError;
		}
		else
		{
			if ( memcmp( data, bf, bfSize ) != 0 )
			{
				LIBLOG_INFO( "Assertion! Read data is invalid! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
					 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
				delete bf;
				return iv_IOError;
			}
			else
			{
				LIBLOG_INFO( "Read data is OK! Sector 0x%llx(%lld), size 0x%llx(%lld)!", 
					 (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );
			}
		}

		delete bf;
		return iv_Success;
	}
#endif // _DebugCache


	virtual eResult readSectors( bc_u64 start_sector, bc_u8 *data, bc_u64 number_of_sectors  ) = 0;
	virtual eResult writeSectors( bc_u64 start_sector, const bc_u8 *data, bc_u64 number_of_sectors  ) = 0;
	virtual eResult sync() = 0;
	
	iVolume()
	{
		m_bWriteProtected = true;
		m_currentReadCachePriority = eRCPNo;
	}

	void setCurrentReadCachePriority( int priority = eRCPNo )
	{
		m_currentReadCachePriority = priority;
	}
 
	virtual ~iVolume()
	{
		m_bWriteProtected = false;
		m_cache.freeChainTable();
	}

	bool isWriteProtected()
	{
		return m_bWriteProtected;
	}

	void setWriteProtected( bool bSet )
	{
		m_bWriteProtected = bSet;
	}

	virtual bc_u64 getSectorCount() = 0;
	virtual bc_u32 getSectorSize() = 0;
};
