/*
 *	Copyright (c) 2012-2014 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File system interface class.
 */

#pragma once

//defined at "../container/intsize.h"
#include <stdlib.h>
#include <string.h>
#include "equstrings.h"
#include "log.h"
#include "cUString.h"
#include "iVolume.h"

#ifdef _WIN32

#define strdup _strdup

// FILETIME
struct bc_file_time {
	bc_u32 lowDateTime;
    bc_u32 highDateTime;
};
#else
//#error "types are undefined"
// FILETIME
struct bc_file_time {
	bc_u32 lowDateTime;
    bc_u32 highDateTime;
};

#endif

enum bc_file_open_mode
{
	eBCModeRead			= 0x01,
	eBCModeWrite		= 0x02,
	eBCModeExecute		= 0x04,
	eBCDispOpenExisting = 0x08,
	eBCDispOpenAlways	= 0x10,
	eBCDispCreateNew	= 0x20,
	eBCDispCreateAlways	= 0x40,
};

// Repeat FRESULT from ./FatFs/ff.h
enum eifsResult {
	eifs_OK = 0,				/* (0) Succeeded */
	eifs_DISK_ERR,				/* (1) A hard error occured in the low level disk I/O layer */
	eifs_INT_ERR,				/* (2) Assertion failed */
	eifs_NOT_READY,				/* (3) The physical drive cannot work */
	eifs_NO_FILE,				/* (4) Could not find the file */
	eifs_NO_PATH,				/* (5) Could not find the path */
	eifs_INVALID_NAME,			/* (6) The path name format is invalid */
	eifs_DENIED,				/* (7) Acces denied due to prohibited access or directory full */
	eifs_EXIST,					/* (8) Acces denied due to prohibited access */
	eifs_INVALID_OBJECT,		/* (9) The file/directory object is invalid */
	eifs_WRITE_PROTECTED,		/* (10) The physical drive is write protected */
	eifs_INVALID_DRIVE,			/* (11) The logical drive number is invalid */
	eifs_NOT_ENABLED,			/* (12) The volume has no work area */
	eifs_NO_FILESYSTEM,			/* (13) There is no valid FAT volume */
	eifs_MKFS_ABORTED,			/* (14) The f_mkfs() aborted due to any parameter error */
	eifs_TIMEOUT,				/* (15) Could not get a grant to access the volume within defined period */
	eifs_LOCKED,				/* (16) The operation is rejected according to the file shareing policy */
	eifs_NOT_ENOUGH_CORE,		/* (17) LFN working buffer could not be allocated */
	eifs_TOO_MANY_OPEN_FILES,	/* (18) Number of open files > _FS_SHARE */
	eifs_INVALID_PARAMETER,		/* (19) Given parameter is invalid */
	eifs_ALLOCATION_ERR,		/* (20) Error in memory allocation */
	eifs_NO_FREE_SPACE			/* (21) No free space for writing */
};

class iFileObject
{
public:

	//
	//	Full path is filled at open file.
	//	Is freed at close but file statistic is saved by the name before.
	//
	cUString m_path;

	bc_u32 m_attributes;
	bc_file_time m_ct;
	bc_file_time m_la;
	bc_file_time m_lw;

	/* Set new file attributes */
	virtual int setFileAttributes( bc_u32 attributes )
	{
		m_attributes = attributes;
		return 0;
	}

    virtual int setCreationTime( bc_file_time ft )
	{
		m_ct = ft;
		return 0;
	}

    virtual int setLastAccessTime( bc_file_time ft )
	{
		m_la = ft;
		return 0;
	}

    virtual int setLastWriteTime( bc_file_time ft )
	{
		m_lw = ft;
		return 0;
	}

	virtual ~iFileObject()
	{
	}

	/* Read data from a file */
	virtual int read( 
		void *buff,		/* Pointer to data buffer */
		bc_u32 btr,		/* Number of bytes to read */
		bc_u32 *br		/* Pointer to number of bytes read */
		) = 0;

	/* Move file pointer of a file object */
	enum eSeekMode { file_begin = 0, cur_pos, file_end }; 
	virtual bc_u64 setPointer( 
		bc_64 ofs,		/* File pointer from point defined by mode */
		eSeekMode mode = file_begin	/* one of enum above */
		) = 0;

	/* Write data to a file */
	virtual int write( 
		const void *buff,	/* Pointer to the data to be written */
		bc_u32 btw,			/* Number of bytes to write */
		bc_u32 *bw			/* Pointer to number of bytes written */
		) = 0;

	/* Flush cached data of a writing file */
	virtual int sync() = 0;

	/* Close an open file object */
	virtual int close() = 0;

	/* Truncate file */
	virtual int setEndOfFile() = 0;

	/* Set new file attributes */
	//virtual int setFileAttributes( bc_u32 attributes ) = 0;
    //virtual int setCreationTime( bc_file_time ft ) = 0;
    //virtual int setLastAccessTime( bc_file_time ft ) = 0;
    //virtual int setLastWriteTime( bc_file_time ft) = 0;

	virtual bc_u64 getSize( bc_u64 *pCurPos = 0 )
	{
		bc_u64 curPos = setPointer( 0, iFileObject::cur_pos );
		bc_u64 size = setPointer( 0, iFileObject::file_end );
		setPointer( curPos, iFileObject::file_begin );
		
		if ( pCurPos != 0 )
			*pCurPos = curPos;

		return size;
	}

	virtual bool isEOF()
	{
		bc_u64 curPos = 0;
		return getSize( &curPos ) == curPos;
	}
};

class iDirectoryObject
{
public:
	char *m_folderPath;

    virtual bc_u32 getFileAttributes() = 0;
    virtual bc_file_time getCreationTime() = 0;
    virtual bc_file_time getLastAccessTime() = 0;
    virtual bc_file_time getLastWriteTime() = 0;
    virtual bc_u64 getFileSize() = 0;
	virtual bc_u64 getCompressedFileSize() = 0;
	virtual const char *getFileName() = 0;
	virtual const bc_u16 *getFileNameW() = 0;
	//virtual bool isRegularFile() = 0;
	//virtual bool isDirectory() = 0;
	virtual bool isRegularFile()	{ return (getFileAttributes() & 0x10) == 0; }	// FILE_ATTRIBUTE_DIRECTORY
	virtual bool isDirectory()		{ return (getFileAttributes() & 0x10) != 0; }
	virtual bool isHidden()			{ return (getFileAttributes() & 0x02) != 0; }	// FILE_ATTRIBUTE_HIDDEN

	virtual ~iDirectoryObject()
	{
		if ( m_folderPath != 0 )
		{
			free( m_folderPath );
			m_folderPath = 0;
		}
	}
	
	virtual const bc_u16 *getFolderPathW() = 0;

	virtual const char *getFolderPath()
	{
		return m_folderPath;
	}
	
	iDirectoryObject( const char *folderPath, int len = -1 )
	{
		m_folderPath = 0;

		if ( folderPath != 0 )
		{
			m_folderPath = strdup( folderPath );

			if ( len > 0 )
				m_folderPath[len] = '\0';
		}
	}
};

class iFileSystem
{
protected:
	BYTE m_drv;

	static iVolume *m_mountedVolumeT[_VOLUMES];

public:

	static int getFirst()
	{
		static bool bInited = false;
		if ( !bInited )
		{
			memset( m_mountedVolumeT, 0, sizeof( m_mountedVolumeT ) );
			bInited = true;
		}

		for( int i = 0; i < _VOLUMES; i++ )
		{
			if ( m_mountedVolumeT[i] == NULL )
				return i;
		}

		return -1;
	}

	virtual iVolume *getVolume()
	{
		return m_volume;
	}

	static iVolume *getMountedVolume( BYTE drv )
	{
		LIBLOG_INFO( "Begin %d", (int)drv );

		if ( drv >= _VOLUMES )
			return NULL;

		LIBLOG_INFO( "End %x", (unsigned int)m_mountedVolumeT[drv] );

		return m_mountedVolumeT[drv];
	}

	static bool setMountedVolume( BYTE drv, iVolume *ptr )
	{
		LIBLOG_INFO( "Begin %d, %x", (int)drv, (unsigned int)ptr  );

		if ( drv >= _VOLUMES )
		{ assert( false ); return false; }

		m_mountedVolumeT[drv] = ptr;
		LIBLOG_INFO( "End %x", (unsigned int)m_mountedVolumeT[drv] );

		return true;
	}

public:

	int m_lastError;
	
	iVolume *m_volume;

	iFileSystem( iVolume *volume )
	{
		m_volume = volume;
		m_lastError = eifs_OK;
	}

	virtual ~iFileSystem()
	{
		// File system cannot destroy volume, it may be used by another file system!!!
//		if ( m_volume != NULL )
//		{
////			BCPutVolume( m_volume );
//			delete m_volume;
//			m_volume = NULL;
//		}
	}

	/* Create File System on the Drive */
	virtual int makeFS() = 0;

	/* Mount/Unmount a logical drive */
	virtual int mountFS() = 0;
	virtual int unMountFS() = 0;

	/**
	 * Open directory object
	 */
	virtual iDirectoryObject *findOpen( const char *dirPath, const char *fileNamePattern ) = 0;

	const char *getFullPath( cUString *adb, const char *filePath )
	{
		adb->Free();

		if ( ( filePath[0] != 0 ) && ( filePath[1] == ':' ) )
			adb->CreateCopy( filePath );
		else
		{
			char drvName[3] = "0:";
			drvName[0] = '0' + m_drv;

			adb->AddString( drvName );

			if ( ( filePath[0] == '\\' ) || (  filePath[0] == '/' ) )
				filePath++;

			adb->AddString( "\\" );

			adb->AddString( filePath );
		}

		return adb->strA();
	}

	const char *getFullPath( cUString *adb, const char *dirPath, const char *fileName )
	{
		if ( dirPath[0] == 0 )
			return getFullPath( adb, fileName );

		getFullPath( adb, dirPath );

		if ( fileName[0] != '\\' )
			adb->AddString( "\\" );

		adb->AddString( fileName );

		return adb->strA();
	}
	/**
	 * Open directory object by path
	 */
	iDirectoryObject *findOpen( const char *filePath )
	{
		char *buff = strdup( filePath );

		LIBLOG_INFO( "Begin %s", buff );

		if ( buff == NULL )
			return NULL;

		size_t pos = findLast( buff, "\\/" );

		const char *fileName;
		const char *dirPath;

		if ( -1 == pos )
		{
			fileName = buff;
			dirPath = "";
		}
		else
		{
			buff[pos] = '\0';
			fileName = buff + pos + 1;
			dirPath = buff;
		}

		bool bIsRoot = filePath[0] == '\0';

		if ( ( fileName[0] == '\0' ) && !bIsRoot )
			fileName = "*";

		iDirectoryObject *dir = findOpen( dirPath, fileName );

		free( buff );

		LIBLOG_INFO( "End %x", (unsigned int)dir );
		return dir;
	}

	/**
	 * Open directory and find first file by pattern
	 */
	iDirectoryObject *findFirstFile( const char *dirPath, const char *fileNamePattern )
	{
		iDirectoryObject *dir = findOpen( dirPath, fileNamePattern );

		if ( dir == NULL )
			return NULL;

		if ( !findNextFile( dir ) )
		{
			findClose( dir );
			return NULL;
		}

		return dir;
	}

	/**
	 * Open directory and find files
	 */
	iDirectoryObject *findFirstFile( const char *filePath )
	{
		LIBLOG_INFO( "Begin '%s'", filePath );

		iDirectoryObject *dir = findOpen( filePath );

		if ( dir == NULL )
		{ LIBLOG_INFO( "Directory object '%s' not found", filePath );	return NULL; }

		bool bIsRoot = filePath[0] == '\0';

		if ( !findNextFile( dir ) && !bIsRoot )
		{
			LIBLOG_INFO( "Next directory object '%s' not found", dir->getFileName() );
			findClose( dir );
			return NULL;
		}

		LIBLOG_INFO( "End '%p'", dir );
		return dir;
	}

	bool fileExists( const char *filePath )
	{
		LIBLOG_INFO( "Begin '%s'", filePath );

		iDirectoryObject *obj = findFirstFile( filePath );

		bool bExist = ( obj != NULL );

		if ( bExist )
			delete obj;

		LIBLOG_INFO( "End '%s' = %s", filePath, bExist ? "True" : "False" );

		return bExist;
	}

	/**
	 * Find next file
	 */
	virtual bool findNextFile( iDirectoryObject *pDirObject ) = 0;

	/**
	 * Close directory object
	 */
	virtual void findClose( iDirectoryObject *pDirObject ) = 0;

	/**
	 * Open or create a file
	 * @param filePath   path to file.
	 * @param mode       one or more from bc_file_open_mode enum.
	 * @return           pointer to file object.
	 */
	virtual iFileObject *openFile( const char *filePath, int mode ) = 0;
	virtual iFileObject *openFile( iDirectoryObject *pDirObject, int mode ) = 0;

	/* Close an open file object */
	virtual void closeFile( iFileObject *pFileObject ) = 0;
	
	/* Delete an existing file or directory */
	virtual int deleteFile( const char *filePath ) = 0;

	/* Create a new directory */
	virtual int createDirectory( const char *dirPath ) = 0;

	/* Change timestamp of the file/dir */
	virtual int setFileTime( const char *path, bc_file_time file_time ) = 0;
	
	/* Rename/Move a file or directory */
	virtual int renameFile( const char *oldPath, const char *newPath ) = 0;

	/* Get number of free bytes on the drive */
	virtual int getFreeSpace( bc_u64 *pFreeSpaceInBytes ) = 0;
};

static inline BYTE getMode( int bcMode )
{
	BYTE ffMode = 0;
	if ( bcMode & eBCModeRead )		ffMode |= FA_READ;
	if ( bcMode & eBCModeWrite )	ffMode |= FA_WRITE;
	if ( bcMode & eBCModeExecute )	ffMode |= FA_READ;

	if ( bcMode & eBCDispOpenExisting )	ffMode |= FA_OPEN_EXISTING;
	if ( bcMode & eBCDispOpenAlways )	ffMode |= FA_OPEN_ALWAYS;
	if ( bcMode & eBCDispCreateNew )	ffMode |= FA_CREATE_NEW;
	if ( bcMode & eBCDispCreateAlways )	ffMode |= FA_CREATE_ALWAYS;

	return ffMode;
}

