/*
 *	Copyright (c) 2015 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	NTFS File system class.
 */

#pragma once

#include "ifs.h"
#include <assert.h>
#include "cADB.h"
#include "equstrings.h"
#include <locale.h>
#include <time.h>
#include "log.h"

class cNtfsFileObject: public iFileObject
{
public:

	void *m_attrCtx;
	bc_u64 m_pos;
	bc_file_open_mode m_mode;

	~cNtfsFileObject()
	{
		clear();
	}

	cNtfsFileObject()
	{
		clear();
	}

	void clear()
	{
		m_attributes = -1;
		memset( &m_lw, 0, sizeof( m_lw ) );
		m_ct = m_la = m_lw;
		m_attrCtx = NULL;
		m_pos = 0;
		m_mode = (bc_file_open_mode)0;
	}

	/* Read data from a file */
	virtual int read(
		void *buff,		/* Pointer to data buffer */
		bc_u32 btr,		/* Number of bytes to read */
		bc_u32 *br		/* Pointer to number of bytes read */
		);

	/* Move file pointer of a file object */
	virtual bc_u64 setPointer(
		bc_64 ofs,		/* File pointer from point defined by mode */
		eSeekMode mode = file_begin	/* one of enum above */
		);

	//FRESULT f_write (FIL*, const void*, UINT, UINT*);	/* Write data to a file */
	virtual int write(
		const void *buff,	/* Pointer to the data to be written */
		bc_u32 btw,			/* Number of bytes to write */
		bc_u32 *bw			/* Pointer to number of bytes written */
		);

	/* Flush cached data of a writing file */
	virtual int sync()
	{
		//
		// It strange but ntfs write attributes operations do not require sync.
		//	The volume sync has to be called at container dismount.
		//
		return 0;
	}

	/* Flush cached data of a writing file */
	virtual int close();

	/* Truncate file */
	virtual int setEndOfFile();

	virtual bc_u64 getSize( bc_u64 *pCurPos = 0 );

	virtual bool isEOF()
	{
		return getSize() == m_pos;
	}

};

#pragma pack (1)
struct __WIN32_FIND_DATAW
{
	bc_u32 dwFileAttributes;
	bc_file_time ftCreationTime;
	bc_file_time ftLastAccessTime;
	bc_file_time ftLastWriteTime;
	bc_u64 fileSize;
	bc_u64 allocatedSize;
	bc_u16 cFileName[ 1024 ];
};
#pragma pack ()


class cNtfsDirectoryObject : public iDirectoryObject
{
public:
	cADB < char > m_path;
	const char *m_findFileName;
	wchar_t *m_folderPathW;
	wchar_t *m_longNameW;

	virtual bc_u32 getFileAttributes()		{ return m_findData.dwFileAttributes; }
	virtual bc_file_time getCreationTime()	{ return m_findData.ftCreationTime; }
	virtual bc_file_time getLastAccessTime(){ return m_findData.ftLastAccessTime; }
	virtual bc_file_time getLastWriteTime() { return m_findData.ftLastWriteTime; }
	virtual bc_u64 getFileSize()			{ return m_findData.fileSize; }
	virtual bc_u64 getCompressedFileSize()	{ return m_findData.allocatedSize; }

	void *m_inode;
	bc_64 m_findPos;
	__WIN32_FIND_DATAW m_findData;

	cNtfsDirectoryObject( const char *folderPath, int len = -1 )
		: iDirectoryObject( folderPath, len )
	{
		m_folderPathW = NULL;
		m_longNameW = NULL;
		m_inode = NULL;
		m_findPos = 0;
		memset( &m_findData, 0, sizeof( m_findData ) );
	}

	void clearLongName()
	{
		deleteWString( &m_longNameW );
	}

	~cNtfsDirectoryObject()
	{
		deleteWString( &m_folderPathW );
		clearLongName();
	}

	void deleteWString( wchar_t **ppStr )
	{
		if ( NULL == ppStr )
			return;

		if ( NULL == *ppStr )
			return;

		delete *ppStr;

		*ppStr = NULL;
	}

	wchar_t *createWString( const char *mbs )
	{
		/* Enable the Japanese locale and codepage */
		char *localeInfo = setlocale(LC_CTYPE, "Japanese_Japan.932");

		size_t requiredSize = mbstowcs(NULL, mbs, 0); // C4996
		/* Add one to leave room for the NULL terminator */
		wchar_t *pwc = new wchar_t[ requiredSize + 1];
		if ( pwc == 0 )
			return 0;

		if ( mbstowcs( pwc, mbs, requiredSize + 1 ) != requiredSize )
		{
			delete pwc;
			return NULL;
		}

		return pwc;
	}

	cUString _gfn;

	virtual const char *getFileName()
	{
		return cUString::convertWcsToUTF8( _gfn.m_bodyA, (const UTF16 *)getFileNameW() );
	}

	virtual const bc_u16 *getFileNameW()
	{
		return m_findData.cFileName;
	}

	virtual const char *getFolderPath()
	{
		if ( m_folderPath == NULL )
			return NULL;

		if ( m_folderPath[1] == ':' )
			return m_folderPath + 2;

		return m_folderPath;
	}

	virtual const bc_u16 *getFolderPathW()
	{
		if ( m_folderPathW == NULL )
			m_folderPathW = createWString( getFolderPath() );

		return (const bc_u16 *)m_folderPathW;
	}

};

class iNTFS : public	iFileSystem
{
	void *m_volCtx;

public:

	iNTFS( iVolume *volume )
		: iFileSystem( volume )
	{
		m_drv = getFirst();
	}

	/*-----------------------------------------------------------------------*/
	/* Create File System on the Drive                                       */
	/*-----------------------------------------------------------------------*/
	virtual int makeFS();

	/**
	 * Mount/Unmount file system
	 */
	virtual int mountFS();
	int unMountFS();


	/**
	 * Open directory object
	 */
	virtual iDirectoryObject *findOpen( const char *dirPath, const char *fileNamePattern );

	virtual bool findNextFile( iDirectoryObject *pDirObject );

	virtual void findClose( iDirectoryObject *pDirObject );

	/**
	 * Open file
	 * @filePath   path to file.
	 * @mode       one or more from bc_file_open_mode enum.
	 */
	virtual iFileObject *openFile( const char *filePath, int mode );

	virtual iFileObject *openFile( iDirectoryObject *pDirObject, int mode );

	/* Close an open file object */
	virtual void closeFile( iFileObject *pFileObject );

	/* Delete an existing file or directory */
	virtual int deleteFile( const char *filePath );

	/* Create a new directory */
	virtual int createDirectory( const char *dirPath );

	/* Change timestamp of the file/dir */
	virtual int setFileTime( const char *path, bc_file_time file_time );

	/* Rename/Move a file or directory */
	virtual int renameFile( const char *oldPath, const char *newPath );

	/* Get number of free clusters on the drive */
	virtual int getFreeSpace( bc_u64 *pFreeSpaceInBytes );
};

