#include "ContainerSource.h"
#include "../user.h"
#include "../file.h"
#include "../volume.h"
#include "../container.h"
#include "../ciphercrtlib.h"
#include <stdint.h>
#include <map>
#include <string>
//#include <ConvertUTF.h>
//#include "../intsize.h"

namespace bcrypt_container{
	static const char* BCAND_FILE_ID = "bcfiles-android";
	static const int SECTOR_SIZE = 512;

	class ContainerImpl: public IContainer{
		bcand::container::IContainerSource& source;
		BCVolume volume;
		bc_key_handle keyHandle;
	public:
		static std::map<std::string, bcand::container::IContainerSource*> openedContainers;

		ContainerImpl(bcand::container::IContainerSource& _source):source(_source){
			CipherFactory::m_gCipher = (CipherBase*) new CipherCrtLib();
		}
		~ContainerImpl(){
			delete CipherFactory::m_gCipher;
			CipherFactory::m_gCipher = 0;
		}

		virtual CONT_IO_STATUS diskInitialize(){
			openedContainers[BCAND_FILE_ID] = &source;
			BCMountParams mountParams = { false, true, "" };
			BCContainerResult res = BCDecodeVolume(BCAND_FILE_ID, (bc_file_mode)kBCFileMode_ReadWrite, BCAND_FILE_ID, false, &mountParams, &volume);
			switch (res){
			case kBCContainer_InvalidPassword:
				return CONT_IO_INVCRED;
			case kBCContainer_Success:
				return CONT_IO_OK;
			default:
				return CONT_IO_ERROR;
			}
		}
		virtual CONT_IO_STATUS diskStatus(){
			return CONT_IO_OK;
		}
		virtual CONT_IO_STATUS diskRead  (      void* buf, __int64_t firstSector, __int64_t sectorCount){
			BCContainerResult res = BCReadVolumeSectors(&volume, firstSector, (bc_u8*)buf, sectorCount);
			if (kBCContainer_Success !=  res){
				return CONT_IO_ERROR;
			}
			return CONT_IO_OK;
		}
		virtual CONT_IO_STATUS diskWrite (const void* buf, __int64_t firstSector, __int64_t sectorCount){
			BCContainerResult res = BCWriteVolumeSectors(&volume, firstSector, (bc_u8*)buf, sectorCount);
			if (kBCContainer_Success !=  res){
				return CONT_IO_ERROR;
			}
			return CONT_IO_OK;
		}

		/**
		 * Executes control function on disk
		 * 	params:
		 * IOCTL_SYNC - Make sure that the disk drive has finished pending write process. When the disk I/O module has a write back cache, flush the dirty sector immediately. This command is not used in read-only configuration. Second param is unused
		 * IOCTL_SECTOR_SIZE - Returns sector size of the drive into the WORD variable pointed by buf
		 * IOCTL_SECTOR_COUNT - Returns number of available sectors on the drive into the _int64_t variable pointed by buf. This command is used by only f_mkfs function to determine the volume size to be created.
		 */
		virtual CONT_IO_STATUS diskIOctl (IO_CTL command, void* buf){
			return CONT_IO_OK;
		}
	};

	std::map<std::string, bcand::container::IContainerSource*> ContainerImpl::openedContainers;

	IContainer* containerCreator(bcand::container::IContainerSource& source){
		return new ContainerImpl(source);
	}

};

unsigned BCUserGenerateSeed(bc_u8* out_buf, unsigned total){
	return 0;
}

bc_file_result BCFileOpen(const char *fileName, bc_file_mode mode, bc_file_handle *handle){
	if (!fileName || std::string(bcrypt_container::BCAND_FILE_ID) != fileName){
		return kBCFileResult_InternalError;
	}
	if (bcrypt_container::ContainerImpl::openedContainers.count(fileName)<1){
		return kBCFileResult_InternalError;
	}
	*handle = bcrypt_container::ContainerImpl::openedContainers[fileName];
    return kBCFileResult_Success;
}

bc_u64 BCFileRead(bc_file_handle *handle, void *buffer, bc_u64 byteCount){
	bcand::container::IContainerSource* source = (bcand::container::IContainerSource*)*handle;
	return source->read(buffer, source->position(), byteCount );
}

bc_u64 BCFileReadWithOffset(bc_file_handle *handle, void *buffer, bc_u64 byteCount, bc_u64 offset){
	return ((bcand::container::IContainerSource*)*handle)
			->read(buffer, offset, byteCount );
}

bc_u64 BCFileWrite(bc_file_handle *handle, const void *buffer_, bc_u64 byteCount){
	bcand::container::IContainerSource* source = (bcand::container::IContainerSource*)*handle;
	return source->write(buffer_, source->position(), byteCount);
}

bc_u64 BCFileWriteWithOffset(bc_file_handle *handle, const void *buffer, bc_u64 byteCount, bc_u64 offset){
	return ((bcand::container::IContainerSource*)*handle)
			->write(buffer, offset, byteCount );
}

BCUserResult BCUserVerifyPassword(const char* query, BCVerifyPasswordCallback callback, void* cookie, BCMountParams * /*mountParams*/, BCPassword* out_pwd){
	if (bcrypt_container::ContainerImpl::openedContainers.count(query)<1){
		return kBCUserCancel;
	}
	using bcand::container::IContainerSource;
	using bcand::container::ICredentials;
	IContainerSource* source = bcrypt_container::ContainerImpl::openedContainers[query];
	ICredentials* credentials = source->requestCredentials(ICredentials::CRED_UNDEFINED);
	bcand::container::PasswordCredentials* pwdCred = (bcand::container::PasswordCredentials*)credentials;
    const char* errorMsg = callback(pwdCred->getPasswordBuf(), pwdCred->getPasswordSize(), cookie);
    if (!errorMsg){
    	out_pwd->set(pwdCred->getPasswordBuf());
    	return kBCUserOk;
    }else{
    	return kBCUserCancel;
    }
}

extern "C" {
#define min(a,b) (a<b?a:b)
size_t BCUserConvertString(const bc_u8* in, size_t inlen, bc_u8* out, size_t outlen, BCUserEncoding fromEncoding, BCUserEncoding toEncoding){
	int resLen = 0;
	if (fromEncoding == toEncoding){
		resLen = min(inlen, outlen);
		memcpy((void*)out, (const void*)in, (unsigned long)resLen);
		return resLen;
	}

#define CONV_DIR(from,to) ((from<<8)|to)
	int covertionDir = CONV_DIR(fromEncoding,toEncoding);
	bc_u8* outSaved = out;

#ifdef _BIG_ENDIAN_DEFINED
	#define	nativeUTF16 kBCUser_UTF16BE
	#define	reversUTF16 kBCUser_UTF16LE
#else
	#define	nativeUTF16 kBCUser_UTF16LE
	#define	reversUTF16 kBCUser_UTF16BE
#endif

//	bc_u8* tIn = 0;

	switch (covertionDir){
	case CONV_DIR(kBCUser_UTF8, kBCUser_UTF16LE):
		return BCUserUTF8ToUTF16LE((const char*)in, inlen, out, outlen);
	case CONV_DIR(kBCUser_UTF8, kBCUser_UTF16BE):
		return BCUserUTF8ToUTF16BE((const char*)in, inlen, out, outlen);
/*
		ConvertUTF8toUTF16((const UTF8**)&in, in+inlen, (UTF16**)&out, (UTF16*)(out+outlen), strictConversion);
		resLen = out - outSaved;
		if (covertionDir & reversUTF16){
			for (bc_u16* utf16Runner = (bc_u16*)outSaved; utf16Runner < (bc_u16*)out; ++utf16Runner){
				const bc_u16 character = *utf16Runner;
				*utf16Runner =  ((character & 0xFF) << 8) | ((character >> 8) & 0xFF);
			}
		}
*/
		break;
/*
	case CONV_DIR(reversUTF16,kBCUser_UTF8):
		tIn = new bc_u8[inlen];
		for (bc_u16* utf16In = (bc_u16*)in, *utf16Out = (bc_u16*)tIn; utf16In < (bc_u16*)(in+inlen); ++utf16In, ++utf16Out ){
			const bc_u16 character = *utf16In;
			*utf16Out =  ((character & 0xFF) << 8) | ((character >> 8) & 0xFF);
		}
		in = tIn;
	case CONV_DIR(nativeUTF16,kBCUser_UTF8):
		ConvertUTF16toUTF8((const UTF16**)&in, (const UTF16*)in+inlen, (UTF8**)&out, (UTF8*)out+outlen, strictConversion);
		resLen = out - outSaved;
		break;
*/
	default:
		break;
	}

//	if (tIn) delete [] tIn;

	return resLen;
}
}

bool BCDirIsRegularFile( bc_dirent *entry ){
	return false;
}

int BCFileClose(bc_file_handle *handle){
	bcand::container::IContainerSource* source = (bcand::container::IContainerSource*)*handle;
	source->seek(0);
	return 0;
}

bc_u64 BCFileSize(const char* path){
	bcand::container::IContainerSource* source = bcrypt_container::ContainerImpl::openedContainers[path];
	return source->sectorCount() * bcrypt_container::SECTOR_SIZE;
}

void BCFileSyncAll(){

}

void BCFileSync(bc_file_handle* handle){

}

bc_u64 BCFileSeek(bc_file_handle *handle, bc_u64 offset, bc_file_seek seek){
	bcand::container::IContainerSource* source = (bcand::container::IContainerSource*)*handle;
	__int64_t size = source->sectorCount();
	__int64_t sectoroffset = -1;
	switch(seek){
	case kBCFileSeek_Begin:
		sectoroffset = offset;
		break;
	case kBCFileSeek_Current:
		sectoroffset = source->position() + offset ;
		break;
	case kBCFileSeek_End:
		sectoroffset = size + offset ;
		break;
	}
	if (bcand::container::CONT_SRC_OK != source->seek(sectoroffset)){
		return -1;
	}
	return sectoroffset * bcrypt_container::SECTOR_SIZE;
}

void BCFileDelete(const char* path){
}

bc_dir_handle BCDirOpen( const char *dirName){
	return 0;
}

bool BCDirRead( bc_dirent *entry, bc_dir_handle dir ){
	return false;
}

void BCDirClose( bc_dir_handle dir ){
}

BCUserResult BCUserCreatePassword(const char* query, BCVerifyPasswordCallback callback, void* cookie, BCPassword* out_pwd){
	return kBCUserOk;
}

BCUserResult BCUserGetNewHiddenPartInfo(void *header, BCProtectExistingHiddenPartCallback protectCallback,
                                     BCVerifyPasswordCallback verifyCallback, void *verifyCookie,
                                     SpaceMap *hiddenMap, BCPassword *newPassword){
	return kBCUserOk;
}

const char *BCDBGetPath(){
	return 0;
}
