#ifndef BCTYPES_H
#define BCTYPES_H

#include "intsize.h"

typedef int (*BCProgressCallback) (const char *message, bc_u64 nbytes, bc_u64 total, void* arg);

typedef bc_u32 bc_alg_id;

/// Data processing mode.
typedef enum {
    kBCProcess_Decrypt = 0,     ///< Decrypt processing mode.
    kBCProcess_Encrypt = 1     ///< Encrypt processing mode.
} bc_proc_type;


/// Block cipher modes.
typedef enum {
	kBCMode_UDF = (bc_u32) 0,			// Undefined mode.	
    kBCMode_CBC = (bc_u32) 0xBC000002,	///< CBC (Cipher Block chaining) mode.
    kBCMode_LRW = (bc_u32) 0xBC000001,	///< LRW (Liskov, Rivest, Wagner) mode.
    kBCMode_XTS = (bc_u32) 0xBC000004,	///< XTS (XEX-based Tweaked Codebook) mode.
    kBCMode_ECB = (bc_u32) 0xBC000008,	///< ECB (Electronic Code Book) mode.
} bc_mode_id;

/// Misc constants. 
/// @warning Keep in sync!
enum {
    kBCCipher_BlockSize   = 512,	///< 
    kBCCipher_PoolSize    = 512,	///< Kernel pool size.
    kBCCipher_AlgNameSize = 16,		///< Length of algorythm name.
    kBCCipher_TotalModes  = 3,		///< Total cipher modes supported by kernel.
    kBCCipher_MaxKeys     = 32		///< Maximum count of keys that can be stored in kernel at one time.
};

/**
 * @brief Registered kernel block cipher definition.
 */
typedef struct BCAlgInfo
{
    int		id;                             ///< Cipher unique id
    char	name[kBCCipher_AlgNameSize];    ///< Cipher name
    size_t	key_length;                     ///< Cipher key length in bits
    size_t	block_size;                     ///< Cipher block size in bytes
    size_t	cbc_key_size;                   ///< CBC mode key size in bytes. 0 if mode is not supported,
    size_t	lrw_key_size;                   ///< LRW mode key size in bytes. 0 if mode is not supported.
    size_t	xts_key_size;                   ///< XTS mode key size in bytes. 0 if mode is not supported.
	bc_u8	reserved[8];					// Cipher specific data		
} BCAlgInfo;


/**
 * @brief Low level internal cipher key hande.
 *		Used by ecb and cbc encrypt/decrypt block function
 */
typedef void * cipher_key_handle;

/**
 * @brief High level external key handle.
 *		The key is extended for container sectors encrypt/decrypt functions in XTS,LRW and CBC modes.
 */
struct struct_bc_key
{
	cipher_key_handle	key_extended;
	cipher_key_handle	xts_context;
	bc_u8	*lrw_context;
	bc_u8	*pool;
	int		lock;
};

#if defined BC_LINUX || defined __APPLE__
typedef int bc_key_handle;
#else
typedef struct struct_bc_key *bc_key_handle;
#endif


#endif // BCTYPES_H
