/*
 *  Copyright 2012 Jetico, Inc. All rights reserved.
 *
 *	CihperBase usage sample:
 *
 *	int main( ... )
 *	{
 *		CipherBase_derived globalCipher();
 *		CipherFactory::m_gCipher = (CipherBase*)&globalCipher;
 *
 *		...
 *
 *		CipherBase *pCipher = CipherFactory::CreateCipher();
 *
 *		int rc = pCipher->init( ... );
 *		rc = pCipher->encrypt(...);
 *
 *		CipherFactory::releaseCipher( pCipher ); *
 *	}
 *
 */

#ifndef CIPHER_BASE_H
#define CIPHER_BASE_H

#include <assert.h>

#include "intsize.h"
#include "memory.h"
#include <exception>
#include "error.h"
#include "bctypes.h"
#include "sector2iv.h"

#include "encmode/bc_alg.h"
#include "../../fs-lib/log.h"

/**
 * @brief	Encryption key entry.
 */

#ifdef _WIN32
#define strcasecmp _stricmp
#endif


/**
 * Exceptions thrown when kernel cipher implementation misbehaves.
 */
class CipherException : public std::exception
{
public:
	
	CipherException(bc_cipher_result rc) throw() : m_rc(rc) {};
	
	virtual const char* what() const throw() {
		return ErrorString(m_rc);
	}

	bc_cipher_result code() const { return m_rc; }

	virtual ~CipherException() throw() {
	}
	
	/**
	 * Returns error string description.
	 * @param result BC kernel function result to process.
	 * @return       String representing short error description.
	 * */
	static inline const char* ErrorString(int result)
	{
		switch(result)
		{
			case kBCCipher_Success:
				return "Success.";
			case kBCCipher_NotInitialized:
				return "Cipher is not initialized.";
			case kBCCipher_NoMemory:
				return "Not enough memory.";
			case kBCCipher_UnknownAlg:
				return "Unknown alg.";
			case kBCCipher_UnsupportedMode:
				return "Unsupported mode.";
			case kBCCipher_InvalidKey:
				return "Invalid key specified.";
			case kBCCipher_InvalidBuffer:
				return "Invalid buffer.";
			case kBCCipher_InvalidArgument:
				return "Invalid argument";
			case kBCCipher_DeviceBusy:
				return "Device busy";
			case kBCCipher_InternalError:
				return "Internal error.";
			default:
				return "Unknown result.";
		}
	}
private:
	
	bc_cipher_result m_rc;
};


/**
 * Combines mode and algorithm and does some error checking before entering kernel.
 * It is safe to copy this object anyway you want.
 * Throws exceptions when communicating with the kernel service fails. If you prefer not to deal with exceptions, use raw c interface declared in BCKernel.h
 * See 'CihperBase usage sample:' above.
 */
class CipherBase
{
public:

	BCAlgInfo   m_alg;
    bc_mode_id  m_mode;

	CipherBase()
	{
		memset(&m_alg, 0, sizeof(m_alg));
		m_mode = kBCMode_UDF;
	}
	virtual ~CipherBase() {}
		
	/**
	 * Initialize cipher configuration.
	 * This is safe to call anytime to reset cipher configuration reusing the same object.
	 * @alg_id	Kernel crypto service algorithm id.
	 * @mode_id	Kernel crypto service mode id.
	 * @return	kBCCipher_UnsupportedMode if given cipher does not support given mode.
	 *		kBCCipher_UnknownAlg if algorithm is not registered in kernel service.
	 */

	int init(bc_alg_id alg_id, bc_mode_id mode_id) 
	{
		int rc;

		rc = lookupAlgId(alg_id, &m_alg);
		if(rc != kBCCipher_Success) {
			LIBLOG_ERROR("Error %d in lookulAlgID 0x%x", rc, (unsigned int)alg_id );
			return rc;
		}

		if(!IsModeSupported(&m_alg, mode_id)) {
			LIBLOG_ERROR("Unsupported mode 0x%x", (unsigned int)mode_id );
			return kBCCipher_UnsupportedMode;
		}

		m_mode = mode_id;
		return kBCCipher_Success;
	}


	int init(BCAlgInfo* alg, bc_mode_id mode_id)
	{
		if(!IsModeSupported(alg, mode_id)) {
			return kBCCipher_UnsupportedMode;
		}

		m_alg = *alg;
		m_mode = mode_id;
		return kBCCipher_Success;
	}
    
    int init(CipherBase * _other) {
        return init(&_other->m_alg, _other->m_mode);
    }

	/// Get cipher block size in bytes.
	size_t blockSize() const {
		return m_alg.block_size;
	}

	/// Get block algorithm key size in bits.
	size_t keyLength() const {
		return m_alg.key_length;
	}

	size_t keySize() const {
		return GetModeKeySize( &m_alg, m_mode );
	}

	size_t keySize( bc_alg_id alg, bc_mode_id mode ) const {

		BCAlgInfo alg_info;
		if ( kBCCipher_Success != lookupAlgId( alg, &alg_info ) ) {
			assert( false );
			return 0;
		}

		return GetModeKeySize( &alg_info, mode );
	}

	/// Get cipher algorithm id.
	bc_alg_id algId() const {
		return m_alg.id;
	}
	
	/// Get cipher mode id.
	bc_mode_id modeId() const {
		return m_mode;
	}

	/// Get cipher algorithm name.
	const char* algName() const {
		return m_alg.name;
	}
	
	/// Get cipher mode name.
	const char* modeName() const {
		switch (m_mode) {
		case kBCMode_CBC: return "CBC";
		case kBCMode_LRW: return "LRW";
		case kBCMode_XTS: return "XTS";
		default: assert(0); return NULL;
		}
	}

	/**
	 * Get mode human readable name for its id.
	 */
	static inline const char* GetModeName(bc_mode_id id)
	{
		switch (id) {
		case kBCMode_CBC: return "CBC";
		case kBCMode_LRW: return "LRW";
		case kBCMode_XTS: return "XTS";
		default: assert(0); break;
		}

		return NULL;
	}

	/**
	 * Gets key size for specified alg and mode.
	 * @param alg       Pointer to alg.
	 * @param mode      Encryption mode id.
	 * @return    Mode key size or zero, if mode is not supported.
	 * */
	static inline size_t GetModeKeySize(const BCAlgInfo *alg, bc_mode_id mode)
	{
		switch (mode)
		{
			case kBCMode_CBC:
				return alg->cbc_key_size;

			case kBCMode_LRW:
				return alg->lrw_key_size;

			case kBCMode_XTS:
				return alg->xts_key_size;

			default:
				return 0;
		}
	}

	/**
	 * Checks mode support by alg.
	 * @param alg       Algorythm to check.
	 * @param mode      Mode to check on algorythm.
	 * @return          true/false value
	 * */
	static inline bool IsModeSupported(const BCAlgInfo *alg, bc_mode_id mode)
	{
		if ( 254 == alg->id ) // Dummy - bcsaDummy
			return true;

		return (0 != GetModeKeySize(alg, mode));
	}

	/**
	 * Get mode id fro its human readable name.
	 */
	static inline bc_mode_id GetModeId(const char* name)
	{
		if (0 == strcasecmp(name, "CBC")) {
			return kBCMode_CBC;
		} else if (0 == strcasecmp(name, "LRW")) {
			return kBCMode_LRW;
		} else if (0 == strcasecmp(name, "XTS")) {
			return kBCMode_XTS;
		} else {
			assert(0);
			return kBCMode_UDF;
		}
	}

	///**
	// * Process data using key handle stored in kernel service memory.
	// * @iv		Initial vector.
	// * @type	Encrypt/decrypt.
	// * @key_handle	Handle to the encryption key in kernel memory.
	// * @pool	Random data pool to xor with data, leave NULL for zero-filled pool.
	// * @data	Data to process.
	// * @nbytes	Number of bytes of data. Should be a multiple of cipher block size.
	// * 
	// * @throws	CipherException on user/kernel communication problems and other things.
	// */
	//bc_cipher_result process( bc_u64 iv, bc_proc_type type, bc_key_handle key_handle, void* data, size_t nbytes )
	//{
	//	bc_cipher_result res;

	//	bc_u8 sync[ 8 ];
	//	iv2sync( sync, iv );

	//	if ( kBCProcess_Encrypt == type )
	//		res = bc_alg_encrypt( this, key_handle, sync, (bc_u8*)data, (bc_u8*)data, nbytes, m_mode );
	//	else
	//		res = bc_alg_decrypt( this, key_handle, sync, (bc_u8*)data, (bc_u8*)data, nbytes, m_mode );

	//	return res;
	//}

	/**
	 * Process data using explicit key data.
	 * Use this carefully when you need some data processed with a one-time key.
	 * @iv		Initial vector.
	 * @type	Encrypt/decrypt.
	 * @key		Key data. Will not be stored in kernel memory.
	 * @keysize	Number of bytes in key data.
	 * @pool	Random data pool to xor with data, leave NULL for zero-filled pool.
	 * @data	Data to process.
	 * @nbytes	Number of bytes of data. Should be a multiple of cipher block size.
	 */
	virtual void process( bc_u64 iv, bc_proc_type type, const bc_u8* key, size_t keysize, const bc_u8* pool, void* data, size_t nbytes) = 0;

    /**
     * Store encryption key in ciher storage. Use BCKernelStoreKey for kernel cipher
     * @param key           Key data to store.
     * @param keylen        Total bytes in key data.
     * @param pool          Pool is saved here with key handle according bc driver style. 
							So encrypt/decrypt with key handle functions are not require the pool again.
     * @param out_handle    Returned handle to stored key.
     */
	virtual bc_cipher_result storeKey(const bc_u8* key, size_t keylen, const bc_u8 *pool, bc_key_handle *out_handle) = 0;

   /**
     * Free encryption key stored by storeKey. Use BCKernelFreeKey for kernel cipher
     * @param handle	    Handle to previosly stored key.
     */
	virtual bc_cipher_result freeKey(bc_key_handle key_handle) = 0;

	/// Encrypt block of data using stored key handle.
	bc_cipher_result encrypt(bc_u64 iv, struct_bc_key* key_handle, void* data, size_t nbytes) {
		bc_u8 sync[ 8 ];
		iv2sync( sync, iv );

		return 	bc_alg_encrypt( this, key_handle, sync, (bc_u8*)data, (bc_u8*)data, nbytes, m_mode );
	//	return this->process(iv, kBCProcess_Encrypt, key_handle, data, nbytes);
	}

	/// Decrypt block of data using stored key handle.
	bc_cipher_result decrypt(bc_u64 iv, struct_bc_key* key_handle, void* data, size_t nbytes) {

		bc_u8 sync[ 8 ];
		iv2sync( sync, iv );

		return 	bc_alg_decrypt( this, key_handle, sync, (bc_u8*)data, (bc_u8*)data, nbytes, m_mode );
	//	return this->process(iv, kBCProcess_Decrypt, key_handle, data, nbytes);
	}
	
	/// Encrypt block of data using supplied key data.
	void encrypt(bc_u64 iv, const bc_u8* key, size_t key_size, const bc_u8* pool, void* data, size_t nbytes) {
		this->process(iv, kBCProcess_Encrypt, key, key_size, pool, data, nbytes);
	}

	/// Decrypt block of data using supplied key data.
	void decrypt(bc_u64 iv, const bc_u8* key, size_t key_size, const bc_u8* pool, void* data, size_t nbytes) {
		this->process(iv, kBCProcess_Decrypt, key, key_size, pool, data, nbytes);
	}

	static void encrypt_ecb( CipherBase *cipher, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, cipher_key_handle key_handle )
	{	cipher->encrypt_ecb( SrcBuffer, DstBuffer, key_handle  );	}

	static void decrypt_ecb( CipherBase *cipher, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, cipher_key_handle key_handle )
	{	cipher->decrypt_ecb( SrcBuffer, DstBuffer, key_handle );	}

	static void encrypt_cbc( CipherBase *cipher, const bc_u8 *IVector, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, size_t Length, cipher_key_handle key_handle )
	{	cipher->encrypt_cbc( IVector, SrcBuffer, DstBuffer, Length, key_handle );	}

	static void decrypt_cbc( CipherBase *cipher, const bc_u8 *IVector, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, size_t Length, cipher_key_handle key_handle )
	{	cipher->decrypt_cbc( IVector, SrcBuffer, DstBuffer, Length, key_handle );	}

	//
	// Abstract class virtual functions...
	//

	/**
     * Creates original cipher, returns pointer to base class.
	 */
 	virtual CipherBase *createCipher() = 0;

	/**
     * Lookup cipher info by id.
     * This function cannot guarantee that found plugin will still be availiable after it returns since kernel module can be unloaded.
     * @param id         Block cipher id.
     * @param out_alg    Initialized algorithm description on success.
     * @return     
     *                   - kBCCipher_UnknownAlg if algorithm is not found, 
     *                   - kBCCipher_InvalidBuffer, if out_alg pointer is NULL,
     *                   - kBCCipher_InternalError if other error occured. In the last case errno is set.
     */

	virtual bc_cipher_result lookupAlgId( bc_alg_id alg_id, BCAlgInfo *alg_info ) const = 0;

	/**
	 * Low level cipher key create/delete and enctypt/decrypt virtual functions
	 */

	virtual bc_cipher_result createCipherKey( const bc_u8 *key, cipher_key_handle *key_handle ) = 0;
	virtual bc_cipher_result deleteCipherKey( cipher_key_handle *key_handle ) = 0;

	virtual void encrypt_ecb( const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, cipher_key_handle key_handle ) = 0;
	virtual void decrypt_ecb( const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, cipher_key_handle key_handle ) = 0;
	virtual void encrypt_cbc( const bc_u8 *IVector, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, size_t Length, cipher_key_handle key_handle ) = 0;
	virtual void decrypt_cbc( const bc_u8 *IVector, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, size_t Length, cipher_key_handle key_handle ) = 0;
	
	/**
     * Enumerate available algorithms and writes it to allocated table.
     * @param pAlgsT        BCAlgInfo table.
     * @param pAlgsQ        Number of items.
     */
	virtual bc_cipher_result getAlgList( BCAlgInfo **pAlgsT, size_t *pAlgsQ ) = 0;
	virtual void freeAlgList( BCAlgInfo *algsT, size_t algsQ ) = 0;

};

//
//	CipherBase static functions correspond this define
//		encrypt_cbc, decrypt_cbc
//
typedef void type_cipher_process_cbc( CipherBase *cipher, const bc_u8 *IVector, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, size_t Length, cipher_key_handle key_handle );
//
//	CipherBase static functions correspond this define
//		encrypt_ecb, decrypt_ecb,
//
typedef void type_cipher_process_ecb( CipherBase *cipher, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, cipher_key_handle key_handle );

/**
 * 1. Defines global cipher derived from CipherBase.
 * 2. Allows to creates cipher objects same as global cipher.
 * See 'CihperBase usage sample:' above.
 */
class CipherFactory
{
public:

	static CipherBase *m_gCipher;

	static CipherBase *CreateCipher()
	{
		if ( m_gCipher == NULL )
			throw CipherException( kBCCipher_NotInitialized );
		
		CipherBase *cipher = m_gCipher->createCipher();

		if ( !cipher )
			throw CipherException( kBCCipher_NoMemory );

		return cipher;
	}

	static void ReleaseCipher( CipherBase **p_cipher )
	{
		if ( ( NULL != p_cipher ) && ( NULL != *p_cipher ) )
		{
			delete *p_cipher;
			*p_cipher = NULL;
		}
	}

	static bc_cipher_result ProcessWithKey( 
				bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type proc,
                const bc_u8* key, size_t key_size,
                bc_u64 iv, const bc_u8* pool,
                bc_u8* data, size_t nbytes )
	{
		BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

		if ( kBCCipher_Success != sCipher->init( alg_id, mode_id ) )
			return kBCCipher_UnknownAlg;

		sCipher->process( iv, proc, key, key_size, pool, data, nbytes );

		return kBCCipher_Success;
	}

	static bc_cipher_result Process(
						bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type type, 
						struct_bc_key* key_handle, bc_u64 iv, // const bc_u8* pool,
						bc_u8* data, size_t nbytes)
	{
		BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

		if ( kBCCipher_Success != sCipher->init( alg_id, mode_id ) )
			return kBCCipher_UnknownAlg;

		bc_cipher_result res;

		bc_u8 sync[ 8 ];
		iv2sync( sync, iv );

		if ( kBCProcess_Encrypt == type )
			res = bc_alg_encrypt( sCipher.get(), key_handle, sync, (bc_u8*)data, (bc_u8*)data, nbytes, sCipher->m_mode );
		else
			res = bc_alg_decrypt( sCipher.get(), key_handle, sync, (bc_u8*)data, (bc_u8*)data, nbytes, sCipher->m_mode );

		return kBCCipher_Success;
	}

	static void FreeKey(bc_key_handle key_handle )
	{
		BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

		sCipher->freeKey( key_handle );
	}

	static bc_cipher_result GetAlgList( BCAlgInfo **pAlgsT, size_t *pAlgsQ )
	{
		BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );
		return sCipher->getAlgList( pAlgsT, pAlgsQ );
	}

	static void FreeAlgList( BCAlgInfo *algsT, size_t algsQ )
	{
		BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );
		sCipher->freeAlgList( algsT, algsQ );
	}

};

/**
 * Enumeration of all supported algorithms.
 */
struct BCEnumAlgorithms {
	BCAlgInfo* algs;
	size_t nalgs;
	
	BCEnumAlgorithms( int defId = -1 )
	{
		bc_cipher_result rc = CipherFactory::GetAlgList(&this->algs, &this->nalgs);

		if(rc != kBCCipher_Success) {
			throw CipherException(rc);
		}

		if ( defId != -1 )
			setFirst( this, defId );
	}

	~BCEnumAlgorithms()
	{
		CipherFactory::FreeAlgList( this->algs, this->nalgs );
	}
	
	bool setFirst( BCEnumAlgorithms *tab, int id )
	{
		for( unsigned i = 1; i < tab->total(); i++ )
		{
			if ( tab->get(i)->id == id )
			{
				BCAlgInfo tmp = tab->algs[i];
				tab->algs[i] = tab->algs[0];
				tab->algs[0] = tmp;
				return true;
			}
		}

		return false;
	}

	BCAlgInfo* get(unsigned idx) {
		if(idx >= nalgs) {
			return NULL;
		}
		
		return algs + idx;
	}

	unsigned total() const {
		return nalgs;
	}
};

/**
 * Enumeration of all supported modes.
 */
struct BCEnumModes {
	enum {
		kBCTotalModes = 3,
	};
	
	bc_mode_id modes[kBCTotalModes];
	
	BCEnumModes()
	{
		modes[0] = kBCMode_XTS;
		modes[1] = kBCMode_CBC;
		modes[2] = kBCMode_LRW;
	}
	
	bc_mode_id get(unsigned idx) {
		assert(idx < kBCTotalModes);
		return modes[idx];
	}
	
	unsigned total() const {
		return kBCTotalModes;
	}
};

#endif // CIPHER_BASE_H
