/*
 *  Copyright 2012 Jetico, Inc. All rights reserved.
 */

#include "ciphercrtlib.h"
#include <string.h>
#include "../certificate/CertificateLib/symmetricAlgs/symmetricAlgs.h"
#include "../certificate/CertificateLib/PGPPkts/pktdefs.h"
#ifndef _WIN32
#include "../certificate/CertificateLib/unix/misc.h"
#endif

#ifndef maxLen
#define maxLen( buff ) ( sizeof( buff ) / sizeof( *buff ) )
#endif

//
//	Algorithm identifiers encoding table.
//	crtLibId	- identifier used by certificate library.
//	bestCryptId	- identifier used as bestcrypt container algorithm id.
//
struct algIds
{
	int crtLibId;
	bc_alg_id bestCryptId;
};

algIds algIdT[] = 
{
	bcsaBFish		, 128, // Blowfish - 256
	pgpsaBlowfish	, 192, // Blowfish - 128
	bcsaBFish448	, 176, // Blowfish - 448
	pgpsaCAST5		, 224,
	bcsaDES			, 64,
	pgpsaTripleDES	, 208,
	pgpsaPlaintext	, 254,
	bcsaGOST		, 170,
	pgpsaIDEA		, 32,
//	bcsaRC6			, 230,
	bcsaRIJN		, 240,
	bcsaSERP		, 232,
	bcsaTFISH		, 160
};

static bc_cipher_result getAlgInfo( BCAlgInfo *algInfo, int crtLibAlgId, bc_alg_id bcAlgId )
{
    if (NULL == algInfo)
        return kBCCipher_InvalidBuffer;
	
	memset( algInfo, 0, sizeof( *algInfo ) );

	if ( !sa_isValidID( crtLibAlgId ) )
		return kBCCipher_UnknownAlg;

	const char *algName = sa_getSymmetricAlgName( (BYTE)crtLibAlgId );

	if ( NULL == algName )
		return kBCCipher_UnknownAlg;
	
	algInfo->id = bcAlgId;
	strncpy_s( algInfo->name, algName, _TRUNCATE );
	algInfo->key_length = sa_getKeySize( crtLibAlgId ) * 8;
	algInfo->block_size = sa_getBlockSize( crtLibAlgId );
	algInfo->cbc_key_size = algInfo->key_length / 8;

	//
	//	All bestcrypt algorithms has KeyLength_LRW value in registry 512
	//
	if ( 0 != algInfo->cbc_key_size )
	{
		algInfo->lrw_key_size = 512 / 8;
		algInfo->xts_key_size = 512 / 8;
	}
	
	*(int*)algInfo->reserved = crtLibAlgId;

	return kBCCipher_Success;
}

bc_cipher_result CipherCrtLib::lookupAlgId( bc_alg_id alg_id, BCAlgInfo *algInfo ) const
{
    if (NULL == algInfo)
        return kBCCipher_InvalidBuffer;

	for ( int i = maxLen( algIdT ); i--; )
	{
		if ( algIdT[i].bestCryptId == alg_id )
			return getAlgInfo( algInfo, algIdT[i].crtLibId, algIdT[i].bestCryptId );
	}

	return kBCCipher_UnknownAlg;
}

bc_cipher_result CipherCrtLib::createCipherKey( const bc_u8 *key, cipher_key_handle *out_handle )
{
	if ( m_alg.id == 0 ) // was not inited...
	{	assert( false ); return kBCCipher_NotInitialized; }

	HSA sa = NULL;

	int crtLibId = *(int*)m_alg.reserved;

	int res = sa_createAlgorithm( &sa, crtLibId, (BYTE*)key, m_alg.key_length / 8 );

	if ( sae_ok != res )
	{	assert( false ); return kBCCipher_InvalidKey; }

	//
	// CAST CBC uses DWORD arrays as sourse and desttination and always change to big endian byte array to be processed by ECB functions.
	//
	if ( ( m_mode == kBCMode_CBC ) && ( m_alg.id == 0xe0 ) )
		sa_setBCDriverCompatible( sa, cSymAlg::eCBCCompatible );

	//
	// GOST CBC uses CFB mode instead.
	//
	if ( ( m_mode == kBCMode_CBC ) && ( m_alg.id == 0xaa ) )
		sa_setBCDriverCompatible( sa, cSymAlg::eCFBCompatible );

	*out_handle = (cipher_key_handle)(HSA)sa;

	return kBCCipher_Success;
}

bc_cipher_result CipherCrtLib::deleteCipherKey( cipher_key_handle *key_handle )
{
	HSA sa = (HSA)*key_handle;

	if ( 0 == sa_getBlockSize( sa ) ) // Just check algorithm handle.
	{ assert( false ); return kBCCipher_InternalError; }

	sa_free( sa );
	
	*key_handle = NULL;

	return kBCCipher_Success;
}


void CipherCrtLib::process( bc_u64 iv, bc_proc_type type, const bc_u8* key, size_t keysize, const bc_u8* pool, void* data, size_t nbytes)
{
    if(pool == NULL) {
        static bc_u8 s_zero_pool[kBCCipher_PoolSize] = { 0 };
        pool = s_zero_pool;
    }

    if(keysize < this->keySize()) {
        throw CipherException(kBCCipher_InvalidKey);
    }

    if ( this->blockSize() == 0 ) {
        if ( this->keySize() == 0 ) // Plain text encryption algorithm.
            return; // Does nothing..
        else
            throw CipherException(kBCCipher_UnknownAlg);
    }

    if(nbytes % this->blockSize()) {
        throw CipherException(kBCCipher_InvalidBuffer);
    }

    bc_key_handle key_handle;

    if ( kBCCipher_Success != storeKey( key, keysize, pool, &key_handle ) ) {
        throw CipherException(kBCCipher_InvalidKey);
    }

    bc_cipher_result res;

    bc_u8 sync[ 8 ];
    iv2sync( sync, iv );

    //bc_cipher_result bc_alg_encrypt ( CipherBase *cipher, struct_bc_key *bc_key, const bc_u8 *iv, const bc_u8 *src, bc_u8 *dst, size_t len, bc_mode_id mode );

    if ( kBCProcess_Encrypt == type )
        res = bc_alg_encrypt( this, (struct_bc_key *)key_handle, sync, (bc_u8*)data, (bc_u8*)data, nbytes, m_mode );
    else
        res = bc_alg_decrypt( this, (struct_bc_key *)key_handle, sync, (bc_u8*)data, (bc_u8*)data, nbytes, m_mode );

    freeKey( key_handle );

    if(kBCCipher_Success != res) {
        throw CipherException(res);
    }
}


void CipherCrtLib::encrypt_ecb( const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, cipher_key_handle key_handle )
{
	HSA sa = (HSA)key_handle;

	if ( sae_ok != sa_encrypt( sa, DstBuffer, SrcBuffer, m_alg.block_size ) )
		throw CipherException( kBCCipher_NotInitialized );
}

void CipherCrtLib::decrypt_ecb( const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, cipher_key_handle key_handle )
{
	HSA sa = (HSA)key_handle;

	if ( sae_ok != sa_decrypt( sa, DstBuffer, SrcBuffer, m_alg.block_size ) )
		throw CipherException( kBCCipher_NotInitialized );
}

void CipherCrtLib::encrypt_cbc( const bc_u8 *IVector, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, size_t Length, cipher_key_handle key_handle )
{
	HSA sa = (HSA)key_handle;

//#ifdef _DEBUG
//	bc_u8 *ptr = (bc_u8 *)SMalloc( Length * 2 );
//
//	if ( ptr == NULL )
//		return;
//
//	memcpy( ptr + Length, SrcBuffer, Length );
//	SrcBuffer = ptr + Length; 
//
//	if ( sae_ok != sa_encryptCBC( sa, IVector, DstBuffer, SrcBuffer, Length ) )
//		throw CipherException( kBCCipher_NotInitialized );
//
//	int res = sa_decryptCBC( sa, IVector, ptr, DstBuffer, Length );
//
//	res = memcmp( SrcBuffer, ptr, Length );
//
//	if ( res != 0 )
//		throw CipherException( kBCCipher_NotInitialized );
//
//	SFree( ptr );
//
//	return;
//#endif

	if ( sae_ok != sa_encryptCBC( sa, IVector, DstBuffer, SrcBuffer, Length ) )
		throw CipherException( kBCCipher_NotInitialized );
}

void CipherCrtLib::decrypt_cbc( const bc_u8 *IVector, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, size_t Length, cipher_key_handle key_handle )
{
	HSA sa = (HSA)key_handle;

//#ifdef _DEBUG
//	bc_u8 *ptr = (bc_u8 *)SMalloc( Length * 2 );
//
//	if ( ptr == NULL )
//		return;
//
//	static bool bChange = false;
//	if ( bChange )
//		sa_setBCDriverCompatible( sa, false );
//
//	memcpy( ptr + Length, SrcBuffer, Length );
//	SrcBuffer = ptr + Length; 
//
//	if ( sae_ok != sa_decryptCBC( sa, IVector, DstBuffer, SrcBuffer, Length ) )
//		throw CipherException( kBCCipher_NotInitialized );
//
//	int res = sa_encryptCBC( sa, IVector, ptr, DstBuffer, Length );
//
//	res = memcmp( SrcBuffer, ptr, Length );
//	if ( res != 0 )
//		throw CipherException( kBCCipher_NotInitialized );
//
//	SFree( ptr );
//
//	return;
//#endif

//#ifdef _DEBUG
//	static bool useCFB = false;
//	if ( useCFB )
//	{
//		sa_initCFB( sa, IVector );
//		if ( sae_ok != sa_decryptCFB( sa, DstBuffer, SrcBuffer, Length ) )
//			throw CipherException( kBCCipher_NotInitialized );
//
//		return;
//	}
//#endif

	if ( sae_ok != sa_decryptCBC( sa, IVector, DstBuffer, SrcBuffer, Length ) )
		throw CipherException( kBCCipher_NotInitialized );
}

bc_cipher_result CipherCrtLib::getAlgList( BCAlgInfo **pAlgsT, size_t *pAlgsQ )
{
	*pAlgsT = NULL;
	*pAlgsQ = maxLen( algIdT );

	if ( 0 == *pAlgsQ )
		return kBCCipher_NotInitialized;

	*pAlgsT = new BCAlgInfo[*pAlgsQ];

	if ( *pAlgsT == NULL )
	{
		*pAlgsQ = 0;
		return kBCCipher_NoMemory;
	}

	for( size_t i = *pAlgsQ; i--; )
	{
		bc_cipher_result res = getAlgInfo( *pAlgsT + i, algIdT[i].crtLibId, algIdT[i].bestCryptId );
		
		if ( kBCCipher_Success != res )
		{
			freeAlgList( *pAlgsT, *pAlgsQ );
			*pAlgsT = NULL;
			*pAlgsQ = 0;

			return res;
		}
	}

	return kBCCipher_Success;
}

void CipherCrtLib::freeAlgList( BCAlgInfo *algsT, size_t algsQ )
{
	memset( algsT, 0, algsQ * sizeof( BCAlgInfo ) );

	delete [] algsT;
}

//virtual
CipherBase *CipherCrtLib::createCipher()
{
	return new CipherCrtLib;
}
