/*
 *  Copyright 2012 Jetico, Inc. All rights reserved.
 */

#ifndef CIPHER_CRT_LIB_H
#define CIPHER_CRT_LIB_H

#include "cipherbase.h"
#include <exception>
//#include "../certificate/CertificateLib/publicKeyAlgs/cPKA.h"
#include "../certificate/CertificateLib/symmetricAlgs/symmetricAlgs.h"

/**
 * Wrapper over certificate lib crypto service.
 * Combines mode and algorithm and does some error checking before encrypting.
 * It is safe to copy this object anyway you want.
 * Throws exceptions when communicating with the kernel service fails. If you prefer not to deal with exceptions, use raw c interface declared in BCKernel.h
 */
class CipherCrtLib : public CipherBase
{
public:
	
	CipherCrtLib()
	{
		crtLib_bWindowsDriverCompatible = true;
	}

	/**
     * Creates original cipher, returns pointer to base class.
	 */
	virtual CipherBase *createCipher();

	/**
     * Lookup cipher info by id.
     * This function cannot guarantee that found plugin will still be availiable after it returns since kernel module can be unloaded.
     * @param id         Block cipher id.
     * @param out_alg    Initialized algorithm description on success.
     * @return     
     *                   - kBCCipher_UnknownAlg if algorithm is not found, 
     *                   - kBCCipher_InvalidBuffer, if out_alg pointer is NULL,
     *                   - kBCCipher_InternalError if other error occured. In the last case errno is set.
     */

	virtual bc_cipher_result lookupAlgId( bc_alg_id alg_id, BCAlgInfo *alg_info ) const;

	/**
	 * Low level cipher key create/delete and enctypt/decrypt virtual functions
	 */

	virtual bc_cipher_result createCipherKey( const bc_u8 *key, cipher_key_handle *key_handle );
	virtual bc_cipher_result deleteCipherKey( cipher_key_handle *key_handle );

	virtual void process( bc_u64 iv, bc_proc_type type, const bc_u8* key, size_t keysize, const bc_u8* pool, void* data, size_t nbytes);

	virtual bc_cipher_result storeKey(const bc_u8* key, size_t keylen, const bc_u8 *pool, bc_key_handle* out_handle)
	{
		return bc_alg_make_key( this, key, keylen, pool, (struct_bc_key **)out_handle );
	}

	virtual bc_cipher_result freeKey( bc_key_handle key_handle)
	{
		return bc_alg_free_key( this, (struct_bc_key*)key_handle );
	}

	virtual void encrypt_ecb( const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, cipher_key_handle key_handle );
	virtual void decrypt_ecb( const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, cipher_key_handle key_handle );
	virtual void encrypt_cbc( const bc_u8 *IVector, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, size_t Length, cipher_key_handle key_handle );
	virtual void decrypt_cbc( const bc_u8 *IVector, const bc_u8 *SrcBuffer, bc_u8 *DstBuffer, size_t Length, cipher_key_handle key_handle );
	
	/**
     * Enumerate available algorithms and writes it to allocated table.
     * @param pAlgsT        BCAlgInfo table.
     * @param pAlgsQ        Number of items.
     */
	virtual bc_cipher_result getAlgList( BCAlgInfo **pAlgsT, size_t *pAlgsQ );
	virtual void freeAlgList( BCAlgInfo *algsT, size_t algsQ );

};

#endif  // CIPHER_CRT_LIB_H
