/*
 * Copyright 2012 Jetico, Inc. All rights reserved.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <map>

#include "container.h"
#include "user.h"
#include "file.h"
#include "password.h"
#include "cipherbase.h"
#include "../certificate/CertificateLib/SMemory/SBlocks.h"
#include "../keygens/keyFiles.h"
#include "../certificate/CertificateLib/x509pkts/pkcs12_funcs.h"

#ifdef _WIN32
#define snprintf _snprintf
//#else
//#define PATH_MAX 255
#endif

cPwdGet *cPwdGet::m_current = NULL;

//////////////////////////////////////////////////////////////////////////
//
// Known keygens 
//
//////////////////////////////////////////////////////////////////////////



extern struct BCKeygen kg_sha;
extern struct BCKeygen kg_ghost;

static const struct BCKeygen* gKeygens[] = {
    &kg_sha,
    &kg_ghost,
};

#define BC_TOTAL_KEYGENS (sizeof(gKeygens) / sizeof(*gKeygens))

bool BCHelperContext::pwdIsEmpty()
{
    if ( pwdGet == NULL )
        return true;
    
    if ( strlen( pwdGet->getPasswordString() ) == 0 )
        return true;
    
    return false;
}

BCContainerResult BCLoadKeygen(BCHelperContext & _helperCtx)
{
    IBCHeader * header = _helperCtx.header;
    if (NULL == header) {
        for (unsigned i = 0; i < BC_TOTAL_KEYGENS; ++i) {
            header = gKeygens[i]->kgload( _helperCtx.channel, _helperCtx.pwdGet != NULL ? _helperCtx.pwdGet->getKeyBlockChannel() : NULL);
            if (!header) {
                continue;
            }
            header->keygen = gKeygens[i];
            _helperCtx.header = header;
            return kBCContainer_Success;
        }
    } else {
        return kBCContainer_Success;
    }
    
    return kBCContainer_UnknownFormat;
}

BCContainerResult BCVerifyPassword(BCHelperContext & _helperCtx)
{
    if ( _helperCtx.pwdGet == NULL )
        return kBCContainer_InternalError;

    BCContainerResult cr = BCLoadKeygen(_helperCtx);
    if (cr != kBCContainer_Success) {
        return cr;
    }
    
    IBCHeader * header = _helperCtx.header;
    
	if ( ( _helperCtx.progressActor != NULL ) && ( !_helperCtx.pwdIsEmpty() ) )
		_helperCtx.progressActor->func("Verifying password", 0, 0); // Call func for the first time to signal that password verification is started

	if (kBCContainer_Success == header->verifyPassword( _helperCtx.pwdGet->getPasswordObject(), &_helperCtx.volumeInfo, _helperCtx.progressActor, &_helperCtx.ptbKey )) {
        _helperCtx.pwdGet->setIsMainVolumePassword(_helperCtx.volumeInfo.hidden == 0);
        LIBLOG_INFO( "Success! %s", header->keygen->name );
        if ( _helperCtx.progressActor != NULL )
            _helperCtx.progressActor->func("Password accepted", 1, 0); // Call func for the last time to signal that password verification is finished
        return kBCContainer_Success;
    }
    
    if ( _helperCtx.progressActor != NULL )
        _helperCtx.progressActor->func("Invalid password", 1, 0); // Call func for the last time to signal that password verification is finished
    
    LIBLOG_INFO( "Warning! %s", "Incorrect password" );
    return kBCContainer_InvalidPassword;
}

BCContainerResult BCGetAndVerifyPassword(BCHelperContext & _helperCtx, bool _forMainVolume)
{
	if ( _helperCtx.pwdGet == NULL )
        return kBCContainer_InvalidArgument;

    BCContainerResult result = kBCContainer_Success;
    bool lastFailed = false;
    do {
        if (result == kBCContainer_InvalidPassword || _helperCtx.pwdIsEmpty()) {
            result = _helperCtx.pwdGet->getPassword( lastFailed );
            if (result != kBCContainer_Success) {
                break;
            }
        }
        
        result = BCVerifyPassword(_helperCtx);
        if ((_helperCtx.progressActor != NULL && _helperCtx.progressActor->m_bCancelled)) {
            result = kBCContainer_Cancelled;
            break;
        }
        if (_forMainVolume && !_helperCtx.pwdGet->isMainVolumePassword()) {
            result = kBCContainer_InvalidPassword;
        }
        lastFailed = true;
    } while (result != kBCContainer_Success && result != kBCContainer_UnknownFormat);
    
    return result;
}

//
//	BCGetAndVerifyNewPassword
//		Returns kBCContainer_Success if password does not exist.
//		Otherwise return error code.
//
BCContainerResult BCGetAndVerifyNewPassword(BCHelperContext & _helperCtx)
{
    if ( _helperCtx.pwdGet == NULL )
        return kBCContainer_InternalError;
    
    BCContainerResult result = kBCContainer_Success;
    bool lastFailed = false;

    while( result == kBCContainer_Success )
	{
        result = _helperCtx.pwdGet->getPassword( lastFailed );
        if (result != kBCContainer_Success) {
            break;
        }
        
        result = BCVerifyPassword(_helperCtx);
        if ((_helperCtx.progressActor != NULL && _helperCtx.progressActor->m_bCancelled)) {
            result = kBCContainer_Cancelled;
            break;
        }

        if (result == kBCContainer_InvalidPassword)
            return kBCContainer_Success;

        lastFailed = true;
    }
    
    return result;
}


static const struct BCKeygen* getKeyGenByID( bc_u16 kgID )
{
    for (unsigned i = 0; i < BC_TOTAL_KEYGENS; ++i) {
		if ( gKeygens[i]->id == kgID ) {
            return gKeygens[i];
        }
    }

	return &kg_ghost;
}

//////////////////////////////////////////////////////////////////////////
//
// Interface
//
//////////////////////////////////////////////////////////////////////////

BCContainerResult BCCreateContainer(BCHelperContext & _helperCtx)
{
	LIBLOG_INFO( "Begin" );

    if (_helperCtx.pwdGet == NULL) {
        return kBCContainer_InternalError;
    }
    
    BCContainerResult result = _helperCtx.pwdGet->getPassword(false);
    if (result != kBCContainer_Success) {
        return result;
    }

	BCContainerContext &containerCtx = _helperCtx.pwdGet->getContainerContext();

	const struct BCKeygen* curKeygen = getKeyGenByID( containerCtx.kgID );

    BCHeaderHandle header(curKeygen->kgcreate());
    if (!header) {
    	LIBLOG_ERROR( "Header is NULL!" );
        return kBCContainer_InternalError;
    }
    header->keygen = curKeygen;
    
    
	result = header->initEx(
			containerCtx.algID, containerCtx.modeID, containerCtx.size, containerCtx.description, _helperCtx.pwdGet->getPasswordObject(),
			containerCtx.kgID, containerCtx.hashAlgID, containerCtx.iterations, containerCtx.salt, containerCtx.saltSize );

    if (kBCContainer_Success != result) {
    	LIBLOG_ERROR( "Error %d in header initialization", result );
        return result;
    }
     
    bc_key_handle key_handle;
    CipherBase * cipher = NULL;

    result = header->decodeKey( _helperCtx.pwdGet->getPasswordObject(), &_helperCtx.volumeInfo, &key_handle, &cipher, &(_helperCtx.ptbKey) );
    if (kBCContainer_Success != result) {
    	LIBLOG_ERROR( "Error %d in decodeKey", result );
        return result;
    }

	result = BCInitVolume( &_helperCtx.volumeInfo, _helperCtx.channel, kBCFileMode_ReadWrite, key_handle, cipher, _helperCtx.outVolume);

    if (kBCContainer_Success != result) {
		CipherFactory::ReleaseCipher( &cipher );
		CipherFactory::FreeKey( key_handle );
    	LIBLOG_ERROR( "Error %d in volume initialization", result );
        return result;
    }
    
    _helperCtx.header = header.delegatePtr();

	LIBLOG_INFO( "End Success" );
    return kBCContainer_Success;
}

eBenchmarkError BCMountContainerBenchmark( bc_u32 hashAlgID, iChannel ** keyFilesT, bc_u32 keyFilesQ, double &kfTime, int iterations, double &iterationsTime  )
{
	cTimer timer;

	cPTB ptb;
	if ( keyFilesQ == 0 )
		ptb.Allocate( 64 );
	else
	if ( !generateSaltBuffer( ptb, keyFilesT, keyFilesQ, hashAlgID, NULL, 0 ) )
		return eBenchmarkError_readKeyFiles;

	kfTime = timer.getTime();

    if (iterations == 0) {
        iterationsTime = 0.0;
        return eBenchmarkError_OK;
    }
    
    timer.Start();

	bc_u8 bf[64];
	if ( x509e_ok !=
			P12_producePseudoRandomBytes(
				bf, sizeof( bf ),
				ptb.buffer, ptb.bufferSize,
				iterations,
				ptb.buffer, ptb.bufferSize,
				P12_prgm4key,
				(bc_u8)hashAlgID
			)
		)
			return eBenchmarkError_keyGeneration;

	iterationsTime = timer.getTime();

	if ( iterationsTime == 0 )
		return eBenchmarkError_iterationCountTooSmall;

	return eBenchmarkError_OK;
}

//bool BCParseContainerFile(const char* file, BCVolumeInfo* out_info)
bool BCParseContainerFile( iChannel *channel, BCVolumeInfo* out_info)
{
    for (unsigned i = 0; i < BC_TOTAL_KEYGENS; ++i) {
        if ( kBCContainer_Success == gKeygens[i]->kgparse( channel, out_info ) ) {
            return true;
        }
    }

    return false;
}

BCContainerResult BCDecodeVolume(BCHelperContext & _helperCtx)
{
	LIBLOG_INFO( "Begin %p", _helperCtx.progressActor );

    _helperCtx.outVolume->reset();
    
    BCContainerResult result = BCGetAndVerifyPassword(_helperCtx);
    if (result != kBCContainer_Success) {
    	LIBLOG_INFO( "Warning! Invalid password!" );
        return result;
    }

    bc_key_handle keyHandle;
	CipherBase * cipher = NULL;

	if ( _helperCtx.ptbKey.bufferSize != 0 )
		result = _helperCtx.header->generateCipherInfo( _helperCtx.ptbKey, &_helperCtx.volumeInfo, &keyHandle, &cipher );
	else
		result = _helperCtx.header->decodeKey(_helperCtx.pwdGet->getPasswordObject(), &_helperCtx.volumeInfo, &keyHandle, &cipher);
	if (result != kBCContainer_Success) {
		LIBLOG_INFO( "Assertion! Error %d in decoding key", result );
		return result;
	}

    result = BCInitVolume(&_helperCtx.volumeInfo, _helperCtx.channel, _helperCtx.mode, keyHandle, cipher, /*pool,*/ _helperCtx.outVolume);
    if (result != kBCContainer_Success) {
    	LIBLOG_INFO( "Assertion! Error %d in BCInitVolume", result );
		CipherFactory::ReleaseCipher( &cipher );
		CipherFactory::FreeKey( keyHandle );
        return result;
    }

	LIBLOG_INFO( "End" );
    return kBCContainer_Success;
}

///////////////////////////////////////
//        Password operations
//////////////////////////////////////

BCContainerResult BCPrepareContainerForOperation( BCHelperContext & _mainVolumeHelperCtx, cPwdGet * _additionalPwdGet)
{
    if (_mainVolumeHelperCtx.pwdGet == NULL) {
        return kBCContainer_InvalidArgument;
    }
    
    BCContainerResult result = BCGetAndVerifyPassword(_mainVolumeHelperCtx, true);
    if (result != kBCContainer_Success) {
        return result;
    }

    BCContainerContext & containerContext = _mainVolumeHelperCtx.pwdGet->getContainerContext();
    if (containerContext.askMore) {
        if (_additionalPwdGet != NULL) {
            BCHelperContext additionalHelperContext = _mainVolumeHelperCtx;
            additionalHelperContext.pwdGet = _additionalPwdGet;
            result = BCGetAndVerifyPassword(additionalHelperContext);
            if (result != kBCContainer_Success) {
                return result;
            }
        } else {
            return kBCContainer_InvalidArgument;
        }
    }
    
    return kBCContainer_Success;
}

BCContainerResult BCAddPassword( BCHelperContext & _mainVolumeHelperCtx, cPwdGet * _volumePwdGet, cPwdGet * _newPwdGet)
{
    if (_mainVolumeHelperCtx.pwdGet == NULL || _mainVolumeHelperCtx.pwdGet->isEmpty() || _newPwdGet == NULL) {
        return kBCContainer_InvalidArgument;
    }
    
    if (_volumePwdGet == NULL || _volumePwdGet->isEmpty()) {
        _volumePwdGet = _mainVolumeHelperCtx.pwdGet;
    }

    BCHelperContext newHelperContext = _mainVolumeHelperCtx;
    newHelperContext.pwdGet = _newPwdGet;
    
    BCContainerResult result = BCGetAndVerifyNewPassword(newHelperContext);
    if (result != kBCContainer_Success) {
        return result;
    }
    
    result = _mainVolumeHelperCtx.header->addPassword(_mainVolumeHelperCtx.pwdGet->getPasswordObject(), _volumePwdGet->getPasswordObject(), _newPwdGet->getPasswordObject());
    
    if (result == kBCContainer_Success) {
        result = _mainVolumeHelperCtx.storeHeader();
    }
    
    return result;
}

BCContainerResult BCChangePassword(BCHelperContext & _mainVolumeHelperCtx, cPwdGet * _volumePwdGet, cPwdGet * _newPwdGet)
{
    if (_mainVolumeHelperCtx.pwdGet == NULL || _mainVolumeHelperCtx.pwdGet->isEmpty() || _newPwdGet == NULL) {
        return kBCContainer_InvalidArgument;
    }
    
    if (_volumePwdGet == NULL || _volumePwdGet->isEmpty()) {
        _volumePwdGet = _mainVolumeHelperCtx.pwdGet;
    }
    
    BCHelperContext newHelperContext = _mainVolumeHelperCtx;
    newHelperContext.pwdGet = _newPwdGet;
    
    BCContainerResult result = BCGetAndVerifyNewPassword(newHelperContext);
    if (result != kBCContainer_Success) {
        return result;
    }
    
    result = _mainVolumeHelperCtx.header->changePassword(_mainVolumeHelperCtx.pwdGet->getPasswordObject(), _volumePwdGet->getPasswordObject(), _newPwdGet->getPasswordObject());
    
    if (result == kBCContainer_Success) {
        result = _mainVolumeHelperCtx.storeHeader();
    }
    
    return result;
}


BCContainerResult BCRemovePassword(BCHelperContext & _mainVolumeHelperCtx, cPwdGet * _oldPwdGet, bool _force)
{
    if (_mainVolumeHelperCtx.pwdGet == NULL || _oldPwdGet == NULL) {
        return kBCContainer_InvalidArgument;
    }
    
    if (_oldPwdGet->isEmpty()) {
        BCHelperContext helperContext = _mainVolumeHelperCtx;
        helperContext.pwdGet = _oldPwdGet;
        BCContainerResult result = BCGetAndVerifyPassword(helperContext);
        if (result != kBCContainer_Success) {
            return result;
        }
    }
    
    if (_oldPwdGet->isMainVolumePassword() && _mainVolumeHelperCtx.header->getMainKeyblocksCount() == 1 && !_force) {
        return kBCContainer_LastKeyBlock;
    }
    
    BCContainerResult result = _mainVolumeHelperCtx.header->removePassword(_mainVolumeHelperCtx.pwdGet->getPasswordObject(), _oldPwdGet->getPasswordObject());
    if (result == kBCContainer_Success) {
        result = _mainVolumeHelperCtx.storeHeader();
    }
    return result;
}


BCContainerResult BCBackupHeader(BCHelperContext & _mainVolumeHelperCtx, iChannel * _backup)
{
    if (_mainVolumeHelperCtx.header == NULL) {
        return kBCContainer_InvalidArgument;
    }
    
    return _mainVolumeHelperCtx.header->storeKeyblockBackup(_backup);
}


BCContainerResult BCRestoreHeader(BCHelperContext & _mainVolumeHelperCtx, iChannel * _backup, bool _prepare)
{
    if (_mainVolumeHelperCtx.pwdGet == NULL || _backup == NULL) {
        return kBCContainer_InvalidArgument;
    }
    
    BCContainerResult result = kBCContainer_Success;
    if (_prepare) {
        _mainVolumeHelperCtx.pwdGet->setKeyBlockChannel(_backup);
        result = BCPrepareContainerForOperation(_mainVolumeHelperCtx);
        if (result != kBCContainer_Success) {
            return result;
        }
        
        result = BCDecodeVolume(_mainVolumeHelperCtx);
    } else {
        result = BCVerifyPassword(_mainVolumeHelperCtx);
        if (result != kBCContainer_Success) {
            return result;
        }
        result = _mainVolumeHelperCtx.storeHeader();
    }
    
    return result;
}


BCContainerResult BCEncryptHeader( BCHelperContext & _helperCtx )
{
    
    BCContainerResult result = _helperCtx.encryptHeader();
    
    if (result == kBCContainer_Success) {
        result = _helperCtx.storeHeader();
    }
    
    return result;
}


BCContainerResult BCDecryptHeader( BCHelperContext & _helperCtx )
{
    BCContainerResult result = _helperCtx.decryptHeader();
    
    if (result == kBCContainer_Success) {
        result = _helperCtx.storeHeader();
    }
    
    return result;
}

BCContainerResult BCWipeHeader( BCHelperContext & _mainVolumeHelperCtx )
{
    if (_mainVolumeHelperCtx.pwdGet == NULL) {
        return kBCContainer_InvalidArgument;
    }
    
    BCContainerResult result = BCPrepareContainerForOperation(_mainVolumeHelperCtx);
    if (result != kBCContainer_Success) {
        return result;
    }
    
    return _mainVolumeHelperCtx.header->wipeDatablock(_mainVolumeHelperCtx.ptbKey);
}

BCContainerResult BCUpdateDescription( BCHelperContext & _helperCtx, const char * _description)
{
    BCContainerResult result = _helperCtx.header->setDescription(_description, _helperCtx.ptbKey);
    
    if (result == kBCContainer_Success) {
        result = _helperCtx.storeHeader();
    }
    
    return result;
}

SpaceMap * BCSpaceMapByTrimming(SpaceMap * _map, bc_u64 _size)
{
    bc_u64 length;
    SpaceMapGetLength(_map, &length);
    
    if (length < _size) {
        return NULL;
    }
    
    std::map<bc_u64, bc_u64> extents;
    SMExtentNode * node = _map->start;
    while (node != NULL) {
        extents.insert(std::make_pair(node->extent.start, node->extent.length));
        node = node->next;
    }
    
    std::map<bc_u64, bc_u64>::reverse_iterator iter;
    iter = extents.rbegin();
    
    unsigned long long nleft = _size;
    SpaceMap * result = SpaceMapCreate();
    
    while (nleft > 0 && iter != extents.rend()) {
        if (nleft >= iter->second) {
            // Add full vector
            SpaceMapAddNew(result, iter->first, iter->second);
            nleft -= iter->second;
        } else {
            // add a tail of the final vector.
            unsigned long long tail_start = iter->first + iter->second - nleft;
            SpaceMapAddNew(result, tail_start, nleft);
            nleft = 0;
        }
        ++iter;
    }
    
    if (result->count > 1) {
        SpaceMapSort(result);
    }
    
    SpaceMapGetLength(result, &length);
    
    assert(length == _size);
    
    return result;
}

BCContainerResult BCCreateHiddenPart( BCHelperContext & _helperCtx, cPwdGet * _hiddenPwdGet, SpaceMap *hiddenExtents )
{
    if (_helperCtx.header == NULL || _helperCtx.pwdGet == NULL || _hiddenPwdGet == NULL) {
        return kBCContainer_InternalError;
    }

	BCContainerResult result;

	//
	//	Do not verify again if key has been already verified and assigned.
	//
	if ( _helperCtx.ptbKey.bufferSize == 0 ) {
		if (_helperCtx.pwdIsEmpty()) {
			result = BCGetAndVerifyPassword(_helperCtx, true);
			if (result != kBCContainer_Success) {
				LIBLOG_INFO( "Warning! Invalid main volume password!" );
				return result;
			}
		}
	}

    BCHelperContext hiddenHelperContext = _helperCtx;
    hiddenHelperContext.pwdGet = _hiddenPwdGet;
    
    result = BCGetAndVerifyNewPassword(hiddenHelperContext);
    if (result != kBCContainer_Success) {
        return result;
    }

    //result = _helperCtx.header->createHiddenKey( hiddenExtents, &_helperCtx.pwdGet->getPasswordObject(), &_hiddenPwdGet->getPasswordObject());
	result = _helperCtx.header->createHiddenKey( hiddenExtents, &_helperCtx.ptbKey, &_hiddenPwdGet->getPasswordObject());
    
    if (result == kBCContainer_Success) {
        result = _helperCtx.storeHeader();
    }
    
    return result;
}


BCContainerResult BCAddSecretSharingScheme( iChannel* container, BCVolume* out_volume)
{
    return kBCContainer_Unsupported;
}


BCContainerResult BCAddKeyPair( iChannel* container, BCVolume* out_volume)
{
    return kBCContainer_Unsupported;
}


////////////////////////////////////////////////////////////////////////////////////
//
//  Reencryption
//
////////////////////////////////////////////////////////////////////////////////////

    
//static bool initRRStateBackup(const char* file)
static bool initRRStateBackup( iChannel* channel )
{
    //bc_file_handle fd;
    //if (kBCFileResult_Success != BCFileOpen(file,  kBCFileMode_ReadWrite, &fd))
	if ( !channel->open( kBCFileMode_ReadWrite ) )
        return false;

    bc_u64 offset = 0;
    //BCFileSeek(&fd, 0, kBCFileSeek_End);
    //if(sizeof(offset) != BCFileWrite(&fd, &offset, sizeof(offset))) {
    //    BCFileClose(&fd);
	channel->setPosition( 0, kBCFileSeek_End );
	if(sizeof(offset) != channel->writeBytes( &offset, sizeof(offset) ) ) {
		channel->close();
        return false;
    }

    //BCFileSync(&fd);
    //BCFileClose(&fd);
	channel->sync();
	channel->close();
    return true;
}

//static bool readRRState(const char* file, bc_u64* offset)
static bool readRRState( iChannel* channel, bc_u64* offset)
{
    //bc_file_handle fd;
    //if (kBCFileResult_Success != BCFileOpen(file, kBCFileMode_Read, &fd))
	if ( !channel->open( kBCFileMode_Read ) )
        return false;

    //bc_u64 size = BCFileSize(file);
    //if (sizeof(*offset) != BCFileReadWithOffset(&fd, offset, sizeof(*offset), size - sizeof(*offset))) {
    //    BCFileClose(&fd);

    bc_u64 size = channel->getSize();
	if (sizeof(*offset) != channel->readBytes( offset, sizeof(*offset), size - sizeof(*offset))) {
		channel->close();
        return false;
    }

    //BCFileSync(&fd);
    //BCFileClose(&fd);
	channel->sync();
	channel->close();
    return true;
}


//static bool writeRRState(const char* file, bc_u64 offset)
static bool writeRRState( iChannel* channel, bc_u64 offset)
{
    //bc_file_handle fd;
    //if (kBCFileResult_Success != BCFileOpen(file, kBCFileMode_ReadWrite, &fd))
	if ( !channel->open( kBCFileMode_ReadWrite ) )
        return false;

    //bc_u64 size = BCFileSize(file);
    //if (sizeof(offset) != BCFileWriteWithOffset(&fd, &offset, sizeof(offset), size - sizeof(offset))) {
    //    BCFileClose(&fd);

    bc_u64 size = channel->getSize();
	if (sizeof(offset) != channel->writeBytes( &offset, sizeof(offset), size - sizeof(offset))) {
		channel->close();
        return false;
    }

    //BCFileSync(&fd);
    //BCFileClose(&fd);
	channel->sync();
	channel->close();
    return true;
}

#define BC_VOLUME_BUFFER_SIZE (1024 * 1024)

//static BCContainerResult reencryptVolume(BCVolume* src, BCVolume* dst, bc_u64 start, const char* file, const char* backup, BCVolumeProgressCallback progress, void* arg)
static BCContainerResult reencryptVolume(BCVolume* src, BCVolume* dst, bc_u64 start, iChannel* channel, iChannel* backup, BCVolumeProgressCallback progress, void* arg)
{
    //bc_file_handle fd;
    //if(kBCFileResult_Success != BCFileOpen(file, kBCFileMode_ReadWrite, &fd))
	if ( !channel->open( kBCFileMode_ReadWrite ) )
        return kBCContainer_IOError;

    char* buffer = (char*) malloc(BC_VOLUME_BUFFER_SIZE);
    if(!buffer) {
        //BCFileClose(&fd);
		channel->close();
        return kBCContainer_NoMemory;
    }

    bc_u64 offset = start;
    bc_u64 size = src->info.size;
    bc_u64 iv;
    BCContainerResult error = kBCContainer_Success;

    while(offset < size) {
        //size_t nbytes = size - offset;
        //nbytes = nbytes > BC_VOLUME_BUFFER_SIZE ? BC_VOLUME_BUFFER_SIZE : nbytes;
		size_t nbytes = u64tou32( size - offset, BC_VOLUME_BUFFER_SIZE );

        //if(nbytes != BCFileReadWithOffset(&fd, buffer, nbytes, src->info.offset + offset)) {
		if(nbytes != channel->readBytes( buffer, nbytes, src->info.offset + offset)) {
            error = kBCContainer_IOError;
            break;
        }

		iv = sector2iv( offset / 512, src->info.use_new_synchro );

        src->cipher->decrypt(iv, (struct_bc_key*)src->key_handle, /*src->pool,*/ buffer, nbytes);
        
		iv = sector2iv( offset / 512, dst->info.use_new_synchro );

        dst->cipher->encrypt(iv, (struct_bc_key*)dst->key_handle, /*dst->pool,*/ buffer, nbytes);

        //if(nbytes != BCFileWriteWithOffset(&fd, buffer, nbytes, dst->info.offset + offset)) {
		if(nbytes != channel->writeBytes( buffer, nbytes, dst->info.offset + offset)) {
            error = kBCContainer_IOError;
            break;
        }

        //BCFileSync(&fd);
		channel->sync();

        offset += nbytes;

        if (!writeRRState(backup, offset)) {
            error = kBCContainer_IOError;
            break;
        }

        if(progress && !progress(offset, size, arg)) {
            // fprintf(stderr, "Reencryption terminated at offset %lu\n", offset);
            error = kBCContainer_Cancelled;
            break;
        }
    }

    //BCFileClose(&fd);
	channel->close();
    free(buffer);

    return error;
}


BCContainerResult BCReencrypt( iChannel* container, iChannel *backupFile, bc_alg_id algid, bc_mode_id modeid, BCVolumeProgressCallback progress, void* arg)
{
#if 0
    BCVolume src;
    BCVolume dst;

    // Get main part password
    BCPassword pwd;
    BCHeaderHandle header(verifyMainPartPassword(container, "Please enter password to reencrypt container", NULL, &pwd));
    if (!header) {
        return kBCContainer_Cancelled;
    }

    //char backupFile[PATH_MAX] = {0};
    //snprintf(backupFile, sizeof(backupFile) - 1, "%s.rrkbb", container);

    const BCKeygen* keygen = header->keygen;
    BCContainerResult error = kBCContainer_Success;

    if (header->isMarkedForReencryption()) {

        // looks like previous reencryption process was interrupted
        // container header we've just read has new key and algorithm 
        // we need to get old key and algorithm from the backup file with the same password and restore the state to continue

        BCHeaderHandle oldHeader(keygen->kgload(backupFile, 0, 0, 0 ));
        if (!oldHeader) {
            error = kBCContainer_UnknownFormat;
            goto LABEL_OUT;
        }

		error = oldHeader->decodeKey(pwd.buf, &src.info, &src.key_handle, &(src.cipher) ); //, src.pool);
        if (error) {
            goto LABEL_OUT;
        }

		error = header->decodeKey(pwd.buf, &dst.info, &dst.key_handle, &(dst.cipher) ); // , dst.pool);
        if (error) {
            goto LABEL_OUT;
        }

        bc_u64 offset;
        if (!readRRState(backupFile, &offset)) {
            error = kBCContainer_IOError;
            goto LABEL_OUT;
        }

        //fprintf(stderr, "Resuming reencryption from offset %lu\n", offset);
        error = reencryptVolume(&src, &dst, offset, container, backupFile, progress, arg);
        if (!error) {
            header->markForReencryption(false);
            header->store(container);
        }

    } else {

        // begin new reencryption process
        error = header->decodeKey(pwd.buf, &src.info, &src.key_handle, &(src.cipher) ); //, src.pool);
        if (error) {
            return error;
        }

        error = header->store(backupFile);
        if (error) {
            goto LABEL_OUT;
        }

        BCHeaderHandle newHeader(keygen->kgcreate());
        if (!newHeader) {
            error = kBCContainer_NoMemory;
            goto LABEL_OUT;
        }

        error = newHeader->init(algid, modeid, src.info.size, src.info.description, pwd.buf);
        if (error) {
            goto LABEL_OUT;
        }

        assert(!src.info.hidden); // we can't reencrypt hidden parts

        error = newHeader->decodeKey(pwd.buf, &dst.info, &dst.key_handle, &(dst.cipher) ); //, dst.pool);
        if (error) {
            goto LABEL_OUT;
        }

        newHeader->markForReencryption(true);

        error = newHeader->store(container);
        if (error) {
            goto LABEL_OUT;
        }

        if (!initRRStateBackup(backupFile)) {
            error = kBCContainer_IOError;
            goto LABEL_OUT;
        }

        // Begin reencryption
        error = reencryptVolume(&src, &dst, 0, container, backupFile, progress, arg);
        if (!error) {
            newHeader->markForReencryption(false);
            newHeader->store(container);
        }
    }
    
    if (!error)
    //    BCFileDelete(backupFile);
		backupFile->deleteFile();

        
LABEL_OUT:
    BCPutVolume(&src);
    BCPutVolume(&dst);

    return error;
#endif
    return kBCContainer_Unsupported;
}

static cPTB &getStaticPtbSeed()
{
	static cPTB ptbSeed;
	return ptbSeed;
}

static int curPos = 0;

bool BCSetSeedBuffer( bc_u8 *seedBuffer, int seedBufferSize )
{
	cPTB &ptbSeed = getStaticPtbSeed();

	LIBLOG_INFO( "Begin Buffer %p, size %d", ptbSeed.buffer, ptbSeed.bufferSize );

	bool res = ptbSeed.WriteBytes( seedBuffer, seedBufferSize ) != 0;

	LIBLOG_INFO( "End Buffer %p, size %d", ptbSeed.buffer, ptbSeed.bufferSize );

	return res;
}

//
//	BCGetSeedBuffer
//		Reads seed from allocated buffer to seedBuffer.
//		Clears the space.
//
bool BCGetSeedBuffer( bc_u8 *seedBuffer, int seedBufferSize )
{
	cPTB &ptbSeed = getStaticPtbSeed();

	LIBLOG_INFO( "Begin Buffer %p, size %d, curPos %d", ptbSeed.buffer, ptbSeed.bufferSize, curPos );

	if ( ptbSeed.bufferSize == 0 )
	{
		LIBLOG_INFO( "Assertion! ptbSeed.bufferSize == 0" );
		return false;
	}

	if ( ptbSeed.bufferSize - curPos < seedBufferSize )
	{
		LIBLOG_INFO( "Assertion! ptbSeed.bufferSize - curPos < seedBufferSize" );
		return false;
	}

	LIBLOG_INFO( "Buffer %p, size %d", seedBuffer, seedBufferSize );

	memcpy( seedBuffer, ptbSeed.buffer + curPos, seedBufferSize );
	memset( ptbSeed.buffer + curPos, 0, seedBufferSize );
	curPos += seedBufferSize;

	return true;
}
