/*
 * Copyright 2010-2014 Jetico, Inc. All rights reserved.
 * Container library interface
 *	Modified - 01.07.2014
 *	iChannel is used insted of file access to container.
 */

#ifndef BC_CONTAINER_H
#define BC_CONTAINER_H

#include "intsize.h"
#include "keygen.h"
#include "volume.h"
#include "error.h"
#include "../certificate/CertificateLib/progressActor.h"
#include "cPwdGet.h"
#include "helpers.h"

#ifdef __cplusplus
extern "C" {
#endif

    /** 
     * Load keygen for container
     */
    BCContainerResult BCLoadKeygen(BCHelperContext & _helperCtx);

    /**
     * Verify password for container
     * @param _helperCtx       Reference to container helper context
     * @param _progressActor   Progress signaling object
     */
    BCContainerResult BCVerifyPassword(BCHelperContext & _helperCtx);
    
    /**
     * Gets password with cPwdGet object from helper context and verify it
     */
    BCContainerResult BCGetAndVerifyPassword(BCHelperContext & _helperCtx, bool _forMainVolume = false);
    
    /**
     * Create a new container header
     * @param helperCtx       Reference to container helper context
     * This function only creates header object with options, got from cPwdGet item of helper context. 
     * Before calling it user must set seed buffre with BCSetSeedBuffer. User must
     * store container header and randomize it. Also he is responsible for header object deletion when it no longer
     * required. No channel object created so it will be NULL in helper context even function is succeeded.
     * @return kBCContainer_Success on success or error code otherwise
     */
    BCContainerResult BCCreateContainer(BCHelperContext & _helperCtx);

	enum eBenchmarkError {
		eBenchmarkError_OK = 0,
		eBenchmarkError_readKeyFiles,
		eBenchmarkError_keyGeneration,
		eBenchmarkError_iterationCountTooSmall
	};

    /**
     * Performs benchmark of container mount for set params
     * @param hashAlgID			Hash algorithm ID
     * @param keyFilesT			Array fo key files channels
     * @param keyFilesQ			Number of key files in array
     * @param keyFilesTime		Double time in seconds for key file procesing
     * @param iterations		Iteraion count for producing key ( recommended 10000 or even more, otherwise time may be zero with error eBenchmarkError_iterationCountTooSmall )
     * @param iterationsTime	Double time in seconds for key producing with assigned iterations count.
     * @return					see eBenchmarkError enum
     */
	eBenchmarkError BCMountContainerBenchmark(
			bc_u32 hashAlgID,
			iChannel ** keyFilesT, bc_u32 keyFilesQ, double &keyFilesTime,
			int iterations, double &iterationsTime
		);
    
    /**
     * Create new hidden part inside container.
     * @param _helperCtx       Reference to container helper context
     * @param _hiddenPwdGet    Instance of cPwdGet to get password and otehr credentials for new hidden part
     * @param _hiddenPartExtents Space map for new hidden part. It must have length and structure that is exactly 
                                 describes hidden part to create. Important! Space map must be alligned to the volume sector size.
     * Before calling this function user must set seed buffer with BCSetSeedBuffer.
     */
    BCContainerResult BCCreateHiddenPart( BCHelperContext & _helperCtx, cPwdGet * _hiddenPwdGet, SpaceMap * _hiddenPartExtents);

    /**
     * Trim existing space map to required size starting from the end.
     * @param _map    Initial space map
     * @param _size   Required size
     * @return New space map that must be freed with SpaceMapFree when it is no longer required
     */
    SpaceMap * BCSpaceMapByTrimming(SpaceMap * _map, bc_u64 _size);

    /**
     * Parse container file and return its main encoded volume description on success.
     * This procedure will not ask passwords so it will fail for containers with encrypted header.
     */
    //bool BCParseContainerFile(const char* file, BCVolumeInfo* out_info);
    bool BCParseContainerFile( iChannel *chanel, BCVolumeInfo* out_info);

    
    /**
     * Create container header backup
     * Main volume password must be obtained and stored into _mainVolumeHelperCtx with BCPrepareContainerForOperation befor calling this function.
     * @param _mainVolumeHelperCtx      Helper context for container. Must contain password getter object with valid main volume
     *                                  password. All operation with container are based on data provided here
     * @param _backup                   Instance of channel object, where new backup should be stored
     */
    BCContainerResult BCBackupHeader(BCHelperContext & _mainVolumeHelperCtx, iChannel * _backup);

    /**
     * Restore container header from backup.
     * It is required to store main volume password into _mainVolumeHelperCtx, but it is not nesessary to call BCPrepareContainerForOperation 
     * separately, it will be asked automatically in case of valid cPwdGet object.
     * Backup restore consist of two parts, firts of which is prepare. At preparation stage (_prepare = true) full decode of main
     * volume key is performed with its valid password, but without actual restore. In the client part container should be mounted
     * with new key and volume information and only on success this function must be colled second time with _prepare parameter set to false.
     * @param _mainVolumeHelperCtx      Helper context for container. Must contain password getter object with valid main volume
     *                                  password. All operation with container are based on data provided here
     * @param _backup                   Instance of channel object, from which container header should be restored
     * @param _prepare                  Defines restore mode (true for prepare, false for actual restore). See comment above.
     */
    BCContainerResult BCRestoreHeader(BCHelperContext & _mainVolumeHelperCtx, iChannel * _backup, bool _prepare);


    /**
     * Encrypt container header.
     * Main volume password must be obtained and stored into helper context with BCPrepareContainerForOperation befor calling this function.
     * @param _helperCtx     Container helper context
     */
    BCContainerResult BCEncryptHeader( BCHelperContext & _helperCtx );

    /**
     * Decrypt container header. 
     * Main volume password must be obtained and stored into helper context with BCPrepareContainerForOperation befor calling this function.
     * @param _helperCtx     Container helper context
     */
    BCContainerResult BCDecryptHeader( BCHelperContext & _helperCtx );

    /**
     * Request main volume password and can ask additional password if required. Suitable for diffrenet password operations
     * such as add, remove and change. To ask additional password 'askMore' member of BCContainerContext must be more then 0.
     * @param _mainVolumeHelperCtx  Helper context for container.
     * @param _additionalPwdGet     Password getter object to request additional password. Its meaning depend on context.
     */
    BCContainerResult BCPrepareContainerForOperation( BCHelperContext & _mainVolumeHelperCtx, cPwdGet * _additionalPwdGet = NULL);
    
    /**
     * Change one of the existing container passwords.
     * Main volume password must be obtained and stored into _mainVolumeHelperCtx with BCPrepareContainerForOperation befor calling this function.
     * @param _mainVolumeHelperCtx      Helper context for container. Must contain password getter object with valid main volume
     *                                  password. All operation with container are based on data provided here
     * @param _volumePwdGet             Password getter object to get password, which will be changed. In case of NULL passwword 
     *                                  getter from _mainVolumeHelperCtx will be used.
     * @param _newPwdGet                Password getter object for password which will change existing one
     */
    BCContainerResult BCChangePassword(BCHelperContext & _mainVolumeHelperCtx, cPwdGet * _volumePwdGet, cPwdGet * _newPwdGet);

    /**
     * Remove one of the existing container passwords.
     * Main volume password must be obtained and stored into _mainVolumeHelperCtx with BCPrepareContainerForOperation befor calling this function.
     * @param _mainVolumeHelperCtx      Helper context for container. Must contain password getter object with valid main volume
     *                                  password. All operation with container are based on data provided here
     * @param _oldPwdGet                Password getter object for password to remove
     * @param _force                    In case of removing last main volume password and this parameter is set to false an error
     *                                  code kBCContainer_LastKeyBlock will be returned. When this parameter is set to true, last
     *                                  keyblock will be removed and kBCContainer_Success will be returned.
     */
    BCContainerResult BCRemovePassword(BCHelperContext & _mainVolumeHelperCtx, cPwdGet * _oldPwdGet, bool _force = false);


    /**
     * Add another container password for any existing volume, including hidden part.
     * Main volume password must be obtained and stored into _mainVolumeHelperCtx with BCPrepareContainerForOperation befor calling this function.
     * @param _mainVolumeHelperCtx      Helper context for container. Must contain password getter object with valid main volume
     *                                  password. All operation with container are based on data provided here
     * @param _volumePwdGet             Password getter object to get volume password to which new password will be added. In case
     *                                  of NULL, passwword will be added to main volume.
     * @param _newPwdGet                Password getter object for new password to add
     */
    BCContainerResult BCAddPassword(BCHelperContext & _mainVolumeHelperCtx, cPwdGet * _volumePwdGet, cPwdGet * _newPwdGet);


    /**
     * Add another secret sharing scheme.
     * @param container     Path to container file
     * @param out_volume    Decoded volume info that had a new sharing scheme added, can be NULL.
     */
    // BCContainerResult BCAddSecretSharingScheme(const char* container, BCVolume* out_volume);
    BCContainerResult BCAddSecretSharingScheme( iChannel* container, BCVolume* out_volume);


    /**
     * Add another private/public key pair.
     * @param container     Path to container file
     * @param out_volume    Decoded volume info that had a new key pair added, can be NULL.
     */
    //BCContainerResult BCAddKeyPair(const char* container, BCVolume* out_volume);
    BCContainerResult BCAddKeyPair( iChannel* container, BCVolume* out_volume);


    /**
     * Change container description.
     * Main volume password must be obtained and stored into helper context with BCPrepareContainerForOperation befor calling this function.
     * @param _helperCtx    Helper context for container.
     * @param _description  New container description.
     */
    BCContainerResult BCUpdateDescription( BCHelperContext & _helperCtx, const char * _description);


    /**
     * Reencrypt container body using new algorithm, mode and key.
     * Hidden part reencryption is not supported.
     * @param container     Path to container file
     * @param algid         New encryption algorithm id
     * @param modeid        New encryption mode id
     * @param progress      Progress reporting callback
     * @param arg           Progress callback argument
     */
    //BCContainerResult BCReencrypt(const char* container, bc_alg_id algid, bc_mode_id modeid, BCVolumeProgressCallback progress, void* arg);
    BCContainerResult BCReencrypt( iChannel* container, iChannel *backupFile, bc_alg_id algid, bc_mode_id modeid, BCVolumeProgressCallback progress, void* arg);


    /**
     * Ask password and decode volume encoded in container.
     * @param _helperCtx     Data structure that contains all infornation, required for container header decode
     * @param _progressActor Object interface, that can signal decoding progress to client application.
     */
    BCContainerResult BCDecodeVolume(BCHelperContext & _helperCtx);

    /** Wipe out container header
     * It is required to store main volume password into _mainVolumeHelperCtx, but it is not nesessary to call BCPrepareContainerForOperation
     * separately, it will be asked automatically in case of valid cPwdGet object.
     * @param _mainVolumeHelperCtx      Helper context for container. Must contain valid password getter object.
     */
    BCContainerResult BCWipeHeader( BCHelperContext & _mainVolumeHelperCtx );
    
//
//	This function store and destroy seed buffer.
//	The buffer is required for key generation for container creating.
//	See BCUserGenerateSeed, BCCreateContainer_pwd
//
//	BCSetSeedBuffer
//		Allocates and fills static buffer.
//
bool BCSetSeedBuffer( bc_u8 *seedBuffer, int seedBufferSize );

//
//	BCGetSeedBuffer
//		Reads seed from allocated buffer to seedBuffer.
//		Clears the space.
//
bool BCGetSeedBuffer( bc_u8 *seedBuffer, int seedBufferSize );

#ifdef __cplusplus
}
#endif

#endif

