#ifndef BC_FILE_H
#define BC_FILE_H

#include <string.h>

#include "intsize.h"

#ifdef BC_LINUX
#include <linux/limits.h>
#include <dirent.h>
typedef int bc_file_handle;
typedef DIR* bc_dir_handle;
#elif _WIN32
#include <windows.h>
typedef HANDLE bc_file_handle;
typedef HANDLE bc_dir_handle;
#elif ANDROID
typedef void* bc_file_handle;
typedef void* bc_dir_handle;
#elif __APPLE__
#include <dirent.h>
typedef int bc_file_handle;
typedef DIR* bc_dir_handle;
#else
    #error "Not supported"
#endif

#include "fileDef.h"

struct bc_dirent
{
	char *d_name;
	bc_u64 d_size;
	bc_u32 d_type; // attributes

	bc_dirent()
	{
		clear();
	}
	
	~bc_dirent()
	{
		free_and_clear();
	}

	void clear()
	{
		d_name = NULL;
		d_size = 0;
		d_type = 0;
	}

	void free_and_clear()
	{
		if ( NULL != d_name )
			delete [] d_name;

		clear();
	}

	bool set( const char *name, bc_u64 size, bc_u32 type )
	{
		free_and_clear();

		size_t len = strlen( name );
		if ( 0 == len )
			return false;

		d_name = new char[ len + 1 ];

		if ( NULL == d_name )
			return false;

		strncpy( (char*)d_name, name, (unsigned long)len );
		d_name[ len ] = '\0';
		d_size = size;
		d_type = type;

		return true;
	}

};

// Unix returned "/home/inso/.bcdb";
const char *BCDBGetPath();

/**
 * Opens specified directory for further processing by BCDirRead.
 * @param dirName		Name of directory to open.
 * @return              Handle to directory object or NULL, use BCDirClose to release the object.
 */
bc_dir_handle BCDirOpen( const char *dirName );

/**
 * Reads next element from specified directory.
 * @param entry			Structure described directory entry.
 * @param dir			Handle of directory object opened by BCDirOpen.
 * @return              Return false if no more entries found.
 */
bool BCDirRead( bc_dirent *entry, bc_dir_handle dir );


/**
 * Closes directory object.
 * @param dir			Handle of directory object opened by BCDirOpen.
 */
void BCDirClose( bc_dir_handle dir );

/**
 * Checks if current entry is regular file.
 * @param entry			Structure described directory entry.
 * @return              Returns false for directory, true for files.
 *		Unix  entry->d_type == DT_REG
 *		Win32 ( entry->d_type & FILE_ATTRIBUTE_DIRCETORY ) == 0
 */
bool BCDirIsRegularFile( bc_dirent *entry );


typedef enum
{
    kBCFileResult_Success    = 0, ///< Operation successful.
    kBCFileResult_NoAccess,       ///< File doesn't exists or access denied.
    kBCFileResult_InternalError   ///< Unspecified error. In this case errno and stderr-output can be checked.
}bc_file_result;

/*#ifdef __cplusplus
extern "C"
{
#endif*/
    /**
     * Opens specified file.
     * @param fileName      Name of file to open.
     * @param mode          Opening mode.
     * @return              Possible returns:
     *                      - kBCFileResult_Success, if succeeded, 
     *                      - kBCFileResult_NoFile, if file was not found,
     *                      - kBCFileResult_NoAccess, if file access denied,
     *                      - kBCFileResult_InternalError, if internal error occured. 
     *                        In this case errno will be set.
     */
    bc_file_result BCFileOpen(const char *fileName, bc_file_mode mode, bc_file_handle *handle);
    
    /**
     * Repositions the offset of the open file.
     * @param handle        File handle returned by BCFileOpen call.
     * @param offset        Offset.
     * @param seek          Describes position, from which offset is defined.
     * @return              Current position of pointer from the beginning of file 
     *                      or (bc_u64)-1, if error occured.
     */
    bc_u64 BCFileSeek(bc_file_handle *handle, bc_u64 offset, bc_file_seek seek);

    /**
     * Returns current offset location from the beginning of file.
     * @param handle        File handle returned by BCFileOpen call.
     * @return              Current position of pointer from the beginning of file
     *                      or (bc_u64)-1, if error occured.
     */
    static inline bc_u64 BCFileGetPosition(bc_file_handle *handle) 
    {
        return BCFileSeek(handle, 0, kBCFileSeek_Current);
    }

    /**
     * Reads a part of file with offset from the beginning of the file.
     * @param handle        File handle returned by BCFileOpen call.
     * @param buffer        Data buffer.
     * @param byteCount     Data buffer size in bytes.
     * @param offset        Number of bytes from the beginning of file 
     *                      that must be skipped before starting reading.
     * @return              Number of bytes actually read or (bc_u64)-1, if error occured.
     */
    bc_u64 BCFileReadWithOffset(bc_file_handle *handle, void *buffer, bc_u64 byteCount, bc_u64 offset);

    /**
     * Reads a part of file starting from current position.
     * @param handle        File handle returned by BCFileOpen call.
     * @param buffer        Data buffer. 
     * @param byteCount     Data buffer size in bytes.
     * @return              Number of bytes actually read or (bc_u64)-1, if error occured.
     */
    bc_u64 BCFileRead(bc_file_handle *handle, void *buffer, bc_u64 byteCount);

    /**
     * Writes buffer to the file starting from the offset from the beginning of file.
     * @param handle File handle returned by BCFileOpen call.
     * @param buffer        Data buffer to write.
     * @param byteCount     Buffer size in bytes.
     * @param offset        Number of bytes from the beginning of file that must be skipped.
     * @return              0 if success, otherwise -1.
     */
    bc_u64 BCFileWriteWithOffset(bc_file_handle *handle, const void *buffer, bc_u64 byteCount, bc_u64 offset);

    /**
     * Writes buffer to the file starting from current location.
     * @param handle        File handle returned by BCFileOpen call.
     * @param buffer        Data buffer to write.
     * @param byteCount     Buffer size in bytes.
     * @return              0, if succeeded, otherwise -1.
     */
    bc_u64 BCFileWrite(bc_file_handle *handle, const void *buffer_, bc_u64 byteCount);

    /**
     * Closes specified file handle.
     * @param handle        Handle to close.
     * @return              0, if succeeded, otherwise -1.
     */
    int BCFileClose(bc_file_handle *handle);

    /**
     * Get file size in bytes.
     */
    bc_u64 BCFileSize(const char* path);

    /**
     * Sync file contents with the disk
     */
    void BCFileSync(bc_file_handle* fd);

    /**
     * Delete given file.
     * @param               File system path.
     */
    void BCFileDelete(const char* path);

    /**
     * Sync all contents with the disk
     */
    void BCFileSyncAll();
/*#ifdef __cplusplus
}
#endif*/
#endif //BC_FILE_H

