/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 */

#ifndef BC_PASSWORD_H
#define BC_PASSWORD_H

#include <cstddef>
#include <cstring>
#include "intsize.h"
#include "iChannel.h"

/// Password sizes in bytes.
enum {
	kBCPasswordMin	= 8,
    kBCPasswordMax	= 512,
	kBCPasswordMaximumKeySize = 64
};


/**
 * Password buffer.
 * Automatically shared data on scope end.
 */
struct BCPassword {
    char buf[kBCPasswordMax];
    size_t length;

    bc_u8 storage_key[kBCPasswordMaximumKeySize];
	bool bReady;

    BCPassword() { clear(); }
    ~BCPassword() { clear(); }

	//	Version 9 table with key files channels
	iChannel**	m_keyFilesT;
	int			m_keyFilesQ;

	//	Table of channels with public keys
	//	Used at creation ant adding slots with public keys.
	iChannel**	m_publicKeyT;
	int			m_publicKeyQ;

	//	Table of channels with secret keys
	//	Used at mounting or verification at editing properties ( like hidden part creation, addding password .. ).
	iChannel**	m_secretKeyT;
	int			m_secretKeyQ;

    void set( const char* pwd )
	{
        clear();
		std::strncpy(buf, pwd, sizeof(buf));
        length = std::strlen(buf);
	}

    void clear() {
		bReady = false;

        std::memset(buf, 0, sizeof(buf));
        length = 0;

	    std::memset(storage_key, 0, sizeof(storage_key));

		m_keyFilesT	 = NULL; m_keyFilesQ  = 0;
		m_publicKeyT = NULL; m_publicKeyQ = 0;
		m_secretKeyT = NULL; m_secretKeyQ = 0;
    }

	bool generateKey( int hash, const bc_u8 *salt, size_t saltSize, bc_u32 iterations );

	bool isPBE()
	{
		return ( length > 0 ) || ( m_keyFilesQ > 0 );
	}
};

struct BCPasswordUTF16
{
	//
	//	Note, Big-endian byte buffer is requied according pkcs #12, B.2 General method. (p.14) . 
	//	See P12_producePseudoRandomBytes.
	//
	bc_u8 *buf;
	size_t size;

	BCPasswordUTF16()
	{
		buf = NULL;
		size = 0;
	}
	
	~BCPasswordUTF16()
	{
		clear();
	}

	void clear()
	{
		clearBuffer( &buf, &size );
	}

	bool set( const char *pwd_utf8, size_t len, bool b_big_endian );
	void clearBuffer( bc_u8 **pp_bf, size_t *p_size );

	static bool generate_key_from_password(int hash_id, const BCPasswordUTF16 *pwd, const bc_u8* salt, size_t saltsize, bc_u32 num_iter, bc_u8* out_key, size_t keysize)
	{
		return pwd->generate_key_from_password( hash_id, salt, saltsize, num_iter, out_key, keysize );
	}

	static bool generate_key_from_password(int hash_id, const char* pwdbuf_utf8, size_t pwdlen, const bc_u8* salt, size_t saltsize, bc_u32 num_iter, bc_u8* out_key, size_t keysize);

	bool generate_key_from_password(int hash_id, const bc_u8* salt, size_t saltsize, bc_u32 num_iter, bc_u8* out_key, size_t keysize) const;
};

#endif

