/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * User authentication session interface.
 */

#ifndef BC_USER_H
#define BC_USER_H

#include <cstdlib>

#include "intsize.h"
#include "password.h"
#include "container.h"
#include "../utf8-utf16/convertUTFSize.h"
#include "../../fs-lib/log.h"

// Default verify password query
#define BC_DEFAULT_PASSWORD_QUERY "Please enter container password:"

// Default new password query
#define BC_DEFAULT_NEW_PASSWORD_QUERY "Please enter new container password:"

/**
 * User session return codes.
 */
typedef enum BCUserResult_{
	kBCUserOk = 0,
	kBCUserBufferTooSmall,
    kBCUserCancel,
    kBCUserError,
	kBCUserNotImpemented
} BCUserResult;


/**
 * User message level.
 */
typedef enum {
	kBCUserMessage_Info = 0,
	kBCUserMessage_Warning,
    kBCUserMessage_Error
} BCUserMessageLevel;

// Encoding tags for conversion
enum BCUserEncoding {
    kBCUser_UTF8    = 0x08,
    kBCUser_UTF16LE = 0x06,
    kBCUser_UTF16BE = 0x60
};


/** 
 * Callback supplied by caller to verify a password.
 * Used to verify existing or new password.
 * Each function that accepts the callback can accept NULL to not perform any verification
 *
 * @param pwd		Password string.
 * @param pwdlen	Length of the password string.
 * @param ctx		Cookie passed by the caller.
 * @return	        Error message to display if password is invalid, NULL otherwise.
 */
typedef const char* (*BCVerifyPasswordCallback) (const char* utf8_pwd, unsigned pwdlen, void* ctx);

/**
 * Callback used to decrypt and protect existing hidden part by substracting it's blocks
 * from free space got from the container.
 * @param header        Container header structure.
 * @param password      Existing hidden part password.
 * @param freeSpaceMap  Space map with extents of new hidden part.
 * @return              BCContainerResult type describing result of operation.
*/
typedef BCContainerResult (*BCProtectExistingHiddenPartCallback) (void *header, const char *password,
                                                                  SpaceMap *freeSpaceMap);


/**
 * Run verify password session.
 * @param callback	    Password verification callback that will be called each time a user tries to enter a password.
 * @param cookie		Password verification callback cookie.
 * @param out_pwd       Password that was successfully verified.
 * @param pwdGetObj     Password query object class with virtual functions. 
 * @param query         Password query, default query will be used if NULL
 * @return	            kBCUserCancel if user gave up to provide a correct password, kBCUserOk otherwise.
 */
static inline BCUserResult BCUserGetAndVerifyPassword( BCVerifyPasswordCallback callback, void* cookie, BCPassword* out_pwd, cPwdGet *pwdGetObj, const char *query = NULL )
{
    assert(false);
#if 0
	pwdGetObj->getPassword( query );

	size_t len = strlen( out_pwd->buf );
	out_pwd->length = len;

	if ( sizeof( out_pwd->buf ) <= len )
		throw CipherException(kBCCipher_InternalError);

	const char *err = (*callback)( out_pwd->buf, len, cookie );

	if ( err == NULL )
		return kBCUserOk;

	// pwdGetObj->pwdError( err ); TODO
#endif
	return kBCUserCancel;
}

#ifdef __cplusplus
extern "C" {
#endif


	/**
	 * Run verify password session.
     * @param query         Password query, default query will be used if NULL
	 * @param callback	    Password verification callback that will be called each time a user tries to enter a password.
	 * @param cookie		Password verification callback cookie.
     * @param out_pwd       Password that was successfully verified.
	 * @return	            kBCUserCancel if user gave up to provide a correct password, kBCUserOk otherwise.
	 */
    static inline BCUserResult BCUserVerifyPassword( const char* query, BCVerifyPasswordCallback callback, void* cookie, BCMountParams *mountParams, BCPassword* out_pwd )
	{
		if ( cPwdGet::m_current == NULL )
		{ assert( false ); return kBCUserNotImpemented; }

		return BCUserGetAndVerifyPassword( callback, cookie, out_pwd, cPwdGet::m_current, query );
	}
	
	
	/**
	 * Run new password session.
     * @param query         New password query, default query will be used if NULL
	 * @param callback	    Password verification callback that will be called each time a user tries to enter a password.
     * @param cookie		Password verification callback cookie.
     * @param out_pwd       Password that was successfully verified.
	 * @return	            kBCUserCancel if user gave up to provide a correct password, kBCUserOk otherwise.
	 */
	static inline BCUserResult BCUserCreatePassword(const char* query, BCVerifyPasswordCallback callback, void* cookie, BCPassword* out_pwd)
	{
    	assert( false );
		// is not implemented
		return kBCUserCancel;
	}
	
	/**
	 * Fire up the seed generation progress dialog.
	 * @param total	        Total bytes of seed value that needs to be generated.
	 * @return	            Number of bytes of seed generated
	 */
	static inline unsigned BCUserGenerateSeed(bc_u8* out_buf, unsigned total)
	{
		if ( BCGetSeedBuffer( out_buf, total ) )
			return total;

		assert( false );
		return false;
	}

    /**
     * Convert UTF16 string to UTF8.
     * @param in_utf16      UTF16LE/UTF16BE encoded input string.
     * @param inlen         Input string buffer size.
     * @param out_utf8      Buffer to hold utf8 converted string.
     * @param out_len       Output buffer size.
     * @return              Number of bytes in output buffer used for converted string or 0 on error, 
	 *						If output buffer is NULL it returns number of bytes required to complete conversion.
     */
    static inline size_t BCUserUTF16LEToUTF8(const bc_u8* in_utf16, size_t inlen, char* out_utf8, size_t outlen)
	{
    	//
    	// multi-lib-enc/keygens/kgghost/kgghost.cpp::156  function-> volumeInfoForKey
    	//

#ifdef _LITTLE_ENDIAN_DEFINED
		// Little-endian does nothing.
#else
    	cPlainTextBlock ptb( in_utf16, inlen );

    	in_utf16 = ptb.m_ptr;

		if ( NULL != out_utf16 )
			swap_bytes( out_utf16, inlen * sizeof( UTF16 ) );
#endif
		size_t sizeInSymbols = convertUTF16toUTF8( (const UTF16*)in_utf16, inlen / sizeof(UTF16), (UTF8 *)out_utf8, outlen / sizeof( UTF8 ) );

		return sizeInSymbols * sizeof( UTF8 );
	}

    static inline size_t BCUserUTF16BEToUTF8(const bc_u8* in_utf16, size_t inlen, char* out_utf8, size_t outlen)
	{
        LIBLOG_INFO( "Assertion!" );
		return 0;
	}

	static inline void swap_bytes( bc_u8* buf, size_t len )
	{
		len &= ~1;

		while( len )
		{
			len -= 2;
			buf[len]   ^= buf[len+1];
			buf[len+1] ^= buf[len];
			buf[len]   ^= buf[len+1];
		}
	}

    /**
     * Same as above but reverse.
     */
    static inline size_t BCUserUTF8ToUTF16LE(const char* in_utf8, size_t inlen, bc_u8* out_utf16, size_t outlen)
	{
		size_t sizeInSymbols = convertUTF8toUTF16( (const UTF8*)in_utf8, inlen, (UTF16 *)out_utf16, outlen / sizeof( UTF16 ) );

#ifdef _LITTLE_ENDIAN_DEFINED
		// Little-endian does nothing.
#else
		if ( NULL != out_utf16 )
			swap_bytes( out_utf16, sizeInSymbols * sizeof( UTF16 ) );
#endif
		
		return sizeInSymbols * sizeof( UTF16 );
	}

    static inline size_t BCUserUTF8ToUTF16BE(const char* in_utf8, size_t inlen, bc_u8* out_utf16, size_t outlen)
	{
		size_t sizeInSymbols = convertUTF8toUTF16( (const UTF8*)in_utf8, inlen, (UTF16 *)out_utf16, outlen / sizeof( UTF16 ) );

#ifdef _LITTLE_ENDIAN_DEFINED
		if ( NULL != out_utf16 )
			swap_bytes( out_utf16, sizeInSymbols * sizeof( UTF16 ) );
#else
		// Big-endian does nothing.
#endif
	
		return sizeInSymbols * sizeof( UTF16 );
	}

	static inline size_t copy_and_swap( bc_u8 *out, size_t outlen, const bc_u8 *in, size_t inlen )
	{
		size_t size = 0;

		if ( NULL == out )
			size = inlen;
		else
		{
			if ( outlen > inlen )
				outlen = inlen;

			memcpy( out, in, outlen );
			swap_bytes( out, outlen );

			size = outlen;
		}

		return size;
	}
    /**
     * Change string encoding.
     * @param msg                   Source message to change encoding of.
     * @param msgSize               Source message size in bytes.
     * @param out_encodedMsg        Buffer to contain target message with changed encoding.
     * @param bufferSize            Buffer length in bytes.
     * @param fromEncoding          Source message encoding.
     * @param toEncoding            Target message encoding.
     * @return                      Number of bytes in output buffer used for converted string or 0 on error, 
	 *								If output buffer is NULL it returns number of bytes required to complete conversion.
     */
    //size_t BCUserConvertString( const bc_u8* in, size_t inlen, bc_u8* out, size_t outlen, BCUserEncoding fromEncoding, BCUserEncoding toEncoding)
	static inline size_t BCUserConvertString( const bc_u8* in, size_t inlen, bc_u8* out, size_t outlen, BCUserEncoding fromEncoding, BCUserEncoding toEncoding)
	{
		size_t size = 0;
		size_t error = 0;

		if ( kBCUser_UTF8 == fromEncoding )
		{
			if ( kBCUser_UTF16LE == toEncoding )
				size = BCUserUTF8ToUTF16LE( (const char*)in, inlen, out, outlen );
			else
			if ( kBCUser_UTF16BE == toEncoding )
				size = BCUserUTF8ToUTF16BE( (const char*)in, inlen, out, outlen );
			else
			{
				assert( false );
				error = 1;
			}
		}
		else
		if ( kBCUser_UTF16LE == fromEncoding )
		{
			if ( kBCUser_UTF16BE == toEncoding )
				size = copy_and_swap( out, outlen, in, inlen );
			else
			if ( kBCUser_UTF8 == toEncoding )
				size = BCUserUTF16LEToUTF8( in, inlen, (char*)out, outlen );
			else
			{
				assert( false );
				error = 2;
			}
		}
		else
		if ( kBCUser_UTF16BE == fromEncoding )
		{
			if ( kBCUser_UTF16LE == toEncoding )
				size = copy_and_swap( out, outlen, in, inlen );
			else
			if ( kBCUser_UTF8 == toEncoding )
				size = BCUserUTF16BEToUTF8( in, inlen, (char*)out, outlen );
			else
			{
				assert( false );
				error = 4;
			}
		}
		else
		{
			assert( false );
			error = 5;
		}

		if ( 0 != error )
			size = 0;

		return size;
	}

	/**
	 * Display a message to the user.
     * @param msg       User message text.
     * @param level     User message type level.
	 */
	void BCUserMessage(const char* msg, BCUserMessageLevel level);

    /**
     * Gets information about a new hidden part. Allows an user to protect existing hidden parts
     * by substracting their extents from the hidden map.
     * @param header            Parameter to pass to protectCallback.
     * @param protectCallback   Used to protect existing hidden part by substracting it's blocks from hiddenMap.
     * @param hiddenMap         Map with extents of future hidden part.
     * @param newPassword       Password for new hidden part.
     */
    static inline BCUserResult BCUserGetNewHiddenPartInfo(void *header, BCProtectExistingHiddenPartCallback protectCallback,
                                         BCVerifyPasswordCallback verifyCallback, void* verifyCookie,
                                         SpaceMap *hiddenMap, BCPassword *newPassword)
    {
    	assert( false );
    	return kBCUserCancel;
    }


    /**
     * Gets public/secret key block by keyID from local database.
     * @param secret            Determinated secret or public key is required.
     * @param keyId             Input buffer with key identifier.
	 * @param keyIdSize			Size of key identifier.
     * @param keyBf             [Out] buffer for public/secret key in x509 or PGP format.
     * @param pKeyBfSize        [In]  pointer to size of allocated buffer for key.
	 *                          [Out] number of copied/required bytes.
	 * @param pKeyPktFormat     [Out] key format, see usage note below and enum kgghost::type_kff
     * @param utf8UserID		[Out] string with user identifier in utf8 format
     * @param pUserIdSize       [In]  size of buffer for user identifier. 
	 *                          [Out] number of bytes copied to buffer.
	 * @return                  kBCUserOk if success,
	 *                          kBCUserBufferTooSmall if output keyBf or utf8UserID buffers are too small 
	 *                           ( pKeyBfSize and pUserIdSize are filled by required sizes )
	 *                          kBCUserCancel if user stop the process.
	 * Usage note:
	 * Usable function to extract corresponded key from the package
	 *		Public kffX509 - x509_certificate::parse	to get x509_publicKeyInfo as value tbsCertificate.subjectPKInfo 
	 *													and then pka_createAlgorithmX509 to create key handle
	 *		Secret kffX509 - P12_PFX_PDU_parse			to get x509_privateKeyInfo in ptbInfo.buffer 
	 *													and then ska_createAlgorithmX509 to create keu handle
	 *		Public kffPGP  - PKT_CreatePublicKeyAlgFromPacket to create public key handle
	 *		Secret kffPGP  - PKT_CreateSecretKeyAlgFromPacket to create secret key handle
     */
	BCUserResult BCGetPKBlockByID( 
		bool secret,
		const bc_u8 *keyId, size_t keyIdSize,
		bc_u8 *keyBf, size_t *pKeyBfSize,		// [OUT] 
		int *pKeyPktFormat,						// [OUT] Format of key packet 
		char *utf8UserID, size_t *pUserIdSize	// 
	);


#ifdef __cplusplus
}
#endif


#endif
