/*
 * Copyright 2010-2014 Jetico, Inc. All rights reserved.
 * Handling of encrypted volume stored in container files
 *
 *	- 01.07.2014 PAV
 *	Uses iChannel interface instead of direct file.
 */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "volume.h"
#include "intsize.h"
//#include "file.h"
#include "iChannel.h"
#include "error.h"
#include "isaac.h"
#include "cipherbase.h"
#include "../certificate/CertificateLib/SMemory/SBlocks.h"
#include "../../fs-lib/log.h"

#define BC_VOLUME_BUFFER_SIZE (1024 * 1024) // Buffer size for randomization and reencryption

BCContainerResult BCInitVolume(struct BCVolumeInfo* info, iChannel *channel, bc_file_mode mode, bc_key_handle key_handle, CipherBase *cipher, struct BCVolume* out_volume)
{
    memset(out_volume, 0, sizeof(*out_volume));

    //out_volume->file = strdup(file);
    //if(!out_volume->file) {
    //    return kBCContainer_NoMemory;
    //}

	out_volume->channel = channel;

	out_volume->mode = mode;

    out_volume->key_handle = key_handle;
	out_volume->cipher = cipher;
    memcpy(&out_volume->info, info, sizeof(BCVolumeInfo));

    if ( info->map )
	{
		out_volume->info.map = SpaceMapCreateCopy(info->map);

		if (!out_volume->info.map)
		{
			//free(out_volume->file);
			//out_volume->file = NULL;
			return kBCContainer_NoMemory;
		}
	}

    return kBCContainer_Success;
}

static bool extendSpaceMap( struct BCVolume* vol )
{
	if ( NULL == vol )
	{
		LIBLOG_INFO( "Assertion! NULL pointer!" );
		return false;
	}

	if ( NULL != vol->info.extent_array )
		return true;

	if ( vol->info.map )
		return ( kSpaceMap_Success == SpaceMapToExtentArray( vol->info.map, &vol->info.extent_array, &vol->info.extent_count ) );

    SpaceMap *space_map = SpaceMapCreate();
	if ( NULL == space_map )
		return false;

	Extent elem;
	elem.start = vol->info.offset;
	elem.length = vol->info.size;

	sm_result res = SpaceMapAddCopy(space_map, &elem);

	if ( kSpaceMap_Success == res )
		res = SpaceMapToExtentArray( space_map, &vol->info.extent_array, &vol->info.extent_count );

	SpaceMapFree( space_map );

	return ( kSpaceMap_Success == res );
}

static bool checkOpenVolume( struct BCVolume* vol )
{
	LIBLOG_INFO( "Begin %p", vol );

	if ( !extendSpaceMap( vol ) )
	{
		LIBLOG_INFO( "Assertion!" );
		return false;
	}

	//if ( NULL == vol->file_handle )
	//	return (kBCFileResult_Success == BCFileOpen(vol->file, vol->mode, &vol->file_handle));
	//return true;

	if ( vol->channel == NULL )
	{
		LIBLOG_INFO( "Assertion!" );
		return false;
	}

	bool ok = vol->channel->open( vol->mode );

	LIBLOG_INFO( "End %d", (int)ok );

	return ok;
}

BCContainerResult BCReadVolumeSectors(struct BCVolume* vol, bc_u64 start_sector, bc_u8 *data, bc_u64 number_of_sectors )
{
	LIBLOG_INFO( "Begin 0x%llx(%lld) 0x%llx(%lld)", (long long)start_sector, (long long)start_sector, (long long)number_of_sectors, (long long)number_of_sectors );

	if ( !checkOpenVolume( vol ) ) //, kBCFileMode_Read | kBCFileMode_Exclude ) )
	{
		LIBLOG_INFO( "Assertion!" );
		return kBCContainer_IOError;
	}

	Extent *physical_extent = NULL;
	bc_u64 physical_extent_count = 0;

	LIBLOG_INFO( "SpaceMapGetPhysicalExtents" );
	sm_result res =
		SpaceMapGetPhysicalExtents(
				vol->info.extent_array, vol->info.extent_count,
				start_sector * BCGetVolumeSectorSize( vol ), number_of_sectors * BCGetVolumeSectorSize( vol ),
				&physical_extent, &physical_extent_count );
	LIBLOG_INFO( "SpaceMapGetPhysicalExtents = %d", res );

	if ( kSpaceMap_Success != res )
	{
		LIBLOG_INFO( "Assertion! SpaceMapGetPhysicalExtents = %d", res );
		return kBCContainer_IOError;
	}

	bc_u64 readBytes = 0;

	for( bc_u64 i = 0; i < physical_extent_count; ++i )
	{
		LIBLOG_INFO( "vol->channel->readBytes, %d, %d", (int)physical_extent->length, (int)physical_extent->start );
		bc_u64 nbytes = 
			vol->channel->readBytes(
				data + readBytes,
				physical_extent->length,
				physical_extent->start);
		LIBLOG_INFO( "vol->channel->readBytes = %d", (int)nbytes );

		if ( BC_FILE_ERR == nbytes )
		{
			LIBLOG_ERROR( "Assertion! Error at chanel reding: buffer pos %p, extent length %x, extent start %x",
				(data + readBytes),
				(unsigned int )(physical_extent->length),
				(unsigned int )(physical_extent->start )
			);
			return kBCContainer_IOError;
		}

		if ( 0 == nbytes )
		{
			LIBLOG_ERROR( "Assertion! Zero bytes at chanel reding: buffer pos %p, extent length %x, extent start %x",
				(data + readBytes),
				(unsigned int )(physical_extent->length),
				(unsigned int )(physical_extent->start )
			);
			break;
		}

		readBytes += nbytes;
	}

	SpaceMapFreePhysicalExtents( &physical_extent );

	if ( readBytes != number_of_sectors * BCGetVolumeSectorSize( vol ) )
	{
		LIBLOG_ERROR( "Assertion! %d != %d.", (int)readBytes, (int)(number_of_sectors * BCGetVolumeSectorSize( vol )) );
		return kBCContainer_IOError;
	}

	bc_u64 iv = sector2iv( start_sector, vol->info.use_new_synchro );

	if(kBCCipher_Success != vol->cipher->decrypt( iv, (struct_bc_key*)vol->key_handle, data, (size_t)readBytes))
	{
		LIBLOG_INFO( "Assertion! Decrypt error." );
        return kBCContainer_InvalidAlgorithm;
    }

	LIBLOG_INFO( "End" );
    return kBCContainer_Success;
}

BCContainerResult BCReadVolumeSector(struct BCVolume* vol, bc_u64 sector, bc_sector_t data)
{
	return BCReadVolumeSectors( vol, sector, data, 1 );
}

BCContainerResult BCWriteVolumeSectors(struct BCVolume* vol, bc_u64 start_sector, const bc_u8 *data, bc_u64 number_of_sectors )
{
	//LIBLOG_INFO( "Begin volume%x, start_sector %d, data %x, sectors %d", (unsigned int)vol, (int)start_sector, (unsigned int)data, (int)number_of_sectors );

	bc_u64 iv = sector2iv( start_sector, vol->info.use_new_synchro );

	bc_u8 sync[ 8 ];
	iv2sync( sync, iv );

	size_t nbytes = (size_t)number_of_sectors * BCGetVolumeSectorSize( vol );
	cPlainTextBlock ptb;
	bc_u8 *encData = ptb.Allocate( nbytes );
	if ( NULL == encData )
	{
		LIBLOG_INFO( "Assertion encData = NULL" );
		return kBCContainer_NoMemory;
	}

	if(kBCCipher_Success != bc_alg_encrypt( vol->cipher, (struct_bc_key*)vol->key_handle, sync, data, encData, nbytes, vol->cipher->m_mode ) )
	{
		LIBLOG_INFO( "Encryption error" );
		return kBCContainer_InvalidAlgorithm;
	}

	if ( !checkOpenVolume( vol ) )//, kBCFileMode_ReadWrite | kBCFileMode_Exclude ) )
	{
		LIBLOG_INFO( "checkOpen(%p) error!", vol );
		return kBCContainer_IOError;
	}

	Extent *physical_extent = NULL;
	bc_u64 physical_extent_count = 0;

	sm_result res =
		SpaceMapGetPhysicalExtents(
				vol->info.extent_array, vol->info.extent_count,
				start_sector * BCGetVolumeSectorSize( vol ), number_of_sectors * BCGetVolumeSectorSize( vol ),
				&physical_extent, &physical_extent_count );

	if ( kSpaceMap_Success != res )
	{
		LIBLOG_INFO( "SpaceMap error %d", (int)res );
		return kBCContainer_IOError;
	}


	bc_u64 writtenBytes = 0;

	for( bc_u64 i = 0; i < physical_extent_count; ++i )
	{
		bc_u64 nbytes = 
			vol->channel->writeBytes(
				encData + writtenBytes,
				physical_extent->length,
				physical_extent->start);

		if ( 0 == nbytes )
		{
			LIBLOG_INFO( "Assertion! Written bytes 0" );
			//assert( false );
			break;
		}

		writtenBytes += nbytes;
	}

	SpaceMapFreePhysicalExtents( &physical_extent );

	if ( writtenBytes != number_of_sectors * BCGetVolumeSectorSize( vol ) )
	{
		LIBLOG_INFO( "Assertion! %d != %d * %d", (int)writtenBytes, (int)number_of_sectors, BCGetVolumeSectorSize( vol ) );
		//assert( false );
		return kBCContainer_IOError;
	}

	//LIBLOG_INFO( "End" );
	return kBCContainer_Success;
}

BCContainerResult BCWriteVolumeSector(struct BCVolume* vol, bc_u64 sector, bc_sector_t data)
{
	return BCWriteVolumeSectors( vol, sector, data, 1 );
}


BCContainerResult BCRandomizeVolumeBody(struct BCVolume* vol, BCVolumeProgressCallback progress,
                                        void* progressArg)
{
	LIBLOG_INFO( "Begin"  );

	if ( !vol->channel->open( kBCFileMode_ReadWrite ) )
		return kBCContainer_IOError;

    char* buffer = (char*) malloc(BC_VOLUME_BUFFER_SIZE);
    memset(buffer, 0, BC_VOLUME_BUFFER_SIZE);

    if(!buffer) {
		vol->channel->close();
        return kBCContainer_NoMemory;
    }

    time_t t;
    time(&t);

    void* isaac = IsaacInit((const char*)&t, sizeof(t));
    if(!isaac) {
		vol->channel->close();
        free(buffer);
		LIBLOG_INFO( "kBCContainer_NoMemory"  );
        return kBCContainer_NoMemory;
    }

    BCContainerResult error = kBCContainer_Success;

	// 04.09.2014 - See ..\versionInfo.txt
	//bc_u64 offset = 0;
	//bc_u64 size = vol->info.size;
    bc_u64 offset = vol->info.offset;
	bc_u64 size = vol->info.size + offset;
    int temp = 0;

	vol->channel->setPosition( offset, kBCFileSeek_Begin );

    while (offset < size)
    {
        size_t nbytes = u64tou32( size - offset, BC_VOLUME_BUFFER_SIZE );

        IsaacGenerate(isaac, buffer, nbytes);

        //if(nbytes != BCFileWrite(&fd, buffer, nbytes)) {
		if( nbytes != vol->channel->writeBytes( buffer, nbytes ) )
		{
            error = kBCContainer_IOError;
    		LIBLOG_INFO( "kBCContainer_IOError"  );
            break;
        }

		vol->channel->sync();

        offset += nbytes;

        if(progress && !progress(offset, size, progressArg)) {
            error = kBCContainer_Cancelled;
    		LIBLOG_INFO( "kBCContainer_Cancelled"  );
            break;
        }
    }

    //BCFileSyncAll();
	vol->channel->sync();

    //BCFileClose(&fd);
	vol->channel->close();
    free(buffer);
    IsaacRelease(isaac);

	LIBLOG_INFO( "End %d", error  );
    return error;
}


BCContainerResult BCReencryptVolume(struct BCVolume* src, struct BCVolume* dst, bc_u64 start, BCVolumeProgressCallback progress, void* arg)
{
    //bc_file_handle fd;
    //if(kBCFileResult_Success != BCFileOpen(src->file, kBCFileMode_ReadWrite | kBCFileMode_Exclude, &fd)) {
    //    return kBCContainer_IOError;
    //}
	if ( !src->channel->open( kBCFileMode_ReadWrite | kBCFileMode_Exclude ) )
		return kBCContainer_IOError;

    char* buffer = (char*) malloc(BC_VOLUME_BUFFER_SIZE);

	if( buffer == NULL )
	{
		//BCFileClose(&fd);
		src->channel->close();
        return kBCContainer_NoMemory;
    }

    bc_u64 offset = start;
    bc_u64 size = src->info.size;
    bc_u64 iv;
    BCContainerResult error = kBCContainer_Success;

    while(offset < size)
	{
        size_t nbytes = u64tou32( size - offset, BC_VOLUME_BUFFER_SIZE );

        //if(nbytes != BCFileReadWithOffset(&fd, buffer, nbytes, src->info.offset + offset)) {
		if(nbytes != src->channel->readBytes( buffer, nbytes, src->info.offset + offset)) {
            error = kBCContainer_IOError;
            break;
        }

 		iv = sector2iv( offset / BCGetVolumeSectorSize( src ), src->info.use_new_synchro );
		src->cipher->decrypt(iv, (struct_bc_key*)src->key_handle, buffer, nbytes);
        
 		iv = sector2iv( offset / BCGetVolumeSectorSize( dst ), dst->info.use_new_synchro );
        dst->cipher->encrypt(iv, (struct_bc_key*)dst->key_handle, buffer, nbytes);

        //if(nbytes != BCFileWriteWithOffset(&fd, buffer, nbytes, dst->info.offset + offset)) {
		if( nbytes != src->channel->writeBytes( buffer, nbytes, dst->info.offset + offset ) ) {
            error = kBCContainer_IOError;
            break;
        }

        //BCFileSyncAll();
		src->channel->sync();

        offset += nbytes;

        if(progress && !progress(offset, size, arg)) {
            error = kBCContainer_Cancelled;
            break;
        }
    }

    //BCFileClose(&fd);
	src->channel->close();
    free(buffer);

    return error;
}

        
void BCPutVolume(struct BCVolume* vol)
{
	if ( NULL != vol->cipher ) {
		vol->cipher->freeKey( vol->key_handle );
		CipherFactory::ReleaseCipher( &(vol->cipher) );
	}

	if ( vol->channel != NULL )
	{
		LIBLOG_INFO( "Channel %p begin", vol->channel );
		vol->channel->close();
		LIBLOG_INFO( "Channel %p end", vol->channel );
	}

    memset(vol, 0, sizeof(*vol));
}

/**
 * Get volume sector count.
 * @param volume        Current volume configuration
 * @return              64 bit value total number of available sectors.
 */
bc_u64 BCGetVolumeSectorCount( struct BCVolume* volume )
{
	bc_u64 len;

	if ( NULL != volume->info.map )
	{
		sm_result res = SpaceMapGetLength( volume->info.map, &len );

		if ( kSpaceMap_Success != res )
		{
			//assert( false );
			LIBLOG_INFO( "Assertion! SpaceMapGetLength returns  %d.", res );
			return 0;
		}
	}
	else
		len = volume->info.size;
	
	len = len / BCGetVolumeSectorSize( volume );

	return len;
}


/**
 * Flush volume data.
 * @param volume        Current volume configuration
 */
void BCFlushVolume( struct BCVolume* volume )
{
	if ( !checkOpenVolume( volume ) ) //, kBCFileMode_ReadWrite | kBCFileMode_Exclude ) )
	{
		//assert( false );
		LIBLOG_INFO( "Assertion! checkOpenVolume returns false." );
		return;
	}

	//BCFileSync( &volume->file_handle );
	volume->channel->sync();
}
