/*
 * Copyright 2012-2014 Jetico, Inc. All rights reserved.
 * Handling of encrypted volume stored in container files
 *
 * Author: Evgeny Yakovlev, Vasily Loginov
 *	- 01.07.2014 - PAV
 *	BCVolume works with container by iChannel interface instead of file.
 */

#ifndef BC_VOLUME_H
#define BC_VOLUME_H

#include "keygen.h"
#include "error.h"
#include "iChannel.h"
#include "cipherbase.h"

#include <stdio.h>

/// Volume format version.
typedef enum {
    kBCVolume_V7 = 7,       // V7 compatibility format
    kBCVolume_V8 = 8,       // V8 current main format
    kBCVolume_Unknown = 0   // Unknown format version
} BCVolumeFormat;



/**
 * Structure describes encrypted volume stored in container file.
 * Each container file encodes at least one main volume and a possible number of additional hidden ones.
 */
struct BCVolume
{
    struct BCVolumeInfo     info;
	iChannel				*channel;
    bc_key_handle           key_handle;
    bc_u32                  should_free_key;
	CipherBase				*cipher;

	/**
	 * checkOpenVolume uses the mode as default to open volume.
	 */
	bc_file_mode            mode;

	BCVolume() {
		// We have to zerois key_handle and cipher, otherwise BCPutVolume fails if the values were not initialised.
        should_free_key = 0;
        reset();
	}
    
    BCVolume & operator=(const BCVolume & _other) {
        info = _other.info;
        channel = _other.channel;
        key_handle = _other.key_handle;
        if (_other.cipher != NULL) {
            cipher = CipherFactory::CreateCipher();
            cipher->init(_other.cipher);
        } else {
            cipher = NULL;
        }
        
        return *this;
    }
    void reset() {
        if (should_free_key > 0
#if defined BC_LINUX || defined __APPLE__
            && key_handle != 0xFFFFF
#endif
            ) {
            CipherFactory::FreeKey( key_handle );
        }
        memset( this, 0, sizeof( *this ) );
#if defined BC_LINUX || defined __APPLE__
        key_handle = 0xFFFFF; // Set to huge value, since 0 can be valid key handle
#endif
    }
    
    ~BCVolume() {
        if (cipher) CipherFactory::ReleaseCipher( &cipher );
        if (should_free_key > 0
#if defined BC_LINUX || defined __APPLE__
            && key_handle != 0xFFFFF
#endif
            ) {
            CipherFactory::FreeKey( key_handle );
        }
    }
};


typedef bc_u8 bc_sector_t[512];

/**
 * Shows progress of reencrypting or creating container process.
 * @param nbytes        Current file position
 * @param total         Total file bytes
 * @return              Return zero to terminate process
 */
typedef int (*BCVolumeProgressCallback) (bc_u64 nbytes, bc_u64 total, void* arg);

#ifdef __cplusplus
extern "C" {
#endif


    /**
     * Init volume structure.
	 *	USe BCPutVolume to free volume resources.
	 *	According to bc drive style pool is saved at key handle, so it is not required to cary it thru sources
     * *** _param pool          Random data pool (optional)
     * @param info          Volume info structure
     * @param channel		Interface for reading container. //* @param file          Backing store file name, copied
	 * @param mode          Default mode to open volume
     * @param key_handle    Encryption key handle in kernel memory
     * @param out_volume    Initialized volume structure on success.
     */
    // BCContainerResult BCInitVolume(struct BCVolumeInfo* info, const char* file, bc_file_mode, bc_key_handle key_handle, CipherBase *cipher, /*char* pool, */ struct BCVolume* out_volume);
    BCContainerResult BCInitVolume(struct BCVolumeInfo* info, iChannel *channel, bc_file_mode, bc_key_handle key_handle, CipherBase *cipher, /*char* pool, */ struct BCVolume* out_volume);


    /**
     * Read and decrypt a volume sector
     * @param vol       Volume
     * @param sector    Sector number
     * @param data      Sector data buffer to store decrypted data
     */
    BCContainerResult BCReadVolumeSector(struct BCVolume* vol, bc_u64 sector, bc_sector_t data);
	BCContainerResult BCReadVolumeSectors(struct BCVolume* vol, bc_u64 start_sector, bc_u8 *data, bc_u64 number_of_sectors );


    /**
     * Encrypt and write a volume sector
     * @param vol       Volume
     * @param sector    Sector number
     * @param data      Sector data buffer to encrypt and write
     */
    BCContainerResult BCWriteVolumeSector(struct BCVolume* vol, bc_u64 sector, bc_sector_t data);
	BCContainerResult BCWriteVolumeSectors(struct BCVolume* vol, bc_u64 start_sector, const bc_u8 *data, bc_u64 number_of_sectors );


    /**
     * Randomize volume data
     * @param vol       Volume
     * @param progress  Caller progress callback
     * @param arg       Progress callback argument
     */
    BCContainerResult BCRandomizeVolumeBody(struct BCVolume* vol, BCVolumeProgressCallback progress, void* progressArg);


    /**
     * Reencrypt volume data
     * @param src           Current volume configuration
     * @param dst           Reencrypted volume configuration
     * @param start         Volume offset where to start reencryption from
     * @param progress      Progress callback
     * @param arg           Progress callback argument
     */
    BCContainerResult BCReencryptVolume(struct BCVolume* src, struct BCVolume* dst, bc_u64 start, BCVolumeProgressCallback progress, void* arg);


    /**
     * Release volume resources.
     */
    void BCPutVolume(struct BCVolume* volume);

	/**
	 * Get volume sector size in bytes.
     * @param volume        Current volume configuration
	 * @return              32 bit sector size
	 */
	static inline bc_u32 BCGetVolumeSectorSize( struct BCVolume* volume )
	{
		return 512;
	}

	/**
	 * Get volume sector count.
     * @param volume        Current volume configuration
	 * @return              64 bit value total number of available sectors.
	 */
	bc_u64 BCGetVolumeSectorCount( struct BCVolume* volume );


	/**
	 * Flush volume data.
     * @param volume        Current volume configuration
	 */
	void BCFlushVolume( struct BCVolume* volume );

#ifdef __cplusplus
}
#endif


#endif


