/*
 * Copyright 2012 Jetico, Inc. All rights reserved.
 * Windows file implementation.
 */

#include "../file.h"
//#include "../exception.h"
#include <stdio.h>
#include <assert.h>

/**
 * Opens specified file.
 * @param fileName      Name of file to open.
 * @param mode          Opening mode.
 * @return              Possible returns:
 *                      - kBCFileResult_Success, if succeeded, 
 *                      - kBCFileResult_NoFile, if file was not found,
 *                      - kBCFileResult_NoAccess, if file access denied,
 *                      - kBCFileResult_InternalError, if internal error occured. 
 *                        In this case errno will be set.
 */
bc_file_result BCFileOpen(const char *fileName, bc_file_mode mode, bc_file_handle *handle)
{
	DWORD dwDesiredAccess = 0;

	if ( mode & kBCFileMode_Read )
		dwDesiredAccess |= GENERIC_READ;

	if ( mode & kBCFileMode_ReadWrite )
		dwDesiredAccess |= GENERIC_READ | GENERIC_WRITE;

	//if ( mode & kBCFileMode_Exclude )
	//	dwDesiredAccess |= GENERIC_EXECUTE;

	DWORD dwCreationDisposition = mode & kBCFileMode_Exclude ? OPEN_EXISTING : OPEN_ALWAYS;

	*handle = CreateFileA( fileName, dwDesiredAccess, FILE_SHARE_READ, NULL, dwCreationDisposition, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( INVALID_HANDLE_VALUE != *handle )
		return kBCFileResult_Success;

	DWORD err = GetLastError();

	*handle = NULL;

	return kBCFileResult_NoAccess;
}

/**
 * Repositions the offset of the open file.
 * @param handle        File handle returned by BCFileOpen call.
 * @param offset        Offset.
 * @param seek          Describes position, from which offset is defined.
 * @return              Current position of pointer from the beginning of file 
 *                      or BC_FILE_ERR, if error occured.
 */
bc_u64 BCFileSeek(bc_file_handle *handle, bc_u64 offset, bc_file_seek seek)
{
	ULARGE_INTEGER uli;
	uli.QuadPart = offset;

	uli.LowPart = SetFilePointer( *handle, uli.LowPart, (PLONG)&(uli.HighPart), seek );

	if ( ( uli.LowPart == -1 ) && ( GetLastError() != ERROR_SUCCESS ) )
		return BC_FILE_ERR;

	return uli.QuadPart;
}

/**
 * Reads a part of file with offset from the beginning of the file.
 * @param handle        File handle returned by BCFileOpen call.
 * @param buffer        Data buffer.
 * @param byteCount     Data buffer size in bytes.
 * @param offset        Number of bytes from the beginning of file 
 *                      that must be skipped before starting reading.
 * @return              Number of bytes actually read or BC_FILE_ERR, if error occured.
 */
bc_u64 BCFileReadWithOffset(bc_file_handle *handle, void *buffer, bc_u64 byteCount, bc_u64 offset)
{
	if ( BCFileSeek( handle, offset, kBCFileSeek_Begin ) == BC_FILE_ERR )
		return BC_FILE_ERR;

	return BCFileRead( handle, buffer, byteCount );
}

/**
 * Reads a part of file starting from current position.
 * @param handle        File handle returned by BCFileOpen call.
 * @param buffer        Data buffer. 
 * @param byteCount     Data buffer size in bytes.
 * @return              Number of bytes actually read or BC_FILE_ERR, if error occured.
 */
bc_u64 BCFileRead(bc_file_handle *handle, void *buffer, bc_u64 byteCount)
{
    if (!buffer)
        return BC_FILE_ERR; 
    if (0 == byteCount)
        return 0;

	DWORD dwRead;

	if ( !ReadFile( *handle, buffer, (DWORD)byteCount, &dwRead, NULL ) || ( dwRead != (DWORD)byteCount ) )
		return BC_FILE_ERR;

	return dwRead;
}

/**
 * Writes buffer to the file starting from the offset from the beginning of file.
 * @param handle File handle returned by BCFileOpen call.
 * @param buffer        Data buffer to write.
 * @param byteCount     Buffer size in bytes.
 * @param offset        Number of bytes from the beginning of file that must be skipped.
 * @return              0 if success, otherwise -1.
 */
bc_u64 BCFileWriteWithOffset(bc_file_handle *handle, const void *buffer, bc_u64 byteCount, bc_u64 offset)
{
	if ( BCFileSeek( handle, offset, kBCFileSeek_Begin ) == BC_FILE_ERR )
		return BC_FILE_ERR;

	return BCFileWrite( handle, buffer, byteCount );
}

/**
 * Writes buffer to the file starting from current location.
 * @param handle        File handle returned by BCFileOpen call.
 * @param buffer        Data buffer to write.
 * @param byteCount     Buffer size in bytes.
 * @return              0, if succeeded, otherwise -1.
 */
bc_u64 BCFileWrite(bc_file_handle *handle, const void *buffer_, bc_u64 byteCount)
{
    if (!buffer_)
        return BC_FILE_ERR; 
    if (0 == byteCount)
        return 0;

	DWORD dwWritten;

	if ( !WriteFile( *handle, buffer_, (DWORD)byteCount, &dwWritten, NULL ) || ( dwWritten != (DWORD)byteCount ) )
		return BC_FILE_ERR;

	return dwWritten;
}

/**
 * Closes specified file handle.
 * @param handle        Handle to close.
 * @return              0, if succeeded, otherwise -1.
 */
int BCFileClose(bc_file_handle *handle)
{
	if( handle == NULL )
	{
		assert( false );
		return -1;
	}

	if ( INVALID_HANDLE_VALUE == *handle )
	{
		assert( false );
		return -1;
	}

	BOOL ok = CloseHandle( *handle );

	*handle = NULL; // INVALID_HANDLE_VALUE;

	return ok ? 0 : -1;
}

/**
 * Get file size in bytes.
 */
bc_u64 BCFileSize(const char* path)
{
	WIN32_FIND_DATAA fd;
	HANDLE hFind = FindFirstFileA( path, &fd );

	if ( hFind == INVALID_HANDLE_VALUE )
		return BC_FILE_ERR;

	ULARGE_INTEGER uli;
	uli.HighPart = fd.nFileSizeHigh;
	uli.LowPart = fd.nFileSizeLow;

	return uli.QuadPart;
}

/**
 * Sync file contents with the disk
 */
void BCFileSync(bc_file_handle* fd)
{
	FlushFileBuffers( *fd );
}

/**
 * Delete given file.
 * @param               File system path.
 */
void BCFileDelete(const char* path)
{
	DeleteFileA( path );
}

void BCFileSyncAll()
{
	_flushall(); // Note the function is affected by Commode.obj.
}

/**
 * Opens specified directory for further processing by BCDirRead.
 * @param dirName		Name of directory to open.
 * @return              Handle to directory object or NULL, use BCDirClose to release the object.
 */
bc_dir_handle BCDirOpen( const char *dirName )
{
	WIN32_FIND_DATAA fd;

	char path[ MAX_PATH ];
	sprintf( path, "%s\\*", dirName );

	bc_dir_handle hFind = FindFirstFileA( path, &fd );

	if ( hFind == INVALID_HANDLE_VALUE )
		return NULL;

	//
	//	We lose information about current diectory :(
	//
	if ( strcmp( fd.cFileName, "." ) != 0 )
	{
		assert( false );
		FindClose( hFind );
		return NULL;
	}

	return hFind;
}

const char *BCDBGetPath()
{
	return "D:\\pav\\BCDB";
}

/**
 * Reads next element from specified directory.
 * @param entry			Structure described directory entry.
 * @param dir			Handle of directory object opened by BCDirOpen.
 * @return              Return false if no more entries found.
 */
bool BCDirRead( bc_dirent *entry, bc_dir_handle dir )
{
	WIN32_FIND_DATAA fd;
	if ( !FindNextFileA( dir, &fd ) )
		return false;

	ULARGE_INTEGER uli;
	uli.LowPart = fd.nFileSizeLow;
	uli.HighPart = fd.nFileSizeHigh;

	return entry->set( fd.cFileName, uli.QuadPart, fd.dwFileAttributes );
}

/**
 * Closes directory object.
 * @param dir			Handle of directory object opened by BCDirOpen.
 */
void BCDirClose( bc_dir_handle dir )
{
	if ( NULL != dir )
		FindClose( dir );
}

/**
 * Checks if current entry is regular file.
 * @param entry			Structure described directory entry.
 * @return              Returns false for directory, true for files.
 *		Unix  entry->d_type == DT_REG
 *		Win32 ( entry->d_type & FILE_ATTRIBUTE_DIRCETORY ) == 0
 */
bool BCDirIsRegularFile( bc_dirent *entry )
{
	if ( NULL == entry )
	{
		assert( false );
		return false;
	}

	return ( entry->d_type & FILE_ATTRIBUTE_DIRECTORY ) == 0;
}
