/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 *
 */

#include "BCAlg.h"
#include "BCKernelDefs.h"

#include "crypto/cipher.h"
#include "crypto/mode.h"

#include <IOKit/IOReturn.h>
#include <IOKit/IOLib.h>


static inline int mode_id2tag(bc_mode_id id) {
	switch (id) {
		case kBCMode_CBC: return kCBC;
		case kBCMode_LRW: return kLRW;
		case kBCMode_XTS: return kXTS;
		//case kBCMode_NOOP: return kNOOP;
		default: BC_ASSERT(0);
	};
}


static inline bool is_alg_valid(const bc_alg* alg) {
	return (alg && alg->mode && alg->cipher);
}


static inline bool is_ctx_valid(const bc_crypto_ctx* ctx) {
	return (ctx && ctx->priv && is_alg_valid(&ctx->alg));
}

	
int bc_alg_init(const bc_cipher* cipher, bc_mode_id mode_id, bc_alg* out_alg)
{
	BC_ASSERT(cipher);
	BC_ASSERT(out_alg);
	
	int mode_tag = mode_id2tag(mode_id);
	
	const bc_mode* mode = bc_lookup_mode(mode_tag);
	if(!mode) {
		BC_LOG_DEBUG("Invalid mode id 0x%x\n", mode_tag);
		return kBCKernel_UnsupportedMode;
	}
	
	if(!mode->is_supported(cipher)) {
		BC_LOG_DEBUG("Cipher 0x%x does not support mode 0x%x\n", cipher->id, mode_tag);
		return kBCKernel_UnsupportedMode;
	}
	
	out_alg->cipher = cipher;
	out_alg->mode = mode;
	
	return kBCKernel_Success;
}


size_t bc_alg_key_size(const bc_alg* alg)
{
	BC_ASSERT(is_alg_valid(alg));
	return alg->mode->key_size(alg->cipher);
}


size_t bc_alg_block_size(const bc_alg* alg)
{
	BC_ASSERT(is_alg_valid(alg));
	return alg->cipher->block_size;
}


int bc_alg_set_key(const bc_alg* alg, const unsigned char* key, size_t keysize, bc_crypto_ctx* out_ctx)
{
	BC_ASSERT(is_alg_valid(alg));
	BC_ASSERT(key);
	
	if(keysize < bc_alg_key_size(alg)) {
		return kBCKernel_InvalidKey;
	}
	
	size_t ctx_size = alg->mode->context_size(alg->cipher);
	void* ctx = IOMalloc(ctx_size);
	if(!ctx) {
		return kBCKernel_NoMemory;
	}
	
	alg->mode->init_context(ctx, alg->cipher, key, keysize);
	
	out_ctx->alg.cipher = alg->cipher;
	out_ctx->alg.mode = alg->mode;
	out_ctx->priv = ctx;

	return kBCKernel_Success;
}


int bc_alg_clone_key(const bc_crypto_ctx* src, bc_crypto_ctx* dst)
{
	BC_ASSERT(is_ctx_valid(src));
	BC_ASSERT(dst);
	
	const bc_alg* alg = &src->alg;
	size_t ctx_size = alg->mode->context_size(alg->cipher);
	void* ctx = IOMalloc(ctx_size);
	if(!ctx) {
		return kBCKernel_NoMemory;
	}
	
	memcpy(ctx, src->priv, ctx_size);
	
	dst->alg = src->alg;
	dst->priv = ctx;
	
	return kBCKernel_Success;
}


void bc_alg_free_key(bc_crypto_ctx* ctx)
{
	if(ctx->priv == NULL) {
		return;
	}
	
	size_t ctx_size = ctx->alg.mode->context_size(ctx->alg.cipher);
	IOFree(ctx->priv, ctx_size);
	
	memset(ctx, 0, sizeof(*ctx));
}


/// Mode processing function type.
typedef void (*proc_func_t) (void* ctx, const bc_cipher* cipher, uint64_t iv, unsigned char* pool, unsigned char* data, size_t nbytes);
//            void (*decrypt) (void* context, const bc_cipher* cipher, uint64_t iv, unsigned char* pool, unsigned char* data, size_t nbytes);

void bc_alg_process(bc_crypto_ctx* ctx, bc_proc_type proc, uint64_t iv, unsigned char * pool, unsigned char* data, size_t nbytes)
{
	size_t nblocks;
	BC_ASSERT((nbytes % bc_alg_block_size(&ctx->alg)) == 0);
	
	nblocks = nbytes / bc_alg_block_size(&ctx->alg);
	if(nblocks == 0) {
		return ;
	}
	
	proc_func_t proc_func = (proc == kBCProcess_Decrypt) ? ctx->alg.mode->decrypt : ctx->alg.mode->encrypt;
	proc_func(ctx->priv, ctx->alg.cipher, iv, pool, data, nblocks);
}


int bc_encrypt(const bc_cipher* cipher, bc_mode_id mode_id, uint64_t iv, const unsigned char* key, unsigned keysize, unsigned char* pool, unsigned char* data, unsigned nbytes)
{
	bc_alg alg;
	int rc = bc_alg_init(cipher, mode_id, &alg);
	if(rc != kBCKernel_Success) {
		return rc;
	}
	
	bc_crypto_ctx ctx;
	rc = bc_alg_set_key(&alg, key, keysize, &ctx);
	if(rc != kBCKernel_Success) {
		return rc;
	}
	
	bc_alg_encrypt(&ctx, iv, pool, data, nbytes);
	bc_alg_free_key(&ctx);
	
	return rc;
}


int bc_decrypt(const bc_cipher* cipher, bc_mode_id mode_id, uint64_t iv, const unsigned char* key, unsigned keysize, unsigned char* pool, unsigned char* data, unsigned nbytes)
{
	bc_alg alg;
	int rc = bc_alg_init(cipher, mode_id, &alg);
	if(rc != kBCKernel_Success) {
		return rc;
	}
	
	bc_crypto_ctx ctx;
	rc = bc_alg_set_key(&alg, key, keysize, &ctx);
	if(rc != kBCKernel_Success) {
		return rc;
	}
	
	bc_alg_decrypt(&ctx, iv, pool, data, nbytes);
	bc_alg_free_key(&ctx);
	
	return rc;
}



