/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 *
 */

#ifndef BC_KERNEL_ALG_H
#define BC_KERNEL_ALG_H

#ifdef BC_KERNEL

//#include "BCKernel.h"
#include "bctypes.h"
//#include "BCKernelDefs.h"
#include "BCDefs.h"
//#include "crypto/cipher.h"

struct bc_cipher;
struct bc_mode;

/// Algorithm definition. Block cipher + mode.
typedef struct bc_alg
{
    const bc_mode*		mode;
    const bc_cipher*	cipher;
} bc_alg;


/**
 * Init algorithm configuration.
 * @cipher	Block cipher to use.
 * @mode	Encryption mode to use.
 * @out_alg	Initialized algorithm on success.
 * @return	kIOReturnUnsupported if mode does not support cipher.
 */
BC_KERNEL_PRIVATE
int bc_alg_init(const bc_cipher* cipher, bc_mode_id mode_id, bc_alg* out_alg);


/// Get algorithm total key size. Will return the size of the combined cipher/mode key in bytes.
BC_KERNEL_PRIVATE
size_t bc_alg_key_size(const bc_alg* alg);


/// Get algorithm block size in bytes.
BC_KERNEL_PRIVATE
size_t bc_alg_block_size(const bc_alg* alg);


/// Encryption context.
typedef struct bc_crypto_ctx
{
	bc_alg		alg;	// cipher + mode
	void*		priv;	// extended key and mode context.
} bc_crypto_ctx; 


/**
 * Set algorithm key and init crypto context.
 *
 * @alg		Configured algorithm to set the key for.
 * @key		Raw key data. Should hold at least enough bytes for this algorithm. Will not be used after return.
 * @keysize	Number of bytes in key buffer.
 * @pool	Random data pool. Required only for CBC mode algorithms.
 * @out_ctx	Initialized crypto context on success.
 *
 * @return	kIOReturnNoMemory on allocation failure.
 */
BC_KERNEL_PRIVATE
int bc_alg_set_key(const bc_alg* alg, const unsigned char* key, size_t keysize, bc_crypto_ctx* out_ctx);


/// Clone crypto context.
BC_KERNEL_PRIVATE
int bc_alg_clone_key(const bc_crypto_ctx* src, bc_crypto_ctx* dst);


/// Free algorithm crypto context.
BC_KERNEL_PRIVATE 
void bc_alg_free_key(bc_crypto_ctx* ctx);


/**
 * Process data using crypto context.
 * Algorithm, mode, key and, possibly, pool in context will be used to process data.
 *
 * @ctx		Context.
 * @data	Data to process inplace.
 * @nbytes	Number of bytes in data. If not alighed to alg block size remainder will not be processed.
 */
BC_KERNEL_PRIVATE
void bc_alg_process(bc_crypto_ctx* ctx, bc_proc_type proc, uint64_t iv, unsigned char* pool, unsigned char* data, size_t nbytes);


/// Encryption processing macro.
static inline void bc_alg_encrypt(bc_crypto_ctx* ctx, uint64_t iv, unsigned char* pool, unsigned char* data, size_t nbytes) {
	bc_alg_process(ctx, kBCProcess_Encrypt, iv, pool, data, nbytes);
}


/// Decryption processing macro.
static inline int bc_alg_decrypt(bc_crypto_ctx* ctx, uint64_t iv, unsigned char* pool, unsigned char* data, size_t nbytes) {
	bc_alg_process(ctx, kBCProcess_Decrypt, iv, pool, data, nbytes);
}


//
//	Quick interface.
//


// Init algorithm, set key, encrypt and free everything in one go.
BC_KERNEL_PRIVATE
int bc_encrypt(const bc_cipher* cipher, bc_mode_id mode_id, uint64_t iv, const unsigned char* key, unsigned keysize, unsigned char* pool, unsigned char* data, unsigned nbytes);


// Init algorithm, set key, decrypt and free everything in one go.
BC_KERNEL_PRIVATE
int bc_decrypt(const bc_cipher* cipher, bc_mode_id mode_id, uint64_t iv, const unsigned char* key, unsigned keysize, unsigned char* pool, unsigned char* data, unsigned nbytes);


#endif // BC_KERNEL
#endif // BC_KERNEL_ALG_H
