/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 */


#include "BCBlockMap.h"
#include "BCDefs.h"

#include <IOKit/IOLib.h>

static inline size_t bvec_bufsize(size_t nvecs) {
	return nvecs * sizeof(bc_bvec);
}


static inline size_t bvec_bufsize(bc_bmap* map) {
	return map->nvecs * sizeof(*map->extents);
}


bool bc_bmap_alloc(bc_bmap* map, size_t nvecs)
{
	map->extents = (bc_bvec*) IOMalloc(bvec_bufsize(nvecs));
	if(!map->extents) {
		return false;
	}
	
	map->nvecs = nvecs;
	memset(map->extents, 0, bvec_bufsize(nvecs));
	return true;
}


bool bc_bmap_load(bc_bmap* map, unsigned char* data)
{
	int nvecs = *(int*)data;
	if(!bc_bmap_alloc(map, nvecs)) {
		return false;
	}
	
	data += sizeof(int);
	
	bc_bvec* src = (bc_bvec*)data;
	bc_bvec* vec = bc_bmap_begin(map);
	bc_bvec* end = bc_bmap_end(map);
	while(vec != end) {
		vec->offset = src->offset;
		vec->length = src->length;
		
		++vec; ++src;
	}
	
#ifdef BC_DEBUG
	return bc_bmap_verify(map);
#else
	return true;
#endif
}


bool bc_bmap_copy(const bc_bmap* src, bc_bmap* dst)
{
	if(!bc_bmap_alloc(dst, src->nvecs)) {
		return false;
	}
	
	memcpy(dst->extents, src->extents, sizeof(*src->extents) * src->nvecs);
	return true;
}


uint64_t bc_bmap_nblocks(const bc_bmap* map)
{
	UInt64 nblocks = 0;
	bc_bvec* vec = bc_bmap_begin(map);
	bc_bvec* end = bc_bmap_end(map);
	
	while(vec != end) {
		nblocks += vec->length;
		++vec;
	}
	
	return nblocks;
}


bool bc_bmap_map_block(const bc_bmap* map, uint64_t block, bc_block_mapping* rc)
{
	BC_ASSERT(bc_bmap_verify(map));
	
	bc_bvec* vec = bc_bmap_begin(map);
	bc_bvec* end = bc_bmap_end(map);
	
	while(vec != end) {
		if(block < vec->length) {
			rc->vec = vec;
			rc->local_offset = block;
			return true;
		}
		
		block -= vec->length;
		++vec;
	}
	
	return false;
}


bool bc_bmap_map_range(const bc_bmap* map, uint64_t block, uint64_t nblocks, bc_block_mapping* first, bc_block_mapping* last)
{
	if(!bc_bmap_map_block(map, block, first)) {
		return 0;
	}
	
	if(!bc_bmap_map_block(map, block + nblocks, last)) {
		return 0;
	}
	
	return (last->vec - first->vec + 1);
}


void bc_bmap_free(bc_bmap* map)
{
	IOFree(map->extents, sizeof(*map->extents) * map->nvecs);
}


#ifdef BC_DEBUG
bool bc_bmap_verify(const bc_bmap* map)
{
	bc_bvec* vec = bc_bmap_begin(map);
	bc_bvec* end = bc_bmap_end(map);
	bc_bvec* next = vec + 1;
	uint64_t vec_end;
	
	while(next != end) {
		// next vector should always start after previous vector ends.
		vec_end = vec->offset + vec->length; // includes one extra block at the end so we atomatically check for non-adjacent vectors too.
		if(next->offset < vec_end) {
			return false;
		}
		
		vec = next;
		next = vec + 1;
	}
	
	return true;
}

void bc_bmap_log(const bc_bmap* map)
{
	IOLog("BCBlockMap: total mapped sectors %llu in %u vectors\n", bc_bmap_nblocks(map), map->nvecs);
	
	bc_bvec* vec = bc_bmap_begin(map);
	bc_bvec* end = bc_bmap_end(map);
	uint64_t bstart = 0;
	
	while(vec != end) {
		IOLog(" [%llu:%llu] -> [%llu:%llu]\n", bstart, bstart + vec->length, vec->offset, vec->offset + vec->length);
		bstart += vec->length;
		++vec;
	}
}
#endif



