/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * Block mapping data structures and algorithms.
 * Separated from disk driver code because controller user client needs to deserialize it.
 */

#ifndef BC_KERNEL
#error "This file can only be included in kernel code"
#endif // BC_KERNEL

#ifndef BC_KERNEL_BLOCK_MAP_H
#define BC_KERNEL_BLOCK_MAP_H

#include <sys/types.h>
//#include "BCKernel.h"
#include "BCKernelDefs.h"

/**
 * Description of single virtual block mapping.
 */
typedef struct 
{
	bc_bvec*	vec;		// vector that mapped virtual block.
	uint64_t	local_offset;	// offset into vector where block was mapped (vec->offset + local_offset == mapped block address).
} bc_block_mapping;


/// Get absolute mapped block address from mapping context.
static inline uint64_t bc_mapped_block_addr(bc_block_mapping* mm) {
	return mm->vec->offset + mm->local_offset;
}


/// Get mapped block vector length from mapping context.
static inline uint64_t bc_mapped_vec_length(bc_block_mapping* mm) {
	return mm->vec->length - mm->local_offset;
}


/// Allocate memory to store nvecs vectors in block map.
bool bc_bmap_alloc(bc_bmap* map, size_t nvecs);


/// Deserealize block map from memory buffer.
bool bc_bmap_load(bc_bmap* map, unsigned char* data);


/// Copy source black map vectors into destination.
bool bc_bmap_copy(const bc_bmap* src, bc_bmap* dst);


/// Get total number of mapped virtual sectors in this block map.
uint64_t bc_bmap_nblocks(const bc_bmap* map);


/// Get first map vector.
static inline bc_bvec* bc_bmap_begin(const bc_bmap* map) {
	return map->extents;
}


/// Get one past last map vector.
static inline bc_bvec* bc_bmap_end(const bc_bmap* map) {
	return (map->extents + map->nvecs);
}


/// Find a vector that maps requested virtual block address and init mapping context.
bool bc_bmap_map_block(const bc_bmap* map, uint64_t block, bc_block_mapping* rc);


/// Init mapping contexts for virtual block range start and end.
bool bc_bmap_map_range(const bc_bmap* map, uint64_t block, uint64_t nblocks, bc_block_mapping* first, bc_block_mapping* last);


/// Free memory that holds block map vectors.
void bc_bmap_free(bc_bmap* map);


#ifdef BC_DEBUG
/// Verify block map consistency.
bool bc_bmap_verify(const bc_bmap* map);

/// Log map structure.
void bc_bmap_log(const bc_bmap* map);
#endif // BC_DEBUG


#endif // BC_KERNEL_BLOCK_MAP_H

