/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 */

#include "BCController.h"
#include "BCDefs.h"
#include "BCScopeLock.h"
#include "BCAlg.h"
#include "BCDiskDriver.h"

#include "crypto/mode.h"
#include <libkern/OSAtomic.h>
#include <libkern/libkern.h>
#include <kern/clock.h>


/// Registered plugin structure.
typedef struct bc_plugin
{
	bc_cipher*	cipher;
	volatile SInt32		usecount;
	OSDictionary*		info_ref;
	bc_plugin*		next;
} bc_plugin;


static bool is_aesni_supported()
{
    uint64_t cpu_aesni = 0;
    
    asm ("movq $1, %%rax    \n\t"
         "cpuid             \n\t"
         :"=c"(cpu_aesni)
         :
         :"%rax","%rbx","%rcx","%rdx"
    );
    
    return (cpu_aesni & (0x1 << 25));
}

#pragma mark -
/////////////////////////////////////////////////////////////////////////////
//
//	Controller IOService lifetime.
//	
//
/////////////////////////////////////////////////////////////////////////////


OSDefineMetaClassAndStructors(BCController, IOService);

bool BCController::init(OSDictionary* props)
{
	if(kIOReturnSuccess != this->init_plugin_list()) {
		return false;
	}
	
	this->init_key_table();
	
	if(!IOService::init(props)) {
		return false;
	}
	
	BC_IOLOG_TRACE;

	return true;
}


void BCController::free()
{
	BC_IOLOG_TRACE;

	BC_ASSERT(this->m_plugins.head == NULL);
	this->free_plugin_list();
	this->free_key_table();

	IOService::free();
}


bool BCController::start(IOService* provider)
{
	BC_IOLOG_TRACE;

	if(kIOReturnSuccess != this->set_plugins_property()) {
		BC_IOLOG("failed setting plugin property set\n");
		return false;
	}
	
	if(!IOService::start(provider)) {
		BC_IOLOG("IOService::start failed\n");
		return false;
	}
	
	this->registerService();
    
    bool cpu_aesni = is_aesni_supported();
    OSBoolean* aesni_prop = OSBoolean::withBoolean(cpu_aesni);
    if(!aesni_prop) {
        return false;
    }
    
    this->setProperty(kBCControllerAESNIAvailableKey, aesni_prop);
    
	return true;
}


void BCController::stop(IOService* provider)
{
	BC_IOLOG_TRACE;
	
	IOService::stop(provider);
}


bool BCController::terminate(IOOptionBits opts /* = 0*/)
{
	BC_IOLOG_TRACE;
	
	IORWLockRead(m_plugins.lock);
	bool rc = (m_plugins.head == NULL);
	IORWLockUnlock(m_plugins.lock);
	
	if(!rc) {
		BC_IOLOG_DEBUG("Has loaded plugins upon terminate\n");
		return false;
	}
	
	return IOService::terminate(opts);
}



#pragma mark -
/////////////////////////////////////////////////////////////////////////////
//
//	Plugin management.
//
/////////////////////////////////////////////////////////////////////////////


static const char* g_mode_keys[] = 
{
	kBCPluginCBCKeySizeKey,		// kCBC
	kBCPluginLRWKeySizeKey,		// kLRW
	kBCPluginXTSKeySizeKey,		// kXTS
};

#define BC_SIZEOF_BITS(var) ( sizeof( (var) ) * 8 )

bc_plugin* BCController::alloc_init_plugin(const bc_cipher* cipher)
{
	enum {
		kBCPluginProp_Id = 0,
		kBCPluginProp_Name,
		kBCPluginProp_KeyLength,
		kBCPluginProp_BlockSize,
		kBCPluginProp_CBCKeySize,
		kBCPluginProp_LRWKeySize,
		kBCPluginProp_XTSKeySize,
		
		kBCPluginProp_Total,
	};
	
	
	bc_plugin* plug = (bc_plugin*) IOMalloc(sizeof(bc_plugin));
	if(!plug) {
		return NULL;
	}
	

	// allocate enough size to hold all properties.
	OSDictionary* info = OSDictionary::withCapacity(kBCPluginProp_Total);
	if(!info) {
		free_plugin(plug);
		return NULL;
	}
		
	bool info_filled = false;
	do {
		OSNumber* id = OSNumber::withNumber(cipher->id, BC_SIZEOF_BITS(cipher->id));
		if(!id) break;
		info->setObject(kBCPluginIdKey, id);
		id->release();
		
		OSNumber* key_size = OSNumber::withNumber(cipher->key_size, BC_SIZEOF_BITS(cipher->key_size));
		if(!key_size) break;
		info->setObject(kBCPluginKeyLengthKey, key_size);
		key_size->release();
		
		OSNumber* block_size = OSNumber::withNumber(cipher->block_size, BC_SIZEOF_BITS(cipher->block_size));
		if(!block_size) break;
		info->setObject(kBCPluginBlockSizeKey, block_size);
		block_size->release();
		
		OSString* name = OSString::withCString(cipher->name);
		if(!name) break;
		info->setObject(kBCPluginNameKey, name);
		name->release();


		OSNumber* zero = OSNumber::withNumber(0x0llu, 32);
		if(!zero) break;

		info_filled = true; // can be reset to false further.

		OSNumber* mode_key_size;
		const bc_mode* mode;
		unsigned i;

		for(i = 0; i < kBCMode_Total; ++i) {
			mode = bc_lookup_mode(i);
			if(mode->is_supported(cipher)) {
				mode_key_size = OSNumber::withNumber(mode->key_size(cipher), 32);
				if(!mode_key_size) {
					info_filled = false;
					break;
				}
				
				info->setObject(g_mode_keys[i], mode_key_size);
				mode_key_size->release();
			} else {
				info->setObject(g_mode_keys[i], zero);
			}
		}
		
		zero->release();
	
	} while(0); // once
	
	if(!info_filled) {
		info->release();
		free_plugin(plug);
		return NULL;
	}
		
	plug->cipher = (bc_cipher*)cipher;
	plug->usecount = 0;
	plug->next = NULL;
	plug->info_ref = info; // extra reference to info dictionary will be released at free_plugin.
	
	return plug;
}


void BCController::free_plugin(bc_plugin* plug)
{
	plug->info_ref->release();
	memset(plug, 0, sizeof(*plug));
	IOFree(plug, sizeof(*plug));
}


IOReturn BCController::init_plugin_list()
{
	m_plugins.lock = IORWLockAlloc();
	if(!m_plugins.lock) {
		return kIOReturnNoMemory;
	}
	
	m_plugins.head = NULL;
	return kIOReturnSuccess;
}


void BCController::free_plugin_list()
{
	IORWLockFree(m_plugins.lock);
	m_plugins.head = NULL;
}


bc_plugin* BCController::find_plugin(int id)
{
	bc_plugin* plug = m_plugins.head;
	while(plug) {
		if(plug->cipher->id == id) {
			return plug;
		}
	
		plug = plug->next;
	}
	
	return NULL;
}


bool BCController::add_plugin_unique(bc_plugin* plug)
{
	bc_plugin* other = this->find_plugin(plug->cipher->id);
	if(other) 
		return false;
	
	plug->next = m_plugins.head;
	m_plugins.head = plug;
	
	return true;
}


bc_plugin* BCController::get_plugin(int id)
{
	BCScopeReadLock lock(m_plugins.lock);
	bc_plugin* plug = this->find_plugin(id);
	if(!plug) {
		return NULL;
	}
	
	OSIncrementAtomic(&plug->usecount);
	return plug;
}


void BCController::put_plugin(int id)
{
	BCScopeReadLock lock(m_plugins.lock);
	bc_plugin* plug = this->find_plugin(id);
	if(!plug) {
		return;
	}
	
	OSDecrementAtomic(&plug->usecount);
}


bool BCController::remove_plugin(bc_plugin* plug) 
{
	if(m_plugins.head == NULL) {
		return false;
	}
	
	if(plug == m_plugins.head) {
		m_plugins.head = plug->next; 
	} else {
		bc_plugin* prev = m_plugins.head;
		while((prev->next != plug) && (prev)) {
			prev = prev->next;
		}
		
		if(!prev) {
			return false;
		}
		
		prev->next = plug->next;
	}
	
	plug->next = NULL;
	return true;
}

bool BCController::publish_plugin(bc_plugin* plug)
{
	OSArray* prop = get_plugins_property();
	BC_ASSERT(prop);

	OSArray* prop_copy = OSArray::withArray(prop);
	if(!prop_copy) {
		return false;
	}
	
	prop_copy->setObject(plug->info_ref);
	return this->setProperty(kBCControllerPluginsPropertyKey, prop_copy);
}


bool BCController::unpublish_plugin(bc_plugin* plug)
{
	OSArray* prop = get_plugins_property();
	BC_ASSERT(prop);
	
	OSArray* prop_copy = OSArray::withArray(prop);
	if(!prop_copy) {
		return false;
	}
	
	int idx = prop_copy->getNextIndexOfObject(plug->info_ref, 0);
	BC_ASSERT(idx != -1);
	prop_copy->removeObject(idx);
	return this->setProperty(kBCControllerPluginsPropertyKey, prop_copy);
}


int BCController::register_plugin(const bc_cipher* cipher, bc_plugin_ref* out_ref)
{
	BC_IOLOG_DEBUG("registering cipher plugin %s (0x%x)\n", cipher->name, cipher->id);
	*out_ref = NULL;
	
	bc_plugin* plug = alloc_init_plugin(cipher);
	if(!plug) {
		BC_IOLOG("no memory to alloc new plugin\n");
		return kIOReturnNoMemory;
	}

	// lock protects dublicate lookup, insert and publishing.
	BCScopeWriteLock(m_plugins.lock);

	IOReturn rc = kIOReturnSuccess;
	do {
		// check for plugin dublicate should have been an assert if we wouldn't publish plugin interface.
		// still it is a primarily a que for debugging perposes so we check it here after allocation.
		// user space should check plugin and print errors if registering went through it.
		if(!this->add_plugin_unique(plug)) {
			BC_IOLOG("cipher id 0x%x is already registered\n", cipher->id);
			rc = kIOReturnBusy;
			break;
		}

		// publish inserted plugin in property array to make it visible to clients.
		if(!this->publish_plugin(plug)) {
			BC_IOLOG("failed publishing plugin\n");
			rc = kIOReturnNoMemory;
			break;
		}
		
	} while(0);
	
	if(kIOReturnSuccess != rc) {
		this->remove_plugin(plug);
		this->free_plugin(plug);
		return rc;
	}
	
	*out_ref = plug;
	return kIOReturnSuccess;
}


int BCController::unregister_plugin(bc_plugin* plug)
{
	BC_IOLOG_DEBUG("unregistering cipher plugin %s (0x%x)\n", plug->cipher->name, plug->cipher->id);

	if(!plug) {
		return kIOReturnBadArgument;
	}
	
	BCScopeWriteLock(m_plugins.lock);
	
	if(plug->usecount > 0) {
		BC_IOLOG("plugin in use\n");
		return kIOReturnBusy;
	}
	
	this->unpublish_plugin(plug);
	this->remove_plugin(plug);
	this->free_plugin(plug);

	return kIOReturnSuccess;
}



#pragma mark -
/////////////////////////////////////////////////////////////////////////////
//
//	Key storage management.
//
/////////////////////////////////////////////////////////////////////////////


/// Key table slot states.
enum {
	kSlot_StateFree		= 0x00,	// Slot is free for new key.
	kSlot_StateOccupied	= 0x01,	// Slot is occupied.
};

enum {
	kSlot_Total		= 32,
};


// Key stored in each slot.
typedef struct {
	unsigned char *		data;
	size_t		size;
} bc_key;


// Key table slot.
typedef struct {
	bc_key			key;
	volatile UInt32		status;
} bc_key_slot;


// Static key table memory.
static bc_key_slot g_key_table[kSlot_Total];


static inline void shred(void* mem, size_t nbytes) {
	memset(mem, 0, nbytes);
	IOFree(mem, nbytes);
}


static inline void shred_key(bc_key* key) {
	shred(key->data, key->size);
	memset(key, 0, sizeof(*key));
}


void BCController::init_key_table()
{
	memset(g_key_table, 0, sizeof(g_key_table));
}


void BCController::free_key_table()
{
	int handle;
	bc_key_slot* slot;
	
	for(handle = 0; handle < kSlot_Total; ++handle) {
		slot = &g_key_table[handle];
		if(OSCompareAndSwap(kSlot_StateOccupied, kSlot_StateFree, &slot->status)) {
			BC_IOLOG_DEBUG("Warning: key slot 0x%x was occupied\n", handle);
			shred_key(&slot->key);
		}
	}
	
	memset(g_key_table, 0, sizeof(g_key_table));
}


// The only thing we protect against is concurrent key store into the same slot.
// So if two threads will try to store two keys they will be guaranteed to get two different key handles back.
IOReturn BCController::store_key(bc_ctl_key* key_in, bc_key_handle* handle_out)
{
	int handle;
	bc_key_slot* slot;
	
	for(handle = 0; handle < kSlot_Total; ++handle) {
		slot = &g_key_table[handle];
		
		// atomically mark slot as occupied if it is currently free.
		// 
		if(OSCompareAndSwap(kSlot_StateFree, kSlot_StateOccupied, &slot->status)) {
			break;
		}
	}
	
	if(handle == kSlot_Total) {
		BC_IOLOG("No free key slots left\n");
		return kBCKernel_NoMemory;
	}
	
	bc_key* new_key = &slot->key;
	
	new_key->data = (unsigned char*)IOMalloc(key_in->size);
	if(!new_key->data) {
		BC_IOLOG("No memory to allocate new key\n");
		goto ALLOC_FAILED;
	}
	
	memcpy(new_key->data, key_in->addr.ptr, key_in->size);
	
	new_key->size = key_in->size;
	*handle_out = handle;
	
	BC_IOLOG_DEBUG("Stored new key with handle %d\n", handle);
	return kIOReturnSuccess;
	
	
ALLOC_FAILED:
	
	OSCompareAndSwap(kSlot_StateOccupied, kSlot_StateFree, &slot->status);
	return kBCKernel_NoMemory;
}


IOReturn BCController::free_key(int handle)
{
	if(handle > kSlot_Total) {
		return kBCKernel_InvalidKey;
	}
	
	
	// once we mark slot as free we can then release memory of some key that already was stored in its place.
	// to avoid this, make copy of the key data pointers.
	bc_key_slot* slot = &g_key_table[handle];
	bc_key copy = slot->key;
	
	if(!OSCompareAndSwap(kSlot_StateOccupied, kSlot_StateFree, &slot->status)) {
		BC_IOLOG("requested handle 0x%x is free\n", handle);
		return kBCKernel_InvalidKey;
	}
	
	shred(copy.data, copy.size);
	BC_IOLOG_DEBUG("freed key handle 0x%x\n", handle);
	return kBCKernel_Success;
}



#pragma mark -


IOReturn BCController::init_alg(bc_alg_id alg_id, bc_mode_id mode_id, bc_alg* out_alg)
{
	bc_plugin* plug = this->get_plugin(alg_id);
	if(!plug) {
		BC_IOLOG_DEBUG("cipher id 0x%x not found\n", alg_id);
		return kBCKernel_UnknownAlg;
	}
	
	int rc = bc_alg_init(plug->cipher, mode_id, out_alg);
	if(kBCKernel_Success != rc) {
		this->put_plugin(alg_id);
	}
	
	return rc;
}


const struct bc_cipher* BCController::get_cipher(bc_alg_id alg_id)
{
	bc_plugin* plug = this->get_plugin(alg_id);
	if(!plug) {
		BC_IOLOG_DEBUG("cipher id 0x%x not found\n", alg_id);
		return NULL;
	}
	
	return plug->cipher;
}


void BCController::put_cipher(const struct bc_cipher* cipher)
{
	this->put_plugin(cipher->id);
}


void BCController::release_alg(bc_alg* alg)
{
	this->put_plugin(alg->cipher->id);
}


IOReturn BCController::set_key(bc_alg* alg, int handle, bc_crypto_ctx* ctx)
{
	if(handle >= kSlot_Total) {
		return kBCKernel_InvalidKey;
	}
	
	bc_key_slot* slot = &g_key_table[handle];
	if(slot->status == kSlot_StateFree) {
		return kBCKernel_InvalidKey;
	}
	
	return bc_alg_set_key(alg, slot->key.data, slot->key.size, ctx);
}


IOReturn BCController::process(bc_ctl_process* arg)
{
	IOReturn rc;
	bc_alg alg;
	bc_crypto_ctx ctx;
	
	rc = this->init_alg(arg->cipher, arg->mode, &alg);
	if(kBCKernel_Success != rc) {
		BC_IOLOG("failed initializing algorithm: 0x%x\n", rc);
		return rc;
	}
	
	if(arg->nbytes % bc_alg_block_size(&alg)) {
		BC_IOLOG("buffer size is unalighed to cipher block size\n");
		this->release_alg(&alg);
		return kBCKernel_InvalidBuffer;
	}
	
	if(arg->with_key) {
		bc_ctl_key* key = &arg->key.data;
		rc = bc_alg_set_key(&alg, (const unsigned char*)key->addr.ptr, key->size, &ctx);
	} else {
		rc = this->set_key(&alg, arg->key.handle, &ctx);
	}
	
	if(kBCKernel_Success != rc) {
		BC_IOLOG("failed setting algorithm key with 0x%x\n", rc);
		this->release_alg(&alg);
		return rc;
	}
	
	bc_alg_process(&ctx, arg->proc, arg->iv, (unsigned char*)arg->pool.ptr, (unsigned char*)arg->data.ptr, arg->nbytes);
	
	bc_alg_free_key(&ctx);
	this->release_alg(&alg);
	return kIOReturnSuccess;
}


IOReturn BCController::benchmark(bc_ctl_benchmark* arg_in, uint64_t* time_out)
{
	bc_alg alg;
	IOReturn rc = this->init_alg(arg_in->alg_id, arg_in->mode_id, &alg);
	if(kBCKernel_Success != rc) {
		BC_IOLOG("failed initializing algorithm: 0x%x\n", rc);
		return rc;
	}
	
	unsigned bufsize = arg_in->bufsiz_mb * 1024 * 1024; // buffer size in megabytes.
	unsigned char* data = (unsigned char*) IOMalloc(bufsize);
	if(!data) {
		BC_IOLOG("failed allocating benchmark data\n");
		this->release_alg(&alg);
		return kIOReturnNoMemory;
	}
	
	memset(data, 0, bufsize);
	unsigned char key[0xff] = {0};
	unsigned char pool[kBCKernel_PoolSize] = {0};
	
	bc_crypto_ctx ctx;
	rc = bc_alg_set_key(&alg, key, sizeof(key), &ctx);
	if(kBCKernel_Success != rc) {
		BC_IOLOG("failed setting algorithm key with 0x%x\n", rc);
		this->release_alg(&alg);
		IOFree(data, bufsize);
		return rc;
	}

	uint64_t iv = 0;
	uint64_t start_time = 0, end_time = 0, mean_time = 0;
	for(unsigned i = 0; i < arg_in->nreps; ++i) {
		clock_get_uptime(&start_time);
		bc_alg_encrypt(&ctx, iv, pool, data, bufsize);
		clock_get_uptime(&end_time);
		
		mean_time += (end_time - start_time);
	}
	
	*time_out = mean_time / arg_in->nreps;
	
	IOFree(data, bufsize);
	bc_alg_free_key(&ctx);
	this->release_alg(&alg);
	
	return kIOReturnSuccess;
}

/*
IOReturn BCController::enable_aeshw(int do_enable)
{
    if(!AESHW_Is_Supported()) {
        return kBCKernel_Unsupported;
    }
    
    BCScopeReadLock lock(m_plugins.lock);
	bc_plugin* plug = this->find_plugin(BC_AESHW_ID);
	if(!plug) {
		return kBCKernel_UnknownAlg;
	}
	
    // This will fail if aes usecount is not 0.
    if(!OSCompareAndSwap(0, 1, (volatile UInt32*) &plug->usecount)) {
        return kBCKernel_Busy;
    }
    
    bc_cipher* cipher = plug->cipher;
    if(do_enable) {
        cipher->encrypt_ecb = AESHW_Encrypt_ECB;
        cipher->decrypt_ecb = AESHW_Decrypt_ECB;
    } else {
        cipher->encrypt_ecb = Encrypt_ECB;
        cipher->decrypt_ecb = Decrypt_ECB;
    }
    
    OSDecrementAtomic(&plug->usecount);
    return kBCKernel_Success;
}
*/


IOReturn BCController::set_disk(bc_ctl_disk* disk)
{
	BC_IOLOG_DEBUG("setting new disk: alg id 0x%x, mode id 0x%x, file path %s, read only %d\n", 
		       disk->alg_id, 
		       disk->mode_id, 
		       (const char*)disk->path.ptr,
		       disk->ro);
	
	IOReturn rc = kIOReturnSuccess;

	// get algorithm context first.
	bc_alg alg;
	bc_crypto_ctx ctx;
	BCDiskDriver* driver = NULL;

	rc = this->init_alg(disk->alg_id, disk->mode_id, &alg);
	if(kBCKernel_Success != rc) {
		BC_IOLOG("failed initializing algorithm with 0x%x\n", rc);
		return rc;
	}
	
	do {
		rc = this->set_key(&alg, disk->key, &ctx);
		if(kBCKernel_Success != rc) {
			BC_IOLOG("failed initializing algorithm with 0x%x\n", rc);
			break;
		}
		
		// setup new disk driver instance.
		driver = new BCDiskDriver;
		if(!driver) {
			BC_IOLOG("Failed allocating new bc disk driver instance\n");
			rc = kBCKernel_NoMemory;
			break;
		}
		
		// TODO: signal 'file not found' error for bad container file paths.
		bc_bmap bmap;
		bmap.nvecs = disk->nvecs;
		bmap.extents = (bc_bvec*)disk->extents.ptr;
		
		if(!driver->init((const char*)disk->path.ptr, &ctx, &bmap, (const char*)disk->pool.ptr, disk->ro, disk->auto_mount)) {
			BC_IOLOG("Failed initializing driver\n");
			rc = kIOReturnInternalError;
			break;
		}
		
		if(!driver->attach(this)) {
			BC_IOLOG("Failed attaching driver\n");
			rc = kIOReturnNotAttached;
			break;
		}	
		
		if(!driver->start(this)) {
			BC_IOLOG("Failed starting driver\n");
			rc = kIOReturnInternalError;
			break;
		}

		driver->release();
		return kIOReturnSuccess;
		
	} while (0);
	
	// failure.
	this->release_alg(&alg);
	bc_alg_free_key(&ctx);
	if(driver) driver->release(); // removing last reference in any error case.
	return rc;
}


void BCController::remove_disk(BCDiskDriver* drv)
{
	BC_IOLOG_TRACE;
	this->release_alg(drv->algorithm());
}



#pragma mark -
/////////////////////////////////////////////////////////////////////////////
//
//	Plugin C hooks.
//	Prototypes declared in BCKernelPlugin.h
//
/////////////////////////////////////////////////////////////////////////////



static BCController* get_instance()
{
	OSDictionary* services = IOService::serviceMatching(kBCControllerMatchKey);
	if(!services) {
		return NULL;
	}
	
	OSIterator* iter = IOService::getMatchingServices(services);
	if(!iter) {
		services->release();
		return NULL;
	}
	
	BCController* serv = NULL;
	OSObject* cur;
	
	while((cur = iter->getNextObject()) != NULL) {
		if( (serv = OSDynamicCast(BCController, cur)) ) {
			break;
		}
	}
	
	services->release();	
	iter->release();
	
	return serv;
}


extern "C" 
int bc_register_plugin(const bc_cipher* cipher, bc_plugin_ref* out_ref)
{
	BCController* serv = get_instance();
	if(!serv) {
		return kIOReturnInternalError;
	}
	
	return serv->register_plugin(cipher, out_ref);
}


extern "C" 
int bc_unregister_plugin(bc_plugin_ref ref)
{
	BCController* serv = get_instance();
	if(!serv) {
		return kIOReturnInternalError;
	}
	
	return serv->unregister_plugin(ref);
}



