/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 */

#ifndef BC_KERNEL_CONTROLLER_H
#define BC_KERNEL_CONTROLLER_H

/// IORegistry lookup service match key.
#define kBCControllerMatchKey		"BCController"

/// Plugin array property
#define kBCControllerPluginsPropertyKey	"plugins"

#define kBCControllerAESNIAvailableKey  "BCAESNIAvailable"  // AES-NI is available flag

/// Cipher plugin properties.
#define kBCPluginIdKey			"id"
#define kBCPluginNameKey 		"name"
#define kBCPluginKeyLengthKey		"key-length"
#define kBCPluginCBCKeySizeKey		"cbc-key-size"
#define kBCPluginLRWKeySizeKey		"lrw-key-size"
#define kBCPluginXTSKeySizeKey		"xts-key-size"
#define kBCPluginBlockSizeKey		"block-size"
#define kBCPluginSupportedModesKey	"modes"

#define kBCPluginCBCModeKey		"CBC"
#define kBCPluginLRWModeKey		"LRW"
#define kBCPluginXTSModeKey		"XTS"


#ifdef BC_KERNEL

#include <IOKit/IOService.h>
#include "BCKernelPlugin.h"
#include "BCCTL.h"

struct bc_alg;
struct bc_crypto_ctx;
class BCDiskDriver;

typedef struct 
{
	bc_plugin_ref	ref;
	IORWLock*	lock;
} bc_plugin_list;


class BCController : public IOService
{
OSDeclareDefaultStructors(BCController)
public:
	
	/// IOKit service constructor.
	bool init(OSDictionary* props);
	
	/// IOKit service destructor.
	void free();
	
	/**
	 * Start servicing provider.
	 * Activates and publishes controller to start accepting user requests.
	 * @provider	ioresources provider.
	 */
	bool start(IOService* privider);
	
	
	/**
	 * Start termination sequence.
	 * By default will fail if any client has this controller open.
	 * Finalize and then stop will be called after that.
	 */
	bool terminate(IOOptionBits opts = 0);

	
	/**
	 * Stop servicing provider.
	 * Deactivate controller and stop accepting requests.
	 */
	void stop(IOService* provider);
	
	
	/**
	 * Register new cipher plugin.
	 * @cipher	Cipher definitions.
	 * @out_ref	Returned plugin reference.
	 */
	IOReturn register_plugin(const bc_cipher* cipher, bc_plugin_ref* out_ref);
	
	
	/**
	 * Unregister cipher plugin.
	 * @ref		Plugin reference.
	 */
	IOReturn unregister_plugin(bc_plugin_ref ref);

	
	/**
	 * Store new encryption key in controller memory.
	 */
	IOReturn store_key(bc_ctl_key* key_in, bc_key_handle* handle_out);
	
	
	/**
	 * Free key from controller memory.
	 */
	IOReturn free_key(bc_key_handle handle);

	
	/// Process data.
	IOReturn process(bc_ctl_process* arg_in);

	
	/// Benchmark algorithm configuration.
	IOReturn benchmark(bc_ctl_benchmark* arg_in, uint64_t* time_out);
	
    // Enable AES-NI hardware acceleration
    //IOReturn enable_aeshw(int do_enable);
	
	/// Set new virtual disk.
	IOReturn set_disk(bc_ctl_disk* disk);
	
	
	/// Remove virtual disk.
	void remove_disk(BCDiskDriver* drv);
	
protected:
	
	typedef struct {
		bc_plugin*	head;
		IORWLock*	lock;
	} bc_plugin_list;
	

	/// Initialize plugin list.
	IOReturn init_plugin_list();
	
	/// Release plugin list resources.
	void free_plugin_list();
	
	/// Allocate and initialize new plugin.
	bc_plugin* alloc_init_plugin(const bc_cipher* cipher);

	/// Free plugin and it's resources.
	void free_plugin(bc_plugin* plug);

	/// Lookup plugin list for matching cipher id.
	bc_plugin* find_plugin(int id);

	
	bc_plugin* get_plugin(int id);
	
	const struct bc_cipher* get_cipher(bc_alg_id alg_id);
	
	void put_cipher(const struct bc_cipher* cipher);

	void put_plugin(int id);
	

	bool add_plugin_unique(bc_plugin* plug);

	bool remove_plugin(bc_plugin* plug);
	
	
	/// Macro to get published plugins property set.
	OSArray* get_plugins_property() {
		return OSDynamicCast(OSArray, this->getProperty(kBCControllerPluginsPropertyKey));
	}
	
	
	/// Macro to allocate and set new plugins property set if it does not exist.
	IOReturn set_plugins_property() {
		if(this->get_plugins_property()) {
			return kIOReturnSuccess;
		}
		
		OSArray* prop = OSArray::withCapacity(10);
		if(!prop) {
			return kIOReturnNoMemory;
		}
		
		bool rc = this->setProperty(kBCControllerPluginsPropertyKey, prop);
		prop->release();
		
		return rc ? kIOReturnSuccess : kIOReturnNoMemory;
	}
	
	
	/// Publish plugin to user space.
	bool publish_plugin(bc_plugin* plug);
	
	/// Unpublish plugin from user space.
	bool unpublish_plugin(bc_plugin* plug);
	
	
	IOReturn init_alg(bc_alg_id alg_id, bc_mode_id mode_id, bc_alg* out_alg);

	IOReturn set_key(bc_alg* alg, int handle, bc_crypto_ctx* ctx);
	
	void release_alg(bc_alg* alg);
	

	void init_key_table();
	void free_key_table();

private:
	
	bc_plugin_list	m_plugins;
};

#endif // BC_KERNEL
#endif // BC_KERNEL_CONTROLLER_H

