/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 *
 */

#include "BCControllerClient.h"
#include "BCController.h"
#include "BCAutoBuffer.h"
#include "BCDefs.h"
#include "BCKernelDefs.h"


OSDefineMetaClassAndStructors(BCControllerUserClient, IOUserClient);


bool BCControllerUserClient::start(IOService* provider)
{
	m_controller = OSDynamicCast(BCController, provider);
	if(!m_controller) {
		return false;
	}
	
	//m_controller->open(this);
	bool rc = IOUserClient::start(provider);
	if(rc) {
		BC_IOLOG_DEBUG("started\n");
	}
	
	return rc;
}


bool BCControllerUserClient::initWithTask(task_t owningTask, void* securityToken, UInt32 type, OSDictionary* properties)
{
	return IOUserClient::initWithTask(owningTask, securityToken, type, properties);
}


IOReturn BCControllerUserClient::clientClose( void )
{
	BC_IOLOG_TRACE;
	this->terminate();
	return kIOReturnSuccess;
}


IOReturn BCControllerUserClient::clientDied( void )
{
	BC_IOLOG_TRACE;
	return IOUserClient::clientDied();
}

IOReturn BCControllerUserClient::externalMethod(uint32_t selector, IOExternalMethodArguments* arguments, IOExternalMethodDispatch* dispatch, OSObject* target, void* reference)
{
	BC_IOLOG_DEBUG("User client request code 0x%x\n", selector);
	
	if(selector == kBCKernel_CTL) {

		struct IOExternalMethodDispatch d = { 
			s_dispatch_ctl, 
			arguments->scalarInputCount,
			arguments->structureInputSize,
			arguments->scalarOutputCount,
			arguments->structureOutputSize 
		};
			
		target = this;
		return IOUserClient::externalMethod(selector, arguments, &d, target, reference);

	} else {
		// called with invalid selector? leave as it is for now.
		return IOUserClient::externalMethod(selector, arguments, dispatch, target, reference);
	}
	
	
	// if dispatch is set parent will call associated method for us.
}


IOReturn BCControllerUserClient::s_dispatch_ctl(OSObject * target, void * reference, IOExternalMethodArguments * arguments)
{
	uint64_t ctlcode = *arguments->scalarInput;
	BCControllerUserClient* obj = (BCControllerUserClient*) target;
	
	BC_LOG_DEBUG("dispatching ioctl code 0x%llx\n", ctlcode);
	switch (ctlcode) {
		case kBCCTL_StoreKey:	
		{
			bc_ctl_key* key = (bc_ctl_key*) arguments->structureInput;
			bc_key_handle* out_handle = (bc_key_handle*) arguments->structureOutput;
			return obj->store_key(key, out_handle);
		}
			
		case kBCCTL_FreeKey:	
		{
			bc_key_handle handle = *(bc_key_handle*) arguments->structureInput;
			return obj->free_key(handle);
		}
			
		case kBCCTL_Process:	
		{
			bc_ctl_process* data = (bc_ctl_process*) arguments->structureInput;
			return obj->process(data);
		}
			
		case kBCCTL_SetDisk:	
		{
			bc_ctl_disk* disk = (bc_ctl_disk*) arguments->structureInput;
			return obj->set_disk(disk);
		}
			
		case kBCCTL_Benchmark:	
		{
			bc_ctl_benchmark* arg = (bc_ctl_benchmark*) arguments->structureInput;
			uint64_t* out_time = (uint64_t*) arguments->structureOutput;
			return obj->benchmark(arg, out_time);
		}

            /*
        case kBCCTL_EnableAESNI:
        {
            int flag = *(int*) arguments->structureInput;
            return obj->enable_aeshw(flag);
        }
            */
            
		default: return kBCCipher_InternalError;
	};
}


IOReturn BCControllerUserClient::store_key(bc_ctl_key* key, bc_key_handle* out_handle)
{
	IOReturn rc;
	
	BCAutoBuffer keybuf;
	
	// allocate required kernel memory.
	rc = keybuf.init_from_user(key->addr.ptr, key->size);
	if(kIOReturnSuccess != rc) {
		BC_IOLOG("failed to init key buffer with 0x%x bytes: 0x%x\n", key->size, rc);
		return rc;
	}
	
	key->addr.ptr = keybuf.mem();
	return this->m_controller->store_key(key, out_handle);
}


IOReturn BCControllerUserClient::free_key(bc_key_handle key_handle)
{
	return this->m_controller->free_key(key_handle);
}


IOReturn BCControllerUserClient::process(bc_ctl_process* arg)
{
	IOReturn rc;
	
	BCAutoBuffer databuf;
	BCAutoBuffer poolbuf;
	
	rc = databuf.init_from_user(arg->data.ptr, arg->nbytes);
	if(kIOReturnSuccess != rc) {
		BC_IOLOG("failed to init data buffer from user space\n");
		return rc;
	}
	
	if(arg->pool.ptr) {
		rc = poolbuf.init_from_user(arg->pool.ptr, kBCKernel_PoolSize);
		if(kIOReturnSuccess != rc) {
			BC_IOLOG("failed to init pool buffer from user space\n");
			return rc;
		}
	}
	
	void* out_data = arg->data.ptr;
	arg->data.ptr = databuf.mem();
	arg->pool.ptr = poolbuf.mem();
	
	if(arg->with_key) {
		BCAutoBuffer keybuf;
		bc_ctl_key* key = &arg->key.data;
		
		rc = keybuf.init_from_user(key->addr.ptr, key->size);
		if(kIOReturnSuccess != rc) {
			BC_IOLOG("failed to init key buffer from user space sized %d: 0x%x\n", key->size, rc);
			return rc;
		}
		
		key->addr.ptr = keybuf.mem();
		rc = m_controller->process(arg); // to prevent autobuffer from releasing the key.
	} else {
		rc = m_controller->process(arg);
	}
		
	if(kIOReturnSuccess != rc) {
		BC_IOLOG("process failed with 0x%x\n", rc);
		return rc;
	}
	
	rc = databuf.copyout(out_data);
	if(kIOReturnSuccess != rc) {
		BC_IOLOG("copyout returned 0x%x\n", rc);
		return rc;
	}
	
	return kIOReturnSuccess;
}


IOReturn BCControllerUserClient::set_disk(bc_ctl_disk* disk)
{
	IOReturn rc;

	BCAutoBuffer namebuf;
	rc = namebuf.init_from_user(disk->path.ptr, disk->pathlen + 1);
	disk->path.ptr = (char*)namebuf.mem();
	if(kIOReturnSuccess != rc) {
		BC_IOLOG("failed copying container file path of length %u\n", disk->pathlen);
		return rc;
	}

	BCAutoBuffer bvecbuf;
	rc = bvecbuf.init_from_user(disk->extents.ptr, disk->nvecs * sizeof(bc_bvec));
	disk->extents.ptr = bvecbuf.mem();
	if(kIOReturnSuccess != rc) {
		BC_IOLOG("failed copying block map of length %zu\n", disk->nvecs * sizeof(bc_bvec));
		return rc;
	}

	BCAutoBuffer poolbuf;
	if(disk->pool.ptr) {
		rc = poolbuf.init_from_user(disk->pool.ptr, kBCKernel_PoolSize);
		disk->pool.ptr = poolbuf.mem();
		if(kIOReturnSuccess != rc) {
			BC_IOLOG("failed copying pool\n");
			return rc;
		}
	}
	
/*
	BCAutoBuffer iconbuf;
	if(disk->icon_path) {
		rc = iconbuf.init_from_user((user_addr_t)disk->icon_path, strlen(disk->icon_path + 1));
		if(kIOReturnSuccess != rc) {
			BC_IOLOG("failed copying icon path\n");
			return rc;
		}
		disk->icon_path = (char*)iconbuf.mem();
	}
*/
	
	
	rc = m_controller->set_disk(disk);
	return rc;
}


IOReturn BCControllerUserClient::benchmark(bc_ctl_benchmark* arg, uint64_t* out_time)
{
	return m_controller->benchmark(arg, out_time);
}

/*
IOReturn BCControllerUserClient::enable_aeshw(int do_enable)
{
    return m_controller->enable_aeshw(do_enable);
}
*/


