/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 */

#include "BCDevice.h"
#include "BCDiskDriver.h"
#include "BCDefs.h"

#define kBCDeviceVendorString	"Jetico, Inc."
#define kBCDeviceVersionString	"9.0"
#define kBCDeviceModelString	"BestCrypt Virtual Disk"
#define kBCDeviceInfoString	"bcdisk"

OSDefineMetaClassAndStructors(BCDevice, IOBlockStorageDevice);

bool BCDevice::init(OSDictionary * properties)
{
	if(!IOBlockStorageDevice::init(properties)) {
		return false;
	}
	
	m_driver = NULL;
	m_locked = false;
	return true;
}


//static 
BCDevice* BCDevice::create()
{
	BCDevice* dev = new BCDevice;
	if(!dev) {
		return NULL;
	}
	
	if(!dev->init()) {
		dev->release();
		return NULL;
	}
	
	return dev;
}

bool BCDevice::attach(IOService* provider)
{
	m_driver = OSDynamicCast(BCDiskDriver, provider);
	if(!m_driver) {
		BC_IOLOG("provider is not bc disk driver\n");
		return false;
	}
	
	if(!IOService::attach(provider)) {
		BC_LOG("IOService::start failed\n");
		return false;
	}
	
	return true;
}


IOReturn BCDevice::doEjectMedia(void)
{
	BC_IOLOG_TRACE;
	if(m_locked) {
		return kIOReturnNotPermitted;
	} else {
		return this->m_driver->eject();
	}
}

IOReturn BCDevice::doFormatMedia(UInt64 byteCapacity) {
	BC_IOLOG_TRACE;
	if((this->m_driver->total_blocks() * kBCKernel_BlockSize) != byteCapacity) {
		BC_IOLOG("invalid byte capacity %llu\n", byteCapacity);
		return kIOReturnNotPermitted;
	}
	
	return kIOReturnSuccess;
}

UInt32 BCDevice::doGetFormatCapacities(UInt64* capacities, UInt32 capacitiesMaxCount) const
{
	BC_IOLOG_TRACE;
	if((capacities != NULL) && (capacitiesMaxCount != 0)) {
		capacities[0] = this->m_driver->total_blocks() * kBCKernel_BlockSize;
	}
	
	return 1;
}


IOReturn BCDevice::doLockUnlockMedia(bool doLock) {
	BC_IOLOG_TRACE;
	m_locked = doLock;
	return kIOReturnSuccess;
}

IOReturn BCDevice::doSynchronizeCache(void) {
	BC_IOLOG_TRACE;
	return kIOReturnSuccess;
}


char* BCDevice::getProductString(void) {
	BC_IOLOG_TRACE;
	static char s_model_string[] = kBCDeviceModelString;
	return s_model_string;
}


char* BCDevice::getRevisionString(void) {
	BC_IOLOG_TRACE;
	static char s_version_string[] = kBCDeviceVersionString;
	return s_version_string;
}


char* BCDevice::getVendorString(void) {
	BC_IOLOG_TRACE;
	static char s_vendor_string[] = kBCDeviceVendorString;
	return s_vendor_string;
}

char* BCDevice::getAdditionalDeviceInfoString(void) {
	BC_IOLOG_TRACE;
	static char s_info_string[] = kBCDeviceInfoString;
	return s_info_string;
}


IOReturn BCDevice::reportBlockSize(UInt64 *blockSize) {
	BC_IOLOG_TRACE;
	*blockSize = kBCKernel_BlockSize;
	return kIOReturnSuccess;
}


IOReturn BCDevice::reportMaxValidBlock(UInt64 *maxBlock) {
	BC_IOLOG_TRACE;
	*maxBlock = this->m_driver->total_blocks() - 1;
	return kIOReturnSuccess;
}


IOReturn BCDevice::reportMaxReadTransfer (UInt64 blockSize, UInt64 *max) {
	BC_IOLOG_TRACE;
	*max = this->m_driver->thoroughput() * blockSize;
	return kIOReturnSuccess;
}


IOReturn BCDevice::reportMaxWriteTransfer(UInt64 blockSize, UInt64 *max) {
	BC_IOLOG_TRACE;
	*max = this->m_driver->thoroughput() * blockSize;
	return kIOReturnSuccess;
}


IOReturn BCDevice::reportRemovability(bool *isRemovable) {
	BC_IOLOG_TRACE;
	*isRemovable = true;
	return kIOReturnSuccess;
}


IOReturn BCDevice::reportEjectability(bool *isEjectable) {
	BC_IOLOG_TRACE;
	*isEjectable = !m_locked;
	return kIOReturnSuccess;
}

IOReturn BCDevice::reportLockability(bool *isLockable) {
	BC_IOLOG_TRACE;
	*isLockable = true;
	return kIOReturnSuccess;
}


IOReturn BCDevice::reportMediaState(bool *mediaPresent,bool *changedState) {
	BC_IOLOG_TRACE;
	*mediaPresent = this->m_driver->is_active();
	*changedState = false;  // TODO
	
	return kIOReturnSuccess;
}

IOReturn BCDevice::reportPollRequirements(bool *pollRequired, bool *pollIsExpensive) {
	BC_IOLOG_TRACE;
	*pollRequired = false;
	*pollIsExpensive = false;
	
	return kIOReturnSuccess;
}


IOReturn BCDevice::getWriteCacheState(bool *enabled) {
	BC_IOLOG_TRACE;
	return kIOReturnUnsupported;
}


IOReturn BCDevice::setWriteCacheState(bool enabled) {
	BC_IOLOG_TRACE;
	return kIOReturnUnsupported;
}


IOReturn BCDevice::reportWriteProtection(bool *isWriteProtected) {
	BC_IOLOG_TRACE;
	
	*isWriteProtected = this->m_driver->is_read_only();
	return kIOReturnSuccess;
}


IOReturn BCDevice::doAsyncReadWrite(IOMemoryDescriptor *buffer,
				       UInt64 block, UInt64 nblks,
				       IOStorageAttributes *attributes,
				       IOStorageCompletion *completion) 
{
    BC_IOLOG_DEBUG("async %-5s request: offset %llu, total %llu\n", (buffer->getDirection() == kIODirectionIn) ? "read" : "write", block, nblks);
    if (attributes && attributes->options) {
		BC_IOLOG_DEBUG("Attributes unsupported but came, options=0x%x\n", (unsigned int)attributes->options);
		//return kIOReturnUnsupported;
	} 
	
	int dir = buffer->getDirection();
	UInt64 end = this->m_driver->total_blocks();
	UInt64 last = block + nblks;
	
	if((dir != kIODirectionIn) && (dir != kIODirectionOut)) {
		BC_IOLOG("invalid io direction flag 0x%x\n", dir);
		return kIOReturnBadArgument;
	}
	
	// trim requests past device size.
	if(block >= end) {
		BC_IOLOG("io request is out of bounds\n");
		return kIOReturnInvalid;
	} else if (last > end) {
		BC_IOLOG_DEBUG("io request spans past device size, truncating\n");
		nblks = end - block;
	}
	
	return this->m_driver->create_async_io_task(buffer, block, nblks, completion);
}

IOReturn BCDevice::doSyncReadWrite(IOMemoryDescriptor *buffer, UInt32 block, UInt32 nblks)
{
	BC_IOLOG_TRACE;
	return kIOReturnUnsupported;
}
