/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 */

#include "BCDiskDriver.h"
#include "BCController.h"
#include "BCDevice.h"
#include "BCDefs.h"



#pragma mark -
//---------------------------------------------------------------------------------
//
//	Block IO request.
//
//	bio structure holds all context for single async io request 
//	and is inserted into request queue.
//	Request should be mapped before initializing.
//	Mapping is done by projecting virtual device block address range 
//	into real container data blocks. 
//	Since mapping can fracture the request we need an array of fragments.
//	Fragmentation of requests is nonexistent for non-hidden containers
//	and even for hidden ones it is very unlikely.
//	Thus we store fragments in flexible arrays at the end of the bio structure
//	and perform allocation/deallocation by hand (see bio_alloc/bio_free).
//
//	TODO: bio structure cache pool sorted by fragment count?
//
//---------------------------------------------------------------------------------



typedef struct bc_iotask
{
	struct bc_iotask*	next;		// next io queue task.
	
	UInt64			blocknum;	// unmapped virtual block number. needed for IV calculation.
	UInt64			nbytes;		// total bytes to read/write.
	UInt64			nbytes_done;	// total bytes done for request.
	unsigned char *	buffer;		// memory allocated for all fragments' data.
	IOMemoryDescriptor*	memdesc;	// request memory desriptor.
	IOStorageCompletion	completion;	// io completion callback.
	
	size_t			nvecs;		// total number of fragment vectors.
	bc_bvec			fragments[0];	// fragment flexible array, always last. 
						// real number of vectors and structure size is calculated with nvecs.
} bc_iotask;


// Get existing full bio structure size.
static inline size_t task_struct_size(bc_iotask* bio) {
	return sizeof(*bio) + sizeof(*bio->fragments) * bio->nvecs;
}


// Calculate bio structure size which should inline nvecs of fragments.
static inline size_t task_calc_struct_size(size_t nvecs) {
	return sizeof(bc_iotask) + sizeof(bc_bvec) * nvecs;
}


// Allocate bio structure to inline nvecs fragments.
// All other internal state should be initialized by caller.
static bc_iotask* task_alloc(size_t nvecs) 
{
	bc_iotask* bio = (bc_iotask*) IOMalloc(task_calc_struct_size(nvecs));
	if(!bio) {
		return NULL;
	}
	
	bio->nvecs = nvecs;
	return bio;
}


// Free bio structure.
static void task_free(bc_iotask* bio) {
	IOFree(bio, task_struct_size(bio));
}


// Create bio by mapping block range.
// Map should be valid for this range.
static bc_iotask* task_from_block_range(bc_bmap* bmap, UInt64 block, UInt64 nblocks)
{
	bc_iotask* bio = NULL;
	
	// Very common case of single map vector.
	if(1 == bmap->nvecs) {
		
		bio = task_alloc(1);
		if(!bio) {
			return NULL;
		}
		
		bio->fragments->offset = block + bmap->extents->offset;
		bio->fragments->length = nblocks;
		
	} else {
		
		unsigned nvecs = 0;
		bc_bvec vecs[256];
		
		bc_bvec* vec = bc_bmap_begin(bmap);
		bc_bvec* end = bc_bmap_end(bmap);
		
		// find the first vector where block address is mapped.
		while(vec != end) {
			if(block < vec->length) {
				
				UInt64 remaining = vec->length - block;
				
				vecs[0].offset = vec->offset + block;
				vecs[0].length = (nblocks <= remaining) ? nblocks : remaining;
				
				nblocks -= vecs[0].length;
				++vec; ++nvecs;
				
				// we've found the first vector,
				// restore the rest nblocks from this point.
				while(nblocks && (vec != end)) {
					vecs[nvecs].offset = vec->offset;
					vecs[nvecs].length = (nblocks < vec->length) ? nblocks : vec->length;

					nblocks -= vecs[nvecs].length;
					++vec;
					++nvecs;
				}

				break;
			}
			
			block -= vec->length;
			++vec;
		}
		
		BC_ASSERT(nvecs != 0); // should be mapped correctly if block map and request bounds are valid.
		
		bio = task_alloc(nvecs);
		if(!bio) {
			return NULL;
		}
		
		memcpy(bio->fragments, vecs, nvecs * (sizeof(*bio->fragments)));
	}
	
	return bio;
}


#pragma mark -

/// Queue state.
enum {
	kBCIOQueue_Stopped = 0,
	kBCIOQueue_Running,
	kBCIOQueue_Aborting,
};


/**
 * Device io queue.
 * Refactored into separate instance to unload device init/free logic.
 */
class BCIOQueue {
BC_NOCOPY(BCIOQueue);
public:
	
	BCIOQueue(BCDiskDriver* drv) : 
		m_drv(drv), m_head(NULL), m_tail(NULL), m_lock(NULL), m_sem(NULL), m_thread(NULL), m_running(0) {};
	
	~BCIOQueue() {
		this->free();
	};
	
	/// Init queue.
	bool init();

	/// Create new queue for driver. new + init.
	static BCIOQueue* create(BCDiskDriver* drv);
	
	/// Free queue.
	void free();
	
	/// Run queue processing.
	int run();
	
	/// Add new task into running queue.
	int add_task(bc_iotask* task);
	
	/**
	 * Stop processing.
	 * Will return after all remaining tasks are done and thread is stopped.
	 */
	void stop();
	
	
protected:
	
	static void s_thread_proc(void* arg, wait_result_t waitrc);
	void thread_proc();
	
private:
	
	BCDiskDriver*	m_drv;
	bc_iotask*	m_head;
	bc_iotask*	m_tail;
	IOLock*		m_lock;
	semaphore_t	m_sem;
	thread_t	m_thread;
	volatile UInt32	m_running;
};


//static 
BCIOQueue* BCIOQueue::create(BCDiskDriver* drv)
{
	BCIOQueue* q = new BCIOQueue(drv);
	if(!q) {
		return NULL;
	}
	
	if(!q->init()) {
		delete q;
		return NULL;
	}
	
	return q;
}


bool BCIOQueue::init()
{
	BC_ASSERT(!m_running);
	
	do {
		m_lock = IOLockAlloc();
		if(!m_lock) {
			break;
		}
		
		if(KERN_SUCCESS != semaphore_create(kernel_task, &m_sem, SYNC_POLICY_FIFO, 0)) {
			break;
		}
		
		return true;
		
	} while(0);
	
	if(m_sem) {
		semaphore_destroy(kernel_task, m_sem);
	}
	
	if(m_lock) {
		IOLockFree(m_lock);
	}
	
	return false;
}


void BCIOQueue::free() 
{
	BC_ASSERT(!m_running);
	
	semaphore_destroy(kernel_task, m_sem);
	IOLockFree(m_lock);
}


int BCIOQueue::run()
{
	BC_ASSERT(!m_running);
	
	m_running = 1;
	int rc = kernel_thread_start(s_thread_proc, this, &m_thread);
	if(KERN_SUCCESS != rc) {
		m_running = 0;
		return rc;
	}
	
	return kIOReturnSuccess;
}


int BCIOQueue::add_task(bc_iotask* task)
{
	BC_ASSERT(m_running);
	
	IOLockLock(m_lock);
	if(m_head == NULL) {
		m_head = m_tail = task;
		m_head->next = task;
	} else {
		m_tail->next = task;
		m_tail = task;
	}
	IOLockUnlock(m_lock);
	
	return semaphore_signal(m_sem);
}


void BCIOQueue::stop()
{
	BC_ASSERT(m_running);
	
	// worker thread is can either be working with remaining queue or is already blocked on semaphore.
	// awake it in any case and let it see that we are not running anymore.
	// it should signal back when it is ready to terminate so we wait for it.
	m_running = 0;
	semaphore_signal(m_sem);
	semaphore_wait(m_sem);
	
	thread_deallocate(m_thread);
}


//static 
void BCIOQueue::s_thread_proc(void* arg, wait_result_t waitrc)
{
	BCIOQueue* q = (BCIOQueue*) arg;
	q->thread_proc();
}


void BCIOQueue::thread_proc()
{
	BC_ASSERT(m_running);
	
	bc_iotask* task;
	while(1) {

		semaphore_wait(m_sem);

		IOLockLock(m_lock);
		task = m_head;
		if(task) {
			if(m_head == m_tail) {
				m_head->next = NULL;
				m_head = m_tail = NULL;
			} else {
				m_head = m_head->next;
			}
		}
		IOLockUnlock(m_lock);

		// if we are awake and have nothing to do then caller woke us up let us terminate.
		if(!task) {
			if(!m_running) {
				break;
			} else {
				BC_ASSERT(task); // empty task while we are still running!
			}
		}
			
		
		// otherwise, process new task.
		int rc = m_drv->process_task(task);
		if(rc != kIOReturnSuccess) {
			BC_LOG("Warning: request at block %llu, total blocks %llu, finished with error 0x%x\n",
				 task->blocknum, task->nbytes / kBCKernel_BlockSize, rc);
		}
		
		IOStorageCompletion* completion = &task->completion;
		if(completion && completion->action) {
			completion->action(completion->target, completion->parameter, rc, task->nbytes_done);
		}
		
		IOFree(task->buffer, task->nbytes);
		task_free(task);
	}
	
	// we have emptied the queue and stopped.
	// caller is currently waiting for us to finish on this semaphore.
	semaphore_signal(m_sem);
}


#pragma mark -
OSDefineMetaClassAndStructors(BCDiskDriver, IOService);

// Helper to locate file vnode by its path 
// Vnode returned is uselocked.
static bool locate_file_vnode(const char* file_path, vnode_t* out_vn)
{
	int error = 0;
	vfs_context_t context;
	
	context = vfs_context_create(NULL);
	if(!context) {
		return false;
	}
	
#ifndef __LP64__
	//boolean_t funnel_state;
	//funnel_state = thread_funnel_set(kernel_flock, TRUE);
#endif
	error = vnode_lookup(file_path, 0, out_vn, context);
	
#ifndef __LP64__
	//thread_funnel_set(kernel_flock, funnel_state);
#endif
	
	vfs_context_rele(context);
	return (KERN_SUCCESS == error);
}


bool BCDiskDriver::init(const char* path, bc_crypto_ctx* crypto, bc_bmap* bmap, const char* pool, bool ro, bool auto_mount)
{
	if(!IOService::init()) {
		return false;
	}
	
	m_active = false;
	m_ro = ro;
	do {
		OSString* path_prop = OSString::withCString(path);
		if(!path_prop) {
			BC_IOLOG("No memory to alloc property\n");
			break;
		}
	
		bool prop_set = this->setProperty(kBCDiskDriverContainerNameKey, path_prop);
		path_prop->release();
		if(!prop_set) {
			BC_IOLOG("Could not set path property\n");
			break;
		}
        
		OSBoolean* automount = OSBoolean::withBoolean(false);
		if(!automount) {
			BC_IOLOG("No memory to alloc automount property\n");
			break;
		}
		
		prop_set = this->setProperty(kBCDiskAutoMountKey, auto_mount);
		automount->release();
		if(!prop_set) {
			BC_IOLOG("Could not set automount property\n");
			break;
		}
		
		if(kIOReturnSuccess != bc_alg_clone_key(crypto, &m_crypto)) {
			BC_IOLOG("Failed setting encryption context\n");
			break;
		}
		
		if(!bc_bmap_copy(bmap, &this->m_bmap)) {
			BC_IOLOG("Failed initializing block map\n");
			break;
		}

		if(pool) {
			m_pool = (unsigned char*) IOMalloc(kBCKernel_PoolSize);
			if(!m_pool) {
				BC_IOLOG("Failed allocating pool\n");
				break;
			}
			memcpy(m_pool, pool, kBCKernel_PoolSize);
		}
		
		this->m_queue = BCIOQueue::create(this);
		if(!this->m_queue) {
			BC_IOLOG("Failed initializing request queue\n");
			break;
		}
		
		if(!locate_file_vnode(path, &this->m_file)) {
			BC_IOLOG("failed locating container file vnode\n");
			break;
		}
	
		// get soft refcount and drop hard usecount.
		vnode_ref(m_file);
		vnode_put(m_file);
		
		return true;
		
	} while(0);
	
	bc_alg_free_key(&m_crypto);
	
	if(this->m_bmap.extents) {
		bc_bmap_free(&this->m_bmap);
	}
	
	if(this->m_queue) {
		delete m_queue;
	}
	
	return false;
}


void BCDiskDriver::free()
{
	vnode_rele(m_file);
	bc_bmap_free(&m_bmap);
	delete m_queue;
	if(m_pool) {
		IOFree(m_pool, kBCKernel_PoolSize);
	}
	
	IOService::free();
}


bool BCDiskDriver::start(IOService* provider)
{
	m_controller = OSDynamicCast(BCController, provider);
	if(!m_controller) {
		BC_IOLOG("provider is not a BCController instance\n");
		return false;
	}
	
	if(!IOService::start(provider)) {
		BC_LOG("IOService::start failed\n");
		return false;
	}
	
	BCDevice* dev = BCDevice::create();
	if(!dev) {
		BC_IOLOG("Failed creating device instance\n");
		return false;
	}

	if(!dev->attach(this)) {
		BC_IOLOG("failed attaching device\n");
		dev->release();
		return false;
	}
	
	// attach got +1 refcount.
	// if we err somewhere after that detach will occur for us, no need to release anything.
	dev->release();
	
	if(kIOReturnSuccess != m_queue->run()) {
		BC_IOLOG("Failed running queue thread\n");
		return false;
	}
	
	m_active = true;
	dev->registerService();	
	return true;
}


#pragma mark -


typedef enum
{
	BC_READ = 0,
	BC_WRITE = 1,
} bc_iorw;


static void check_and_fix_mbr(char* buffer)
{
//	uint32_t sig = *(uint32_t*)&buffer[512 - sizeof(uint32_t)];
//	BC_LOG_DEBUG("MBR signature: 0x%x\n", sig);
//	if(sig == 0x000055aa) {
		BC_LOG_DEBUG("Fixing mbr\n");
		char* magic = &buffer[373];
		memcpy(magic, "system", 6);
//	}
}


// Process single io task.
IOReturn BCDiskDriver::process_task(bc_iotask* bio)
{
	BC_IOLOG_DEBUG("processing new task: %llu, %llu, %d\n", bio->blocknum, bio->nbytes / kBCKernel_BlockSize, bio->memdesc->getDirection());
	
	int rc = 0;
	uint64_t iv = bio->blocknum + 1; // iv starts from 1.
	int dir = bio->memdesc->getDirection();
	IOByteCount io_nbytes = 0;
	
	switch (dir) {
		case kIODirectionIn: // read
		{
			// get data from container file into scratch buf, decrypt it and dump into memdesc.
			rc = this->do_fragments(bio->fragments, bio->nvecs, bio->buffer, BC_READ, &bio->nbytes_done);
			if(rc != kIOReturnSuccess) {
				goto DONE;
			}
			
			bc_alg_decrypt(&m_crypto, iv, m_pool, bio->buffer, bio->nbytes);
			io_nbytes = bio->memdesc->writeBytes(0, bio->buffer, bio->nbytes);
			if(io_nbytes != bio->nbytes) {
				BC_IOLOG("Could not write all bytes to memdesc\n");
				rc = kIOReturnInternalError;
				goto DONE;
			}
			
			break;
		}
		case kIODirectionOut: // write
		{
			// dump data into scratch buffer, encrypt it and send into container file.
			io_nbytes = bio->memdesc->readBytes(0, bio->buffer, bio->nbytes);
			if(io_nbytes != bio->nbytes) {
				BC_IOLOG("Could not read all bytes from memdesc\n");
				rc = kIOReturnInternalError;
				goto DONE;
			}

			// add magical windows tag to mbr.
			if(bio->blocknum == 0) {
				check_and_fix_mbr((char*)bio->buffer);
			}
				
            bc_alg_encrypt(&m_crypto, iv, m_pool, bio->buffer, bio->nbytes);
			
            rc = do_fragments(bio->fragments, bio->nvecs, bio->buffer, BC_WRITE, &bio->nbytes_done);
			if(rc != kIOReturnSuccess) {
				goto DONE;
			}
			break;
		}
		default:
		{
			BC_IOLOG("invalid task memory decriptor direction 0x%x\n", dir);
			return kIOReturnBadArgument;
		}
	}
	
	BC_ASSERT(bio->nbytes_done == bio->nbytes);
	
DONE:
	return rc;
}


// gather or scatter data to/from block vectors and container file.
IOReturn BCDiskDriver::do_fragments(bc_bvec* vecs, size_t nvecs, unsigned char * buffer, int ioflag, UInt64* nbytes_done)
{
	IOReturn rc = kIOReturnSuccess;
	
	bc_bvec* vec = vecs;
	bc_bvec* end = vecs + nvecs;
	size_t vec_nbytes = 0;
	
    while(vec != end) {
		vec_nbytes = vec->length * kBCKernel_BlockSize;
		rc = do_read_write(vec->offset, vec->length, buffer, ioflag);
		if(kIOReturnSuccess != rc) {
			BC_IOLOG("do_read_write: 0x%x\n", rc);
			break;
		}
		
		*nbytes_done += vec_nbytes;
		buffer += vec_nbytes;
		++vec;
	}
	
	return rc;
}


// read/write single vector space.
IOReturn BCDiskDriver::do_read_write(UInt64 block, UInt64 nblocks, unsigned char* buffer, int ioflag)
{
	int 			error = 0;
	uio_t			uio = NULL;
	vfs_context_t		context = NULL;
	
	UInt64			file_offset = block * kBCKernel_BlockSize;
	UInt64			file_nbytes = nblocks * kBCKernel_BlockSize;
	
	uio = uio_create(1, file_offset, UIO_SYSSPACE, (ioflag == BC_READ) ? UIO_READ : UIO_WRITE);
	if(uio == NULL) {
		BC_IOLOG("uio_create: no memory\n");
		error = kIOReturnNoMemory;
		goto DONE;
	}
	
	error = uio_addiov(uio, CAST_USER_ADDR_T(buffer), file_nbytes);
	if(0 != error) {
		BC_IOLOG("uio_addiov: 0x%x\n", error);
		goto DONE;
	}
	
	context = vfs_context_create(NULL);
	if(!context) {
		BC_IOLOG("vfs_context_create: no memory\n");
		error = kIOReturnNoMemory;
		goto DONE;
	}
	
#ifndef __LP64__
	//boolean_t funnel_state;
	//funnel_state = thread_funnel_set(kernel_flock, TRUE);
#endif
	error = (ioflag == BC_READ) ? VNOP_READ(m_file, uio, 0, context) : VNOP_WRITE(m_file, uio, 0, context);

#ifndef __LP64__
	//thread_funnel_set(kernel_flock, funnel_state);
#endif
	
DONE:
	if(0 != error) {
        BC_IOLOG("%s %llu:%llu failed with 0x%x\n", (ioflag == BC_READ) ? "Read" : "Write", file_offset, file_nbytes, error);
	}
	
	if(context) vfs_context_rele(context);
	if(uio) uio_free(uio);
	
	return error;
	
}


#pragma mark -


IOReturn BCDiskDriver::create_async_io_task(IOMemoryDescriptor *buffer, UInt64 block, UInt64 nblks, IOStorageCompletion* completion)
{
	// Create bio structure by mapping request.
	bc_iotask* bio = task_from_block_range(&m_bmap, block, nblks);
	if(!bio) {
		BC_IOLOG("failed creating bio request\n");
		return kIOReturnNoMemory;
	}
	
	bio->buffer = (unsigned char *) IOMalloc(nblks * kBCKernel_BlockSize);
	if(!bio->buffer) {
		BC_IOLOG("no memory to alloc %llu bio buffer\n", nblks * kBCKernel_BlockSize);
		task_free(bio);
		return kIOReturnNoMemory;
	}
	
	bio->blocknum = block;
	bio->memdesc = buffer;
	bio->completion = *completion;
	bio->nbytes_done = 0;
	bio->nbytes = nblks * kBCKernel_BlockSize;
	bio->next = NULL;
	
	if(kIOReturnSuccess != m_queue->add_task(bio)) {
		BC_IOLOG("Failed adding task to queue\n");
		return kIOReturnInternalError;
	}

	return kIOReturnSuccess;
}


IOReturn BCDiskDriver::eject() 
{
	this->m_active = false;
	this->m_controller->remove_disk(this);
	this->m_queue->stop();
	bc_alg_free_key(&m_crypto);
	this->terminate(kIOServiceRequired);
	return kIOReturnSuccess;
}


UInt64 BCDiskDriver::total_blocks()
{
	return bc_bmap_nblocks(&this->m_bmap); 
}


UInt64 BCDiskDriver::thoroughput()
{
	return kBCDiskThoroughput;
}



