/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 */

#ifndef BC_KERNEL_DISK_DRIVER_H
#define BC_KERNEL_DISK_DRIVER_H


#define kBCDiskDriverContainerNameKey	"container-file"
#define kBCDiskAutoMountKey		"autodiskmount" // this property will be read by diskarbitrationd to see if it should automatically mount new block device.


enum {
	// This value seems to interfere with BSD DKIOCGETMAXBLOCKCOUNTWRITE ioctl.
	// On 10.5 the diskutil seems to use it to calculate maximum number of blocks it can write when paritioning the disk.
	// Supply 256 blocks to make ioctl return 128kb which is what HDIX is doing.
	kBCDiskThoroughput = 0x100,
};


#ifdef BC_KERNEL
#include <IOKit/IOService.h>
#include <IOKit/storage/IOStorage.h>

/* a hack for vnode.h include idiosyncrasies */
#ifndef _KAUTH_CRED_T
#  include <sys/ucred.h>
#endif
#include <sys/vnode.h>

#include "BCBlockMap.h"
#include "BCAlg.h"


class BCController;
class BCDevice;
class BCIOQueue;
struct bc_iotask;


/**
 * Drives virtual container block devices.
 * Uses BCDevice instances as interface nubs for storage level.
 * Can be extended in the future to support different virtual device configs and backends.
 */
class BCDiskDriver : public IOService
{
OSDeclareDefaultStructors(BCDiskDriver);
public:

	/**
	 * Initialize disk driver.
	 * Will call IOService::init here.
	 * @path	Container file path.
	 * @crypto	Initialized crypto context for container.
	 */
	bool init(const char* path, bc_crypto_ctx* crypto, bc_bmap* bmap, const char* pool, bool ro, bool auto_mount);

	
	/**
	 * Will release vnode reference and other allocated resources.
	 */
	void free();

	
	/**
	 * Will initialize new bc device nub instance, prepare request processing and publish device.
	 * @provider	BCController instance.
	 */
	bool start(IOService* provider);

	

public:
	
	
	//-----------------------------------------------------------------------------------------
	//
	//	Methods below are block storage device procedure implementations.
	//	Device nub instance piggybacks on them when receiving requests from block layer.
	//
	//-----------------------------------------------------------------------------------------
	
	
	/**
	 * Create new async IO task.
	 * Will initialize new task and put it in queue.
	 * Request correctness should be checked by nub instance.
	 */
	IOReturn create_async_io_task(IOMemoryDescriptor *buffer, UInt64 block, UInt64 nblks, IOStorageCompletion* completion);
	
	
	/**
	 * Eject disk.
	 * Halt all processing and terminate.
	 */
	IOReturn eject();
	
	
	/**
	 * Get device size in blocks.
	 */
	UInt64 total_blocks();
	
	
	/**
	 * Get maximum size of single io request.
	 */
	UInt64 thoroughput();
	
	
	bool is_active() {
		return m_active;
	}
	
	bc_alg* algorithm() {
		return &m_crypto.alg;
	}
	
	bool is_read_only() {
		return m_ro;
	}
	
protected:
	
	friend class BCIOQueue;
	
	IOReturn process_task(bc_iotask* bio);

	IOReturn do_fragments(bc_bvec* vecs, size_t nvecs, unsigned char * buffer, int ioflag, UInt64* nbytes_done);
	
	IOReturn do_read_write(UInt64 block, UInt64 nblocks, unsigned char * buffer, int ioflag);

	
private:
	
	
	BCController*			m_controller;
	const char*			m_name;		// display name.
	vnode_t				m_file;		// container file vnode.
	bc_bmap				m_bmap;		// encrypted data blocks map.
	bc_crypto_ctx		m_crypto;
	unsigned char *		m_pool;		// random data pool.
	BCIOQueue*			m_queue;	// request queue.
	bool				m_active;
	bool				m_ro;
};

#endif // BC_KERNEL
#endif // BC_KERNEL_DISK_DRIVER
