/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 *
 */

#include "BCIOQueue.h"

BCIOQueue::BCIOQueue() : m_head(NULL), m_tail(NULL), m_lock(NULL), m_running(0), m_sem(NULL), m_worker(NULL)
{
}


BCIOQueue::~BCIOQueue()
{
	BC_ASSERT(!m_running);
	thread_deallocate(m_worker);
	semaphore_destroy(kernel_task, m_sem);
	IOLockFree(m_lock);
}


BCIOQueue* BCIOQueue::create()
{
	BCIOQueue* q = new BCIOQueue;
	if(!q) {
		return NULL;
	}
	
	if(!q->init()) {
		delete q;
		return NULL;
	}

	return q;
}

int BCIOQueue::init()
{
	BC_ASSERT(!m_running);
	
	int rc;
	
	do {
		m_lock = IOLockAlloc();
		if(!m_lock) {
			rc = kIOReturnNoMemory;
			break;
		}

		rc = semaphore_create(kernel_task, &m_sem, SYNC_POLICY_FIFO, 0);
		if(rc != KERN_SUCCESS) {
			break;
		}
	
		return kIOReturnSuccess;
	} while (0);
	
	if(m_sem) {
		semaphore_destroy(kernel_task, m_sem);
	}

	if(m_lock) {
		IOLockFree(m_lock);
	}

	return rc;
}


int BCIOQueue::run()
{
	BC_ASSERT(!m_running);
	
	int rc = kernel_thread_start(s_worker_proc, this, &m_worker);
	if(KERN_SUCCESS != rc) {
		return rc;
	}
	
	m_running = 1;
	return kIOReturnSuccess;
}


int BCIOQueue::add_task(BCIOTask* task)
{
	BC_ASSERT(m_running);
	
	IOLockLock(m_lock);
	if(m_head == NULL) {
		m_head = m_tail = task;
		m_head->m_next = task;
	} else {
		m_tail->m_next = task;
		m_tail = task;
	}
	IOLockUnlock(m_lock);
	
	return semaphore_signal(m_sem);
}


void BCIOQueue::s_worker_proc(void* arg, wait_result_t waitrc)
{
	BCIOQueue* q = (BCIOQueue*) arg;
	q->worker_proc();
}


void BCIOQueue::worker_proc()
{
	BCIOTask* task;
	
	while(m_running) {
		
		semaphore_wait(m_sem);
	
		if(!m_running)
			break;
		
		IOLockLock(m_lock);
		task = m_head;
		if(task) {
			if(m_head == m_tail) {
				m_head->m_next = NULL;
				m_head = m_tail = NULL;
			} else {
				m_head = m_head->m_next;
			}
		}
		IOLockUnlock(m_lock);
		
		task->process();
		this->on_task_processed(task);
	}
}


int BCIOQueue::stop()
{
	BC_ASSERT(m_running);
	m_running = 0;
	semaphore_signal(m_sem);
	return kIOReturnSuccess;
}


int BCIOQueue::abort()
{
	return this->stop();
}	
