/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 *
 */

#ifndef BC_KERNEL_IOQUEUE_H
#define BC_KERNEL_IOQUEUE_H

#ifdef BC_KERNEL

#include "BCDefs.h"


#if 0
// Allocate bio structure to inline nvecs fragments.
// All other internal state should be initialized by caller.
static bc_bio* bio_alloc(size_t nvecs) 
{
	bc_bio* bio = (bc_bio*) IOMalloc(bio_calc_struct_size(nvecs));
	if(!bio) {
		return NULL;
	}
	
	bio->nvecs = nvecs;
	return bio;
}


// Free bio structure.
static void bio_free(bc_bio* bio) {
	IOFree(bio, bio_struct_size(bio));
}


// Create bio by mapping block range.
// Map should be valid for this range.
static bc_bio* bio_from_block_range(bc_bmap* bmap, UInt64 block, UInt64 nblocks)
{
	bc_bio* bio = NULL;
	
	// Very common case of single map vector.
	if(1 == bmap->nvecs) {
		
		bio = bio_alloc(1);
		if(!bio) {
			return NULL;
		}
		
		bio->fragments->offset = block + bmap->vecs->offset;
		bio->fragments->length = nblocks;
		
	} else {
		
		bc_block_mapping rng_start;
		bc_block_mapping rng_end;
		
		int nvecs = bc_bmap_map_range(bmap, block, nblocks, &rng_start, &rng_end);
		BC_ASSERT(nvecs != 0); // should be mapped correctly if block map and request bounds are valid.
		
		bio = bio_alloc(nvecs);
		if(!bio) {
			return NULL;
		}
		
		// check another common case of mapping into single fragment.
		if(1 == nvecs) {
			bio->fragments->offset = rng_start.vec->offset + rng_start.local_offset;
			bio->fragments->length = rng_end.local_offset;
		} else {
			
			// hardest and rarest case - request was torn down into multiple fragments.
			bc_bvec* fragment = bio->fragments;
			
			// map first fragment
			fragment->offset = rng_start.vec->offset + rng_start.local_offset;
			fragment->length = rng_start.vec->length - rng_start.local_offset;
			
			// copy all vectors in between.
			bc_bvec* vec = rng_start.vec;
			bc_bvec* end = rng_end.vec;
			while(vec != end) {
				memcpy(++fragment, ++vec, sizeof(*fragment));
			}
			
			// finish last vector.
			fragment->offset = rng_end.vec->offset;
			fragment->length = rng_end.local_offset;
		}
	}
	
	return bio;
}
#endif // 0


/**
 * Generic io task.
 * Queue element.
 */
class IBCIOTask {
public:
	
	IBCIOTask() : m_next(NULL) {};
	
	virtual ~IBCIOTask() {};

	
	/**
	 * Process this task.
	 * Called when io queue worker thread starts with this task.
	 */
	virtual void process() = 0;
	
	
	/**
	 * Abort this task.
	 * Called when io queue is told to abort operation.
	 */
	virtual int abort() = 0;
	
	
protected:
	
	friend class BCIOQueue;
	IBCIOTask*	m_next;
};


typedef struct bc_bio
{
	struct bc_bio*		next;		// next io queue task.
	
	UInt64			blocknum;	// unmapped virtual block number. needed for IV calculation.
	UInt64			nbytes;		// total bytes to read/write.
	UInt64			nbytes_done;	// total bytes done for request.
	unsigned char*		buffer;		// memory allocated for all fragments' data.
	IOMemoryDescriptor*	memdesc;	// request memory desriptor.
	IOStorageCompletion	completion;	// io completion callback.
	
	size_t			nvecs;		// total number of fragment vectors.
	bc_bvec			fragments[0];	// fragment flexible array, always last.
} bc_bio;


// Get existing full bio structure size.
static inline size_t bio_struct_size(bc_bio* bio) {
	return sizeof(*bio) + sizeof(*bio->fragments) * bio->nvecs;
}


// Calculate bio structure size which should inline nvecs of fragments.
static inline size_t bio_calc_struct_size(size_t nvecs) {
	return sizeof(bc_bio) + sizeof(bc_bvec) * nvecs;
}





/**
 * Generic io request processing queue.
 * Caller acts as producer and adds jobs. Queue has a consumer thread that blocks on semaphore and waits for new tasks to process.
 * Currently exists as a separate instance to serve two purposes:
 * 1)	To unload disk driver from queue management.
 * 2)	To allow experimenting with different request processing approaches and disk driver variants without having to reproduce entire generic io task logistics.
 *
 * Queue has two states:
 * 1)	Running. 
 *	Queue is actively processing tasks and accepts new tasks through add_task.
 *	This state is entered by calling run.
 * 2)	Stopped.
 *	Queue is not accepting any more tasks and finishes existing ones if any.
 *	Can be entered through stop and abort. The former will try to finish all current tasks normally and latter will call abort on each task.
 */
class BCIOQueue {
BC_NOCOPY(BCIOQueue);
public:
	
	BCIOQueue();
	
	~BCIOQueue();
	
	
	/// new + init.
	static BCIOQueue* create();
	
	
	/**
	 * Init this queue.
	 * Will initialize all internal state.
	 * @return	kIOReturnSuccess on success, error code otherwise.
	 */
	int init();
	
	
	/**
	 * Start worker thread.
	 * @return	kernel_thread_start result.
	 */
	int run();
	
	
	/**
	 * Enqueue new task.
	 * Queue should be running.
	 * @return	kIOReturnAborted if queue is not running.
	 *		kIOReturnSuccess on successfull enqueue.
	 *		Other errors when something fails.
	 */
	int add_task(BCIOTask* task);
	
	
	/**
	 * Stop processing.
	 * After calling this no new tasks will be accepted and it will return after all current tasks will have been processed.
	 * See abort for more radical termination.
	 */
	int stop();
	
	
	/**
	 * Stop processing and abort all current tasks.
	 * No new tasks will be accepted and abort will be called on existing ones.
	 */
	int abort();

protected:
	
	
	/// Worker thread glue and actual processing loop.
	static void s_worker_proc(void* arg, wait_result_t waitrc);
	void worker_proc();
	
	
	/**
	 * Optional callback called by worker thread when task is done processing.
	 */
	virtual void on_task_processed(IBCIOTask* task) {};
	
	
	/**
	 * Optional callback called by worker thread when task is aborted.
	 */
	virtual void on_task_aborted(IBCIOTask* task) {};
	
private:
	
	BCIOTask*	m_head;
	BCIOTask*	m_tail;
	IOLock*		m_lock;
	
	volatile UInt32	m_running;	// queue state.
	semaphore_t	m_sem;		// up's each time new task is added, down's when worker dequeues new task.
	thread_t	m_worker;	// consumer worker thread.
	
};


#endif // BC_KERNEL
#endif // BC_KERNEL_IOQUEUE_H
