/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 *
 * BCKernel userspace lib implementation.
 */

#include <IOKit/IOKitLib.h>

#include "BCKernel.h" // we're implementing this.
#include "BCKernelDefs.h"

#include "BCCTL.h"
#include "BCController.h"
#include "BCContainerGuardKext.h"
#include "BCKeyboardFilterKext.h"

//#include <stdlib.h>
#include <string.h>
#include <CoreFoundation/CoreFoundation.h>
#include <unistd.h>
#include <fcntl.h>
#include <map>
#include <vector>

/////////////////////////////////////////////////////////////////////////
//
//	BCController IOService connection.
//
/////////////////////////////////////////////////////////////////////////


static inline bc_cipher_result result_from_kernel(IOReturn rc)
{
	if(rc == kIOReturnSuccess) {
		return kBCCipher_Success;
	} else if(rc & BC_RESULT_KERNEL_MASK) {
		return (bc_cipher_result)(rc & ~BC_RESULT_KERNEL_MASK);
	} else {
		return kBCCipher_InternalError;
	}
}


static io_connect_t open_controller(void)
{
	CFMutableDictionaryRef dict = IOServiceMatching(kBCControllerMatchKey);
	if(!dict) {
		return IO_OBJECT_NULL;
	}
	
	io_iterator_t iter;
	kern_return_t rc = IOServiceGetMatchingServices(kIOMasterPortDefault, dict, &iter);
	if(KERN_SUCCESS != rc) {
		return IO_OBJECT_NULL;
	}
	
	io_service_t serv = IOIteratorNext(iter);
	IOObjectRelease(iter);

	if(serv == IO_OBJECT_NULL) {
		return IO_OBJECT_NULL;
	}

	io_connect_t port;
	rc = IOServiceOpen(serv, mach_task_self(), 0, &port);
	IOObjectRelease(serv);

	if(KERN_SUCCESS != rc) {
		return IO_OBJECT_NULL;
	}
	
	return port;
}


static void close_controller(io_connect_t port)
{
	IOServiceClose(port);
}


// Find BCController IORegistry entry
static io_registry_entry_t find_controller_regentry(void)
{
	io_iterator_t iter;
	kern_return_t rc = IORegistryCreateIterator(kIOMasterPortDefault, "IOService", kIORegistryIterateRecursively, &iter);
	if(KERN_SUCCESS != rc) {
		return IO_OBJECT_NULL;
	}
	
	io_name_t name;
	io_registry_entry_t entry;
	while(IO_OBJECT_NULL != (entry = IOIteratorNext(iter))) {
		rc = IORegistryEntryGetNameInPlane(entry, "IOService", name);
		if(KERN_SUCCESS != rc) {
			return IO_OBJECT_NULL;
		}
		
		if(!strcmp(kBCControllerMatchKey, name)) {
			break;
		}
	}
	
	IOObjectRelease(iter);
	return entry;
}


static void try_reload_kernel()
{
    // This might fail if job is already loaded.
    system("launchctl load /Library/LaunchDaemons/com.jetico.BestCrypt.helper.plist");
    
    // We can check return here, but it will be better to
    // simply check kernel modules status one more time.
    system("launchctl start com.jetico.BestCrypt.helper");
}


bc_cipher_result BCKernelIsInitialized(void)
{
	io_connect_t port = open_controller();
	if(IO_OBJECT_NULL == port) {
        try_reload_kernel();
        port = open_controller();
        if (IO_OBJECT_NULL == port) {
            return kBCCipher_NotInitialized;
	    }
    }
	
	close_controller(port);
	return kBCCipher_Success;
}



/////////////////////////////////////////////////////////////////////////
//
//	Plugins IORegistry entry parsing.
//
/////////////////////////////////////////////////////////////////////////


// Static mode id/IORegistry key mapping table.
/*static const struct
{
	bc_mode_id	id;
	CFStringRef	key;
} g_modes[] = {
	{ kBCMode_CBC, CFSTR(kBCPluginCBCKeySizeKey) },
	{ kBCMode_LRW, CFSTR(kBCPluginLRWKeySizeKey) },
	{ kBCMode_XTS, CFSTR(kBCPluginXTSKeySizeKey) }
};

#define BC_TOTAL_MODES ( sizeof(g_modes) / sizeof(*g_modes) )


static int map_mode(bc_mode_id id)
{
	switch (id) {
		case kBCMode_CBC: return 0;
		case kBCMode_LRW: return 1;
		case kBCMode_XTS: return 2;
		default: return -1;
	}
}


size_t bc_alg_key_size_for_mode(const bc_alg * alg, bc_mode_id id)
{
	int mode_idx = map_mode(id);
	if(mode_idx == -1) {
		return 0;
	} else {
		return alg->key_size[mode_idx];
	}
} */

static void fill_plugin_info(BCAlgInfo * alg, CFDictionaryRef info)
{
	memset(alg, 0, sizeof(*alg));
	
	CFNumberRef id = (CFNumberRef)CFDictionaryGetValue(info, CFSTR(kBCPluginIdKey));
	CFStringRef name = (CFStringRef)CFDictionaryGetValue(info, CFSTR(kBCPluginNameKey));
	CFNumberRef key_size = (CFNumberRef)CFDictionaryGetValue(info, CFSTR(kBCPluginKeyLengthKey));
	CFNumberRef block_size = (CFNumberRef)CFDictionaryGetValue(info, CFSTR(kBCPluginBlockSizeKey));
		
	CFNumberGetValue(id, kCFNumberIntType, &alg->id);
	CFNumberGetValue(key_size, kCFNumberIntType, &alg->key_length);
	CFNumberGetValue(block_size, kCFNumberIntType, &alg->block_size);
	CFStringGetCString(name, alg->name, sizeof(alg->name), kCFStringEncodingASCII);

	/*unsigned i;
	CFNumberRef mode_key_size;
	for(i = 0; i < BC_TOTAL_MODES; ++i) {
		mode_key_size = CFDictionaryGetValue(info, g_modes[i].key);
        
        CFNumberGetValue(mode_key_size, kCFNumberIntType, &alg->key_size[i]);
		if(alg->key_size[i] != 0) { // unsupported modes are marked as having key size set to 0 by kernel service.
			alg->supported_modes |= g_modes[i].id;
		}
	}*/
    
    CFNumberRef modeKeySizeRef = (CFNumberRef)CFDictionaryGetValue(info, CFSTR(kBCPluginCBCKeySizeKey));
    CFNumberGetValue(modeKeySizeRef, kCFNumberIntType, &alg->cbc_key_size);
    modeKeySizeRef = (CFNumberRef)CFDictionaryGetValue(info, CFSTR(kBCPluginLRWKeySizeKey));
    CFNumberGetValue(modeKeySizeRef, kCFNumberIntType, &alg->lrw_key_size);
    modeKeySizeRef = (CFNumberRef)CFDictionaryGetValue(info, CFSTR(kBCPluginXTSKeySizeKey));
    CFNumberGetValue(modeKeySizeRef, kCFNumberIntType, &alg->xts_key_size);
}


// Plugin lookup predicate signature.
typedef int(*lookup_predicate_t)(CFDictionaryRef info, void* ctx);


// Lookup algorithm plugin in IORegistry.
static bc_cipher_result lookup_plugin(lookup_predicate_t pred, void* pred_ctx, BCAlgInfo * out_alg)
{
	bc_cipher_result rc = kBCCipher_Success;
	
	io_registry_entry_t regentry = IO_OBJECT_NULL;
	CFArrayRef plugins = NULL;
	size_t total_plugins = 0;
	
	regentry = find_controller_regentry();
	if(IO_OBJECT_NULL == regentry) {
		rc = kBCCipher_NotInitialized;
		goto OUT;
	}
	
	plugins = (CFArrayRef)IORegistryEntryCreateCFProperty(regentry, CFSTR(kBCControllerPluginsPropertyKey), kCFAllocatorDefault, 0);
	if(!plugins) {
		rc = kBCCipher_InternalError;
		goto OUT;
	}
	
	total_plugins = CFArrayGetCount(plugins);
	if(total_plugins == 0) {
		rc = kBCCipher_UnknownAlg;
		goto OUT;
	}
	
	CFDictionaryRef info;
	while(total_plugins--) {
		info = (CFDictionaryRef)CFArrayGetValueAtIndex(plugins, total_plugins);
		if(pred(info, pred_ctx)) {
			fill_plugin_info(out_alg, info);
			goto OUT;
		}
	}

	rc = kBCCipher_UnknownAlg;
	
OUT:
	if(regentry != IO_OBJECT_NULL) IOObjectRelease(regentry);
	if(plugins) CFRelease(plugins);
	return rc;
}


// Predicate helper for looking up algorithm id.
static int lookup_pred_id(CFDictionaryRef info, void* ctx) 
{
	bc_alg_id lhv, rhv;
	
	CFNumberRef id_ref = (CFNumberRef)CFDictionaryGetValue(info, CFSTR(kBCPluginIdKey));
	CFNumberGetValue(id_ref, kCFNumberIntType, &lhv);
	
	rhv = *(bc_alg_id*) ctx;
	
	return (lhv == rhv);
}


// Predicate helper for looking up algorithm name.
static int lookup_pred_name(CFDictionaryRef info, void* ctx) 
{
	CFStringRef lhv = (CFStringRef)CFDictionaryGetValue(info, CFSTR(kBCPluginNameKey));
	CFStringRef rhv = CFStringCreateWithCStringNoCopy(kCFAllocatorDefault, (const char*) ctx, kCFStringEncodingASCII, kCFAllocatorNull);
	
	CFComparisonResult rc = CFStringCompare(lhv, rhv, kCFCompareCaseInsensitive);
	CFRelease(rhv);
	
	return (rc == kCFCompareEqualTo) ? 1 : 0;
}
	

//bc_kern_result bc_kernel_lookup_alg_id(bc_alg_id id, bc_alg* out_alg)
bc_cipher_result BCKernelLookupAlgById(bc_alg_id id, BCAlgInfo* out_alg)
{
	return lookup_plugin(lookup_pred_id, &id, out_alg);
}

bc_cipher_result BCKernelLookupAlgByName(const char* name, BCAlgInfo* out_alg)
//bc_kern_result bc_kernel_lookup_alg_name(const char* name, bc_alg* out_alg)
{
	return lookup_plugin(lookup_pred_name, (void*) name, out_alg);
}

bc_cipher_result BCKernelListAlgs(BCAlgInfo** out_palgs, size_t *out_total)
{
	io_registry_entry_t regentry = IO_OBJECT_NULL;
	CFArrayRef plugins = NULL;
	size_t total_plugins = 0;
	
	regentry = find_controller_regentry();
	if (IO_OBJECT_NULL == regentry) {
		return kBCCipher_NotInitialized;
	}
	
	plugins = (CFArrayRef)IORegistryEntryCreateCFProperty(regentry, CFSTR(kBCControllerPluginsPropertyKey), kCFAllocatorDefault, 0);
	if (!plugins) {
        IOObjectRelease(regentry);
		return kBCCipher_InternalError;
	}
	
	total_plugins = CFArrayGetCount(plugins);
	if (total_plugins == 0) {
        CFRelease(plugins);
        IOObjectRelease(regentry);
		return kBCCipher_UnknownAlg;
	}
	
	BCAlgInfo * list = (BCAlgInfo *)malloc(sizeof(BCAlgInfo) * total_plugins);
	if (!list) {
        CFRelease(plugins);
        IOObjectRelease(regentry);
		return kBCCipher_NoMemory;
	}
	
	*out_palgs = list;
	*out_total = total_plugins;
	
	CFDictionaryRef info;
	while(total_plugins--) {
		info = (CFDictionaryRef)CFArrayGetValueAtIndex(plugins, total_plugins);
		fill_plugin_info(list, info);
		++list;
	}
	
	return kBCCipher_Success;
}


void BCKernelFreeAlgList(BCAlgInfo * list)
{
    if (list) free(list);
}


/////////////////////////////////////////////////////////////////////////
//
//	BCController CTL.
//
/////////////////////////////////////////////////////////////////////////


static bc_cipher_result controller_ctl(int ctlcode, void* data_in, size_t insize, void* data_out, size_t outsize)
{
	io_connect_t port = open_controller();
	if(IO_OBJECT_NULL == port) {
		return kBCCipher_NotInitialized;
	}
	
	uint64_t ctl_u64 = ctlcode;
	
	int rc = IOConnectCallMethod(port, 
				     kBCKernel_CTL, 
				     &ctl_u64, 1, 
				     data_in, insize, 
				     NULL, NULL, 
				     data_out, &outsize);
	
	close_controller(port);
	return result_from_kernel(rc);
}

std::map<bc_u32, std::vector<bc_u8> > g_pool;

bc_cipher_result BCKernelStoreKey(const char* key, size_t keylen, const bc_u8 *pool, bc_key_handle* out_handle)
{
	if (keylen == 0) {
		return kBCCipher_InvalidKey;
	}
	
	bc_ctl_key ctl;
	memset(&ctl, 0, sizeof(ctl));
	
	ctl.addr.ptr = key;
	ctl.size = keylen;
    
    bc_cipher_result cr = controller_ctl(kBCCTL_StoreKey, &ctl, sizeof(ctl), out_handle, sizeof(*out_handle));
    
    // TODO move pool store into kernel driver
    if (cr == kBCKernel_Success) {
        std::vector<bc_u8> newPool;
        newPool.resize(kBCKernel_PoolSize);
        memcpy(newPool.data(), pool, kBCKernel_PoolSize);
        g_pool.insert(std::make_pair((bc_u32)*out_handle, newPool));
        
    }
    
    return cr;
}

bc_cipher_result BCKernelFreeKey(bc_key_handle handle)
{
    std::map<bc_u32, std::vector<bc_u8> >::iterator iter = g_pool.find(handle);
    if (iter != g_pool.end()) {
        g_pool.erase(iter);
    }
    
    return controller_ctl(kBCCTL_FreeKey, &handle, sizeof(handle), NULL, 0);
}

static const char s_zero_pool[kBCKernel_PoolSize] = {0};

bc_cipher_result BCKernelProcess(bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type proc,
                                 bc_key_handle key_handle, uint64_t iv, const char* pool,
                                 char* data, size_t nbytes)
//bc_kern_result bc_kernel_process(bc_alg_id alg_id, bc_mode_id mode_id, bc_key_handle key_handle, bc_proc_type proc, uint64_t iv, const char* pool, char* data, size_t nbytes)
{
	bc_ctl_process ctl;
	memset(&ctl, 0, sizeof(ctl));
	
	ctl.cipher = alg_id;
	ctl.mode = mode_id;
	ctl.with_key = 0;
	ctl.key.handle = key_handle;
	ctl.iv = iv;
	ctl.pool.ptr = pool != NULL ? pool : s_zero_pool;
	ctl.data.ptr = data;
	ctl.nbytes = nbytes;
	ctl.proc = proc;
	
	return controller_ctl(kBCCTL_Process, &ctl, sizeof(ctl), NULL, 0);
}

bc_cipher_result BCKernelProcessWithKey(bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type proc,
                                        const char* key, size_t keysize,
                                        uint64_t iv, const char* pool,
                                        char* data, size_t nbytes)
/*bc_kern_result bc_kernel_process_with_key(bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type proc,
						 const char* key, size_t keysize,
						 uint64_t iv, const char* pool,
						 char* data, size_t nbytes)*/
{
	bc_ctl_process ctl;
	memset(&ctl, 0, sizeof(ctl));

	ctl.cipher = alg_id;
	ctl.mode = mode_id;
	ctl.with_key = 1;
	ctl.iv = iv;
    ctl.pool.ptr = pool != NULL ? pool : s_zero_pool;
	ctl.data.ptr = data;
	ctl.nbytes = nbytes;
	ctl.proc = proc;

	bc_ctl_key* in_key = &ctl.key.data;
	in_key->addr.ptr = key;
	in_key->size = keysize;
    
#ifdef BC_DEBUG
	fprintf(stderr, "bc_ctl_process size is %d\n", sizeof(ctl));
#endif
	return controller_ctl(kBCCTL_Process, &ctl, sizeof(ctl), NULL, 0);
}


bc_cipher_result bc_kernel_benchmark(bc_alg_id alg_id, bc_mode_id mode_id, unsigned bufsiz_mb, unsigned nreps, uint64_t* mean)
{
	bc_ctl_benchmark ctl;
	memset(&ctl, 0, sizeof(ctl));
	
	ctl.alg_id = alg_id;
	ctl.mode_id = mode_id;
	ctl.bufsiz_mb = bufsiz_mb;
	ctl.nreps = nreps;
	
	return controller_ctl(kBCCTL_Benchmark, &ctl, sizeof(ctl), mean, sizeof(*mean));
}
bc_cipher_result BCKernelSetupDisk(const char* file, int alg, int mode, bc_key_handle key_handle, SpaceMap *extents, int options, char* out_devname)
//bc_kern_result bc_kernel_set_disk(const char* path, bc_bmap* block_map, bc_alg_id alg_id, bc_mode_id mode_id, bc_key_handle key, const char* pool, int ro, int autoMount)
{
	bc_ctl_disk ctl;
	memset(&ctl, 0, sizeof(ctl));

	ctl.path.ptr	= file;
	ctl.pathlen	= strlen(file);
	ctl.alg_id	= alg;
	ctl.mode_id	= (bc_mode_id)mode;
	ctl.key		= key_handle;
    
    // TODO move pool store into kernel driver
    std::map<bc_u32, std::vector<bc_u8> >::iterator iter = g_pool.find(key_handle);
    if (iter != g_pool.end()) {
        ctl.pool.ptr	= iter->second.data();
    } else {
        return kBCCipher_InternalError;
    }
    
    ctl.ro		= options & kBCDisk_ReadOnly; //ro;
    ctl.auto_mount	= options & kBCDisk_Automount; //autoMount;
    
    if (extents == NULL) {
        return kBCCipher_InvalidArgument;
    }
    
    bc_bvec exts[extents->count];
    SMExtentNode * extNode = extents->start;
    int i = 0;
    while (extNode) {
        exts[i].offset = extNode->extent.start / kBCKernel_BlockSize;
        exts[i++].length = extNode->extent.length / kBCKernel_BlockSize;
        extNode = extNode->next;
    }
    ctl.nvecs	= extents->count;
    ctl.extents.ptr = exts;
    
	//ctl.nvecs	= block_map->nvecs;
	//ctl.extents.ptr = block_map->extents;
    
	return controller_ctl(kBCCTL_SetDisk, &ctl, sizeof(ctl), NULL, 0);
}


/////////////////////////////////////////////////////////////////////////
//
//	AES-NI Acceleration.
//
/////////////////////////////////////////////////////////////////////////



#define BC_CONFIG_FOLDER "/Library/Application Support/BestCrypt/"

static bc_cipher_result read_kernel_option(const char* name, int* value)
{
    // write the flag to config file in our application support folder.
    int fd = open(name, O_RDONLY);
    if(fd < 0) {
        return kBCCipher_InternalError;
    }
    
    char val;
    read(fd, &val, sizeof(val));
    close(fd);
    
    *value = (val == '1');
    
    return kBCCipher_Success;
}

static bc_cipher_result update_kernel_option(const char* name, int value)
{
    // write the flag to config file in our application support folder.
    int fd = open(name, O_RDWR | O_TRUNC);
    if(fd < 0) {
        return kBCCipher_InternalError;
    }
    
    const char* val = (value ? "1" : "0");
    
    write(fd, val, 1);
    close(fd);
    
    return kBCCipher_Success;
}


extern bc_kern_result bc_kernel_is_aesni_available(int* value)
{
    bc_kern_result rc;
    io_registry_entry_t regentry = find_controller_regentry();
	if(IO_OBJECT_NULL == regentry) {
		return kBCKernel_NotInitialized;
	}
    
	CFBooleanRef prop = (CFBooleanRef)IORegistryEntryCreateCFProperty(regentry, CFSTR(kBCControllerAESNIAvailableKey), NULL, 0);
    IOObjectRelease(regentry);
    if(!prop) {
		return kBCKernel_InternalError;
	}
    
    *value = CFBooleanGetValue(prop);
    CFRelease(prop);
    
    return kBCKernel_Success;
}


extern bc_kern_result bc_kernel_enable_aesni(int do_enable)
{
    // write the flag to config file in our application support folder.
    int fd = open("/Library/Application Support/BestCrypt/bc_aesni", O_RDWR | O_TRUNC);
    if(fd < 0) {
        return kBCKernel_InternalError;
    }
    
    const char* value = (do_enable ? "1" : "0");
    
    write(fd, value, 1);
    close(fd);
    
    return kBCKernel_Success;
}


extern bc_kern_result bc_kernel_is_aesni_enabled(int* is_enabled)
{
    // write the flag to config file in our application support folder.
    int fd = open("/Library/Application Support/BestCrypt/bc_aesni", O_RDONLY);
    if(fd < 0) {
        return kBCKernel_InternalError;
    }
    
    char value;
    read(fd, &value, sizeof(value));
    close(fd);
    
    *is_enabled = (value == '1');
    
    return kBCKernel_Success;
}

bool BCKernelIsAESNISupported()
{
    int flag = 0;
    bc_kernel_is_aesni_available(&flag);
    return (flag == 1);
}

bool BCKernelIsAESNIEnabled()
{
    int flag = 0;
    bc_kernel_is_aesni_enabled(&flag);
    return (flag == 1);
}

bool BCKernelEnableAESNI()
{
    return (kBCKernel_Success == bc_kernel_enable_aesni(1));
}

bool BCKernelDisableAESNI()
{
    return (kBCKernel_Success == bc_kernel_enable_aesni(0));
}


/////////////////////////////////////////////////////////////////////////
//
//	Container guard.
//
/////////////////////////////////////////////////////////////////////////


extern bc_cipher_result bc_kernel_is_container_guard_enabled(int* is_enabled)
{
    size_t oldlen = sizeof(*is_enabled);
    int error = sysctlbyname(BC_GUARD_SYSCTL_NAME, is_enabled, &oldlen, NULL, 0);
    if(error) {
        return kBCCipher_InternalError;
    } else {
        return kBCCipher_Success;
    }
}


extern bc_cipher_result bc_kernel_enable_container_guard(int do_enable)
{
    // Update sysctl value
    int error = sysctlbyname(BC_GUARD_SYSCTL_NAME, NULL, NULL, &do_enable, sizeof(do_enable));
    if(error) {
        return kBCCipher_InternalError;
    } else {
        return kBCCipher_Success;
    }
    
    // Update persistent config file
    int fd = open("/Library/Application Support/BestCrypt/bc_cguard", O_RDWR | O_TRUNC);
    if(fd < 0) {
        return kBCCipher_InternalError;
    }
    
    const char* value = (do_enable ? "1" : "0");
    
    write(fd, value, 1);
    close(fd);
    
    return kBCCipher_Success;
}


extern bc_cipher_result bc_kernel_is_kbfilter_enabled(int* is_enabled)
{
    return read_kernel_option("/Library/Application Support/BestCrypt/bc_kbfilter", is_enabled);
}


extern bc_cipher_result bc_kernel_enable_kbfilter(int do_enable)
{
    return update_kernel_option("/Library/Application Support/BestCrypt/bc_kbfilter", do_enable);
}

