/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 *
 */


#include <stdio.h>
//#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/disk.h>

#include "BCKernel.h"


static const char* g_devpath = NULL;

enum {
	kBCTest_Passed = 0,
	kBCTest_Failed = 1,
	kBCTest_Error = -1,
};


static void print_usage() {
		printf("Usage: bcdiskver -r|-n device\n");
}


static int generate_testbuf(unsigned char* buf, size_t nbytes)
{
	int fd = open("/dev/random", O_RDONLY);
	if(fd < 0) {
		return errno;
	}
	
	int rc = read(fd, buf, nbytes);
	if(rc != nbytes) {
		return errno;
	}
	
	return 0;
}


static int do_random_test(int devfd, uint64_t nblocks)
{
	int rc;
	unsigned char block[kBCKernel_BlockSize] = {0};
	unsigned char buf[kBCKernel_BlockSize] = {0};

	rc = generate_testbuf(block, sizeof(block));
	if(0 != rc) {
		printf("Failed generating test block data with %d\n", rc);
		return kBCTest_Error;
	}
	
	printf("Starting random disk test with device %s, nblocks %llu\n", g_devpath, nblocks);
	printf("Test block:\n");
	
	size_t i;
	uint32_t* dwords = (uint32_t*) block;
	for(i = 1; i <= (sizeof(block) / sizeof(*dwords)); ++i) {
		printf("0x%08x ", *(dwords + i - 1));
		if(0 == (i % 8)) printf("\n");
	}
	
	
	uint64_t p = 0;
	uint64_t cur_block;
	for(cur_block = 0; cur_block < nblocks; ++cur_block) {
		rc = pwrite(devfd, block, sizeof(block), cur_block * sizeof(block));
		if(rc != sizeof(block)) {
			printf("Failed writing block number %llu with %d\n", cur_block, errno);
			return kBCTest_Error;
		}
		
		rc = pread(devfd, buf, sizeof(buf), cur_block * sizeof(buf));
		if(rc != sizeof(buf)) {
			printf("Failed reading block number %llu with %d\n", cur_block, errno);
			return kBCTest_Error;
		}
		
		if(0 != memcmp(block, buf, sizeof(buf))) {
			printf("Check failed at block number %llu\n", cur_block);
			return kBCTest_Failed;
		}
		
		memset(buf, 0, sizeof(buf));
		
		p = ((cur_block + 1) * 100) / nblocks;
		
		printf("checked block %llu / %llu  (%3llu%%)		\r", cur_block + 1, nblocks, p);
	}

	return kBCTest_Passed;
}


static int do_secnum_test(int devfd, uint64_t nblocks)
{
	int rc;
	printf("Starting secnum disk test with device %s, nblocks %llu\n", g_devpath, nblocks);
	
	unsigned char block[kBCKernel_BlockSize] = {0};
	
	uint64_t p = 0;
	uint64_t cur_block;
	for(cur_block = 0; cur_block < nblocks; ++cur_block) {
		*(uint64_t*) block = cur_block;
		
		rc = pwrite(devfd, block, sizeof(block), cur_block * sizeof(block));
		if(rc != sizeof(block)) {
			printf("Failed writing block number %llu with %d\n", cur_block, errno);
			return kBCTest_Error;
		}

		memset(block, 0, sizeof(block));
		
		rc = pread(devfd, block, sizeof(block), cur_block * sizeof(block));
		if(rc != sizeof(block)) {
			printf("Failed reading block number %llu with %d\n", cur_block, errno);
			return kBCTest_Error;
		}
		
		if(cur_block != *(uint64_t*)block) {
			printf("Check failed at block number %llu\n", cur_block);
			return kBCTest_Failed;
		}
		
		memset(block, 0, sizeof(block));
		
		p = ((cur_block + 1) * 100) / nblocks;
		printf("checked block %llu / %llu  (%3llu%%)		\r", cur_block + 1, nblocks, p);
	}
	
	return kBCTest_Passed;
}


int main(int argc, char** argv)
{
	if(argc < 3) {
		print_usage();
		return EXIT_FAILURE;
	}
	
	g_devpath = argv[2];
	
	int devfd = open(g_devpath, O_RDWR);
	if(devfd < 0) {
		printf("Failed opening device file %s: %d\n", g_devpath, errno);
		return EXIT_FAILURE;
	}
	
	struct stat devstat;
	int rc = fstat(devfd, &devstat);
	if(rc != 0) {
		printf("Failed fstat with %d\n", errno);
		return EXIT_FAILURE;
	}
	
	if(!S_ISBLK(devstat.st_mode)) {
		printf("File %s is not a special block device file\n", g_devpath);
		return EXIT_SUCCESS;
	}
	
	uint64_t nblocks;
	if(-1 == ioctl(devfd, DKIOCGETBLOCKCOUNT, &nblocks)) {
		printf("Failed getting device size\n");
		return EXIT_FAILURE;
	}
	
	
	const char* opt = argv[1];
	if(0 == strcmp(opt, "-r")) {
		rc = do_random_test(devfd, nblocks);
	} else if(0 == strcmp(opt, "-n")) {
		rc = do_secnum_test(devfd, nblocks);
	} else {
		print_usage();
		return EXIT_FAILURE;
	}
	
	if(rc == kBCTest_Passed) {
		printf("\nTest successful\n");
		return EXIT_SUCCESS;
	} else if(rc == kBCTest_Failed) {
		printf("\nTest failed\n");
		return EXIT_FAILURE;
	} else {
		printf("\nError occured during test: %d\n", rc);
		return EXIT_FAILURE;
	}
}

