/*
 *	Copyright (c) 2010-2014 Jetico, Inc., Finland
 *	All rights reserved.
 */

#include "datablock.h"
#include "../../certificate/CertificateLib/hashAlgs/hashAlgs.h"
#include "../../certificate/CertificateLib/PGPPkts/pktdefs.h"
#include "../keyFiles.h"
#include "../../container/standard.h"
#include "pke.h"
#include "pbe.h"

using namespace kgghost;


// rebuild contigious empty space starting at specified keyslot.
// used to describe free space and possible hidden key with key descriptor.
static void kmap_rebuild_empty_key(BCKeySlot* kmap, unsigned keynum, unsigned maxkeys, key_descriptor_t* out_kdesc)
{
	BCKeySlot* slot = &kmap[keynum];
	assert(slot->type == kBCKeyType_Empty);

	out_kdesc->base = keynum;
	out_kdesc->total = 1; // including the first slot we just got as an argument.
	//out_kdesc->type = kKBType_Undefined;

	while(++keynum < maxkeys) {
		slot = &kmap[keynum];
		if(slot->type != kBCKeyType_Empty) {
			break;
		}

		out_kdesc->total++;
	}
}


// rebuild known key descriptor starting at specified position.
static void kmap_rebuild_key(BCKeySlot* kmap, unsigned keynum, unsigned maxkeys, key_descriptor_t* out_kdesc)
{
	BCKeySlot* slot = &kmap[keynum];
	assert((slot->type != kBCKeyType_Empty) && (slot->type != kBCKeyType_Part));

	out_kdesc->base = keynum;
	out_kdesc->total = 1;
	out_kdesc->type = slot->type;

	while(keynum < maxkeys) {
		slot = &kmap[++keynum];
		if(slot->type != kBCKeyType_Part) {
			break;
		}

		out_kdesc->total++;
	}
}

// record key descriptor in key map.
static void kmap_record_key(BCKeySlot* kmap, key_descriptor_t* kdesc, size_t size = -1 )
{
	assert((kdesc->base + kdesc->total) <= kBCV8_MaxKeys);

	BCKeySlot* head = &kmap[kdesc->base];
	head->type = kdesc->type;
	head->size = ( size == -1 ?  kdesc->total * kKeySlotSize : size );
	head->param = 0;

	unsigned keynum;
	BCKeySlot* slot;
	for(keynum = 1; keynum < kdesc->total; ++keynum) {
		slot = &kmap[keynum + kdesc->base];
		slot->type = kBCKeyType_Part;
	}
}


// Try to locate and describe contigious free slot area of specified size.
static bool kmap_get_free_space(BCKeySlot* kmap, size_t nslots, size_t maxSlots, key_descriptor_t* out_kdesc)
{
	if(nslots > maxSlots) {
		return false;
	}

	// find contigious vector of empty blocks of enough size
	unsigned keynum = 0;
	while(keynum < maxSlots) {
		BCKeySlot* slot = &kmap[keynum];
		if(slot->type == kBCKeyType_Empty) {
			kmap_rebuild_empty_key(kmap, keynum, maxSlots, out_kdesc);
			if(out_kdesc->total >= nslots) {
				out_kdesc->total = nslots;
				return true;
			} else {
				keynum += out_kdesc->total;
			}
		} else {
			++keynum;
		}
	}

	return false;
}


static void kmap_rebuild_empty_key_reversed(BCKeySlot* kmap, unsigned keynum, key_descriptor_t* out_kdesc)
{
	BCKeySlot* slot = &kmap[keynum];
	assert(slot->type == kBCKeyType_Empty);

	out_kdesc->base = keynum;
	out_kdesc->total = 1; // including the first slot we just got as an argument.

	while(keynum != 0) {
		slot = &kmap[--keynum];
		if(slot->type != kBCKeyType_Empty) {
			break;
		}
		out_kdesc->base--;
		out_kdesc->total++;
	}
}


// Get free space for hidden key.
// Look from the end of the map which is marked by the maxslot argument.
static bool kmap_get_free_space_reversed(BCKeySlot* kmap, size_t need_slots, size_t have_slots, key_descriptor_t* out_kdesc)
{
	if(need_slots > have_slots) {
		return false;
	}

	int slotnum = have_slots - 1;
	while(slotnum >= 0) {
		BCKeySlot* slot = &kmap[slotnum];
		if(slot->type == kBCKeyType_Empty) {
			kmap_rebuild_empty_key_reversed(kmap, slotnum, out_kdesc);
			if(out_kdesc->total >= need_slots) {
				// correct base index to point to the space we need excatly from the end.
				out_kdesc->base += out_kdesc->total - need_slots;
				out_kdesc->total = need_slots;
				return true;
			} else {
				slotnum -= out_kdesc->total;
			}
		} else {
			--slotnum;
		}
	}

	return false;

}


// Mark referenced key slots as empty.
static void kmap_mark_empty(BCKeySlot* kmap, key_descriptor_t* kdesc)
{
	for(unsigned i = 0; i < kdesc->total; ++i) {
		BCKeySlot* slot = &kmap[kdesc->base + i];
		slot->size = 0;
		slot->param = 0;
		slot->type = kBCKeyType_Empty;
	}
}


// Merge src map into dst. Dst will container all non-empty descriptors of both maps.
static void kmap_merge( const BCKeySlot* src, BCKeySlot* dst)
{
	for(unsigned i = 0; i < kBCV8_MaxKeys; ++i) {
		BCKeySlot* slot = &dst[i];
		if(slot->type == kBCKeyType_Empty) {
			memcpy(slot, &src[i], sizeof(*slot));
		}
	}
}


// Get number of key slots required to contain encoded key data of this size.
static size_t slots_for_key_size(size_t keysize)
{
	return (keysize + kKeySlotSize - 1) / kKeySlotSize;
}


namespace kgghost 
{

	//void DataBlock_Init(DATA_BLOCK* db, size_t dbsize, int alg_id, int mode_id, int hash_id, unsigned long long datasize)
	//{
	//	shred((bc_u8*)db, dbsize);
	//	db_header* header = (db_header*) db;
	//	memset(header, 0, sizeof(*header));
	//	header->alg_id		= alg_id;
	//	header->mode_id		= mode_id;
	//	header->hash_id		= hash_id;
	//	header->wKeyGenId	= kKGID;
	//	header->setVersion( BCV8_VERSION_TAG );
	//	header->data_start	= kBCV8_InitialDBSize;
	//	// 04.09.2014 - See ..\versionInfo.txt
	//	header->data_length	= datasize - dbsize;
	//	//header->data_length	= datasize;
	//	memcpy(header->signature, DB_HEADER_SIGNATURE, kBCV8_SignatureSize);
	//	memcpy(header->volumeLabel, DB_HEADER_VOLUME_LABEL, sizeof(header->volumeLabel));
	//}

	const int saltPos = 0;

	void DataBlock_Init(
			DATA_BLOCK* db, size_t dbsize, int alg_id, int mode_id, int hash_id, unsigned long long datasize,
			bc_u16 kgID, bc_u32 iterations, const bc_u8 *salt, int saltSize )
	{
		shred((bc_u8*)db, dbsize);
		db_header* header = (db_header*) db;
		memset(header, 0, sizeof(*header));
		header->alg_id		= alg_id;
		header->mode_id		= mode_id;
		header->hash_id		= hash_id;

		if ( kgID == (bc_u16)-1 )
			kgID = kKGID_V5;

		if ( iterations == 0 )
			iterations = 16*1024;

		header->wKeyGenId	= kgID;

		if ( kgID >= kKGID_V5 )
		{
			header->iterations	= iterations;
			memcpy( db->keys[saltPos].data, salt, saltSize );
			header->keymap[saltPos].type = kBCKeyType_Salt;
			header->keymap[saltPos].size = max( 1, saltSize );
		}

		header->setVersion( BCV8_VERSION_TAG );
		header->data_start	= kBCV8_InitialDBSize;
		// 04.09.2014 - See ..\versionInfo.txt
		header->data_length	= (datasize - dbsize) & 0xFFFFFFFFFFFFF000; // Bound size to 4K
		//header->data_length	= datasize;
		memcpy(header->signature, DB_HEADER_SIGNATURE, kBCV8_SignatureSize);
		memcpy(header->volumeLabel, DB_HEADER_VOLUME_LABEL, sizeof(header->volumeLabel));
	}

	size_t DataBlock_CapacityForSize(size_t dbsize)
	{
		if(dbsize <= sizeof(DATA_BLOCK)) {
			return 0;
		}

        size_t capacity = ((dbsize - sizeof(DATA_BLOCK)) / kKeySlotSize);
        return capacity < kBCV8_MaxKeys ? capacity : kBCV8_MaxKeys;
	}


	size_t DataBlock_SizeForCapacity(size_t nkeys)
	{
		return (sizeof(DATA_BLOCK) + (nkeys * kKeySlotSize));
	}


    BCContainerResult DataBlock_Extend(size_t & dbsize)
    {
        if (dbsize + kBCV8_DataBlockExtentionUnit <= kBCV8_MaximumDataBlockSize) {
            dbsize += kBCV8_DataBlockExtentionUnit;
            return kBCContainer_Success;
        }
        return kBCContainer_NoKeySpace;
    }
    
	bool DataBlock_IsHeaderEncrypted( const DATA_BLOCK* db )
	{
		db_header* hdr = (db_header*) db;
		return (0 != memcmp(hdr->signature, DB_HEADER_SIGNATURE, sizeof(hdr->signature)));
	}


	static void DataBlock_CalcHeaderDigest(db_header* hdr, int hash_id, bc_u8* out_digest)
	{
		db_header copy;
		memcpy(&copy, hdr, sizeof(copy));

		// nullify some service header fields that do not affect header digest.
		memset(copy.jmpCode, 0, sizeof(copy.jmpCode));
		memset(copy.occupiedTag, 0, sizeof(copy.occupiedTag));
		memcpy(copy.volumeLabel, DB_HEADER_VOLUME_LABEL, sizeof(copy.volumeLabel));
		memcpy(copy.signature, DB_HEADER_SIGNATURE, sizeof(copy.signature));

		calculate_digest(hash_id, &copy, sizeof(copy), out_digest);
	}


	template < class T >
	static bool decrypt_header(T* enblock, const bc_u8* pool, CipherBase* cipher, const key_data_t* key)
	{
		bc_u64 iv = *(bc_u64*) enblock->iv;
		cipher->decrypt(iv, key->buf, sizeof(key->buf), pool, (bc_u8*) enblock, sizeof(*enblock) - sizeof(enblock->iv));

		bc_u8 digest[sizeof(enblock->digest)] = {0};
		DataBlock_CalcHeaderDigest((db_header*) enblock, key->hash_id, digest);

		return (0 == memcmp(digest, enblock->digest, sizeof(enblock->digest)));
	}

	template < class T >
	void db_calcHeaderDigest( DATA_BLOCK* db, const key_data_t* key )
	{
		T *enblock = (T*)db;
		DataBlock_CalcHeaderDigest((db_header*)enblock, ((db_header*)enblock)->hash_id, enblock->digest);

		// digest is always encrypted.
		bc_u64 iv = *(bc_u64*) enblock->iv;

		BCScopedPtr < CipherBase >  sCipher( CipherFactory::CreateCipher() );

		if ( kBCCipher_Success != sCipher->init(key->alg_id, key->getMode()))
			assert( false );
		else
			sCipher->encrypt(iv, key->buf, sizeof(key->buf), db->pool, (bc_u8*)enblock->digest, sizeof(enblock->digest));
	}


	void DataBlock_UpdateDigest(DATA_BLOCK* db, const key_data_t* key)
	{
		int digestSize = 0;
		if ( ha_getDigestSize( &digestSize, key->hash_id  ) != hae_ok ) // db->header.header.hash_id
		{
			assert( false );
			return;
		}

		if ( digestSize <= 32 )
			db_calcHeaderDigest < db_enc_block32 > ( db, key );
		else
			db_calcHeaderDigest < db_enc_block64 > ( db, key );
	}


	template < class T >
	bool db_verifyDigest( DATA_BLOCK* db, const key_data_t* key )
	{
		T *enblock = (T*)db;
		bc_u64 iv = *(bc_u64*) enblock->iv;
    
		BCScopedPtr < CipherBase >  sCipher( CipherFactory::CreateCipher() );
		if ( kBCCipher_Success != sCipher->init(key->alg_id, key->getMode())) {
		   assert( false ); // we are using decoded key here, algorithm should be valid
		}

		bc_u8 digest[kBCV8_DigestSize64] = {0};

		if(DataBlock_IsHeaderEncrypted(db)) {
			T copy;
			memcpy(&copy, (T*) db, sizeof(copy));

			return decrypt_header(&copy, db->pool, sCipher.get(), key);
		} else {
			DataBlock_CalcHeaderDigest((db_header*) db, key->hash_id, digest);

			//// windows version decrypt ecrypted digest
			//bc_u8 decryptedDigest[kBCV8_DigestSize] = {0};
			//memcpy( decryptedDigest, enblock->digest, sizeof(decryptedDigest) );

			//sCipher->decrypt(iv, key->buf, sizeof(key->buf), db->pool, decryptedDigest, sizeof(decryptedDigest));
			//return (0 == memcmp(digest, decryptedDigest, sizeof(digest)));

			//while( true )
			//{
			//	memcpy( decryptedDigest, enblock->digest, sizeof(decryptedDigest) );

			//	sCipher->decrypt(iv, key->buf, sizeof(key->buf), db->pool, decryptedDigest, sizeof(decryptedDigest));

			//	int res = memcmp(digest, decryptedDigest, sizeof(digest));

			//	sCipher->encrypt(iv, key->buf, sizeof(key->buf), db->pool, decryptedDigest, sizeof(decryptedDigest));

			//	res = memcmp(enblock->digest, decryptedDigest, sizeof(digest));
			//}

			// encrypt generated digest to compare with encrypted header digest directly
			sCipher->encrypt(iv, key->buf, sizeof(key->buf), db->pool, digest, sizeof(digest));
			return (0 == memcmp(digest, enblock->digest, sizeof(enblock->digest)));
		}
	}

	bool DataBlock_VerifyDigest(DATA_BLOCK* db, const key_data_t* key)
	{
		int digestSize = 0;
		if ( ha_getDigestSize( &digestSize, key->hash_id ) != hae_ok ) // db->header.header.hash_id
		{
			assert( false );
			return false;
		}

		if ( digestSize <= 32 )
			return db_verifyDigest < db_enc_block32 >( db, key );
		else
			return db_verifyDigest < db_enc_block64 >( db, key );
	}


	template < class T >
	void db_encryptHeader( DATA_BLOCK* db, const key_data_t* key)
	{
		T *encp = (T*)db;
		DataBlock_CalcHeaderDigest(&encp->header, encp->header.hash_id, encp->digest);

		BCScopedPtr < CipherBase >  sCipher( CipherFactory::CreateCipher() );
		if ( kBCCipher_Success != sCipher->init(key->alg_id, key->getMode()))
			assert( false ); // we are using decoded key here, algorithm should be valid

		bc_u64 iv = *(bc_u64*) encp->iv;
		sCipher->encrypt(iv, key->buf, sizeof(key->buf), db->pool, (bc_u8*) encp, sizeof(*encp) - sizeof(encp->iv));
	}

	void DataBlock_EncryptHeader(DATA_BLOCK* db, const key_data_t* key)
	{
		int digestSize = 0;
		if ( ha_getDigestSize( &digestSize, key->hash_id ) != hae_ok ) // db->header.header.hash_id
		{
			assert( false );
			return;
		}

		if ( digestSize <= 32 )
			db_encryptHeader < db_enc_block32 > ( db, key );
		else
			db_encryptHeader < db_enc_block64 > ( db, key );
	}


	template < class T >
	bool db_decryptHeader( DATA_BLOCK* db, const key_data_t* key)
	{
		T *eb = (T*)db;
		T decrypted;
        memcpy(&decrypted, eb, sizeof(decrypted));

		BCScopedPtr < CipherBase >  sCipher( CipherFactory::CreateCipher() );
		if ( kBCCipher_Success != sCipher->init(key->alg_id, key->getMode())) {
			assert( false );
			return false;
		}

		if(!decrypt_header(&decrypted, db->pool, sCipher.get(), key)) {
			return false;
		}

        memcpy((db_header*)db, (db_header*)&decrypted, sizeof(db_header));
		DataBlock_UpdateDigest(db, key);

		return true;
	}


	bool DataBlock_DecryptHeader(DATA_BLOCK* db, const key_data_t* key)
	{
		int digestSize = 0;
		if ( ha_getDigestSize( &digestSize, key->hash_id ) != hae_ok ) // db->header.header.hash_id
		{
			assert( false );
			return false;
		}

		if ( digestSize <= 32 )
			return db_decryptHeader < db_enc_block32 > ( db, key );
		else
			return db_decryptHeader < db_enc_block64 > ( db, key );
	}

	bool getSalt_v5( const DATA_BLOCK* db, int hash, cPTB &ptbKFSaltOut, iChannel **keyFilesT, size_t keyFilesQ, const bc_u8 **pSalt, size_t *pSaltSize )
	{
		bool is_encrypted = DataBlock_IsHeaderEncrypted(db);

		if ( !is_encrypted && ( db->header.header.wKeyGenId != kKGID_V5 ) )
		{
			*pSalt = NULL;
			*pSaltSize = 0;
			return false;
		}

		*pSalt = db->keys[saltPos].data;

		int size;
		
		if ( ha_getDigestSize( &size, hash ) != hae_ok )
		{
			assert( false );
			*pSaltSize = kBCV8_DigestSize64;
		}
		else
			*pSaltSize = size;

		if ( !is_encrypted )
		{
			assert( db->header.header.keymap[saltPos].type == kBCKeyType_Salt );
		}

		if ( generateSaltBuffer( ptbKFSaltOut, keyFilesT, keyFilesQ, hash, *pSalt, *pSaltSize ) )
		{
			*pSalt = ptbKFSaltOut.buffer;
			*pSaltSize = ptbKFSaltOut.bufferSize;
		}

		return true;
	}

	bool checkSpaceAndFillKeyMap( BCKeySlot* kMap, key_descriptor_t &kdesc, int kbType, size_t slots, size_t maxSlots, bool bHidden )
	{
		// locate free key blocks for new key.
		// we look for space for main part keys from the first key block and for hidden part keys from the last.
		bool has_space;

		if ( bHidden )
			has_space = kmap_get_free_space_reversed(kMap, slots, maxSlots, &kdesc);
		else
			has_space = kmap_get_free_space(kMap, slots, maxSlots, &kdesc);

		kdesc.type = kbType;

		kmap_record_key( kMap, &kdesc ); //, slots );

		return has_space;
	}

	//
	//	Hidden part key block has additional verification block.
	//	It is has to be reserved early and filled by fillOriginalMapAndUpdateHidden later.
	//
	int getHPVerificationBlockSize( const key_data_t* key )
	{
		if ( !key->is_hidden() )
			return 0;

		int digestSize;
		if ( ha_getDigestSize( &digestSize, key->hash_id ) != hae_ok )
		{ assert( false ); }

		if ( digestSize > 32 )
			return sizeof( verification_block_64 );
		else
			return sizeof( verification_block_32 );
	}

	void fillOriginalMapAndUpdateHidden( const key_data_t* key, DATA_BLOCK* db, key_descriptor_t &kdesc, size_t encoded_size, BCKeySlot* hiddenMap )
	{
		bc_u8*	kbdata = (bc_u8*) &db->keys[kdesc.base];

		BCKeySlot* origMap = ((db_header*)db)->keymap;

		if(!key->is_hidden())
			kmap_record_key( origMap, &kdesc, encoded_size);
		else
		{
			// for hidden key add verification block.
			verification_data verdata;
			int hash = key->hash_id;
			BCScopedPtr < CipherBase >  sCipher( CipherFactory::CreateCipher() );
			if ( kBCCipher_Success != sCipher->init(key->alg_id, key->getMode()))
				assert( false ); 

			int digestSize;
			if ( ha_getDigestSize( &digestSize, hash ) != hae_ok )
			{ assert( false ); }

			if ( digestSize > 32 )
			{
				verification_block_64* verblock = (verification_block_64*) (kbdata + encoded_size);
				KGEncryptBlock64(&verdata, *(bc_u64*) verblock->iv, sCipher.get(), hash, key->buf, sizeof(key->buf), db->pool, &verblock->verdata);
			}
			else
			{
				verification_block_32* verblock = (verification_block_32*) (kbdata + encoded_size);
				KGEncryptBlock32(&verdata, *(bc_u64*) verblock->iv, sCipher.get(), hash, key->buf, sizeof(key->buf), db->pool, &verblock->verdata);
			}

			if(hiddenMap) {
				kmap_record_key(hiddenMap, &kdesc ); // , maxSlots);
			}
		}
	}

	BCContainerResult makePKEBlock( DATA_BLOCK* db, BCKeySlot* virtualKMap, BCKeySlot* hiddenMap, size_t maxSlots, const key_data_t* key, BCPassword* pwd, bool bCheckSizeOnly )
	{
		//// Just test...
		//pke_set_t set;
		//if ( pwd->m_secretKeyQ != 0 )
		//{
		//	cPTB ptbSK;
		//	pwd->m_secretKeyT[0]->readToBlock( ptbSK );

		//	cPassword p( pwd->buf );
		//	ParseSecretKey( ptbSK.buffer, ptbSK.bufferSize, kffX509, &p, &set.ska);
		//}

		for( int i = 0; i < pwd->m_publicKeyQ; i++ )
		{
			cPTB ptbKey;
			if ( !pwd->m_publicKeyT[i]->readToBlock( ptbKey ) )
				return kBCContainer_InvalidPassword;

			unsigned kbSize = 0;
			int res = pke_encode( key, ptbKey.buffer, ptbKey.bufferSize, NULL, kbSize );
			if ( res != x509e_bufferTooSmall )
				return kBCContainer_InternalError;

			size_t slots = ( kbSize + getHPVerificationBlockSize( key ) + kKeySlotSize - 1 )/ kKeySlotSize;
			
			key_descriptor_t kdesc;
			if ( !checkSpaceAndFillKeyMap( virtualKMap, kdesc, kKBType_PKE, slots, maxSlots, key->is_hidden() ) )
				return kBCContainer_NoKeySpace;

			//
			//	Virtual key map is filled.
			//	So we can pass it to check for others block creators.
			//
			if ( bCheckSizeOnly )
				continue;

			bc_u8*	kbdata = (bc_u8*) &db->keys[kdesc.base];
			assert( kbSize <= kdesc.total * kKeySlotSize );

			res = pke_encode( key, ptbKey.buffer, ptbKey.bufferSize, kbdata, kbSize ); //, set.ska );

			if ( res != x509e_ok )
				return kBCContainer_InternalError;

			//if ( set.ska != NULL )
			//{
			//	key_data_t decKey;
			//	res = pke_decode( kbdata, kbSize, pwd, NULL, key->hash_id, &decKey );
			//}

			fillOriginalMapAndUpdateHidden( key, db, kdesc, kbSize, hiddenMap );
		}

		return kBCContainer_Success;
	}

	BCContainerResult makePBEBlock( DATA_BLOCK* db, BCKeySlot* virtualKMap, BCKeySlot* hiddenMap, size_t maxSlots, const key_data_t* key, BCPassword* pwd, bool bCheckSizeOnly )
	{
		key_descriptor_t kdesc;

		size_t slots = ( getPBEBlockSize( key ) + getHPVerificationBlockSize( key ) + kKeySlotSize - 1 )/ kKeySlotSize;
		
		if ( !checkSpaceAndFillKeyMap( virtualKMap, kdesc, kKBType_PBE, slots, maxSlots, key->is_hidden() ) )
			return kBCContainer_NoKeySpace;

		//
		//	Virtual key map is filled.
		//	So we can pass it to check for others block creators.
		//
		if ( bCheckSizeOnly )
			return kBCContainer_Success;

		bc_u8*	kbdata = (bc_u8*) &db->keys[kdesc.base];
		size_t	kbsize = kdesc.total * kKeySlotSize;

		const bc_u8 *salt = NULL;
		size_t saltSize = 0;
		cPTB ptbTmp;
		int hash = key->hash_id;
		getSalt_v5( db, hash, ptbTmp, pwd->m_keyFilesT, pwd->m_keyFilesQ, &salt, &saltSize );

		if ( salt != NULL ) // db->header.header.wKeyGenId == kKGID_V5 )
			pwd->generateKey( hash, salt, saltSize, db->header.header.iterations );

		BCScopedPtr < CipherBase >  sCipher( CipherFactory::CreateCipher() );
		if ( kBCCipher_Success != sCipher->init(key->alg_id, key->getMode()))
			assert( false ); 

		size_t encoded_size = KGEncodeKeyBlock(key, pwd, sCipher.get(), hash, kbdata, kbsize);

		fillOriginalMapAndUpdateHidden( key, db, kdesc, encoded_size, hiddenMap );

		return kBCContainer_Success;
	}
	
	BCContainerResult DataBlock_EncodeKey(DATA_BLOCK* db, size_t & dbsize, const key_data_t* key,
                               BCPassword* pwd, BCKeySlot* hiddenMap, bool bCheckSizeOnly )
	{
		LIBLOG_INFO( "Begin %d, %d", (int)dbsize, (int)bCheckSizeOnly );

		assert(!DataBlock_IsHeaderEncrypted(db));

		BCKeySlot* origMap = ((db_header*)db)->keymap;

		// merge current and hidden maps
		BCKeySlot mergedMap[kBCV8_MaxKeys];
		memset(mergedMap, 0, sizeof(mergedMap));

		kmap_merge(origMap, mergedMap);

		if ( hiddenMap )
			kmap_merge(hiddenMap, mergedMap);

		BCContainerResult res = kBCContainer_InvalidArgument;

        while (true) {
            size_t maxSlots = DataBlock_CapacityForSize(dbsize);
            if ( pwd->isPBE() ) {
                res = makePBEBlock( db, mergedMap, hiddenMap, maxSlots, key, pwd, bCheckSizeOnly );
                if (res == kBCContainer_NoKeySpace && DataBlock_Extend(dbsize) == kBCContainer_Success) {
                    continue;
                }
                if ( res != kBCContainer_Success ) {
                    LIBLOG_INFO( "makePBEBlock failed %d", res );
                    return res;
                }
            } else {
                LIBLOG_INFO( "Is not PBE" );
            }
            
            if ( pwd->m_publicKeyQ != 0 ) {
                res = makePKEBlock( db, mergedMap, hiddenMap, maxSlots, key, pwd, bCheckSizeOnly );
                if (res == kBCContainer_NoKeySpace && DataBlock_Extend(dbsize) == kBCContainer_Success) {
                    continue;
                }
                if ( res != kBCContainer_Success ) {
                    LIBLOG_INFO( "makeKEEBlock faled %d", res );
                    return res;
                }
            }
            break;
        }

		LIBLOG_INFO( "End %d", res );
		return res;
	}

	static bool check_verification_block( bc_u8 *data, int dataSize, CipherBase* cipher, key_data_t* key, bc_u8 *pool )
	{
		int digestSize;

		if ( ha_getDigestSize( &digestSize, key->hash_id ) != hae_ok )
		{ assert( false ); return false; }

		verification_data verdata;

		if ( digestSize > 32 )
		{
			const verification_block_64* verblock = (const verification_block_64*)data;

			if ( !KGDecryptBlock64(&verblock->verdata, *(bc_u64*)verblock->iv, cipher, key->hash_id, key->buf, sizeof(key->buf), pool, &verdata) )
				return false;
		}
		else
		{
			//
			//	In 32 bit mode verification blocks for KG v4 and v5 are compatible .
			//
			verification_block_32* verblock = (verification_block_32*)data;

			if ( dataSize < sizeof(*verblock) )

			if ( !KGDecryptBlock32(&verblock->verdata, *(bc_u64*)verblock->iv, cipher, key->hash_id, key->buf, sizeof(key->buf), pool, &verdata ) )
				return false;
		}

		// verify decrypted verdata
		
		return verdata.check();
	}


	// Decode key from keyblock referenced by descriptor.
    static bool decode_keyblock(DATA_BLOCK* db, const BCPassword* pwd, CipherBase* cipher, int hash, key_descriptor_t* kdesc, key_data_t* out_key )
	{
		bc_u8* kbdata	= (bc_u8*) &db->keys[kdesc->base];
		unsigned kbsize	= kdesc->total * kKeySlotSize;

		size_t decoded_size = KGDecodeKeyBlock(kbdata, kbsize, kdesc->type, pwd, cipher, hash, out_key);
		if (0 == decoded_size) {
            return false;
        }

		// If key is hidden verify it. See verification_data comments.
		if(out_key->is_hidden() &&
			!check_verification_block( kbdata + decoded_size, (int)(kbsize - decoded_size), cipher, out_key, db->pool ) )
				return false;

		// update decoded key info.
        // TODO add verification block size to slots count calculation
		kdesc->total = (unsigned)slots_for_key_size(decoded_size);

		return true;
	}

#define _slotFirst

	static bool showProgress(
			cProgressActor *progressActor,
			const char *algName, int algNum, int algTotal,
			const char *modeName,
			int hashId, int hashNum, int hashTotal,
			int slotNum, int slotTotal,
			const char *version = 0 )
	{
		static int count = 0;
		count++;

		if ( progressActor == NULL )
			return true;

		const char *hashName = ha_getHashAlgName( hashId );

#ifdef _ANDROID
		const char *newLine = "<br/>";
#else
		const char *newLine = "\n";
#endif

		char str[256];
#ifdef _slotFirst
		sprintf( str,
				"Slot (%d/%d) %s%s"
				"Algorithm (%d/%d) %s(%s)%s"
				"Hash (%d/%d) %s",
				slotNum + 1, slotTotal, (version ? version : ""), newLine,
			algNum + 1, algTotal, algName, modeName, newLine,
			hashNum + 1, hashTotal, hashName
			);
#else
		sprintf( str,
				"Algorithm %s(%s) (%d/%d) %s%s"
				"Hash %s (%d/%d)%s"
				"Slot (%d/%d)",
			algName, modeName, algNum + 1, algTotal, (version ? version : ""), newLine,
			hashName, hashNum + 1, hashTotal, newLine,
			slotNum + 1, slotTotal
			);
#endif

		// Continue == 0
		bool bContinue = progressActor->func( str, 0, 0 ) == 0;
		if ( !bContinue )
			progressActor->m_bCancelled = true;

		return bContinue;
	}

	static bool walkNotEncrypted(
			DATA_BLOCK* db,
			BCPassword* pwd, iChannel**	keyFilesT, size_t keyFilesQ,
			key_descriptor_t* out_desc, key_data_t* out_key,
			size_t maxSlots, CipherBase *sCipher
		)
	{
		db_header*	header = (db_header*) db;
		BCKeySlot*	kmap = header->keymap;

		if( kBCCipher_Success != sCipher->init(header->alg_id, header->getMode()) )
			return false;

        int hash = header->hash_id;
		bc_u32 iterations = header->getIterationCount();

		const bc_u8 *salt = NULL;
		size_t saltSize = 0;
		cPTB ptbTmp;
		getSalt_v5( db, hash, ptbTmp, keyFilesT, keyFilesQ, &salt, &saltSize );

		if ( salt != NULL ) // db->header.header.wKeyGenId == kKGID_V5 )
			pwd->generateKey( hash, salt, saltSize, iterations );

		for( unsigned slotnum = 0; slotnum < maxSlots; slotnum++ )
		{
			// for encrypted header we have got (possible) cipher and hash from the caller but we don't have a key map.
			// we have to employ the same strategy as for looking hidden keys except to treat all key space as hidden.
			out_desc->base	= slotnum;
			out_desc->total	= maxSlots - slotnum;
			out_desc->type	= kKBType_Unknown;

			// for open header we can use key map for more precise decoding.
			BCKeySlot* slot = &kmap[slotnum];

			if ( ( slot->type == kBCKeyType_Part ) || ( slot->type == kBCKeyType_Salt ) )
                continue;

			// assert(slot->type != kBCKeyType_Part);

			if(slot->type == kBCKeyType_Empty) {
				// look for hidden key.
				out_desc->type = kKBType_Unknown;
				kmap_rebuild_empty_key(kmap, slotnum, maxSlots, out_desc);
			} else {
				kmap_rebuild_key(kmap, slotnum, maxSlots, out_desc);
			}

            if ( decode_keyblock(db, pwd, sCipher, hash, out_desc, out_key ) )
				return true;
		}

		return false;
	}
	
	//
    // if header is encrypted we will have to iterate all possible alg/mode/hash configurations
    // there is a reason to iterate cipher configurations for each key block and not the other way aroud:
    // if password is correct it will most probably decode some very first key blocks.
	//
	static bool walkEncryptedV8(
			DATA_BLOCK* db, BCPassword* pwd, key_descriptor_t* out_desc, key_data_t* out_key, cProgressActor *progressActor,
			size_t slotFirst, size_t slotLast, size_t maxSlots, CipherBase *sCipher, BCEnumAlgorithms &algs, BCEnumModes &modes
		)
	{
		if ( ( progressActor != NULL ) && ( progressActor->m_bCancelled ) )
			return false;
		
		BCEnumHashes        hashes( true );

#ifdef _slotFirst
		for( unsigned slotnum = slotFirst; slotnum < slotLast; ++slotnum )
#endif
		{
			for(unsigned algnum = 0; algnum < algs.total(); ++algnum)
			{
				BCAlgInfo* alg = algs.get(algnum);

				for(unsigned modenum = 0; modenum < modes.total(); ++modenum)
				{
					bc_mode_id mode = modes.get(modenum);

					if ( !sCipher->IsModeSupported( alg, mode ) ) {
						continue; // mode not supported.
					}

					sCipher->init(alg, mode);

					for( unsigned hashnum = 0; hashnum < hashes.total(); ++hashnum )
					{
						int hash = hashes.get(hashnum);
						pwd->bReady = false;; // Bug #1824. v8 encrypted headers not recognized?

#ifndef _slotFirst
						for( slotnum = slotFirst; slotnum < slotLast; ++slotnum )
#endif
						{
							if ( !showProgress(
									progressActor,
									alg->name, algnum, algs.total(),
									sCipher->modeName(),
									hash, hashnum, hashes.total(),
									slotnum, maxSlots, "V8" ) )
								return false;

							// for encrypted header we got (possible) cipher and hash from the caller but we don't have a key map.
							// we have to employ the same strategy as for looking hidden keys except to treat all key space as hidden.
							out_desc->base	= slotnum;
							out_desc->total	= maxSlots - slotnum;
							out_desc->type	= kKBType_Unknown;

							if ( decode_keyblock(db, pwd, sCipher, hash, out_desc, out_key ) )
								return true;
						}
					}
				}
			}
		}

		return false;
	}

	static bool walkEncryptedV9(
			DATA_BLOCK* db,
			BCPassword* pwd, iChannel**	keyFilesT, size_t keyFilesQ,
			key_descriptor_t* out_desc, key_data_t* out_key, cProgressActor *progressActor,
			size_t slotFirst, size_t slotLast, size_t maxSlots, CipherBase *sCipher, BCEnumAlgorithms &algs, BCEnumModes &modes
		)
	{
		if ( ( progressActor != NULL ) && progressActor->m_bCancelled )
			return false;

        int hash = -1;
		const bc_u8 *salt = NULL;
		size_t saltSize = 0;
		unsigned slotnum = 0;
		
		BCEnumHashes        hashes( false, bchaWhirlpool512 );
		BCEnumIterations	iterations;

		for( unsigned hashnum = 0; hashnum < hashes.total(); ++hashnum )
		{
			hash = hashes.get(hashnum);

			cPTB ptbTmp;
			if ( !getSalt_v5( db, hash, ptbTmp, keyFilesT, keyFilesQ, &salt, &saltSize ) )
				return false;

			for( unsigned iterationN = 0; iterationN < iterations.total(); ++iterationN )
			{
				pwd->generateKey( hash, salt, saltSize, iterations.get( iterationN ) );

				for( slotnum = slotFirst; slotnum < slotLast; ++slotnum )
				{
					if ( slotnum == saltPos )
						continue;

					for(unsigned algnum = 0; algnum < algs.total(); ++algnum)
					{
						BCAlgInfo* alg = algs.get(algnum);

						for(unsigned modenum = 0; modenum < modes.total(); ++modenum)
						{
							bc_mode_id mode = modes.get(modenum);

							if ( !sCipher->IsModeSupported( alg, mode ) )
								continue; // mode not supported.

							sCipher->init(alg, mode);


							if ( !showProgress(
									progressActor,
									alg->name, algnum, algs.total(),
									sCipher->modeName(),
									hash, hashnum, hashes.total(),
									slotnum, maxSlots ) )
								return false;

							// for encrypted header we got (possible) cipher and hash from the caller but we don't have a key map.
							// we have to employ the same strategy as for looking hidden keys except to treat all key space as hidden.
							out_desc->base	= slotnum;
							out_desc->total	= maxSlots - slotnum;
							out_desc->type	= kKBType_Unknown;

							if ( decode_keyblock( db, pwd, sCipher, hash, out_desc, out_key ) )
								return true;
						}
					}
				}
			}
		}

		return false;
	}

    bool DataBlock_DecodeKey(
			DATA_BLOCK* db, size_t dbsize,
			BCPassword* pwd,
			BCKeySlot* hiddenMap,
			key_descriptor_t* out_desc,
			key_data_t* out_key,
			cProgressActor *progressActor
		)
	{
    	LIBLOG_INFO("Begin %p, %lu", db, dbsize );

		const size_t _4K = 1024 * 4;
		size_t minSlots = DataBlock_CapacityForSize( _4K );
		size_t maxSlots = DataBlock_CapacityForSize(dbsize);

		if ( maxSlots == 0 )
			return false;

		iChannel**	keyFilesT = pwd->m_keyFilesT;
		size_t		keyFilesQ = pwd->m_keyFilesQ;

		BCScopedPtr < CipherBase >  sCipher( CipherFactory::CreateCipher() );

        if ( DataBlock_IsHeaderEncrypted(db) )
		{
			BCEnumAlgorithms    algs( 240 ); // bcsaRIJN );
			BCEnumModes         modes;

            if (progressActor) progressActor->init(6624, 41668);
            if (
				!walkEncryptedV8( db, pwd,						out_desc, out_key, progressActor,			0,        1, minSlots, sCipher.get(), algs, modes ) &&
				!walkEncryptedV9( db, pwd, keyFilesT, keyFilesQ,out_desc, out_key, progressActor,			0, minSlots, minSlots, sCipher.get(), algs, modes ) &&
				!walkEncryptedV8( db, pwd,						out_desc, out_key, progressActor,			1, minSlots, minSlots, sCipher.get(), algs, modes ) &&
				!walkEncryptedV9( db, pwd, keyFilesT, keyFilesQ,out_desc, out_key, progressActor, minSlots +1, maxSlots, maxSlots, sCipher.get(), algs, modes ) &&
				!walkEncryptedV8( db, pwd,						out_desc, out_key, progressActor, minSlots +1, maxSlots, maxSlots, sCipher.get(), algs, modes )
			   )
				return false;
		}
		else
		if ( !walkNotEncrypted( db, pwd, keyFilesT, keyFilesQ, out_desc, out_key, maxSlots, sCipher.get() ) )
	        return false;

	    // if key is hidden mark it on the map
	    if (out_key->is_hidden() && hiddenMap) {
		    kmap_record_key(hiddenMap, out_desc);
	    }

		LIBLOG_INFO( "End success" );
	    return true;
	}

	void DataBlock_DeleteKey(DATA_BLOCK* db, size_t dbsize, key_descriptor_t* desc, BCKeySlot* hiddenMap)
	{
		bc_u8*	kbdata = (bc_u8*) &db->keys[desc->base];
		size_t	kbsize = desc->total * kKeySlotSize;

		shred(kbdata, kbsize);
		BCKeySlot* kmap = ((db_header*)db)->keymap;
		kmap_mark_empty(kmap, desc);
        if (hiddenMap) kmap_mark_empty(hiddenMap, desc);
	}

} // namespace kgghost

