/*
 *	Copyright (c) 2010 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	KGGHOST serialized key storage format and decoding interface.
 */

#ifndef KGGHOST_DATA_BLOCK_H
#define KGGHOST_DATA_BLOCK_H

#include <stdlib.h>

#include "../../container/password.h"
#include "keyblock.h"
#include "db_header.h"
#include "kgghost.h"

namespace kgghost {
	
	
#pragma pack(push, 1)
	
    enum {
        kBCKeyType_Part = -1,
        kBCKeyType_Empty = 0,
        kBCKeyType_PBE = 1,
        kBCKeyType_SSS = 2,
        kBCKeyType_PKE = 3,
		kBCKeyType_HP = 4,		// Busied by hidden part, it used for temporary key maps only.
		kBCKeyType_Salt = 5		// KeyGen V5 save salt at key map, see DATA_BLOCK::updateSalt
    };

	/**
	 * Hidden key verification block.
	 * Stored right next to encoded hidden key encrypted with it.
	 * Any possible hidden key is verified by correctly decypting this block.
	 */
	struct verification_data {
		bc_u16		signature;
		bc_u8		hash_id;
		bc_u32		alg_id;
		bc_u32		mode_id;
		
		verification_data() {
			signature = 0xBC08;
			hash_id = 0;
			alg_id = 0;
			mode_id = 0;
		}

		bool check()
		{
			return signature == 0xBC08;
		}
	};
	
	
	typedef struct {
		KGEncryptedBlock32 < verification_data > verdata;
		bc_u8 iv[kIVSize];
	} verification_block_32;
	
	typedef struct {
		KGEncryptedBlock64 < verification_data > verdata;
		bc_u8 iv[kIVSize];
	} verification_block_64;

	typedef struct {
		bc_u8		data[kKeySlotSize];
	} KEY_BLOCK;
	
	
	/**
	 * Complete data block structure.
	 */
	typedef struct {
		db_enc_block32	header;
		bc_u8			padding[1024 - sizeof(db_enc_block32)];
		// ---------------------------------------------------------------------------------------------
		// 1KB boundary
		
		bc_u8			pool[kPoolSize];
#ifdef _WIN32
#pragma warning ( disable : 4200 )
#endif
		KEY_BLOCK		keys[0];	// real number of keys depends on total data block size.
	} DATA_BLOCK;

	typedef struct {
		db_enc_block64	header;
		bc_u8			padding[1024 - sizeof(db_enc_block64)];
		// ---------------------------------------------------------------------------------------------
		// 1KB boundary
		
		bc_u8			pool[kPoolSize];
#ifdef _WIN32
#pragma warning ( disable : 4200 )
#endif
		KEY_BLOCK		keys[0];	// real number of keys depends on total data block size.
	} DATA_BLOCK_64;
#pragma pack(pop)
	
	
	// Describes the encoded key layout and type in data block.
	typedef struct key_descriptor 
	{
		unsigned int	base;	// first slot number
		unsigned int	total;	// total slots including base.
		int		        type;	// key type that is stored here. can be undefined.
		
		// Init empty descriptor.
        key_descriptor() : base(0), total(0), type(kKBType_Unknown) {}
		
		// Init specific descriptor.
        key_descriptor(unsigned int _base, unsigned int _total, int _type) : base(_base), total(_total), type(_type) {}
		
	} key_descriptor_t;


	/// Initialize a data block memory of given size.
	void DataBlock_Init(
			DATA_BLOCK* db, size_t dbsize, int alg_id, int mode_id, int hash_id, unsigned long long datasize,
			bc_u16 kgID = kKGID, bc_u32 iterations = 256, const bc_u8 *salt = NULL, int saltSize = 0
		);
	
	/// Tell how many possible keys slots can a data block of this size have.
	size_t DataBlock_CapacityForSize(size_t dbsize);
	
	/// Tell data block structure memory size if it holds this much keys.
	size_t DataBlock_SizeForCapacity(size_t nkeys);
	
	/// Tell if header is possibly encrypted.
	bool DataBlock_IsHeaderEncrypted( const DATA_BLOCK* db );
	
	/**
     * Decrypt data block header
     *
     * @param db            Data block
     * @param key           Main part key that can decrypt the header
     */
	bool DataBlock_DecryptHeader(DATA_BLOCK* db, const key_data_t* key);
	
	/**
     * Encrypt data block header
     *
     * @param db            Data block
     * @param key           Main part key that can encrypt the header
     */
	void DataBlock_EncryptHeader(DATA_BLOCK* db, const key_data_t* key);
	
	/**
     * Recalculate and encrypt data block header digest
     *
     * @param db            Data block
     * @param key           Main part key with which to encrypt the digest
     */
	void DataBlock_UpdateDigest(DATA_BLOCK* db, const key_data_t* key);
	
	/**
     * Verify data block digest
     *
     * @param db            Data block
     * @param key           Main part key with which to decrypt the digest
     */
	bool DataBlock_VerifyDigest(DATA_BLOCK* db, const key_data_t* key);

	/**
	 * Encode key into data block.
	 * Header should be decrypted.
     *
	 * @param db             Data block that will store the key.
	 * @param dbsize         Data block size.
	 * @param key            Key to encode.
	 * @param pwd            Password to use for encoding.
     * @param hiddenMap      Optional map with known hidden key locations so we can avoid overwriting them.
	 * @param bCheckSizeOnly If true it does not modify data block, just check size.
	 *
	 * @return               kBCContainer_NoKeySpace if there are no spaces for new key.
	 *							Caller function may assign new dbsize and call again.
	 *							Failure can occur if we there is absolutely no space for the key.
	 */
    BCContainerResult DataBlock_EncodeKey(DATA_BLOCK* db, size_t & dbsize, const key_data_t* key, BCPassword* pwd, BCKeySlot* hiddenMap, bool bCheckSizeOnly = false );

	/**
	 * Decode key from data block.
     * If header is encrypted we will iterate all possible alg/mode/hash configurations.
     * Caller can look for the one that matched in decoded key info.
	 *
	 * @param db            Data block that stored the key.
	 * @param dbsize        Size of the data block.
	 * @param password      Initial password.
	 * @param hiddenMap     Optional map of known hidden key locations, will be updated if key is hidden.
     * @param out_desc      On success will be updated with information about decoded key location.
	 * @param out_key       On success will contain decoded key data.
	 *
	 * @return              false if no key could be decoded for these parameters.
	 */
    bool DataBlock_DecodeKey(DATA_BLOCK* db, size_t dbsize, BCPassword* pwd, BCKeySlot* hiddenMap, key_descriptor_t* out_desc, key_data_t* out_key, cProgressActor *porgressActor = 0 );
	
	/**
	 * Erase key data and map entries for specified descriptor.
	 * Will fill addressed key blocks by random data and mark the whole key map region as empty.
	 *
	 * @db		Data block.
	 * @dbsize	Data block size in bytes.
	 * @kdesc	Key descriptor addressing the encoded area to erase.
	 */
	void DataBlock_DeleteKey(DATA_BLOCK* db, size_t dbsize, key_descriptor_t* desc, BCKeySlot* hiddenMap);
	
} // namespace kgghost


#endif

