/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 *
 *  Small procedural hashing interface.
 */

#ifndef KGGHOST_HASH_H
#define KGGHOST_HASH_H

#include <stdlib.h>
#include <assert.h>
#include "../../certificate/CertificateLib/hashAlgs/hashAlgs.h"

#include "../../container/intsize.h"
#include "../../container/password.h"

namespace kgghost {
	
	
	/**
	 * Calculate message digest.
	 * @hash_id	Hash algorithm id to use.
	 * @message	Message data.
	 * @nbytes	Number of bytes in message to calculate digest for.
	 * @out_digest	Returned calculated digest.
	 */
	void calculate_digest(int hash_id, const void* message, size_t nbytes, bc_u8* out_digest);
	
	
	/**
	 * Overwrite data with random garbage.
	 * Here because implementation will use SHA256 to generate random garbage.
	 * @data	Data to overwrite.
	 * @nbytes	Nuber of bytes of data.
	 */
	void shred(void* data, size_t nbytes);
	
	
	/**
	 * Generate encryption key from password and seed.
	 * @hash_id	Hash algorithm to use for key generation.
	 * @password	NULL-terminated password string.
	 * @salt	Password salt buffer.
	 * @saltsize	Total salt bytes.
	 * @num_iter	Number of hash iteration to perform.
	 * @out_key	Returned key buffer.
	 * @keysize	Size of the key in bytes.
	 */
	//bool generate_key_from_password(int hash_id, const char* pwdbuf_utf8, size_t pwdlen, const bc_u8* salt, size_t saltsize, unsigned num_iter, bc_u8* out_key, size_t keysize);
	//bool generate_key_from_password(int hash_id, const BCPasswordUTF16 *pwd, const bc_u8* salt, size_t saltsize, unsigned num_iter, bc_u8* out_key, size_t keysize);
	
	
	/**
	 * Generate container encryption key data.
	 * @hash_id	Hash algorithm to use for key generation.
	 * @seed	Random seed buffer.
	 * @seedlen	Number of seed bytes.
	 * @out_key	Initialized key buffer.
	 * @keylen	Number of bytes of key to generate.
	 */
	void generate_key(int hash_id, const bc_u8* seed, size_t seedlen, bc_u8* out_key, size_t keylen);
	
	
	/**
	 * Enumeration of all supported hash algorithms.
	 */
	struct BCEnumHashes {
		
		unsigned kTotalHashes;
		
		int *hashes; // [kTotalHashes];
		int *bf;	// Allocated buffer.
		
		BCEnumHashes( bool bV8Only, int defId = -1 );
		~BCEnumHashes();
		
		unsigned total() const {
			return kTotalHashes;
		}
		
		int get(unsigned idx) {
			assert(idx < kTotalHashes);
			return hashes[idx];
		}

		void set(unsigned idx, int val ) {
			assert(idx < kTotalHashes);
			hashes[idx] = val;
		}
	};

	/**
	 * Enumeration of all supported iterations for encrypted header.
	 */
	struct BCEnumIterations
	{
		static const unsigned kTotalAvailableIterations = 2;
		int iterationT[kTotalAvailableIterations];

		BCEnumIterations();

		unsigned total() const {
			return kTotalAvailableIterations;
		}
		
		int get(unsigned idx) {
			assert(idx < kTotalAvailableIterations);
			return iterationT[idx];
		}

		void set(unsigned idx, int val ) {
			assert(idx < kTotalAvailableIterations);
			iterationT[idx] = val;
		}
	};
}


#endif
