/*
 *	Copyright (c) 2010-2012 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Key block coders register.
 *
 *	Revision: 05.11.2012
 *		Updated for CipherBase implementation.
*/

#include "keyblock.h"
#include "pbe.h"
#include "sss.h"
#include "pke.h"

using namespace kgghost;


typedef unsigned (*kb_decode_proc) (const bc_u8* kbdata, unsigned kbsize, const BCPassword* pwd, CipherBase* cipher, int hash, key_data_t* out_key);
typedef bool (*kb_verify_proc) (const bc_u8* kbdata, unsigned kbsize, const BCPassword* pwd, CipherBase* cipher, int hash);

struct kb_decoder {
    int type;
    kb_decode_proc decode;
    kb_verify_proc verify;
};

static const kb_decoder gDecoders[] = {
    { kKBType_PBE, pbe_decode, pbe_verify },
    { kKBType_PKE, pke_decode, pke_verify },
    { kKBType_SSS, sss_decode, sss_verify },
};

#define TOTAL_CODERS (sizeof(gDecoders) / sizeof(*gDecoders))

static const kb_decoder* decoder_for_type(int type) 
{
	const kb_decoder* coder;
	unsigned i;
	for(i = 0; i < TOTAL_CODERS; ++i) {
		coder = &gDecoders[i];
		if(coder->type == type) {
			return coder;
		}
	}
		
	return NULL;
}


namespace kgghost
{

    unsigned KGEncodeKeyBlock(const key_data_t* key, const BCPassword *pwd, CipherBase* cipher, int hash, bc_u8* kbdata, unsigned kbsize)
    {
        return pbe_encode(key, pwd, cipher, kbdata, kbsize);
    }


    unsigned KGDecodeKeyBlock(const bc_u8* kbdata, unsigned kbsize, int & kbtype, const BCPassword *pwd, CipherBase* cipher, int hash, key_data_t* out_key)
    {
        if (kbtype != kKBType_Unknown)
		{
            const kb_decoder* decoder = decoder_for_type(kbtype);

			if ( decoder == NULL )
			{
				assert( false );
				return 0;
			}
            
            unsigned decodedSize = decoder->decode(kbdata, kbsize, pwd, cipher, hash, out_key);
            if (decodedSize > 0) kbtype = decoder->type;
            
            return decodedSize;
        }
		else
		{
	        for (unsigned i = 0; i < TOTAL_CODERS; ++i)
			{
                unsigned decodedSize = gDecoders[i].decode(kbdata, kbsize, pwd, cipher, hash, out_key);
                if (decodedSize > 0) {
                    kbtype = gDecoders[i].type;
                    return decodedSize;
                }
                
			}

            return 0;
        }
    }


    bool KGVerifyPassword(const bc_u8* kbdata, unsigned kbsize, int kbtype, const BCPassword* pwd, CipherBase* cipher, int hash)
    {
        if (kbtype != kKBType_Unknown) {
            const kb_decoder* decoder = decoder_for_type(kbtype);
            return decoder->verify(kbdata, kbsize, pwd, cipher, hash);
        } else {
	        for (unsigned i = 0; i < TOTAL_CODERS; ++i) {
                if (gDecoders[i].verify(kbdata, kbsize, pwd, cipher, hash)) {
                    return true;
                }
            }

            return false;
        }
    }


    bool KGChangePassword(const bc_u8* kbdata, unsigned kbsize, int kbtype, const BCPassword *pwd, const BCPassword *newpwd, CipherBase* cipher, int hash)
    {
        return false;
    }
}

