/*
 *	Copyright (c) 2010 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Key block coders register and interface.
 */

#ifndef KGGHOST_KEYBLOCK_H
#define KGGHOST_KEYBLOCK_H

#include <cstring>
#include <assert.h>

#include "../../container/cipherbase.h"
#include "kgghost.h"
#include "hash.h"

#pragma pack(push, 1)

namespace kgghost 
{

    // Type of the encoded key block
	enum {
		kKBType_Unknown   = 0,    // Unknown key block, will try every supported method
		kKBType_PBE       = 1,    // Key block encoded by password
		kKBType_PKE       = 2,    // Key block encoded by public key algorithm
		kKBType_SSS       = 3,    // Key block encoded by secret sharing scheme
	};


    /**
     * Encode key using password based encryption
     * 
     * @param key           Key data to encode
     * @param pwd           Password
     * @param cipher        Configured block cipher
     * @param hash          Hash algorithm id
     * @param kbdata        Key block buffer to store encoded key in
     * @param kbsize        Size of the key block buffer
     * @return              Size of the encoded key block on success or 0 on failure.
     */
    unsigned KGEncodeKeyBlock(const key_data_t* key, const BCPassword* pwd, CipherBase* cipher,
                              int hash, bc_u8* kbdata, unsigned kbsize);

	
    /**
     * Decode key data from key block
     * 
     * @param kbdata        Encoded key block bytes
     * @param kbsize        Encoded key block size in bytes
     * @param kbtype        Type of the key block (PBE, SSS, PKE) if known.
     *                      If type is unknown (kKBType_Unknown) proc will try every possible decoding method.
     * @param pwd           Initial password
     * @param cipher        Configured block cipher
     * @param hash          Hash algorithm id
     * @param out_key       Decoded key data on success
     * @return              On success real encoded key block size is returned, or 0 on failure
     */
    unsigned KGDecodeKeyBlock(const bc_u8* kbdata, unsigned kbsize, int & kbtype, const BCPassword* pwd,
                              CipherBase* cipher, int hash, key_data_t* out_key);


    /**
     * Verify password.
     * This procedure will only verify initial password without decoding the whole key.
     * 
     * @param kbdata        Encoded key block bytes
     * @param kbsize        Encoded key block size in bytes
     * @param kbtype        Type of the key block (PBE, SSS, PKE) if known.
     *                      If type is unknown (kKBType_Unknown) proc will try every possible decoding method.
     * @param pwd           Initial password
     * @param cipher        Configured block cipher
     * @param hash          Hash algorithm id
     */
    bool KGVerifyPassword(const bc_u8* kbdata, unsigned kbsize, int kbtype, const BCPassword* pwd, CipherBase* cipher, int hash);

    bool KGChangePassword(const bc_u8* kbdata, unsigned kbsize, int kbtype, const BCPassword* pwd,
                          const BCPassword* newpwd, CipherBase* cipher, int hash);

    /**
     * Helper structure to hold encrypted data of type T padded to block boundary with its digest.
     * Used in key block encoding algorithms for password verification
     */
    template < typename T >
    struct KGEncryptedBlock32 {
        enum { kPaddedDataSize = (sizeof(T) + kEncryptedBlockSize - 1) & ~(kEncryptedBlockSize - 1) };
        
        bc_u8 data[kPaddedDataSize];
        bc_u8 digest[kDigestSize32];

        KGEncryptedBlock32() {
			assert(0 == (sizeof(*this) % kEncryptedBlockSize));
        }
    };

    template < typename T >
    struct KGEncryptedBlock64 {
        enum { kPaddedDataSize = (sizeof(T) + kEncryptedBlockSize - 1) & ~(kEncryptedBlockSize - 1) };
        
        bc_u8 data[kPaddedDataSize];
        bc_u8 digest[kDigestSize64];

        KGEncryptedBlock64() {
			assert(0 == (sizeof(*this) % kEncryptedBlockSize));
        }
    };

    // Decrypt and verify data of type T stored in encrypted block
    template < typename T >
    bool KGDecryptBlock32(const KGEncryptedBlock32<T>* block, bc_u64 iv, CipherBase* cipher, int hash, const bc_u8* key, unsigned keysize, const bc_u8* pool, T* data) 
    {
        KGEncryptedBlock32<T> copy;
        memcpy(&copy, block, sizeof(copy));
        try {
            cipher->decrypt(iv, key, keysize, pool, &copy, sizeof(copy));
        }
        catch (std::exception& e) {
            e.what();
            return false;
        }

        memcpy(data, copy.data, sizeof(*data));

        bc_u8 digest[kDigestSize32] = {0};
        calculate_digest(hash, data, sizeof(*data), digest);

        return (0 == memcmp(digest, copy.digest, sizeof(digest)));
    }

    template < typename T >
    bool KGDecryptBlock64(const KGEncryptedBlock64<T>* block, bc_u64 iv, CipherBase* cipher, int hash, const bc_u8* key, unsigned keysize, const bc_u8* pool, T* data) 
    {
        KGEncryptedBlock64<T> copy;
        memcpy(&copy, block, sizeof(copy));
        try {
            cipher->decrypt(iv, key, keysize, pool, &copy, sizeof(copy));
        }
        catch (std::exception& e) {
            e.what();
            return false;
        }

        memcpy(data, copy.data, sizeof(*data));

        bc_u8 digest[kDigestSize64] = {0};
        calculate_digest(hash, data, sizeof(*data), digest);

        return (0 == memcmp(digest, copy.digest, sizeof(digest)));
    }

    // Encrypt and store data of type T in encrypted block
    template < typename T >
    void KGEncryptBlock32(const T* data, bc_u64 iv, CipherBase* cipher, int hash, const bc_u8* key, unsigned keysize, const bc_u8* pool, KGEncryptedBlock32<T>* block)
    {
        shred(block, sizeof(*block));
        memcpy(block->data, data, sizeof(*data));
        calculate_digest(hash, data, sizeof(*data), block->digest);

        try {
            cipher->encrypt(iv, key, keysize, pool, block, sizeof(*block));
        } catch (std::exception& e) {
            memset(block, 0, sizeof(*block));
            e.what();
        } 
    }

    template < typename T >
    void KGEncryptBlock64(const T* data, bc_u64 iv, CipherBase* cipher, int hash, const bc_u8* key, unsigned keysize, const bc_u8* pool, KGEncryptedBlock64<T>* block)
    {
        shred(block, sizeof(*block));
        memcpy(block->data, data, sizeof(*data));
        calculate_digest(hash, data, sizeof(*data), block->digest);

        try {
            cipher->encrypt(iv, key, keysize, pool, block, sizeof(*block));
        } catch (std::exception& e) {
            memset(block, 0, sizeof(*block));
            e.what();
        } 
    }
}

#pragma pack(pop)

#endif

