/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 *
 * KGHOST key generator interface implementation.
 * This is the top level of key generator implementation.
 * It initates user sessions and applies policies to data block key encoding/decoding mechanisms based on user input.
 */

#include "datablock.h"
#include "kgghost.h"

#include "../../container/file.h"
#include "../../certificate/CertificateLib/PGPPkts/pktdefs.h"

using namespace kgghost;

//
//	IsEncrypted
//		The function allows ambiguity so we have to check if cont 
//		info is really encrypted after encryption!!!
//
bool db_header::IsEncrypted()
{
	return ( strncmp( volumeLabel, "BC_KeyGenID", sizeof( volumeLabel ) )!= 0 ) ;
}

//
//	KG-Ghost v5 allows to change iteration
//	1. It changes key gen identifier, so old key generator cannot open this container.
//		Otherwise it will report invalid password.
//	2. Uses the version field as iteration count. But keyGenId DWORD files uses as WORD wKeyGenId, WORD wVesrion.
//	3. Key generator ID - 5 does not allow to encrypt header, otherwise enumeration becames imposible.
//
//	See http://nsk.jetico.com/redmine/projects/best-crypt-for-windo/wiki/Wiki
//
bool db_header::setKeyGenIdV5( bc_u32 count )
{
	if ( IsEncrypted() )
	{	assert( false ); return false; }

	if ( wKeyGenId != kgghost::kKGID_V5 )
	{
		wKeyGenId = kKGID_V5;
		wVersion = (WORD)version;
		assert( wVersion == version );
	}

	iterations = count;

	return true;
}

bc_u32 db_header::getIterationCount()
{
	if ( !IsEncrypted() && ( wKeyGenId == kKGID_V5 ) )
		return iterations;
	else
		return 256;
}

void db_header::setVersion( bc_u16 ver )
{
	if ( wKeyGenId == kKGID_V5 )
		wVersion = ver;
	else
	{
		version = ver;
		wVersion = 0;
	}
}

bool db_header::useNewSyncro()
{
	if ( wKeyGenId > kKGID )
		return true;

	return getVersion() != 0;
}


int db_header::getVersion()
{
	if ( wKeyGenId == kKGID_V5 )
		return wVersion;
	else
		return version;
}


typedef struct {
    DATA_BLOCK*     db;
    unsigned        dbsize;
    CipherBase*     cipher;
    int             hash;
    key_data_t*     key;
} HeaderKeyAuth;


static BCContainerResult GetContainerResultBySMResult(sm_result result)
{
    if (kSpaceMap_Success == result)
        return kBCContainer_Success;

    if (kSpaceMap_NoMem == result)
        return kBCContainer_NoMemory;

    if (kSpaceMap_InvalidArg == result)
        return kBCContainer_InvalidArgument;

    return kBCContainer_InternalError;
}


static bool generateKey(int algid, int modeid, int hashid, bc_u64 offset, bc_u64 datasize, unsigned keysize, key_data_t* out_key)
{
    bc_u8 seed[kSeedLength] = {0};
    if (sizeof(seed) != BCUserGenerateSeed(seed, sizeof(seed))) {
		LIBLOG_INFO( "Assertion! Seed values is not prepared" );
        return false;
    }

    generate_key(hashid, seed, kSeedLength, out_key->buf, kMaximumKeySize);

    out_key->alg_id     = algid;
    out_key->mode_id    = modeid;
    out_key->hash_id    = hashid;
    out_key->offset     = offset;
    out_key->datasize   = datasize;
    
    if (modeid == kBCMode_CBC)
        out_key->size   = sizeof(key_data_t) + keysize - kMaximumKeySize;
    else // LRW, XTS
        out_key->size   = sizeof(key_data_t);

    return true;
}

KGGHOST_Header::KGGHOST_Header( iChannel *channel, bc_u8* datablock, unsigned size, bool bEncrypted )
{
    shred(m_db, sizeof(m_db));
    memcpy(m_db, datablock, size);
    shred(m_dbDecrypted, sizeof(m_dbDecrypted));
    memcpy(m_dbDecrypted, datablock, size);
    m_dbsize = size;
	m_channel = channel;
	memset( m_hiddenMap, 0, sizeof( m_hiddenMap ) );
}

void KGGHOST_Header::mainVolumeInfo(struct BCVolumeInfo* info)
{
    DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
    bool is_encrypted   = DataBlock_IsHeaderEncrypted(db);
    if ( is_encrypted && (!DataBlock_IsHeaderEncrypted( (DATA_BLOCK*)m_dbDecrypted )) ) {
        db = (DATA_BLOCK*)m_dbDecrypted;
        is_encrypted = false;
    }
    assert(!is_encrypted);

    db_header* header   = (db_header*) db;
    
    memset(info, 0, sizeof(*info));

    info->alg       = header->alg_id;
	info->mode      = header->getMode();
    info->hash      = header->hash_id;
    info->size      = header->data_length;
    info->offset    = header->data_start;
    info->hidden    = 0;

	//
	//	For sector encrypt/decrypt functions BestCrypt driver uses synchro sequence is 8 byte.
	//	Old version used 16 bit sector number and packed 4 same number to 8 bytes.
	//	Current version uses 64bit sector number as synchro. See sector2iv function.
	//
	//	If header->verson is zero so the conatiner does not support new synchro.
	//	Current BestCrypt writes in header->verson BC version 7, 8 ...
	//	Linux uses 3 defined as BCV8_VERSION_TAG
	//
	info->use_new_synchro = header->version != 0;

    BCUserConvertString((bc_u8*)header->description, sizeof(header->description), (bc_u8*)info->description, sizeof(info->description), kBCUser_UTF16LE, kBCUser_UTF8);
}

BCContainerResult KGGHOST_Header::volumeInfoForKey(key_data_t* key, struct BCVolumeInfo* info)
{
	db_header* header = (db_header*) m_db;
	db_header* decrypedHeader = (db_header*) m_dbDecrypted;
    DATA_BLOCK *db = (DATA_BLOCK*)m_db;
    sm_result smres;

    memset(info, 0, sizeof(*info));

    info->alg       = key->alg_id;
	info->mode      = key->getMode();
    info->hash      = key->hash_id;
    info->size      = key->datasize;
    info->offset    = key->offset;
    info->hidden    = key->is_hidden();

	bool bIsEncrypted = ( DataBlock_IsHeaderEncrypted(db) == TRUE );

    if(!key->is_hidden())
    {
		if ( !bIsEncrypted )
			BCUserConvertString((const bc_u8*)header->description, sizeof(header->description), (bc_u8*)info->description, sizeof(info->description), kBCUser_UTF16LE, kBCUser_UTF8);
		else
		if ( !decrypedHeader->IsEncrypted() )
			BCUserConvertString((const bc_u8*)decrypedHeader->description, sizeof(decrypedHeader->description), (bc_u8*)info->description, sizeof(info->description), kBCUser_UTF16LE, kBCUser_UTF8);


        info->map = SpaceMapCreate();
        if (!info->map) {
            return kBCContainer_NoMemory;
        }

        smres = SpaceMapAddNew(info->map, info->offset, info->size);
        if (kSpaceMap_Success != smres)
        {
            SpaceMapFree(info->map);

            LIBLOG_INFO( "Assertion! Space map adding error %d", smres );

            return GetContainerResultBySMResult(smres);
        }

        //fprintf(stderr, "Usual key found. Map contains %lu extents.", info->map->count);
        LIBLOG_INFO( "Usual key found. Map contains %d extents", (int)info->map->count );
    }
    else
    {
        // Get space map for hidden part.
        assert(key->datasize == 4096); // Mark of hidden key.
		assert( m_channel != NULL );

		if ( !m_channel->open( kBCFileMode_Read ) )
            return kBCContainer_IOError;

        // Read space map from file.
        bc_u8 clusterData[kBCV8_HiddenPartSpaceMapSize];

		bc_u64 bytesRead = m_channel->readBytes( clusterData, (bc_u64)kBCV8_HiddenPartSpaceMapSize, info->offset);
		m_channel->close();

        if (kBCV8_HiddenPartSpaceMapSize != bytesRead) {
            return kBCContainer_IOError;
        }

       // Decrypt space map.
		bc_cipher_result rc = CipherFactory::ProcessWithKey(key->alg_id, key->getMode(), kBCProcess_Decrypt, key->buf, sizeof(key->buf) /*key->size*/, kBCV8_HiddenPartMapIV, db->pool, clusterData, kBCV8_HiddenPartSpaceMapSize);
        if (kBCCipher_Success != rc) {
            return kBCContainer_InternalError;
        }

		int digestSize = 0;
		if ( ha_getDigestSize( &digestSize, key->hash_id  ) != hae_ok )
		{	assert( false ); return kBCContainer_InternalError; }

		if ( digestSize < 32 )
			digestSize = 32;
		else
		if ( digestSize > 64 )
			return kBCContainer_InternalError;

		bc_u8 digest[64] = {0};
		bc_u8 *spaceMapData = spaceMapData = clusterData + digestSize;

		calculate_digest(key->hash_id, spaceMapData, kBCV8_HiddenPartSpaceMapSize - digestSize, digest);

		if ( 0 != memcmp(digest, clusterData, digestSize) )
			return kBCContainer_Damaged;

        smres = SpaceMapFromBytes((void*)spaceMapData, kBCV8_HiddenPartSpaceMapSize, &info->map);
        // fprintf(stderr, "Hidden key found. Map contains %lu extents.", info->map->count);

        if (kSpaceMap_Success != smres) {
            return GetContainerResultBySMResult(smres);
        }
    }

	if ( !bIsEncrypted )
		info->use_new_synchro = header->useNewSyncro(); 
	else
	if ( !decrypedHeader->IsEncrypted() )
		info->use_new_synchro = decrypedHeader->useNewSyncro();
	else
		info->use_new_synchro = true;

    return kBCContainer_Success;
}

BCContainerResult KGGHOST_Header::setDescription(const char * description, const cPTB & mainKey)
{
    DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
    bool is_encrypted   = DataBlock_IsHeaderEncrypted(db);
    
    key_data_t * key	= (key_data_t *)mainKey.buffer;
    if (description == NULL || mainKey.bufferSize != sizeof( key_data_t ) || key->is_hidden()) {
        return kBCContainer_InvalidArgument;
    }
    
    if (is_encrypted) {
        if (!DataBlock_IsHeaderEncrypted( (DATA_BLOCK*)m_dbDecrypted)) {
            db = (DATA_BLOCK*)m_dbDecrypted;
        } else {
            memcpy(m_dbDecrypted, m_db, m_dbsize);
            if (!DataBlock_DecryptHeader((DATA_BLOCK*)m_dbDecrypted, key)) {
                return kBCContainer_InvalidPassword;
            }
            db = (DATA_BLOCK*)m_dbDecrypted;
        }
    }
    db_header * header   = (db_header*) db;

    memset(header->description, 0, sizeof(header->description));
    if (0 == BCUserConvertString((const bc_u8*)description, strlen(description) + 1,
                                 (bc_u8*)header->description, sizeof(header->description),
                                 kBCUser_UTF8, kBCUser_UTF16LE)) {
        return kBCContainer_InternalError;
    }
    DataBlock_UpdateDigest((DATA_BLOCK*)db, key);
    
    if (is_encrypted) {
        memcpy(m_db, m_dbDecrypted, m_dbsize);
        DataBlock_EncryptHeader((DATA_BLOCK*)m_db, key);
        
    }
    return kBCContainer_Success;
}

void KGGHOST_Header::markForReencryption(bool mark)
{
    assert(!this->isEncrypted());

    db_header* header = (db_header*) m_db;

    if (mark) {
        strncpy(header->volumeLabel, DB_REENCRYPTED_LABEL, sizeof(header->volumeLabel));
    } else {
        strncpy(header->volumeLabel, DB_HEADER_VOLUME_LABEL, sizeof(header->volumeLabel));
    }
}

//virtual
const char *KGGHOST_Header::getKeygenName()
{
    db_header* header = (db_header*) m_db;
	if ( header->IsEncrypted() || ( header->wKeyGenId == kKGID_V5 ) )
		return "KG-Ghost";
	else
		return "KG-Ghost v.8";
}

/**
 * Returns true if header encryption ot decryption is allowed
 */
//virtual
bool KGGHOST_Header::isHeaderEncryptionAllowed()
{
    db_header* header = (db_header*) m_db;

	if ( header->IsEncrypted() )
		return true;

	if ( header->wKeyGenId == kKGID )
		return true;
	
	int iterations = header->getIterationCount();

	if (256 == iterations || 16*1024 == iterations)
		return true;

	return false;
}

/**
 * Returns true if container can be reencrypted
 */
//virtual
bool KGGHOST_Header::isReencryptionAllowed()
{
    db_header* header = (db_header*) m_db;

	if ( header->IsEncrypted() )
		return false;

	if ( header->wKeyGenId == kKGID )
		return true;
	
	int iterations = header->getIterationCount();

	if (256 == iterations || 16*1024 == iterations)
		return true;

	return false;
}

bool KGGHOST_Header::isMarkedForReencryption()
{
    assert(!this->isEncrypted());

    db_header* header = (db_header*) m_db;
    return (0 == strncmp(header->volumeLabel, DB_REENCRYPTED_LABEL, sizeof(header->volumeLabel)));
}

bool KGGHOST_Header::isEncrypted()
{
    DATA_BLOCK* db = (DATA_BLOCK*) m_db;
    return DataBlock_IsHeaderEncrypted(db);
}

//virtual
int KGGHOST_Header::setBusiedName( const char *compName, int len )
{
	if ( isEncrypted() )
		return -1;
	// Uses same function for both SHA, GHOST key generators
	return -1;
}

//virtual
int KGGHOST_Header::getBusiedName( char *compName, int len )
{
	if ( isEncrypted() )
		return -1;

	return -1;
}

BCContainerResult KGGHOST_Header::initEx(
		bc_alg_id algid, bc_mode_id modeid, bc_u64 size, const char* description, BCPassword & password,
		bc_u16 kgID, int hash, bc_u32 iterations, const bc_u8 *salt, int saltSize
	)
{
	//LIBLOG_INFO( "Begin %x, %x, %d", (unsigned int)algid, (unsigned int)modeid, (int)size );

    DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
	db_header* header   = (db_header*) m_db;

    DataBlock_Init(
			db, m_dbsize, algid, modeid, hash, size,
			kgID, iterations, salt, saltSize
		);

    if (description) {
        BCUserConvertString((const bc_u8*)description, strlen(description) + 1, (bc_u8*)header->description, sizeof(header->description), kBCUser_UTF8, kBCUser_UTF16LE);
    }

	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

	if ( kBCCipher_Success != sCipher->init(algid, modeid) ) {
		return kBCContainer_InvalidAlgorithm;
	}

    key_data_t key;

	if ( !generateKey( algid, modeid, hash, header->data_start, header->data_length, (unsigned)sCipher->keySize(), &key ) ) {
        return kBCContainer_InternalError;
    }

	BCContainerResult res = DataBlock_EncodeKey( (DATA_BLOCK*)db, m_dbsize, &key, &password, NULL);
    if ( res != kBCContainer_Success ) {
        return res;
    }

	DataBlock_UpdateDigest((DATA_BLOCK*)db, &key);

    memcpy(m_dbDecrypted, db, m_dbsize);
    //LIBLOG_INFO( "End %d", res );

    return res;
}


static bool allocateSpaceForMap(SpaceMap *map, bc_u64 *out_offset)
{
    if (map->start->extent.length > kBCV8_HiddenPartSpaceMapSize)
    {
        Extent *first = &map->start->extent;
        bc_u64 offset = first->start;

        first->length -= kBCV8_HiddenPartSpaceMapSize;
        first->start += kBCV8_HiddenPartSpaceMapSize;

        if (0 == first->length)
        {
            sm_result smres = SpaceMapRemove(map, first);
            if (kSpaceMap_Success != smres)
                return false;
        }

        (*out_offset) = offset;
        return true;
    }

    return false;
}

static bool saveSpaceMapByOffset( iChannel *channel, CipherBase* cipher, bc_u8 *pool, key_data_t *key, SpaceMap *map, bc_u64 offset)
{
	if ( channel == NULL )
	{ assert( false ); return false; }

    bc_u8 buffer[kBCV8_HiddenPartSpaceMapSize];
    memset(buffer, 0, kBCV8_HiddenPartSpaceMapSize);

    int digestSize = 0;
    if ( ha_getDigestSize( &digestSize, key->hash_id  ) != hae_ok )
	{ assert( false ); return false; }
    
    if ( digestSize < 32 ) digestSize = 32;
    else if ( digestSize > 64 )
	{ assert( false ); return false; }
    
    bc_u8 * spaceMapData = buffer + digestSize;
    
    // Convert space map to bytes.
    sm_result smres = SpaceMapToBytes(map, (void*)spaceMapData, kBCV8_HiddenPartSpaceMapSize - digestSize);
    if (kSpaceMap_Success != smres)
        return false;

    calculate_digest(key->hash_id, spaceMapData, kBCV8_HiddenPartSpaceMapSize - digestSize, buffer);
    
    // Encrypt bytes by specified key.
	cipher->encrypt(kBCV8_HiddenPartMapIV, key->buf, sizeof(key->buf)/*key->size*/, pool, (void*)buffer, kBCV8_HiddenPartSpaceMapSize);

	if ( !channel->open( kBCFileMode_ReadWrite ) )
        return false;

	bc_u64 bytesRead = channel->writeBytes( buffer, kBCV8_HiddenPartSpaceMapSize, offset);
	channel->close();

    if (bytesRead != kBCV8_HiddenPartSpaceMapSize)
        return false;
    
    return true;
}


BCContainerResult KGGHOST_Header::createHiddenKey(SpaceMap *spaceMap, cPTB *ptbMainKey, BCPassword* hiddenPassword)
{
    DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
    bool is_encrypted   = DataBlock_IsHeaderEncrypted(db);

    if ( ( ptbMainKey == NULL ) || ( ptbMainKey->bufferSize != sizeof(key_data_t) ) ) {
		return kBCContainer_InternalError;
    }
	key_data_t &mainKey = *((key_data_t *)ptbMainKey->buffer);
    
    // decrypt header with the main key to get the space map
    if (is_encrypted && !DataBlock_DecryptHeader((DATA_BLOCK*)db, &mainKey)) {
        return kBCContainer_InternalError;
    }

    db_header* header   = (db_header*) db;
	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

	if ( kBCCipher_Success != sCipher->init(header->alg_id, header->getMode()) ) {
		return kBCContainer_InvalidAlgorithm;
	}

    // Calculate offset of the spacemap store.
    bc_u64 offset;
    if (!allocateSpaceForMap(spaceMap, &offset))
        return kBCContainer_InternalError;

    // generate and encode new hidden key
	key_data_t hiddenKey;
	if (!generateKey(mainKey.alg_id, mainKey.getMode(), mainKey.hash_id, offset, 4096, (unsigned)sCipher->keySize(), &hiddenKey))
        return kBCContainer_Cancelled;

    BCContainerResult res = DataBlock_EncodeKey((DATA_BLOCK*)db, m_dbsize, &hiddenKey, hiddenPassword, m_hiddenMap );
	if ( res != kBCContainer_Success ) return res;

    // Save the space map by offset.
	if (!saveSpaceMapByOffset( m_channel, sCipher.get(), db->pool, &hiddenKey, spaceMap, offset))
        return kBCContainer_IOError;

    // update digest and encrypt the header if it was encrypted
	DataBlock_UpdateDigest((DATA_BLOCK*)db, &mainKey);

    if (is_encrypted) {
        memcpy(m_dbDecrypted, db, m_dbsize);
        DataBlock_EncryptHeader(db, &mainKey);
    }

    return kBCContainer_Success;
}

void KGGHOST_Header::updateDataBlockSize()
{
    if (m_channel == NULL || !m_channel->open(kBCFileMode_Read)) {
        return;
    }

    DATA_BLOCK * db      = (DATA_BLOCK*) m_db;
    if (DataBlock_IsHeaderEncrypted(db)) {
        db = (DATA_BLOCK*) m_dbDecrypted;
        if (DataBlock_IsHeaderEncrypted(db)) {
            return;
        }
    }
    db_header * header   = (db_header*) db;
    
    bc_u64 fileSize	= m_channel->getSize();
    m_channel->close();
    
    bc_u64 extKbOffset = 0;
    bc_u32 extKbSize = 0;
    
    extKbOffset	= header->data_length + header->data_start;
    if(extKbOffset > fileSize) {
        // this is possible when we are dealing with backup headers.
        extKbSize = 0;
    } else {
        extKbSize = u64tou32( fileSize - extKbOffset, kBCV8_MaximumExtKeyBlockSize );
    }

    m_dbsize = kBCV8_InitialDataBlockSize + extKbSize;
}

BCContainerResult KGGHOST_Header::verifyPassword( BCPassword & password, struct BCVolumeInfo* out_info, cProgressActor *progressActor, cPTB *keyOut )
{
    DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
    //db_header* header   = (db_header*) db;
	bool is_encrypted   = DataBlock_IsHeaderEncrypted(db);

	//
	//	KG-Ghost does not empty pasword.
	//	The "just_test" is used by experimental version of BCArchive and may be removed.
	//
	if ( ( password.length == 0 ) || ( strcmp( password.buf, "just_test" ) == 0 ) )
		return kBCContainer_InvalidPassword;

    if ( is_encrypted && (!DataBlock_IsHeaderEncrypted( (DATA_BLOCK*)m_dbDecrypted )) ) {
        db = (DATA_BLOCK*)m_dbDecrypted;
        is_encrypted = false;
    }

    key_descriptor_t kdesc;
    key_data_t key;

    if( !DataBlock_DecodeKey(db, m_dbsize, &password, m_hiddenMap, &kdesc, &key, progressActor ) )
	{
       LIBLOG_INFO( "Warning! Password is invalid" );
        return kBCContainer_InvalidPassword;
    }

    BCContainerResult result = volumeInfoForKey(&key, out_info);
    if (result != kBCContainer_Success) {
        return result;
    }

    if ( is_encrypted ) {
        memcpy( m_dbDecrypted, m_db, m_dbsize );
        if (DataBlock_DecryptHeader((DATA_BLOCK*)m_dbDecrypted, &key)) {
            updateDataBlockSize();
        }
	}

	if ( NULL != keyOut )
		keyOut->Assign( (bc_u8*)&key, sizeof( key ) );

    return kBCContainer_Success;
}

BCContainerResult KGGHOST_Header::generateCipherInfo( const cPTB &keyIn, struct BCVolumeInfo* out_info, bc_key_handle* out_handle, CipherBase **cipher_out )
{
	LIBLOG_INFO( "Begin" );

	if ( NULL == cipher_out )
		return kBCContainer_InvalidArgument;

	if ( keyIn.bufferSize != sizeof( key_data_t ) )
		return kBCContainer_InvalidArgument;

	key_data_t *key = ( key_data_t * )keyIn.buffer;

	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

	if ( kBCCipher_Success != sCipher->init( key->alg_id, key->getMode() ) )
	    return kBCContainer_InvalidAlgorithm;

	DATA_BLOCK* db = (DATA_BLOCK*) m_db;
	if( kBCCipher_Success != sCipher->storeKey( key->buf, sizeof(key->buf), db->pool, out_handle ) )
        return kBCContainer_NoMemory;

    BCContainerResult result = volumeInfoForKey( key, out_info );
    if (result != kBCContainer_Success) {
        *cipher_out = NULL;
        CipherFactory::FreeKey( *out_handle );
        return result;
    }

    *cipher_out = sCipher.delegatePtr();

	LIBLOG_INFO("End Success");
	return kBCContainer_Success;
}

BCContainerResult KGGHOST_Header::decodeKey(BCPassword & password, struct BCVolumeInfo* out_info, bc_key_handle* out_handle, CipherBase **cipher_out, cPTB *keyOut )
{
	LIBLOG_INFO("Begin %p, %lu", m_db, m_dbsize );

	DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
	bool is_encrypted   = DataBlock_IsHeaderEncrypted(db);

	if ( is_encrypted && (!DataBlock_IsHeaderEncrypted( (DATA_BLOCK*)m_dbDecrypted )) )	{
		db = (DATA_BLOCK*)m_dbDecrypted;
		is_encrypted = false;
	}

    db_header* header   = (db_header*) db;

    key_descriptor_t kdesc;
    key_data_t key;

	if ( NULL == cipher_out )
		return kBCContainer_InvalidArgument;

	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

    if(!is_encrypted) {
		if ( kBCCipher_Success != sCipher->init( header->alg_id, header->getMode() ) ) {
		    return kBCContainer_InvalidAlgorithm;
	    }
    }

    if(!DataBlock_DecodeKey(db, m_dbsize, &password, m_hiddenMap, &kdesc, &key)) {
        return kBCContainer_InvalidPassword;
    }

	if(!key.is_hidden() && !DataBlock_VerifyDigest(db, &key)) {
		return kBCContainer_Damaged;
	}

	//
	//	certificate Lib key handle is pointer to algorithm class with key.
	//	So we have to completely initialise sCipher to create the key handle.
	//	As result crtlib encrypts without alg and mode id only by key handle
	//	BCCipher at kernel encryption finds algorithm by id and provides it with key handle as number.
	//	The initializing seems is redundant for kernel style, but it is definitely useful to check 
	//	key's id and mode after decryption.
	//
	if ( kBCCipher_Success != sCipher->init( key.alg_id, key.getMode() ) ) {
	    return kBCContainer_InvalidAlgorithm;
    }

	if( kBCCipher_Success != sCipher->storeKey( key.buf, sizeof(key.buf), db->pool, out_handle ) ) {
        return kBCContainer_NoMemory;
    }

    BCContainerResult result = volumeInfoForKey( &key, out_info );
    if (result != kBCContainer_Success) {
        *cipher_out = NULL;
        CipherFactory::FreeKey( *out_handle );
        return result;
    }

    *cipher_out = sCipher.delegatePtr();

    if ( is_encrypted ) {
        memcpy( m_dbDecrypted, m_db, m_dbsize );
        if (DataBlock_DecryptHeader( (DATA_BLOCK*)m_dbDecrypted, &key)) {
            updateDataBlockSize();
        }
    }

	if ( NULL != keyOut )
		keyOut->Assign( (bc_u8*)&key, sizeof( key ) );

	LIBLOG_INFO("End Success");
	return kBCContainer_Success;
}

BCContainerResult KGGHOST_Header::passwd(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & password, BCKeyOp opflag)
{
	DATA_BLOCK*	    db = (DATA_BLOCK*) m_db;
	bool		    is_encrypted = DataBlock_IsHeaderEncrypted(db);

    // Key data and descriptor we will work with

    key_data_t          mainKey;    // In case of encrypted header this will container decoded main key
    key_descriptor_t    mainKDesc;
    
    if (is_encrypted) {
        if (!DataBlock_IsHeaderEncrypted( (DATA_BLOCK*)m_dbDecrypted))	{
            db = (DATA_BLOCK*)m_dbDecrypted;
        } else {
            assert(false);
            return kBCContainer_InternalError;
        }
    }

    if (!DataBlock_DecodeKey(db, m_dbsize, &mainPassword, NULL, &mainKDesc, &mainKey)) {
        return kBCContainer_InvalidPassword;
    }
    
    if (!DataBlock_VerifyDigest(db, &mainKey)) {
        return kBCContainer_Damaged;
    }

    key_descriptor_t    kdesc;
    key_data_t          key;
    if (!DataBlock_DecodeKey(db, m_dbsize, &volumePassword, m_hiddenMap, &kdesc, &key)) {
        return kBCContainer_InvalidPassword;
    }
    // Invariants at this point: 1) key is decoded and 2) header is decrypted if needed.
    assert(key.size != 0);
    
    BCContainerResult result = kBCContainer_Success;
    switch (opflag) {

        // change password.
        case kBCKey_Recode:
        {
            DataBlock_DeleteKey(db, m_dbsize, &kdesc, m_hiddenMap);
            result = DataBlock_EncodeKey(db, m_dbsize, &key, &password, m_hiddenMap);

            break;
        }

        // add password.
        case kBCKey_Duplicate:
        {
            result = DataBlock_EncodeKey(db, m_dbsize, &key, &password, m_hiddenMap);
            break;
        }

        // remove password.
        case kBCKey_Delete:
        {
            DataBlock_DeleteKey(db, m_dbsize, &kdesc, m_hiddenMap);
            break;
        }

        default: 
            return kBCContainer_Unsupported;
    }

    DataBlock_UpdateDigest(db, &mainKey);

    // if header was decrypted store a copy and encrypt it back with the new changes.
    if (is_encrypted) {
        if ((DATA_BLOCK*)m_dbDecrypted != db) memcpy(m_dbDecrypted, db, m_dbsize);
        memcpy(m_db, m_dbDecrypted, m_dbsize);
        DataBlock_EncryptHeader((DATA_BLOCK*)m_db, &mainKey);
    }

    return result;
}

BCContainerResult KGGHOST_Header::addPassword(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & newPassword)
{
    return this->passwd(mainPassword, volumePassword, newPassword, kBCKey_Duplicate);
}

BCContainerResult KGGHOST_Header::changePassword(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & newPassword)
{
    return this->passwd(mainPassword, volumePassword, newPassword, kBCKey_Recode);
}

BCContainerResult KGGHOST_Header::removePassword(BCPassword & mainPassword, BCPassword & oldPassword)
{
    BCPassword dummy;
    return this->passwd(mainPassword, oldPassword, dummy, kBCKey_Delete);
}

bc_u16 KGGHOST_Header::getMainKeyblocksCount()
{
    DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
    bool is_encrypted   = DataBlock_IsHeaderEncrypted(db);
    
    if ( is_encrypted && (!DataBlock_IsHeaderEncrypted( (DATA_BLOCK*)m_dbDecrypted )) )
    {
        db = (DATA_BLOCK*)m_dbDecrypted;
        is_encrypted = false;
    }

    if (is_encrypted) {
        return 0;
    }
    
    BCKeySlot * kmap = ((db_header*)db)->keymap;
    size_t mainSlotsCount = DataBlock_CapacityForSize( 4096 );

    bc_u16 keyblocksCount = 0;
    for (bc_u16 i=0; i<mainSlotsCount; ++i) {
        if (kmap[i].type != kBCKeyType_Part &&
            kmap[i].type != kBCKeyType_Salt &&
            kmap[i].type != kBCKeyType_Empty &&
            kmap[i].type != kBCKeyType_HP) {
            ++keyblocksCount;
        }
    }
    return keyblocksCount;
}

BCContainerResult KGGHOST_Header::encrypt(BCPassword & password)
{
    DATA_BLOCK*     db = (DATA_BLOCK*) m_db;
    db_header*      header = (db_header*) db;

	if (DataBlock_IsHeaderEncrypted(db)) {
		return kBCContainer_Success;
	}

	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );
	if ( kBCCipher_Success != sCipher->init( header->alg_id, header->getMode() ) ) {
		return kBCContainer_InvalidAlgorithm;
	}

    key_descriptor_t kdesc;
    key_data_t key;

    if (!DataBlock_DecodeKey(db, m_dbsize, &password, m_hiddenMap, &kdesc, &key)) {
        return kBCContainer_InvalidPassword;
    }

    if (key.is_hidden()) {
        return kBCContainer_InvalidPassword;
    }

    if(!DataBlock_VerifyDigest(db, &key)) {
        return kBCContainer_Damaged;
    }

    memcpy(m_dbDecrypted, m_db, m_dbsize);

    DataBlock_EncryptHeader(db, &key);
    return kBCContainer_Success;
}

//virtual
BCContainerResult KGGHOST_Header::encrypt( const cPTB &  _mainKey)
{
    if (isEncrypted()) {
        return kBCContainer_Success;
    }
    
	if ( _mainKey.bufferSize != sizeof( key_data_t ) )
		return kBCContainer_InvalidArgument;

    key_data_t	*key	= (key_data_t *)_mainKey.buffer;
    DATA_BLOCK	*db		= (DATA_BLOCK*) m_db;

    if (key->is_hidden()) {
        return kBCContainer_InvalidPassword;
    }

    if(!DataBlock_VerifyDigest(db, key)) {
        return kBCContainer_Damaged;
    }

    memcpy(m_dbDecrypted, m_db, m_dbsize);
    
    DataBlock_EncryptHeader(db, key);
    return kBCContainer_Success;
}

 
BCContainerResult KGGHOST_Header::decrypt(BCPassword & password)
{
    DATA_BLOCK*	    db = (DATA_BLOCK*) m_db;

    if(!DataBlock_IsHeaderEncrypted(db)) {
        return kBCContainer_Success;
    }

    key_descriptor_t kdesc;
    key_data_t key;

    if (!DataBlock_DecodeKey(db, m_dbsize, &password, m_hiddenMap, &kdesc, &key)) {
        return kBCContainer_InvalidPassword;
    }

    if (key.is_hidden()) {
        return kBCContainer_InvalidPassword;
    }

    if (!DataBlock_DecryptHeader(db, &key)) {
        return kBCContainer_Damaged;
    }

    return kBCContainer_Success;
}

//virtual
BCContainerResult KGGHOST_Header::decrypt( const cPTB &  _mainKey )
{
	if ( _mainKey.bufferSize != sizeof( key_data_t ) )
		return kBCContainer_InvalidArgument;

    key_data_t	*key	= (key_data_t *)_mainKey.buffer;
    DATA_BLOCK	*db		= (DATA_BLOCK*) m_db;

    if(!DataBlock_IsHeaderEncrypted(db)) {
        return kBCContainer_Success;
    }
	
    if (key->is_hidden()) {
        return kBCContainer_InvalidPassword;
    }

    if (!DataBlock_DecryptHeader(db, key)) {
        return kBCContainer_Damaged;
    }

    return kBCContainer_Success;
}

BCContainerResult KGGHOST_Header::wipeDatablock(const cPTB &  _mainKey)
{
    shred(m_db, sizeof(m_db));
    m_dbsize = kBCV8_InitialDataBlockSize;
    
    return storeDatablock();
}

BCContainerResult KGGHOST_Header::storeDatablock()
{
    if ( m_channel == NULL || !m_channel->open( kBCFileMode_ReadWrite ) )
        return kBCContainer_IOError;

    db_header * header = NULL;
    bool is_encrypted = DataBlock_IsHeaderEncrypted((DATA_BLOCK*)m_db);
    if ( is_encrypted ) {
        if (!DataBlock_IsHeaderEncrypted( (DATA_BLOCK*)m_dbDecrypted )) {
            header = (db_header*)m_dbDecrypted;
            is_encrypted = false;
        }
    } else {
        header = (db_header*) m_db;
    }
    
    if (is_encrypted && m_dbsize > kBCV8_InitialDataBlockSize) {
        m_channel->close();
        return kBCContainer_InvalidPassword;
    }
    
    // Check if current size of the file is larger then required. If so randomize end of the file and truncate it
    bc_u64 newFileSize = header->data_start + header->data_length + (m_dbsize - kBCV8_InitialDataBlockSize);
    bc_u64 oldFileSize = m_channel->getSize();
    if (newFileSize < oldFileSize) {
        bc_u64 unusedSize = oldFileSize - newFileSize;
        bc_u8 * buf = (bc_u8 *)malloc(unusedSize);
        shred(buf, unusedSize);
        
        if (m_channel->writeBytes(buf, unusedSize, newFileSize) != unusedSize) {
            free(buf);
            m_channel->close();
            return kBCContainer_IOError;
        }
        free(buf);
        m_channel->close();
        if (m_channel->truncate(newFileSize) != newFileSize) {
            return  kBCContainer_InternalError;
        }
        m_channel->open( kBCFileMode_ReadWrite );
    }
    
    // write initial data block header
    if (kBCV8_InitialDataBlockSize != m_channel->writeBytes( m_db, kBCV8_InitialDataBlockSize)) {
        m_channel->close();
        return kBCContainer_IOError;
    }
    
    // write the rest to the end of the file
    bc_u64 extKbSize = m_dbsize - kBCV8_InitialDataBlockSize;
    if (extKbSize) {
        bc_u64 extKbOffset = header->data_start + header->data_length;
        if (extKbSize != m_channel->writeBytes( m_db + kBCV8_InitialDataBlockSize, extKbSize, extKbOffset)) {
            m_channel->close();
            return kBCContainer_IOError;
        }
    }
    
    m_channel->close();
    
    return kBCContainer_Success;
}

BCContainerResult KGGHOST_Header::store( iChannel* channel )
{
    if ( channel == NULL || !channel->open( kBCFileMode_ReadWrite ) ) {
        return kBCContainer_IOError;
    }

    channel->close();
    m_channel = channel;
    
    return storeDatablock();
}

BCContainerResult KGGHOST_Header::storeKeyblockBackup(iChannel * channel)
{
    if ( !channel->open( kBCFileMode_ReadWrite ) ) {
        LIBLOG_INFO( "Error in opening channel" );
        return kBCContainer_InvalidArgument;
    }

    bc_u32 blockSize = (bc_u32)m_dbsize;
    if (channel->writeBytes(&blockSize, sizeof(blockSize)) != sizeof(blockSize)) {
        channel->close();
        return kBCContainer_IOError;
    }

    if (channel->writeBytes(m_db, blockSize) != blockSize) {
        channel->close();
        return kBCContainer_IOError;
    }

    /* Write hidden sector */
    if (channel->writeBytes(m_db, 512) != 512) {
        channel->close();
        return kBCContainer_IOError;
    }

    channel->close();
    return kBCContainer_Success;
}

KGGHOST_Header::~KGGHOST_Header()
{
    shred(m_db, sizeof(m_db));
    shred(m_dbDecrypted, sizeof(m_dbDecrypted));
}


static IBCHeader* kgghost_create(void)
{
    bc_u8 db[kBCV8_MaximumDataBlockSize] = {0};
    unsigned dbsize = kBCV8_InitialDataBlockSize;
    return new KGGHOST_Header( NULL, db, dbsize, false);
}

//
//	readDBFromKeyBlockBackup
//	Just read key block from packed key block file.
//		Key block file consists of:
//			1. 4 bytes - key block size 
//			2. the key block
//			3. 512 bytes hidden sector
//
static bool readDBFromKeyBlockBackup( bc_u8 *db, unsigned &dbSize, iChannel *channel )
{
	if ( !channel->open( kBCFileMode_Read ) )
	{
    	LIBLOG_INFO( "Error in opening channel" );
		return false;
	}

	memset( db, 0, kBCV8_MaximumDataBlockSize );
	//
	//	Read block size first...
	//
	bc_u32 blockSize;
	bc_u64 read = channel->readBytes( &blockSize, sizeof( blockSize ) );
	if ( read != sizeof( blockSize ) )
	{
		LIBLOG_ERROR( "Error in reading channel!" );
		return false;
	}

	if ( blockSize > kBCV8_MaximumDataBlockSize )
	{
		LIBLOG_ERROR( "Error! Block size too large!" );
		channel->close();
		return false;
	}


	if ( blockSize == 0 )
	{
		LIBLOG_ERROR( "Error! Block size is zero!" );
		channel->close();
		return false;
	}
	
	dbSize = blockSize;

	//
	//	Read data block...
	//
	read = channel->readBytes( db, dbSize );
	if ( read != dbSize )
	{
		LIBLOG_ERROR( "Error in reading channel!" );
		channel->close();
		return false;
	}

	//
	//	Read hidden sector...
	//		Skipped.
	//
	unsigned hsSize = 512;

	bc_u64 fileSize = channel->getSize();

	bool ok = ( fileSize == ( sizeof( blockSize ) + blockSize + hsSize ) );

	channel->close();
	return ok;
}

static bool readDBFromContainer( bc_u8 *db, unsigned &dbSize, iChannel *channel )
{
    db_header* header = (db_header*) db;

	if ( !channel->open( kBCFileMode_Read ) )
	{
    	LIBLOG_INFO( "Error in opening channel" );
		return false;
	}

	if (kBCV8_InitialDataBlockSize != channel->readBytes( header, kBCV8_InitialDataBlockSize))
	{
    	LIBLOG_INFO("Error in reading channel" );
		channel->close();
        return false;
    }

	// if header is invalid we can possibly have an encrypted header.
	// but we can't know for sure because there might be another keygen that will recognize this container.
    bool encrypted = false;
	if (memcmp( header->signature, DB_HEADER_SIGNATURE, kBCV8_SignatureSize) != 0)
	{
    	LIBLOG_INFO("Header is encrypted" );
        encrypted = true;
    }
	else if (header->wKeyGenId == kKGID_V5)
	{
    	LIBLOG_INFO("Header version %d, id %d", header->version, header->wKeyGenId );
	}
	else if ((header->getVersion() != 3) || (header->wKeyGenId != kKGID))
	{
    	LIBLOG_INFO("Invalid header version %d, id %d", header->version, header->wKeyGenId );
		channel->close();
        return false;
    }

	// see how much, if any, extended key data we need to read and allocate enough data to hold everything.
	bc_u64 fileSize	= channel->getSize();
    bc_u64 extKbOffset = 0;
    bc_u32 extKbSize = 0;

    if (!encrypted) {
        extKbOffset	= header->data_length + header->data_start;
    	if(extKbOffset > fileSize) {
	    	// this is possible when we are dealing with backup headers.
		    extKbSize = 0;
	    } else {
			extKbSize = u64tou32( fileSize - extKbOffset, kBCV8_MaximumExtKeyBlockSize );
	    }
    } else {
        extKbSize = u64tou32( fileSize - kBCV8_InitialDataBlockSize, kBCV8_MaximumExtKeyBlockSize );
        extKbOffset = fileSize - extKbSize;
    }

    if(extKbSize > kBCV8_MaximumExtKeyBlockSize)
    {
    	LIBLOG_INFO("Extended block too large %d", extKbSize );
		channel->close();
        return false;
    }

    if (extKbSize > 0)
    {
		if (extKbSize != channel->readBytes( (bc_u8 *)header + kBCV8_InitialDataBlockSize, extKbSize, extKbOffset))
		{
	    	LIBLOG_INFO("Error inreading extended block" );
			channel->close();
            return false;
        }
    }
	
	dbSize = kBCV8_InitialDataBlockSize + extKbSize;

	channel->close();

	return true;
}

//
//	kgghost_load
//		Parses keyBlockChannel as key block backup (.kbb file ) or standard container ( .jbc ).
//
static IBCHeader* kgghost_load( iChannel *channel, iChannel *keyBlockChannel = 0 )
{
	bc_u8 db[kBCV8_MaximumDataBlockSize] = {0};
	unsigned dbSize = 0;
	
	if ( keyBlockChannel != NULL )
	{
		if ( !readDBFromKeyBlockBackup( db, dbSize, keyBlockChannel ) &&
			 !readDBFromContainer( db, dbSize, keyBlockChannel )
			)
				return NULL;
	}
	else
	if ( !readDBFromContainer( db, dbSize, channel ) )
		return NULL;
 
    return new KGGHOST_Header( channel, db, dbSize, false);
}

BCContainerResult kgghost_parse( iChannel *channel, struct BCVolumeInfo* info)
{
    BCHeaderHandle header(kgghost_load( channel ));
    if (!header || header->isEncrypted()) {
        return kBCContainer_UnknownFormat;
    }

    header->mainVolumeInfo(info);
    return kBCContainer_Success;
}


struct BCKeygen kg_ghost = {
    kKGID,
    "kgghost",
    "Default BestCrypt version 8 key generator",
    BCV8_VERSION_TAG,
    kBCKeygen_EncryptedHeaders,

    kgghost_create,
    kgghost_load,
    kgghost_parse,
};


