/*
 * Copyright (c) 2005-2010 Jetico, Inc., Finland
 * All rights reserved.
 *
 * Contains common keygen auth session definitions.
 */

#ifndef KGGHOST_H
#define KGGHOST_H

#include <stdlib.h>
#include <string>
//#include <cstring>

#include "../../container/keygen.h"
#include "../../container/intsize.h"
#include "../../container/user.h"
#include "../../container/cipherbase.h"
#include "db_header.h"
#include "hash.h"
#include "../../certificate/CertificateLib/SMemory/SBlocks.h"
#include "../../certificate/CertificateLib/PGPPkts/pktdefs.h"


namespace kgghost {
	
	// Common definitions.
	enum {
        kKGID               = 4,   // Keygen id
		kKGID_V5			= 5,   // Keygen id for BestCrypt version 5
        kKeySlotSize        = 256, // Single encoded key slot.
        kDigestSize32       = 32,  // Digest buffer length.
        kDigestSize64       = 64,  // Digest buffer length.
        kIVSize	            = 16,  // IV buffer length.
        kSeedLength         = 128, // random seed buffer size for key generation.
        kEncryptedBlockSize = 16,  // Block cipher encryption block size.
        kMaximumKeySize     = 64,  // Maximum stored key size.
        kPoolSize           = 512  // Random data pool size.
	};
	
	
    #pragma pack(push, 1)
	// This is the key data encoded to and decoded from key blocks by coder algorithms.
	typedef struct key_data {
		bc_u8		size;		// real size of the structure in bytes
		bc_u8		hash_id;	// Digest algorithm used for key block verification ( same as pgp id ( f.e. pgphaSHA1 or bchaSHA256 see pktdefs.h )
		bc_u32		alg_id;		// Algorithm is used for encryption of header, contInfo, hiddenContInfo, keyMap and container body
		bc_u32		mode_id;	// The mode is used for encryption for encryption algorithm above
		bc_u64		offset;		// container data offset that is encrypted with this key.
		bc_u64		datasize;	// used as a hidden key marker, will be set to 4096 if the key is hidden.
		
		bc_u8		buf[kMaximumKeySize];
		
		key_data() {
			memset(this, 0, sizeof(key_data));
		}
		
		~key_data() {
			memset(this, 0, sizeof(key_data));
		}
		
		/// Get total key structure size that contains key with the given length.
		static inline size_t calc_size_for_key_length(size_t keylen) {
			return sizeof(key_data) + keylen;
		}

		bool is_hidden() const {
			return offset != kBCV8_InitialDataBlockSize;
		}

		bc_mode_id getMode() const {
			return (bc_mode_id)mode_id;
		}
		
	} key_data_t;
    #pragma pack(pop)	
	
    /**
     * kgghost keygen interface implementation.
     */
    class KGGHOST_Header : public IBCHeader {
    public:
        KGGHOST_Header( iChannel *channel, bc_u8 *datablock, unsigned size, bool bEncrypted);

        virtual void mainVolumeInfo(struct BCVolumeInfo* info);
        virtual bool isEncrypted();
        virtual bool isValid()
        {
        	return true; // It may be encrypted.
        }

        virtual int setBusiedName( const char *compName, int len );
        virtual int getBusiedName( char *compName, int len );

        virtual BCContainerResult init(	bc_alg_id algid, bc_mode_id modeid, bc_u64 size, const char* description, BCPassword & password )
		{
			return initEx( algid, modeid, size, description, password );
		}

		virtual BCContainerResult initEx(
			bc_alg_id algid, bc_mode_id modeid, bc_u64 size, const char* description, BCPassword & password,
			bc_u16 kgID = kKGID, int hash = bchaSHA256, bc_u32 iterations = 256, const bc_u8 *salt = NULL, int saltSize = 0 ); // bchaWhirlpool512

        //virtual BCContainerResult createHiddenKey(SpaceMap *spaceMap, BCPassword* mainPassword, BCPassword* hiddenPassword);
        virtual BCContainerResult createHiddenKey(SpaceMap *spaceMap, cPTB *mainKey, BCPassword* hiddenPassword);
        virtual BCContainerResult verifyPassword(BCPassword & password, struct BCVolumeInfo* out_info, cProgressActor *progressActor, cPTB *keyOut = NULL );
        virtual BCContainerResult decodeKey(BCPassword & password, struct BCVolumeInfo* out_info, bc_key_handle* out_handle, CipherBase **cipher_out, cPTB *keyOut = NULL );
		virtual BCContainerResult generateCipherInfo( const cPTB &keyIn, struct BCVolumeInfo* out_info, bc_key_handle* out_handle, CipherBase **cipher_out );
        virtual BCContainerResult addPassword(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & newPassword);
        virtual BCContainerResult changePassword(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & newPassword);
        virtual BCContainerResult removePassword(BCPassword & mainPassword, BCPassword & oldPassword);
        virtual BCContainerResult store( iChannel* channel );
		virtual BCContainerResult encrypt( const cPTB &  _mainKey );
        virtual BCContainerResult encrypt(BCPassword & password);
        virtual BCContainerResult decrypt( const cPTB &  _mainKey );
        virtual BCContainerResult decrypt(BCPassword & password);
        virtual BCContainerResult setDescription(const char* description, const cPTB & mainKey);
        virtual void markForReencryption(bool mark);
        virtual bool isMarkedForReencryption();

		/**
		 * Get name of keygen used
		 */
		virtual const char * getKeygenName();
	    
		/**
		 * Returns true if header encryption or decryption is allowed
		 */
		virtual bool isHeaderEncryptionAllowed();
		/**
		 * Returns true if container can be reencrypted
		 */
		virtual bool isReencryptionAllowed();
        
        virtual bc_u16 getMainKeyblocksCount();

        virtual BCContainerResult storeKeyblockBackup(iChannel *channel);
        
        virtual BCContainerResult wipeDatablock(const cPTB &  _mainKey);

        virtual ~KGGHOST_Header();

    protected:

        BCContainerResult volumeInfoForKey(key_data_t* key, struct BCVolumeInfo* info);
        BCContainerResult passwd(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & password, BCKeyOp opflag);
        void updateDataBlockSize();
        BCContainerResult storeDatablock();

    private:
        
        bc_u8		m_db[kBCV8_MaximumDataBlockSize];   // data block buffer
		bc_u8		m_dbDecrypted[kBCV8_MaximumDataBlockSize];

        size_t	m_dbsize;                           // size of data block as read from file
		iChannel*	m_channel;							// channel interface to container.

		BCKeySlot	m_hiddenMap[kBCV8_MaxKeys];
    };
}

#endif

