/*
 *	Copyright (c) 2004-2012 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Password based key block encoder.
 *
 *	Revision: 05.11.2012
 *		Updated for CipherBase implementation.
 */

#include "keyblock.h"

#define PBE_HASH_ITERATIONS	0x100
#define PBE_SALT_SIZE		8

using namespace kgghost;

// Encoded key representation format stored in keyblock.
// We encode it by generating the storage key from password and encrypting into keyblock.
#pragma pack(push, 1)

typedef struct {
    bc_u8				            salt[PBE_SALT_SIZE];    // randomly generated salt for password.
    KGEncryptedBlock32<key_data_t>	keyblock;               // encoded key.
    bc_u8				            iv[kIVSize];            // unencrypted initial vector.
} pbe_format;

// PBE key block format for version 5
typedef struct {
    KGEncryptedBlock64<key_data_t>	keyblock;               // encoded key.
    bc_u8				            iv[kIVSize];            // unencrypted initial vector.
} pbe_format_v5_64;

typedef struct {
    KGEncryptedBlock32<key_data_t>	keyblock;               // encoded key.
    bc_u8				            iv[kIVSize];            // unencrypted initial vector.
} pbe_format_v5_32;

#pragma pack(pop)


namespace kgghost {

	unsigned getPBEBlockSize( const key_data_t* key )
	{
		int digestSize;
		if ( ha_getDigestSize( &digestSize, key->hash_id ) != hae_ok )
		{ assert( false ); }

		if ( digestSize > 32 )
			return sizeof( pbe_format_v5_64 );
		else
			return sizeof( pbe_format_v5_32 );
	}


    unsigned pbe_encode(const key_data_t* key, const BCPassword* pwd, CipherBase* cipher, bc_u8* kb_data, unsigned kb_size)
    {
	    if(kb_size < sizeof(pbe_format)) {
		    return 0;
	    }

	    // shredding the key block will have a side effect of generating random data for salt and iv.
	    shred(kb_data, kb_size);
	    pbe_format* pbe = (pbe_format*) kb_data;

		//BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );
		//if ( kBCCipher_Success != sCipher->init( key->alg_id, key->getMode() ) )
		//{
		//	assert( false );
		//	return 0;
		//}
	

		//
		//	V5 keygen generates storage key from salt + key files + password.
		//
		if ( pwd->bReady )
		{
			int digestSize;

			if ( ha_getDigestSize( &digestSize, key->hash_id ) != hae_ok )
			{
				assert( false );
				return 0;
			}

			if ( digestSize <= 32 )
			{
				pbe_format_v5_32* pbe32 = (pbe_format_v5_32*) kb_data;

				KGEncryptBlock32( key, *(bc_u64*)pbe32->iv, cipher, key->hash_id, pwd->storage_key, sizeof(pwd->storage_key), NULL, &pbe32->keyblock );
				return sizeof(pbe_format_v5_32);
			}
			else
			{
	    	    pbe_format_v5_64* pbe64 = (pbe_format_v5_64*) kb_data;

				KGEncryptBlock64(key, *(bc_u64*)pbe64->iv, cipher, key->hash_id, pwd->storage_key, sizeof(pwd->storage_key), NULL, &pbe64->keyblock);
				return sizeof(pbe_format_v5_64);
			}

			return 0;
		}

	    bc_u8 storage_key[kMaximumKeySize] = {0};
		bool ok = BCPasswordUTF16::generate_key_from_password(key->hash_id, pwd->buf, pwd->length, pbe->salt,
                                   sizeof(pbe->salt), PBE_HASH_ITERATIONS,
                                   storage_key, sizeof(storage_key));
	
		if ( ok )
			KGEncryptBlock32(key, *(bc_u64*)pbe->iv, cipher, key->hash_id, storage_key,
                       sizeof(storage_key), NULL, &pbe->keyblock);

		shred( storage_key, sizeof( storage_key ) );

		if ( !ok )
			return 0;

	    return sizeof(pbe_format);
    }

    unsigned pbe_decode(const bc_u8* kb_data, unsigned kb_size, const BCPassword* pwd, CipherBase* cipher, int hash, key_data_t* out_key)
    {
	    if(kb_size < sizeof(pbe_format)) {
		    return 0;
	    }

		if ( pwd->bReady )
		{
			int digestSize;

			if ( ha_getDigestSize( &digestSize, hash ) != hae_ok )
			{
				assert( false );
				return 0;
			}

			if ( digestSize <= 32 )
			{
				const pbe_format_v5_32* pbe32 = (const pbe_format_v5_32*) kb_data;

				if ( KGDecryptBlock32(&pbe32->keyblock, *(bc_u64*) pbe32->iv, cipher, hash, pwd->storage_key, sizeof(pwd->storage_key), NULL, out_key))
					return sizeof(pbe_format_v5_32); // pbe_format);
			}
			else
			{
	    	    const pbe_format_v5_64* pbe64 = (const pbe_format_v5_64*) kb_data;

				if ( KGDecryptBlock64(&pbe64->keyblock, *(bc_u64*) pbe64->iv, cipher, hash, pwd->storage_key, sizeof(pwd->storage_key), NULL, out_key))
					return sizeof(pbe_format_v5_64);
			}

			return 0;
	    }
	
	    const pbe_format* pbe = (const pbe_format*) kb_data;

	    bc_u8 storage_key[kMaximumKeySize] = {0};
		bool ok = BCPasswordUTF16::generate_key_from_password(hash, pwd->buf, pwd->length, pbe->salt, sizeof(pbe->salt), PBE_HASH_ITERATIONS, storage_key, sizeof(storage_key));
	
		unsigned res = 0;

        if (ok && KGDecryptBlock32(&pbe->keyblock, *(bc_u64*) pbe->iv, cipher, hash, storage_key, sizeof(storage_key), NULL, out_key)) {
			res = sizeof(pbe_format);
	    }

	    shred(storage_key, sizeof(storage_key));

		return res;
    }


    bool pbe_verify(const bc_u8* kb_data, unsigned kb_size, const BCPassword* pwd,
                    CipherBase* cipher, int hash)
    {
	    // for password encoded keys verification and decoding is the same thing since we don't ask anything else.
	    key_data_t tmpkey;
	    return (pbe_decode(kb_data, kb_size, pwd, cipher, hash, &tmpkey) != 0);
    }

}

