/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 *
 * KGSHA keygen library interface.
 */

#include <stdlib.h>
#include <string.h>

#include "../../container/keygen.h"
#include "../../container/cipherbase.h"
#include "../../container/user.h"
#include "../../container/file.h"

#include "kgsha.h"
#include "hash.h"
#include "defs.h"
#include "datablock.h"

KGSHA_Header::KGSHA_Header(HiddenSector* hs, DATA_BLOCK* db)
{
    memcpy(&m_hs, hs, sizeof(*hs));

    if ( db != NULL )
    	memcpy(&m_db, db, sizeof(m_db));
    else
    	memset( &m_db, 0, sizeof(m_db));
}

void KGSHA_Header::mainVolumeInfo(struct BCVolumeInfo* info)
{
    memset(info, 0, sizeof(*info));

    info->alg      = m_hs.algId;
    info->mode     = kBCMode_CBC;      // we only support CBC
    info->hash     = m_db.header.hashAlgorithmId;
    info->size     = m_hs.br.totalSectorsLong * 512LL;
    info->offset   = m_hs.dwDataOffset;

    strncpy(info->description, m_hs.description, sizeof(m_hs.description));
}

BCContainerResult KGSHA_Header::volumeInfoForKey(const kgsha_key* key, struct BCVolumeInfo* out_info)
{
    assert(key && out_info);

    memset(out_info, 0, sizeof(*out_info));

    out_info->alg       = getEncAlgId();
    out_info->mode      = kBCMode_CBC;
    out_info->hash      = m_db.header.hashAlgorithmId;
    out_info->hidden    = key->is_hidden;
    out_info->size      = m_hs.br.totalSectorsLong * 512LL;
	out_info->use_new_synchro = m_hs.version != 0;

    if (key->is_hidden)	{
		// 20.10.2015 Offset was not muliplied by sector size.
        out_info->offset = key->data_offset * 512LL;
        out_info->size -= out_info->offset;
    } else {
        out_info->offset = m_hs.dwDataOffset;
        strncpy(out_info->description, m_hs.description, sizeof(m_hs.description));
    }
    
    out_info->map = SpaceMapCreate();
    if (!out_info->map) {
        return kBCContainer_NoMemory;
    }
    
    sm_result smres = SpaceMapAddNew(out_info->map, out_info->offset, out_info->size);
    if (kSpaceMap_Success != smres) {
        SpaceMapFree(out_info->map);
        LIBLOG_INFO( "Assertion! Space map adding error %d", smres );
        
        return kBCContainer_InternalError;
    }
    return kBCContainer_Success;
}

BCContainerResult KGSHA_Header::setDescription(const char* description, const cPTB & mainKey)
{
    memset(m_hs.description, 0, sizeof(m_hs.description));
    strncpy(m_hs.description, description, sizeof(m_hs.description));
    
    return kBCContainer_Success;
}

void KGSHA_Header::markForReencryption(bool mark)
{
    if (mark) {
        strncpy(m_hs.br.volumeLabel, BCV7_REENCRYPTED_LABEL, sizeof(m_hs.br.volumeLabel));
    } else {
        strncpy(m_hs.br.volumeLabel, BCV7_VOLUME_LABEL, sizeof(m_hs.br.volumeLabel));
    }
}

//virtual
const char * KGSHA_Header::getKeygenName()
{
	switch ( m_hs.keygenId )
	{
		case 2: return "SHA-1";
		case 3: return "SHA-256";
		default:
			return "SHA";
	}
}

bool KGSHA_Header::isMarkedForReencryption()
{
    return (0 == strncmp(m_hs.br.volumeLabel, BCV7_REENCRYPTED_LABEL, sizeof(m_hs.br.volumeLabel)));
}

//virtual
bool KGSHA_Header::isValid()
{
    if ( 0 != strncmp(m_hs.br.OEMid, "LOCOS94 ", 8) )
    	return false;

	if ( (0 == strncmp(m_hs.br.volumeLabel, BCV7_VOLUME_LABEL, sizeof(m_hs.br.volumeLabel))) ) return true;
	if ( (0 == strncmp(m_hs.br.volumeLabel, BCV7_REENCRYPTED_LABEL, sizeof(m_hs.br.volumeLabel))) ) return true;

	return false;
}

enum { busiedNamePos = 15, busiedNameLen = 28 };

//
// getValidLen
//		returns length before first space ' ' or maxLen
//
static int getValidLen( int origLen, const char *computerName, int maxLen )
{
	if ( origLen <= maxLen )
		return origLen;

	for( int i = 0; i < maxLen; i++ )
	{

		if ( computerName[i] == ' ' )
			return i;
	}

	return maxLen;
}

int KGSHA_Header::setBusiedName( const char *computerName, int len )
{
 	LIBLOG_INFO( "Begin '%s' length %d", computerName, len );

 	bc_u8 *firstByte = m_hs.br.jmpCode;

	bc_u8 *pos = firstByte + busiedNamePos;

	if ( computerName[0] == '\0' )
		len = 0;
	else
	if ( len == -1 )
		len = strlen(computerName);

	// Clear old name
	memset( pos, 0, busiedNameLen );
	firstByte[0] = 0;

	if ( len > 0 ) // Set new computer name
	{
		len = getValidLen( len, computerName, busiedNameLen / 2 );

		size_t res = BCUserUTF8ToUTF16LE( computerName, len, pos, busiedNameLen );
		firstByte[0] = 'B';
		LIBLOG_INFO( "BCUserUTF8ToUTF16LE(%s, %d, %p:%p, %d) = %d  %c, %c.",
				computerName, len, firstByte, pos, busiedNameLen, res, firstByte[0], firstByte[15]  );

		len = (int)res;
	}
	else
	{
		LIBLOG_INFO( "0x%x, 0x%x.",
				(unsigned int)firstByte[0], (unsigned int)firstByte[15]  );
	}

 	LIBLOG_INFO( "End length %d", len );
	return len;
}

int KGSHA_Header::getBusiedName( char *computerName, int len )
{
 	LIBLOG_INFO( "Begin %p length %d", computerName, len );

	if ( len > busiedNameLen )
		len = busiedNameLen;

	if ( m_hs.br.jmpCode[0] == 0 )
	{
		computerName[0] = 0;
		return 0;
	}

	bc_u8 *pos = (bc_u8*)&m_hs + busiedNamePos;

	size_t res = BCUserUTF16LEToUTF8( pos, busiedNameLen, computerName, len );

 	LIBLOG_INFO( "End '%s' %d", computerName, len );

	computerName[len-1] = '\0';
	return len;
}

BCContainerResult KGSHA_Header::init(bc_alg_id algid, bc_mode_id modeid, bc_u64 size, const char* description, BCPassword & password)
{
    // Init blank header and data block structures
    memset(&m_hs, '0', sizeof(HiddenSector));
    memset(&m_db, '0', sizeof(DATA_BLOCK));

    strncpy(m_hs.br.OEMid, "LOCOS94 ", 8);
    strncpy(m_hs.br.volumeLabel, "CRYPTED_DSK", 11);

    m_hs.br.jmpCode[0] = 0xEB;
    m_hs.br.jmpCode[1] = 0x3C;
    m_hs.br.jmpCode[2] = 0x90;
    m_hs.dwKeySize = sizeof(DATA_BLOCK);
    m_hs.dwDataOffset = (512 + sizeof(DATA_BLOCK) + 511) & ~511;
    m_hs.br.totalSectorsLong = (bc_u32)(size / 512) ;
    m_hs.algId = algid;
    m_hs.keygenId = kKGSHA256_ID;
    m_hs.version = 7;

    if(description) {
        size_t len = strlen(description);
        if (len != 0) {
		    memcpy(m_hs.description, description, len > sizeof(m_hs.description) ? sizeof(m_hs.description) : len);
	    }
    }

    DataBlock_Init(&m_db, algid, kSHA256_ID);
    bool ok = DataBlock_Verify(&m_db);
	assert( ok );

    // Init container key and encode it with the password
    if (modeid != kBCMode_CBC) {
        return kBCContainer_InvalidAlgorithm;
    }
        
	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

	if ( kBCCipher_Success != sCipher->init(algid, modeid) ) {
		return kBCContainer_InvalidAlgorithm;
	}

    bc_u8 seed[kKGSHA_SeedSize] = {0};
    if (sizeof(seed) != BCUserGenerateSeed(seed, sizeof(seed))) {
        return kBCContainer_Cancelled;
    }

    kgsha_password pwd;
    DataBlock_GenerateStorageKey(&m_db, password.buf, strlen(password.buf), &pwd);

    kgsha_key key;
    DataBlock_GenerateKey(&m_db, seed, false, m_hs.dwDataOffset, &key);

    DataBlock_StoreKeyWithPassword(&m_db, &key, &pwd, sCipher.get());
    return kBCContainer_Success;
}

BCContainerResult KGSHA_Header::createHiddenKey(SpaceMap *map, // BCPassword* mainPassword,
												cPTB *ptbMainKey,
												BCPassword* hiddenPassword)
{
	if ( getEncAlgId() == kDummy_ID )
	    return kBCContainer_Success;

	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

	if ( kBCCipher_Success != sCipher->init(getEncAlgId(), kBCMode_CBC)) {
        return kBCContainer_InvalidAlgorithm;
    }

    bc_u8 seed[kKGSHA_SeedSize] = {0};
    if (sizeof(seed) != BCUserGenerateSeed(seed, sizeof(seed))) {
        return kBCContainer_Cancelled;
    }

    if ((!map) || (!map->start))
        return kBCContainer_InvalidArgument;

    // The data offset is equals to the start of the first extent.
    bc_u64 offset = map->start->extent.start;

    kgsha_password pwd;
    DataBlock_GenerateStorageKey(&m_db, hiddenPassword->buf, hiddenPassword->length, &pwd);

    kgsha_key key;
    DataBlock_GenerateKey(&m_db, seed, true, (bc_u32)offset, &key);

    if(!DataBlock_StoreKeyWithPassword(&m_db, &key, &pwd, sCipher.get())) {
        return kBCContainer_NoKeySpace;
    }
    
    return kBCContainer_Success;
}

int KGSHA_Header::getEncAlgId()
{
	// SHA key gen saves encryption key gen id at data block.
	int encAlg = m_db.header.encrAlgorithmId;

	if ( ( 0 == strncmp(m_hs.br.volumeLabel, "CRYPTED_DSK", sizeof(m_hs.br.volumeLabel))) && 
		 ( m_hs.keygenId == kDummy_ID ) )
				encAlg = m_hs.algId;

	return encAlg;
}

BCContainerResult KGSHA_Header::verifyPassword(BCPassword & password, struct BCVolumeInfo* out_info, cProgressActor *progressActor, cPTB *keyOut )
{
	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

	if ( kBCCipher_Success != sCipher->init( getEncAlgId(), kBCMode_CBC ) ) {
		return kBCContainer_InvalidAlgorithm;
	}

	kgsha_key key;
	memset( &key, 0, sizeof( key ) );

	if ( getEncAlgId() != kDummy_ID )
	{
		kgsha_password pwd;
		DataBlock_GenerateStorageKey(&m_db, password.buf, strlen(password.buf), &pwd);

		bc_u8 pool[512] = {0};
		if (!DataBlock_LoadKeyWithPassword(&m_db, &pwd, sCipher.get(), &key, pool)) {
			return kBCContainer_InvalidPassword;
		}
	}

	volumeInfoForKey(&key, out_info);

	if ( NULL != keyOut )
		keyOut->Assign( (bc_u8*)&key, sizeof( key ) );

	return kBCContainer_Success;
}

//virtual
BCContainerResult KGSHA_Header::generateCipherInfo( const cPTB &keyIn, struct BCVolumeInfo* out_info, bc_key_handle* out_handle, CipherBase **cipher_out )
{
	if ( !cipher_out )
		return kBCContainer_InvalidArgument;

	if ( sizeof( kgsha_key ) != keyIn.bufferSize )
		return kBCContainer_InvalidArgument;

	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

	if ( kBCCipher_Success != sCipher->init( getEncAlgId(), kBCMode_CBC)) {
        return kBCContainer_InvalidAlgorithm;
    }

	kgsha_key &key = *((kgsha_key*)keyIn.buffer);

	if ( getEncAlgId() == kDummy_ID ) {
		if ( NULL != out_handle )
			*out_handle = 0;
	} else {
        bc_u8 pool[kKGSHA_PoolSize] = {0};
        memcpy(pool, m_db.pool, kKGSHA_PoolSize);
        sCipher->decrypt(key.iv, key.data, kKGSHA_KeySize, NULL, pool, kKGSHA_PoolSize);
        
        if (kBCCipher_Success != sCipher->storeKey( key.data, sizeof(key.data), pool, out_handle))
            return kBCContainer_NoMemory;
    }

	*cipher_out = sCipher.delegatePtr();

	volumeInfoForKey(&key, out_info);

	return kBCContainer_Success;
}

BCContainerResult KGSHA_Header::decodeKey(BCPassword & password, struct BCVolumeInfo* out_info, bc_key_handle* out_handle, CipherBase **cipher_out, cPTB *keyOut )
{
	if ( !cipher_out ) {
		return kBCContainer_InvalidArgument;
	}

	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

	if ( kBCCipher_Success != sCipher->init( getEncAlgId(), kBCMode_CBC)) {
        return kBCContainer_InvalidAlgorithm;
    }

	kgsha_key key;

	if ( getEncAlgId() == kDummy_ID )
	{
		if ( NULL != out_handle )
			*out_handle = 0;

		memset( &key, 0, sizeof( key ) );
	}
	else
	{
		kgsha_password pwd;
		DataBlock_GenerateStorageKey(&m_db, password.buf, strlen(password.buf), &pwd);

		bc_u8 pool[512] = {0};

		if (!DataBlock_LoadKeyWithPassword(&m_db, &pwd, sCipher.get(), &key, pool)) {
			return kBCContainer_InvalidPassword;
		}

		if (kBCCipher_Success != sCipher->storeKey( key.data, sizeof(key.data), pool, out_handle)) {
			return kBCContainer_NoMemory;
		}
	}

	*cipher_out = sCipher.delegatePtr();

	volumeInfoForKey(&key, out_info);

	if ( NULL != keyOut )
		keyOut->Assign( (bc_u8*)&key, sizeof( key ) );

	return kBCContainer_Success;
}

//BCContainerResult KGSHA_Header::passwd(BCPassword & password, BCPassword & newPassword, BCKeyOp opflag)
BCContainerResult KGSHA_Header::passwd(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & password, BCKeyOp opflag)
{
    if (!DataBlock_Verify(&m_db)) {
        return kBCContainer_Damaged;
    }
    
	BCScopedPtr < CipherBase > sCipher( CipherFactory::CreateCipher() );

	if ( kBCCipher_Success != sCipher->init( getEncAlgId(), kBCMode_CBC)) {
        return kBCContainer_InvalidAlgorithm;
    }
    
	if ( getEncAlgId() == kDummy_ID )
	    return kBCContainer_Success;

    kgsha_password pwd;
    DataBlock_GenerateStorageKey(&m_db, mainPassword.buf, strlen(mainPassword.buf), &pwd);

    kgsha_key key;
    bc_u8 pool[512] = {0};

    KEY_BLOCK* kb = DataBlock_LoadKeyWithPassword(&m_db, &pwd, sCipher.get(), &key, pool);
    if(!kb) {
        return kBCContainer_InvalidPassword;
    }

    switch (opflag) {
        // remove password.
        case kBCKey_Delete: {
            KeyBlock_Erase(kb);
            break;
        }

        // change password.
        case kBCKey_Recode: {
            kgsha_password new_pwd;
            DataBlock_GenerateStorageKey(&m_db, password.buf, strlen(password.buf), &new_pwd);

            KeyBlock_EncodeKey(&key, new_pwd.storage_key, sCipher.get(), m_db.header.hashAlgorithmId, kb);
            break;
        }

        // add password.
        case kBCKey_Duplicate: {
            kgsha_password new_pwd;
            DataBlock_GenerateStorageKey(&m_db, password.buf, strlen(password.buf), &new_pwd);
            
            if (!DataBlock_StoreKeyWithPassword(&m_db, &key, &new_pwd, sCipher.get())) {
                return kBCContainer_NoKeySpace;
            }
            
            break;
        }
            
        default: {
            return kBCContainer_Unsupported;
        }
    };
    
    return kBCContainer_Success;
}

BCContainerResult KGSHA_Header::addPassword(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & newPassword) {
    return this->passwd(mainPassword, volumePassword, newPassword, kBCKey_Duplicate);
}

BCContainerResult KGSHA_Header::changePassword(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & newPassword) {
    return this->passwd(mainPassword, volumePassword, newPassword, kBCKey_Recode);
}

BCContainerResult KGSHA_Header::removePassword(BCPassword & mainPassword, BCPassword & oldPassword) {
    BCPassword dummy;
    return this->passwd(mainPassword, dummy, oldPassword, kBCKey_Delete);
}

//BCContainerResult KGSHA_Header::store(const char* file)
BCContainerResult KGSHA_Header::store( iChannel* channel )
{
    //bc_file_handle handle;
    //if ( kBCFileResult_Success != BCFileOpen(file, kBCFileMode_ReadWrite, &handle))
	if ( !channel->open( kBCFileMode_ReadWrite ) )
        return kBCContainer_IOError;

    //if (sizeof(HiddenSector) != BCFileWrite(&handle, &m_hs, sizeof(HiddenSector))) {
    //    BCFileClose(&handle);
	if (sizeof(HiddenSector) != channel->writeBytes( &m_hs, sizeof(HiddenSector))) {
		channel->close();
        return kBCContainer_IOError;
    }

    //if (sizeof(DATA_BLOCK) != BCFileWrite(&handle, &m_db, sizeof(DATA_BLOCK))) {
    //    BCFileClose(&handle);
	if (sizeof(DATA_BLOCK) != channel->writeBytes( &m_db, sizeof(DATA_BLOCK))) {
		channel->close();
        return kBCContainer_IOError;
    }

    //BCFileClose(&handle);
	channel->close();
    return kBCContainer_Success;
}

KGSHA_Header::~KGSHA_Header()
{
    memset(&m_hs, 0, sizeof(m_hs));
    memset(&m_hs, 0, sizeof(m_hs));
}



///////////////////////////////////////////////////////////////////////
//
//  Keygen exports.
//
///////////////////////////////////////////////////////////////////////


static IBCHeader* kgsha_load( iChannel *channel, iChannel *keyBlockChannel = 0 )
{
    HiddenSector hs;
    DATA_BLOCK db;

    //bc_file_handle handle;
    // bc_file_result error = BCFileOpen(file, kBCFileMode_Read, &handle);
    //if (kBCFileResult_Success != error)

	if ( !channel->open( kBCFileMode_Read ) )
	{
    	__android_log_print(ANDROID_LOG_INFO, "kgsha_load", "Error in opening channel" );
		return NULL;
	}

    //bc_u64 bytesRead = BCFileRead(&handle, &hs, sizeof(hs));
	bc_u64 bytesRead = channel->readBytes( &hs, sizeof(hs));
    if (sizeof(hs) != bytesRead) {
        //BCFileClose(&handle);
    	__android_log_print(ANDROID_LOG_INFO, "kgsha_load", "Error in reading channel, bytesRead %d != sizeof(hs) %d", (int)bytesRead, (int)sizeof(hs) );
		channel->close();
        return NULL;
    }
    
    if (0 != strncmp(hs.br.OEMid, "LOCOS94 ", sizeof(hs.br.OEMid))) {
		channel->close();
        return NULL;
    }

	if ( kDummy_ID == hs.keygenId )
	{
		channel->close();
		memset( &db, 0, sizeof( db ) );
	}
	else
	{
		//bytesRead = BCFileRead(&handle, &db, sizeof(db));
		bytesRead = channel->readBytes( &db, sizeof( db ) );
		channel->close();

		if (sizeof(db) != bytesRead) {
    		__android_log_print(ANDROID_LOG_INFO, "kgsha_load", "Error in reading channel, bytesRead %d != sizeof(db) %d", (int)bytesRead, (int)sizeof(db) );
		   return NULL;
		}

		if ((kKGSHA_ID != hs.keygenId) && (kKGSHA256_ID != hs.keygenId)) {
    		__android_log_print(ANDROID_LOG_INFO, "kgsha_load", "Key gen id = %d", hs.keygenId );
			return NULL;
		}

		if (!DataBlock_Verify(&db)) {
    		__android_log_print(ANDROID_LOG_INFO, "kgsha_load", "Error in data block verification" );
			return NULL;
		}
	}

    return new KGSHA_Header(&hs, &db);
}


static IBCHeader* kgsha_create(void)
{
    HiddenSector hs;
    DATA_BLOCK db;

    return new KGSHA_Header(&hs, &db);
}


//static BCContainerResult kgsha_parse( const char* file, struct BCVolumeInfo* info)
static BCContainerResult kgsha_parse( iChannel *channel, struct BCVolumeInfo* info)
{
    BCHeaderHandle header( kgsha_load( channel ) );
    if (!header) {
        return kBCContainer_UnknownFormat;
    }

    header->mainVolumeInfo(info);
    return kBCContainer_Success;
}

struct BCKeygen kg_sha = 
{
    kKGSHA256_ID,
    "kgsha/kgsha256",
    "Compatibility with BestCrypt version 7",
    7,
    0,
    kgsha_create,
    kgsha_load,
    kgsha_parse
};

