/*
 * Copyright (c) 2012 Jetico, Inc., Finland
 * All rights reserved.
 */

#ifndef SPACEMAP_H
#define SPACEMAP_H

#include "defs.h"
#include "extent.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef enum
{
    kSpaceMap_Success = 0,
    kSpaceMap_InvalidArg,
    kSpaceMap_NotFound,
    kSpaceMap_NoMem
} sm_result;

typedef struct _SMExtentNode
{
    Extent                  extent;
    struct _SMExtentNode    *next; // TODO: Error! This must be changed to aligned pointer.
    bc_u64                  _align; // TODO: Remove this.
} SMExtentNode;

typedef struct _SpaceMap
{
    SMExtentNode            *start;
    SMExtentNode            *end;
    bc_u64                  count;
    bc_u64                  needsNormalization; // Type 'bc_u64' is used for proper alignment.
}
SpaceMap;



const char *SpaceMapGetErrorString(sm_result result);

SpaceMap *SpaceMapCreate(void);

SpaceMap *SpaceMapCreateCopy(SpaceMap *map);

sm_result SpaceMapToExtentArray(SpaceMap *map, Extent **out_array, bc_u64 *out_count);

sm_result SpaceMapFreeArray(Extent **array);

/**
  @brief Saves map to the buffer specified.
  @p map        SpaceMap to save.
  @p buffer     Buffer to write extents to.
  @p bufferSize Maximum available buffer size.
*/
sm_result SpaceMapToBytes(SpaceMap *map, void *buffer, bc_u64 bufferSize);


/**
  @brief Loads map from byte buffer.
  @p buffer     Buffer with extents.
  @p bufferSize Size of the buffer.
  @p out_map    Pointer to map to load data to it.
*/
sm_result SpaceMapFromBytes(void *buffer, bc_u64 bufferSize, SpaceMap **out_map);

void SpaceMapFree(SpaceMap *map);

sm_result SpaceMapGetLength(SpaceMap *map, bc_u64 *out_length);

sm_result SpaceMapGetExtentSpace(Extent *array, bc_u64 count, bc_u64 *out_length);

/**
  @brief Adds extent to the map.
  @p map        SpaceMap to add the extent.
  @p extent     Extent to add.
  @warning      Extent will be added as a pointer and will not be copied.
*/
sm_result SpaceMapAddNode(SpaceMap *map, SMExtentNode *item);

sm_result SpaceMapAddCopy(SpaceMap *map, const Extent *extent);

sm_result SpaceMapAddNew(SpaceMap *map, bc_u64 start, bc_u64 length);

sm_result SpaceMapRemove(SpaceMap *map, Extent *extent);

/**
  @brief Returns kSpaceMap_Success if specified map contains specified extent.
  @p map    Space map to check.
  @p extent Extent to use.
  @p exact  If true, pointers will be compared, otherwise values.
*/
sm_result SpaceMapContainsExtent(SpaceMap *map, const Extent *extent, _Bool exact);

sm_result SpaceMapExtentCount(SpaceMap *map, bc_u64 *out_count);

sm_result SpaceMapSort(SpaceMap *map);

/**
  @brief Removes zeroed extents and unions intercepted.
  @p map Map to process.
*/
sm_result SpaceMapNormalize(SpaceMap *map);

sm_result SpaceMapUnion(SpaceMap *result_map, SpaceMap *another_map);

sm_result SpaceMapSubstract(SpaceMap *from_map, SpaceMap *what_map);

sm_result SpaceMapGetPhysicalExtents(Extent *virtual_extents, bc_u64 virtual_count, bc_u64 io_offset,
                                     bc_u64 io_count, Extent **out_physical_extents, bc_u64 *out_extent_count);

void SpaceMapFreePhysicalExtents(Extent **physical_extents);

void SpaceMapPrint(SpaceMap *map);

#ifdef __cplusplus
}
#endif

#endif // SPACEMAP_H
