/*
 * Copyright 2012 Jetico, Inc. All rights reserved.
 * Help conversion functions implementation.
 */
#include "convertUTFSize.h"
#include <assert.h>

/**
 *	Calculate conversion size.
 *	@param utf8bf    utf8 buffer for converting
 *	@param size      buffer size 
 *	@param flags     strictConversion = 0 ( do not change invalid symbols ), lenientConversion = 1 ( change invalid symbols to default )
 *	@return          size in UTF16 symbols of required output buffer for converting
 */
unsigned int getSizeUTF8toUTF16( const UTF8* utf8bf, unsigned int size, ConversionFlags flags )
{
#define bfSize 64
	UTF16 bf[bfSize];

	unsigned int sizeInSymbols = 0;

	const UTF8* sourceStart = utf8bf;
	const UTF8* sourceEnd = utf8bf + size;

	for(;;)
	{
		unsigned int lSizeInSymbols;

		UTF16* targetStart = bf;
		UTF16* targetEnd = bf + bfSize;

		ConversionResult res = ConvertUTF8toUTF16( &sourceStart, sourceEnd, &targetStart, targetEnd, flags );
		
		lSizeInSymbols = (unsigned int)(targetStart - bf);
		
		sizeInSymbols += lSizeInSymbols;

		switch( res )
		{
			case conversionOK: 	return sizeInSymbols;
			case targetExhausted: continue;
			default:
				return 0;
		}
	}

	return 0;
}

unsigned int getSizeUTF16toUTF8( const UTF16* utf16bf, unsigned int size, ConversionFlags flags )
{
#define bfSize 64
	UTF8 bf[bfSize];

	unsigned int sizeInSymbols = 0;

	const UTF16* sourceStart = utf16bf;
	const UTF16* sourceEnd = utf16bf + size;

	for(;;)
	{
		unsigned int lSizeInSymbols;

		UTF8* targetStart = bf;
		UTF8* targetEnd = bf + bfSize;

		ConversionResult res = ConvertUTF16toUTF8( &sourceStart, sourceEnd, &targetStart, targetEnd, flags );

		lSizeInSymbols = (unsigned int)(targetStart - bf);

		sizeInSymbols += lSizeInSymbols;

		switch( res )
		{
			case conversionOK: 	return sizeInSymbols;
			case targetExhausted: continue;
			default:
				return 0;
		}
	}

	return 0;
}


/**
 *	Calculate conversion size.
 *	@param in_utf8   input UTF8 string for converting
 *	@param inlen     input string length 
 *	@param out_utf16 output utf8 string for converting
 *	@param outlen    output string length in UTF16 symbols
 *	@return          size in UTF16 symbols of required output buffer for converting
 */
unsigned int convertUTF8toUTF16( const UTF8* in_utf8, unsigned int inlen, UTF16 *out_utf16, unsigned int outlen )
{
	const UTF8* sourceStart = in_utf8;
	const UTF8* sourceEnd = in_utf8 + inlen;
	ConversionFlags flags = strictConversion;
	UTF16* targetStart = out_utf16;
	UTF16* targetEnd = (out_utf16 + outlen);
	ConversionResult res;

	if ( out_utf16 == 0 )
		return getSizeUTF8toUTF16( sourceStart, inlen, flags );

	res = ConvertUTF8toUTF16( &sourceStart, sourceEnd, &targetStart, targetEnd, flags );

	if ( conversionOK != res )
	{
		assert( 0 );
		return 0;
	}

	return (unsigned int)(targetStart - out_utf16);
}

unsigned int convertUTF16toUTF8( const UTF16* in_utf16, unsigned int inlen, UTF8 *out_utf8, unsigned int outlen )
{
	const UTF16* sourceStart = in_utf16;
	const UTF16* sourceEnd = in_utf16 + inlen;
	ConversionFlags flags = strictConversion;
	UTF8* targetStart = out_utf8;
	UTF8* targetEnd = (out_utf8 + outlen);
	ConversionResult res;

	if ( out_utf8 == 0 )
		return getSizeUTF16toUTF8( sourceStart, inlen, flags );

	res = ConvertUTF16toUTF8( &sourceStart, sourceEnd, &targetStart, targetEnd, flags );

	if ( conversionOK != res )
	{
		assert( 0 );
		return 0;
	}

	return (unsigned int)(targetStart - out_utf8);
}
