#include <QDebug>
#include <QTimer>
#include <QFileDialog>
#include <QScrollBar>
#include <QFontMetrics>
#include <QResizeEvent>
#include <QDesktopServices>
#include <QUrl>
#include <QDir>
#include <QMessageBox>

#include "containerinfowidget.h"
#include "ui_containerinfowidget.h"

#include "entity/containerdescriptor.h"
#include "operations/bcoperation.h"
#include "StringUtils.h"


static void SetClippedTextToLabel(QLabel *label, QString text)
{
#if 0
    QFontMetrics metrix(label->font());
    int width = label->width() - 2;
    QString clippedText = metrix.elidedText(text, Qt::ElideRight, width);
    label->setText(clippedText);
#else
    label->setText(text);
#endif

}

static void SetTextOrDefault(QLabel *label, const QString& text, const QString& defaultText)
{
    if (text.count())
    {
        SetClippedTextToLabel(label, text);
    }
    else
    {
        SetClippedTextToLabel(label, defaultText);
    }
}


ContainerInfoWidget::ContainerInfoWidget(ContainerDescriptor *d) :
    QWidget(0),
    ui(new Ui::ContainerInfoWidget),
    descriptor(d)
{
    ui->setupUi(this);
    ui->ErrorLabel->hide();
    ui->ErrorTextLabel->hide();

    ui->headerEncrypted_nameField->hide();
    ui->headerEncryptedLabel->hide();

    connect(descriptor, SIGNAL(changed()), this, SLOT(onUpdate()));
    connect(ui->descriptionEditButton, SIGNAL(clicked()), this, SIGNAL(edit_description()));
    connect(ui->cancelButton, SIGNAL(clicked()), this, SIGNAL(cancel()));

    onUpdate();

    ui->descriptionEditButton->hide(); // underconstruction
    
    job_finished_timer = new QTimer(this);

    onStopProgressbar();
}

ContainerInfoWidget::~ContainerInfoWidget()
{
    delete ui;
}


void ContainerInfoWidget::onUpdate()
{
    // Fill controls with descriptor parameters.
    SetTextOrDefault(this->ui->containerLocationLabel, descriptor->GetPath(), "-");

    SetTextOrDefault(this->ui->containerNameLabel, descriptor->GetFileName(), "-");

    SetTextOrDefault(this->ui->containerSizeLabel, QString::fromStdString(StringUtils::humanSizeString(descriptor->size)), "-");
    SetTextOrDefault(this->ui->containerDescriptionLabel, descriptor->description, "No description");
    SetTextOrDefault(this->ui->containerMountPointLabel, descriptor->GetMountPoint(), "-");
    SetTextOrDefault(this->ui->deviceLabel, descriptor->GetDeviceName(), "-");

    if (descriptor->IsLinked())
        SetTextOrDefault(this->ui->containerMountPointLabel, descriptor->GetMountPoint(), "Linked only");

    if (! descriptor->IsExists())
        this->ui->errorLabel->setText("File not found.");
    else if (! descriptor->IsParsed())
    {
        this->ui->errorLabel->setText("Encrypted or corrupted header.");
        this->ui->containerAlgorithmLabel->setText("-");
    }
    else
    {
        ui->errorLabel->setText("");
        this->ui->containerAlgorithmLabel->setText(QString("%0-%1").arg(descriptor->algName).arg(descriptor->modeName));
    }

    if (! descriptor->error.isEmpty()) {
        this->ui->errorLabel->setText(descriptor->error);
    }

    // this->ui->mountLocationText->setText(descriptor->lastMountPoint);
}


void ContainerInfoWidget::onShowError(QString error)
{
    ui->ErrorTextLabel->setText(error);
    ui->ErrorLabel->show();
    ui->ErrorTextLabel->show();
}


void ContainerInfoWidget::onProgress(QString text, int n, int m)
{
    ui->progressBar->show();
    ui->cancelButton->show();
    ui->progressLabel->show();

    if (! text.isEmpty())
        ui->progressLabel->setText(text);

    ui->progressBar->setValue(n);
    ui->progressBar->setMaximum(m);
}


void ContainerInfoWidget::onBGJobFinished(bool success)
{
    job_finished_timer->singleShot(2000, this, SLOT(onStopProgressbar()));
}


void ContainerInfoWidget::onOpenLocationSlot()
{
    QString dir = this->ui->containerMountPointLabel->text();

    if (QDir(dir).exists())
    {
        QDesktopServices::openUrl(QUrl(QString("file:///") + dir));
    }
    else
    {
        QMessageBox::information(this, "Incorrect Path", "Location '" + dir
                                 + "' was not found.", QMessageBox::Ok);
    }
}


void ContainerInfoWidget::onStopProgressbar()
{
    ui->progressBar->setValue(-1);
    ui->progressBar->setMaximum(-1);
    ui->progressBar->setEnabled(false);
    ui->cancelButton->hide();
    ui->progressLabel->hide();
}
