#include "containertreewidget.h"
#include "ui_containertreewidget.h"

#include "model/containerlistmodel.h"
#include "operations/bcoperation.h"
#include "entity/containerdescriptor.h"
#include "containerlistitem.h"

#include <QLayoutItem>
#include <QKeyEvent>
#include <QMutex>

ContainerTreeWidget::ContainerTreeWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::ContainerTreeWidget)
{
    ui->setupUi(this);
    this->containerModel_ = 0;
    this->selectedView_ = 0;
    this->containerViews_ = new QList<ContainerListItem*>();

    this->deleteFileAction_ = new QAction(tr("Delete container file"), this->ui->removeToolButton);
    this->clearListAction_ = new QAction(tr("Clear list"), this->ui->removeToolButton);

    this->ui->removeToolButton->addAction(this->clearListAction_);
    this->ui->removeToolButton->addAction(this->deleteFileAction_);

    connect(this->clearListAction_, SIGNAL(triggered()), this, SLOT(ClearButtonClickSlot()));
    connect(this->deleteFileAction_, SIGNAL(triggered()), this, SIGNAL(OnDelete()));
}

ContainerTreeWidget::~ContainerTreeWidget()
{
    delete ui;

    qDeleteAll(*(this->containerViews_));
    delete this->containerViews_;
}

void ContainerTreeWidget::InitModel(ContainerListModel *model)
{
    Q_ASSERT(!this->containerModel_);
    this->containerModel_ = model;
    connect(model, SIGNAL(changed()), this, SLOT(DataChangedSlot()), Qt::DirectConnection);
    this->UpdateViewList();
}

bool ContainerTreeWidget::SelectDescriptor(ContainerDescriptor *descriptor)
{
    ContainerListItem *cur;
    for (int i = 0; i < this->containerViews_->count(); i++)
    {
        cur = this->containerViews_->at(i);
        if (cur->GetDescriptor() == descriptor)
        {
            this->SelectView(cur);
            return true;
        }
    }
    return false;
}


void ContainerTreeWidget::UpdateViewList()
{
    QString selectedContainerPath;
    if (this->selectedView_)
        selectedContainerPath = this->GetSelectedDescriptor()->GetPath();
    this->selectedView_ = 0;

    if (this->containerViews_->count())
    {
        ContainerListItem *cur = 0;
        QLayoutItem *item = 0;

        for (int i=0; i < this->containerViews_->count(); i++)
        {
            cur = this->containerViews_->at(i);
            disconnect(cur, SIGNAL(OnSelected(ContainerListItem*)), this, SLOT(SelectedViewChangedSlot(ContainerListItem*)));
            disconnect(cur, SIGNAL(OnEject(ContainerListItem*)), this, SLOT(SelectedViewEjectClickedSlot(ContainerListItem*)));
            disconnect(cur, SIGNAL(OnMount(ContainerListItem*)), this, SLOT(SelectedViewMountClickedSlot(ContainerListItem*)));
            disconnect(cur, SIGNAL(OnSelectNext(ContainerListItem*)), this, SLOT(SelectedViewSelectNextSlot(ContainerListItem*)));
            disconnect(cur, SIGNAL(OnSelectPrevious(ContainerListItem*)), this, SLOT(SelectedViewSelectPreviousSlot(ContainerListItem*)));
        }

        while (this->ui->itemLayout->count() > 1)
        {
            item = this->ui->itemLayout->takeAt(0);
            delete item->widget();
            delete item;
        }

        this->containerViews_->clear();
    }

    ContainerListModel::list_t descriptors = this->containerModel_->GetDescriptors();
    int indexToSelect = 0;

    for (int i = 0; i < descriptors.count(); i++)
    {
        ContainerDescriptor *descriptor = descriptors.at(i);
        ContainerListItem *view = new ContainerListItem(this->ui->scrollArea, descriptor);

        if (descriptor->GetPath() == selectedContainerPath)
            indexToSelect = i;

        connect(view, SIGNAL(OnSelected(ContainerListItem*)), this, SLOT(SelectedViewChangedSlot(ContainerListItem*)));
        connect(view, SIGNAL(OnEject(ContainerListItem*)), this, SLOT(SelectedViewEjectClickedSlot(ContainerListItem*)));
        connect(view, SIGNAL(OnMount(ContainerListItem*)), this, SLOT(SelectedViewMountClickedSlot(ContainerListItem*)));
        connect(view, SIGNAL(OnSelectNext(ContainerListItem*)), this, SLOT(SelectedViewSelectNextSlot(ContainerListItem*)));
        connect(view, SIGNAL(OnSelectPrevious(ContainerListItem*)), this, SLOT(SelectedViewSelectPreviousSlot(ContainerListItem*)));

        this->containerViews_->append(view);
        this->ui->itemLayout->insertWidget(i, view);
    }

    this->SelectContainerByIndex(indexToSelect);
}

void ContainerTreeWidget::SelectContainerByIndex(int index)
{
    if (index >= 0 && index < this->containerViews_->count())
    {
        this->SelectView(this->containerViews_->at(index));
    }
    else
    {
        this->SelectView(0);
    }
}



void ContainerTreeWidget::SelectedViewChangedSlot(ContainerListItem *view)
{
    this->SelectView(view);
}

void ContainerTreeWidget::SelectedViewEjectClickedSlot(ContainerListItem *view)
{
    this->SelectView(view);
    emit this->OnEject();
}

void ContainerTreeWidget::SelectedViewMountClickedSlot(ContainerListItem */*view*/)
{
    // Mount emits on double-click, so view should be selected by first click.
    emit this->OnMount();
}

void ContainerTreeWidget::SelectedViewSelectNextSlot(ContainerListItem *view)
{
    int i;
    for (i = 0; i < this->containerViews_->count() - 1; i++)
    {
        if (this->containerViews_->at(i) == view)
        {
            this->SelectView(this->containerViews_->at(i + 1));
            break;
        }
    }
}

void ContainerTreeWidget::SelectedViewSelectPreviousSlot(ContainerListItem *view)
{
    int i;
    for (i = 1; i < this->containerViews_->count(); i++)
    {
        if (this->containerViews_->at(i) == view)
        {
            this->SelectView(this->containerViews_->at(i -1));
            break;
        }
    }
}

void ContainerTreeWidget::DataChangedSlot()
{
    this->UpdateViewList();
}

void ContainerTreeWidget::ClearButtonClickSlot()
{
    this->containerModel_->Clear();
}

void ContainerTreeWidget::on_removeToolButton_clicked()
{
    ContainerDescriptor *descriptor = this->GetSelectedDescriptor();
    if (descriptor)
    {
        this->containerModel_->Remove(descriptor);
    }
}

void ContainerTreeWidget::SelectView(ContainerListItem *view)
{
    for (int i = 0; i < this->containerViews_->count(); i++)
    {
        this->containerViews_->at(i)->Deselect();
    }

    if (view)
    {
        view->Select();
    }

    this->selectedView_ = view;

    emit this->OnSelectedDescriptorChanged(this->GetSelectedDescriptor());
    //this->mutex_->unlock();
}

void ContainerTreeWidget::keyPressEvent(QKeyEvent *event)
{
    if (!this->selectedView_)
        return;

    ContainerDescriptor *descriptor = this->selectedView_->GetDescriptor();

    if (!descriptor)
        return;

    switch(event->key())
    {
        case Qt::Key_Return:
        case Qt::Key_Enter:
            if (descriptor->IsExists())
            {
                event->accept();
                if (descriptor->IsLinked())
                    emit this->OnEject();
                else
                    emit this->OnMount();
            }
            break;

        case Qt::Key_Up:
            event->accept();
            this->SelectedViewSelectPreviousSlot(this->selectedView_);
            break;

        case Qt::Key_Down:
            event->accept();
            this->SelectedViewSelectNextSlot(this->selectedView_);
            break;

        default:
            break;
    }
}

