#include <QDir>
#include <QFile>
#include <QDebug>

#include "ContainerCreateDialog.h"
#include "ui_ContainerCreateDialog.h"

#include "certificate/CertificateLib/hashAlgs/hashAlgs.h"
#include "keygens/kgghost/kgghost.h"
#include "operations/QtPwdGet.h"
#include "operations/bcoperation.h"
#include "selectcipherwidget.h"
#include "filesystemwidget.h"
#include "dialogs/keyfilesdialog.h"
#include "entity/containerdescriptor.h"
#include "filesystemwidget.h"
#include "container/helpers.h"
#include "FilesystemUtils.h"
#include "StringUtils.h"
#include "model/qt_helpers.h"


ContainerCreateDialog::ContainerCreateDialog(QWidget *parent) :
    QDialog(parent)
    , ui(new Ui::ContainerCreateDialog)
    , pwdGet(0)
    , password_is_ok(false)
    , mount_point_is_specified_by_user(false)
{
    ui->setupUi(this);

    // brief tab
    setDefaultFilename();
    setDefaultSize();
    onUpdateFreeSpaceInfo();
    onContainerSizeChanged();

    // container body preferences tab
    FileSystemWidget::ReloadFilesystems(ui->fsCombo, true /* include 'no fs' option */);
    algList = new QList<BCAlgInfo>();
    SelectCipherWidget::FillAlgList(algList, ui->algCombo);

    // container header preferences tab
    fillHashAlgs();

    onShowBriefTab();

    validatePassword();

    connect(ui->sizeUnitsCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(onContainerSizeChanged()));
    connect(ui->sizeSpinBox, SIGNAL(editingFinished()), this, SLOT(onContainerSizeChanged()));
}

ContainerCreateDialog::~ContainerCreateDialog()
{
    delete ui;
}

void ContainerCreateDialog::setDefaultFilename()
{
    QDir d(QDir::homePath());
    QString filename;

    int i = 1;

    do {
        filename = QString("Container%1.jbc").arg(i++);
    } while (d.exists(filename));

    ui->filepathEdit->setText(d.filePath(filename));
}

void ContainerCreateDialog::setDefaultSize()
{
    ui->sizeUnitsCombo->addItem("MB");
    ui->sizeUnitsCombo->addItem("GB");
    ui->sizeUnitsCombo->addItem("TB");
    ui->sizeUnitsCombo->setCurrentIndex(0);

    ui->sizeSpinBox->setValue(30);
}

void ContainerCreateDialog::onUpdateFreeSpaceInfo()
{
    freespace_bytes = FilesystemUtils::getFreeSpace(ui->filepathEdit->text().toStdString());

    qDebug() << "Free disk space: " << freespace_bytes << " bytes";

    QString s = QString("%1 Free").arg(QString::fromStdString(StringUtils::humanSizeString(freespace_bytes)));

    ui->freeSpaceLabel->setText(s);
}


//! set message under both password fields
void ContainerCreateDialog::setUnderpasswordMessage(QString text, bool create_btn_enabled)
{
    ui->creationTimeLabel->setText(text);
    ui->mountTimeLabel->setText(text);
    ui->createButton_11->setEnabled(create_btn_enabled);
    ui->createButton_22->setEnabled(create_btn_enabled);
    ui->createButton_33->setEnabled(create_btn_enabled);
}

void ContainerCreateDialog::validatePassword()
{
    password_is_ok = false;

    if (ui->passEdit->text().isEmpty()) {
        setUnderpasswordMessage("", false);
        return;
    }

    if (ui->passEdit->text().size() < 8) {
        setUnderpasswordMessage("<font color='red'>Password is too short</font>", false);
        return;
    }

    if (ui->passEdit->text().size() > 255) {
        setUnderpasswordMessage("<font color='red'>Password is too long</font>", false);
        return;
    }

    if (ui->passEdit->text() != ui->confirmEdit->text()) {
        setUnderpasswordMessage("<font color='red'>Passwords do not match</font>", false);
        return;
    }

    password_is_ok = true;
    setUnderpasswordMessage("", true);

    onContainerSizeChanged();
    onUpdateMountBenchmark(0);
}


void ContainerCreateDialog::onUpdateMountBenchmark(int /*unused*/)
{
    if (!password_is_ok)
        return;

    int iterations = ui->iterationSpin->value();

    // TODO: port mount benchmard from Mac version
    // m_mountIterationsTime = BCContainer::performIterationsBenchmark(hash.id);
    // m_mountKeyFilesTime = BCContainer::performKeyFilesBenchmark(hash.id, channels);
    // double mountTime = ((double)iterations) / 10000.0 * m_mountIterationsTime + m_mountKeyFilesTime;

    double mountTime = ((double)iterations) / 10000.0;
    QString time_str = QString::fromStdString(StringUtils::humanTimeInterval(mountTime));

    ui->mountTimeLabel->setText(QString("Estimated mounting time: %1").arg(time_str));
}


void ContainerCreateDialog::onContainerSizeChanged()
{
    // Get container size.
    QString sizeStr;
    QString sizeUnit = this->ui->sizeUnitsCombo->currentText().at(0); // Get first letter (M/G/T).
    sizeStr = sizeStr.append("%0%1").arg(this->ui->sizeSpinBox->value()).arg(sizeUnit);
    container_size = StringUtils::sizeStringToBytes(sizeStr.toStdString().c_str());
    qDebug() << "ContainerCreateDialog: size is" << container_size << "bytes";

    bool container_size_was_corrected = false;

    if (container_size < BCTool::minimalContainerSize) {
        BCOperationShowWarningString("Container size is too small. The value was increased");
        container_size = BCTool::minimalContainerSize;
        container_size_was_corrected = true;
    }

    if (freespace_bytes > 0 && container_size > freespace_bytes) {
        BCOperationShowWarningString("Container size is too large. The value was decreased to maximum available.");
        container_size = (freespace_bytes - _1MB) & (~0x1ffff); // reset lower 5 bits- round down to megabytes.
        container_size_was_corrected = true;
    }

    // set this value back to the dialog
    if (container_size_was_corrected) {
        std::string ss = StringUtils::humanSizeString(container_size);
        float mb;
        char units[30];
        if (sscanf(ss.c_str(), "%f %s", &mb, units) == 2) {
            ui->sizeSpinBox->setValue(int(mb));
            int i = ui->sizeUnitsCombo->findText(units);
            ui->sizeUnitsCombo->setCurrentIndex(i);
        }
    }

    // TODO: use real disk benchmark data here
    // for now just pick  40 Mb/sec speed

    double seconds = container_size / 40000000;

    QString time_str = QString::fromStdString(StringUtils::humanTimeInterval(seconds));

    ui->creationTimeLabel->setText(QString("Estimated creation time: %1").arg(time_str));
}


void ContainerCreateDialog::on_algCombo_currentIndexChanged(int index)
{
    int alg_id = ui->algCombo->itemData(index).toInt();
    SelectCipherWidget::updateModeCombo(algList, ui->modeCombo, alg_id);
}

void ContainerCreateDialog::on_mountPointBrowse_clicked()
{
    QString p = BCHelperGetLocation("Please specify mount point", QDir::homePath());
    if (! p.isEmpty()) {
        mount_point_is_specified_by_user = true;
        ui->mountLocationEdit->setText(p);
    }
}

void ContainerCreateDialog::onFilesystemOptionsChanged(int /*unused*/)
{
    // The trick here is that comboBox::changed(int) and checkbox::changed(int) are
    // connected to the same (this) slot. The value is ignored anyway

    bool format = (ui->fsCombo->currentIndex() != 0);
    ui->fsLabelEdit->setEnabled(format);
    ui->mountCheckBox->setEnabled(format);

    bool mount = format & ui->mountCheckBox->isChecked();
    ui->mountAtSpecifiedCheckbox->setEnabled(mount);

    bool location = mount & ui->mountAtSpecifiedCheckbox->isChecked();
    ui->mountLocationEdit->setEnabled(location);
    ui->mountPointBrowse->setEnabled(location);
}

void ContainerCreateDialog::onShowBriefTab()
{
    ui->stackedWidget->setCurrentWidget(ui->briefTab);
}

void ContainerCreateDialog::onShowBodyTab()
{
    ui->stackedWidget->setCurrentWidget(ui->bodyTab);
}

void ContainerCreateDialog::onShowHeaderTab()
{
    ui->stackedWidget->setCurrentWidget(ui->headerTab);
}

void ContainerCreateDialog::onPassEdited(const QString &val)
{
    ui->passEdit->setText(val);
    ui->passEdit_3->setText(val);
    validatePassword();
}

void ContainerCreateDialog::onConfirmEdited(const QString &val)
{
    ui->confirmEdit->setText(val);
    ui->confirmEdit_3->setText(val);
    validatePassword();
}

void ContainerCreateDialog::on_keyFilesButton_clicked()
{
    keyfilesdialog dlg(this, keyfiles);

    if (dlg.exec() != QDialog::Accepted)
        return;

    keyfiles = dlg.getKeyFiles();
    ui->keyFilesButton->setText(dlg.getButtonCaption(keyfiles.size()));
}

void ContainerCreateDialog::fillHashAlgs()
{
    int algCount = 0;
    int *algTab;

    ha_getIDsOfAvailableAlgAlgorithms( &algCount, &algTab );

    for (int i = 3; i < algCount; ++i) {
        const char * algName = ha_getHashAlgName(algTab[i]);
        ui->hashAlgBox->addItem(algName, QVariant::fromValue(algTab[i]));
    }

    // no need to free 'algTab' - statically allocated

    // select 'Whirlpool' or 'SHA-512' by  default
    int def = ui->hashAlgBox->findText("Whirlpool-512");
    if (def == -1) {
        def = ui->hashAlgBox->findText("SHA-512");
    }
    if (def >= 0) {
        ui->hashAlgBox->setCurrentIndex(def);
    }
}

void ContainerCreateDialog::onGenerateSalt()
{
    int digestSize = 0;
    ha_getDigestSize(&digestSize, hash_alg_id);

    salt.resize(digestSize);
    for(int i = 0 ; i < salt.size() ; i++)
        salt[i] = qrand();

    ui->saltEdit->setText(salt.toHex());
    ui->saltLabel->setText(QString("Salt (%1 bytes):").arg(digestSize));
}

void ContainerCreateDialog::on_hashAlgBox_currentIndexChanged(int index)
{
    hash_alg_id = ui->hashAlgBox->itemData(index).toInt();
    onGenerateSalt();
}

QString ContainerCreateDialog::getFilename() const
{
    return ui->filepathEdit->text();
}

bc_alg_id ContainerCreateDialog::getEncryptionAlgId() const
{
    int ai = ui->algCombo->currentIndex();
    return ui->algCombo->itemData(ai).toInt();
}

QString ContainerCreateDialog::getFilesystemType() const
{
    return FileSystemWidget::GetFileSystem(ui->fsCombo);
}

QString ContainerCreateDialog::getVolumeLabel() const
{
    return ui->fsLabelEdit->text();
}


ContainerDescriptor *ContainerCreateDialog::getData()
{
    QString algName = ui->algCombo->currentText();
    bc_alg_id algId = getEncryptionAlgId();

    int mi = ui->modeCombo->currentIndex();
    QString modeName = ui->modeCombo->itemText(mi);
    bc_mode_id modeId = static_cast<bc_mode_id>(ui->modeCombo->itemData(mi).toInt());

    QString filePath = ui->filepathEdit->text();
    QString description = ui->descriptionEdit->text();

    QString fileSystem = getFilesystemType();

    return ContainerDescriptor::CreateForNewContainer(filePath, container_size, algId, algName,
                                                      modeId, modeName, fileSystem, description);
}


void ContainerCreateDialog::fillContainerContext(BCContainerContext &cc)
{
    cc.algID = getEncryptionAlgId();

    int mi = ui->modeCombo->currentIndex();
    cc.modeID = static_cast<bc_mode_id>(ui->modeCombo->itemData(mi).toInt());

    cc.size = container_size;

    QByteArray dd = ui->descriptionEdit->text().toUtf8();
    strncpy(cc.description, dd.data(), sizeof(cc.description));

    Q_ASSERT(salt.size() <= sizeof(cc.salt));
    memcpy(cc.salt, salt.data(), salt.size());
    cc.saltSize = salt.size();

    cc.kgID = kgghost::kKGID_V5;

    cc.iterations = ui->iterationSpin->value();

    int hashidx = ui->hashAlgBox->currentIndex();
    cc.hashAlgID = ui->hashAlgBox->itemData(hashidx).toInt();

    QByteArray path = ui->filepathEdit->text().toUtf8();
    strncpy(cc.path, path.data(), sizeof(cc.path));

    cc.encryptHeader = ui->encryptHeaderCheckbox->isChecked();

    cc.askMore = false;
}


void ContainerCreateDialog::dumpParamsTo(BCToolCreateParams *createParams, BCContainerContext &cc, BCToolMountParams *mountParams)
{
    createParams->fsType = getFilesystemType().toStdString();
    createParams->volumeLabel = getVolumeLabel().toStdString();

    if (ui->mountCheckBox->isChecked()) {
        createParams->mount_after_creation = true;
        if (ui->mountAtSpecifiedCheckbox->isChecked())
            mountParams->mountPoint =  ui->mountLocationEdit->text().toStdString();
    }

    fillContainerContext(cc);
}


void ContainerCreateDialog::setQtPwdGet(QtPwdGet *obj)
{
    pwdGet = obj;
}


void ContainerCreateDialog::onNeedPassword(bool prev_failed)
{
    assert(pwdGet);
    // OK was clicked in creation dialog already
    if (! pwdGet->setPasswordBundle(ui->passEdit->text(), keyfiles)) {
        setUnderpasswordMessage(pwdGet->getLastError(), false);
    }
}

void ContainerCreateDialog::on_encryptHeaderCheckbox_stateChanged(int arg1)
{
    if (arg1) {
        ui->iterationSpin->setValue(16384);
        ui->iterationSpin->setEnabled(false);
    }
    else {
        ui->iterationSpin->setEnabled(true);
    }
}

void ContainerCreateDialog::on_filepathBrowseButton_clicked()
{
    QString s = BCHelperSaveFile(ui->filepathEdit->text(), tr("Enter Container File Name"), BC_FILETYPES_AVAILABLE);
    if (! s.isEmpty()) {
        ui->filepathEdit->setText(s);
    }
}

void ContainerCreateDialog::on_filepathEdit_textChanged(const QString &arg1)
{
    if (! mount_point_is_specified_by_user)
        ui->mountLocationEdit->setText(getDefaultMountPointFor(arg1));
}
