#include <QFileDialog>
#include <QTimer>

#include "ContainerMountDialog.h"
#include "ui_ContainerMountDialog.h"

#include "keyfilesdialog.h"
#include "model/qt_helpers.h"
#include "bctool.h"
#include "operations/QtPwdGet.h"

ContainerMountDialog::ContainerMountDialog(QWidget *parent, BCToolMountParams *mp) :
    QDialog(parent),
    ui(new Ui::ContainerMountDialog),
    mountParams(mp)
{
    ui->setupUi(this);
    this->ui->errorLabel->setText("");
    this->ui->okButton->setDefault(true);

    QObject::connect(this->ui->passEdit, SIGNAL(textChanged(QString)), this, SLOT(onPassTextChangedSlot()));
    QObject::connect(this->ui->advancedSettingsToggle, SIGNAL(linkActivated(QString)), this, SLOT(onAdvancedSettingsToggle()));
    this->ui->keyblock_layout->setEnabled(false);

    setParamsFrom(mountParams);

    onAdvancedSettingsToggle();

    dlg_close_timer = new QTimer(this);
    // connect(dlg_close_timer, SIGNAL(timeout()), this, SLOT(close()));
}

ContainerMountDialog::~ContainerMountDialog()
{
    delete ui;
}

void ContainerMountDialog::showError(const QString &error)
{
    this->ui->errorLabel->setText(error);
    BCHelperFixSize(this);
    this->ui->passEdit->selectAll();
    QTimer::singleShot(0, this->ui->passEdit, SLOT(setFocus()));
}


void ContainerMountDialog::on_mountPointText_textChanged(const QString &arg1)
{
    QDir d(arg1);
    // Directory for mounting should be empty
    if (d.exists() && d.entryInfoList(QDir::NoDotAndDotDot|QDir::AllEntries).count() > 0) {
        showError("Mount target directory is not empty");
        ui->okButton->setEnabled(false);
        return;
    }
    ui->errorLabel->setText("");
    ui->okButton->setEnabled(true);
}


void ContainerMountDialog::setTitle(const char *query, bool prev_failed)
{
    if (query)
        QDialog::setWindowTitle(QString(query));

    if (prev_failed)
        showError("Incorrect password");

    dlg_close_timer->stop();
}

void ContainerMountDialog::onPassTextChangedSlot()
{
    this->ui->errorLabel->setText("");
}

void ContainerMountDialog::setParamsFrom(const BCToolMountParams *mountParams)
{
    // copy params from mountParams struct to Qt widgets in form
    ui->containerFilenameText->setText(QString::fromStdString(mountParams->filename));
    ui->mountPointText->setText(QString::fromStdString(mountParams->mountPoint));
    ui->mountReadOnlyCheckbox->setChecked(mountParams->mountReadOnly);
    BCHelperFixSize(this);
}

void ContainerMountDialog::dumpParamsTo(BCToolMountParams *mountParams)
{
    // Copy mount option values from dialog to struct.
    mountParams->mountPoint = ui->mountPointText->text().toStdString();
    mountParams->mountReadOnly = this->ui->mountReadOnlyCheckbox->checkState() == Qt::Checked;
    mountParams->useDefaultMountPoint = false;
}

void ContainerMountDialog::on_mountLocationButton_clicked()
{
    QString p = BCHelperGetLocation("Please specify mount point", QDir::homePath());
    if (! p.isEmpty())
        ui->mountPointText->setText(p);
}

void ContainerMountDialog::onAdvancedSettingsToggle()
{
    if (ui->advancedOptions->isHidden()) {
        ui->advancedOptions->setVisible(true);
        ui->advancedSettingsToggle->setText("<a href=#here>Hide Advanced Settings</a>");
        ui->keyFilesButton->setVisible(true);
    }
    else {
        ui->advancedOptions->setVisible(false);
        ui->advancedSettingsToggle->setText("<a href=#here>Show Advanced Settings</a>");
        ui->keyFilesButton->setVisible(false);
    }
    BCHelperFixSize(this);
}


void ContainerMountDialog::on_keyFilesButton_clicked()
{
    keyfilesdialog dlg(this, keyfiles);

    if (dlg.exec() != QDialog::Accepted)
        return;

    keyfiles = dlg.getKeyFiles();
    ui->keyFilesButton->setText(dlg.getButtonCaption(keyfiles.size()));
}

void ContainerMountDialog::savePasswordBundleTo(QtPwdGet *pwdGet) const
{
    if (! pwdGet->setPasswordBundle(ui->passEdit->text(), keyfiles, ui->keyblockPathText->text())) {
        ui->errorLabel->setText(pwdGet->getLastError());
    }
}


void ContainerMountDialog::on_keyblockBrowseButton_clicked()
{
    QStringList r = BCHelperLoadFiles("Please specify keyblock backup file", "*.*", false);
    if (r.size() > 0)
        ui->keyblockPathText->setText(r[0]);
}

void ContainerMountDialog::on_okButton_clicked()
{
    dumpParamsTo(mountParams);

    QDir d(QString::fromStdString(mountParams->mountPoint));
    if (! d.exists())
        d.mkpath(".");

    accepted();

    dlg_close_timer->singleShot(1000, this, SLOT(hide()));
}
