#include <cerrno>
#include <QDebug>

#include "containerdescriptor.h"
#include "containerinfowidget.h"
#include "operations/BackgroundJob.h"
#include "operations/bcoperation.h"

ContainerDescriptor::ContainerDescriptor(QObject *parent, const QString& path) :
    QObject(parent),
    fileinfo(path),
    container_info_widget(0), bg_job(0)
{
    this->state_ = cBCContainerState_Creation;
    this->algId = -1;
    this->modeId = -1;
    this->size = 0;
    this->isParsed = false;
    this->mountReadOnly = false;
}

ContainerDescriptor::~ContainerDescriptor()
{
}


ContainerDescriptor * ContainerDescriptor::CreateForNewContainer(QString filePath,
    bc_u64 size, int algId, QString algName, int modeId, QString modeName,
    QString fileSystem, QString description)
{
    ContainerDescriptor *result = new ContainerDescriptor(0, filePath);
    result->size = size;
    result->algId = algId;
    result->algName = QString(algName);
    result->modeId = modeId;
    result->modeName = QString(modeName);
    result->fileSystem = fileSystem;
    result->description = description;
    result->state_ = kBCContainerState_Unlinked;

    return result;
}


bool ContainerDescriptor::IsExists() const
{
    fileinfo.refresh();
    bool exists = fileinfo.exists() && fileinfo.isFile();
    return exists;
}


void ContainerDescriptor::runJob(BackgroundJob *job)
{
    Q_ASSERT(job);
    Q_ASSERT(bg_job == 0);

    bg_job = job;

    connect(bg_job, SIGNAL(finished(bool)), container_info_widget, SLOT(onBGJobFinished(bool)));
    connect(bg_job, SIGNAL(finished(bool)), this, SLOT(onBackgroundJobFinished(bool)));

    // progress and cancel signals
    connect(bg_job, SIGNAL(progress(QString,int,int)), container_info_widget, SLOT(onProgress(QString,int,int)));

    connect(this->container_info_widget, SIGNAL(cancel()), bg_job, SLOT(onCancelation()));

    bg_job->start();
}


void ContainerDescriptor::onBackgroundJobFinished(bool success)
{
    qDebug() << "ContainerDescriptor: background job on " << GetFileName() << " finished with " << success;

    delete bg_job;
    bg_job = 0;
}

void ContainerDescriptor::SetParsed(bool isParsed)
{
    this->isParsed = isParsed;
    emit changed();
}

void ContainerDescriptor::Mount(const QString& device, const QString& mountPoint)
{
    this->device_ = device;
    this->mountPoint_ = QString(mountPoint);
    this->state_ = kBCContainerState_Mounted;
    emit changed();
}

// TODO: delete (not used)
void ContainerDescriptor::Unmount()
{
    this->mountPoint_.clear();
    this->state_ = kBCContainerState_Linked;
    emit changed();
}

void ContainerDescriptor::Link(const QString& device)
{
    this->device_ = device;
    this->mountPoint_.clear();
    this->state_ = kBCContainerState_Linked;
    emit changed();
}

void ContainerDescriptor::Unlink()
{
    this->mountPoint_.clear();
    this->device_.clear();
    this->state_ = kBCContainerState_Unlinked;
    emit changed();
}

void ContainerDescriptor::update()
{
    qDebug() << "ContainerDescriptor::update(): device:" << GetDeviceName() << ", filename:" << GetFileName();

    if (GetFileName().isEmpty()) {
        qDebug() << "ContainerDescriptor: no container filename. Probably deleted";
    }

    BCOperationUpdateDescriptor();

    UpdateMountState();
}


void ContainerDescriptor::getInfoFrom(BCVolumeInfo &containerInfo)
{
    ContainerDescriptor *descriptor = this;
    Q_ASSERT(descriptor);

    BCAlgInfo algInfo;
    bc_cipher_result kernelResult;

    kernelResult = BCKernelLookupAlgById(containerInfo.alg, &algInfo);

    if (kBCCipher_Success != kernelResult) {
        descriptor->error = "Unable to get algorithm by its nummeric ID";
        return;
    }

    descriptor->algId = algInfo.id;
    descriptor->size = containerInfo.size;
    descriptor->algName = QString(algInfo.name);
    descriptor->modeId = containerInfo.mode;
    descriptor->modeName = QString(BCKernelModeName(containerInfo.mode));
    descriptor->description = QString::fromUtf8(containerInfo.description);
    // descriptor->encrypted = containerInfo.encrypted;   need helperContext->header->isEncrypted() to get this info
    descriptor->SetParsed(true);

    qDebug() << Q_FUNC_INFO << "emit changed()";
    emit changed();
}


void ContainerDescriptor::BCOperationUpdateDescriptor()
{
    ContainerDescriptor *descriptor = this;
    Q_ASSERT(descriptor);

    BCVolumeInfo containerInfo;

    std::string path(descriptor->GetPath().toStdString());
    BCFileChannel fch(path.c_str());

    if (descriptor->IsExists() && BCParseContainerFile(&fch, &containerInfo))
    {
        getInfoFrom(containerInfo);
    }
    else
    {
        descriptor->SetParsed(false);
    }
}

void ContainerDescriptor::UpdateMountState()
{
    ContainerDescriptor *descriptor = this;

    if (!descriptor->IsExists())
    {
        descriptor->Unlink();
        return;
    }

    std::vector<bc_mount_entry> mountEntries;
    bool succeeded = BCTool::GetMountInfo(&mountEntries);

    if (!succeeded)
    {
        // Error had already been showed to user inside previous function call.
        return;
    }

    std::vector<bc_mount_entry>::iterator i = mountEntries.begin();
    std::vector<bc_mount_entry>::iterator end = mountEntries.end();
    std::string descriptorPath = descriptor->GetPath().toStdString();
    char descriptorAbsPath[PATH_MAX], tmpAbsPath[PATH_MAX];

    if (!realpath(descriptorPath.c_str(), descriptorAbsPath))
    {
        int error = errno;
        BCOperationShowErrorString(QString("Error while resolving path %0: %1").arg(descriptorPath.c_str()).arg(strerror(error)));
        return;
    }

    for (; i != end; i++)
    {
        bc_mount_entry entry = *i;

        if (!entry.isFileExists)
            continue;

        if (strlen(entry.containerPath) == 0)
            continue;

        if (!realpath(entry.containerPath, tmpAbsPath))
        {
            int error = errno;
            BCOperationShowErrorString(QString("Error while resolving path '%0': '%0'").arg(entry.containerPath).arg(strerror(error)));
            return;
        }
        if (0 == strcmp(descriptorAbsPath, tmpAbsPath))
        {
            // Found container linked to device.
            if (strlen(entry.mountPoint))
            {
                // This container is also mounted, not only linked.
                descriptor->Mount(entry.deviceName, entry.mountPoint);
                return;
            }
            else
            {
                // This container is linked, but not mounted.
                descriptor->Link(entry.deviceName);
                return;
            }
        }
    }

    // All mount entries had been checked, but no such file found.
    descriptor->Unlink();
}
