#ifndef CONTAINERDESCRIPTOR_H
#define CONTAINERDESCRIPTOR_H

#include <QString>
#include <QFileInfo>

#include "container/intsize.h"

class ContainerInfoWidget;
class BackgroundJob;
class BCVolumeInfo;


/**
 * @brief Represents container parameters that will be used in user interface.
 */
class ContainerDescriptor : public QObject
{
    Q_OBJECT

public:
    ContainerDescriptor(QObject *parent, const QString &path);
    ~ContainerDescriptor();

    /// @brief Container mount status.
    typedef enum
    {
        cBCContainerState_Creation, ///< Is under creation (not a valid container file yet)
        kBCContainerState_Unlinked, ///< Container is not linked.
        kBCContainerState_Linked,   ///< Container is linked to device, but not mounted.
        kBCContainerState_Mounted   ///< Container is linked to device and mounted to some folder.
    } bc_container_state;

    bool IsExists() const;

    QString GetPath() const { return fileinfo.absoluteFilePath(); }

    QString GetFileName() const { return fileinfo.fileName(); }

    QString GetFileBaseName() const { return fileinfo.baseName(); }

    const QString& GetMountPoint() const
    {
        return this->mountPoint_;
    }

    const QString& GetDeviceName() const
    {
        return this->device_;
    }

    bc_container_state GetState() const
    {
        return this->state_;
    }

    bool IsMounted() const
    {
        return this->state_ == kBCContainerState_Mounted;
    }

    bool IsLinked() const
    {
        return (this->state_ == kBCContainerState_Linked) || (this->state_ == kBCContainerState_Mounted);
    }

    bool IsParsed() const
    {
        return this->isParsed;
    }


    QString             modeName;
    QString             algName;
    QString             description;
    QString             lastMountPoint;
    QString             fileSystem;
    int                 algId;
    int                 modeId;
    bool                mountReadOnly;
    bc_u64              size;

    QString             error;  //! Contains error text if any while operate with the container

    ContainerInfoWidget *container_info_widget;

    void runJob(BackgroundJob *job);

    bool isBusy() const { return bool(bg_job); }

    static ContainerDescriptor *CreateForNewContainer(QString filePath, bc_u64 size, int algId, QString algName,
                                                      int modeId, QString modeName, QString fileSystem,
                                                      QString description);

    //! Set actual mount state for a container by data from kernel module
    void UpdateMountState();

    void update();

    void getInfoFrom(BCVolumeInfo &containerInfo);

public slots:
    void onBackgroundJobFinished(bool success);

signals:
    void changed();


private:
    mutable QFileInfo           fileinfo;
    QString             device_;
    QString             mountPoint_;
    bool                isParsed;

    bc_container_state  state_;

    BackgroundJob        *bg_job;   //<! active background operation on that container

    //! parse a container header; set alg, mode, ... fields
    void BCOperationUpdateDescriptor();

    void SetParsed(bool isParsed);

    void Link(const QString& device);
    void Mount(const QString& device, const QString& mountPoint);
    void Unlink();
    void Unmount();
};

#endif // CONTAINERDESCRIPTOR_H
