#include <QEvent>
#include <QTimer>
#include <QMessageBox>
#include <QFileDialog>
#include <QDesktopServices>
#include <QUrl>
#include <QIcon>
#include <QMotifStyle>
#include <QSystemTrayIcon>
#include <QSystemSemaphore>
#include <QCloseEvent>
#include <QDebug>

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "model/containerlistmodel.h"
#include "model/hotkeylistener.h"
#include "model/qt_helpers.h"
#include "operations/bcoperation.h"
#include "operations/MountController.h"
#include "operations/LinkController.h"
#include "operations/CreationController.h"
#include "operations/HiddenCreateController.h"
#include "operations/AddPasswordController.h"
#include "operations/ChangePasswordController.h"
#include "operations/RemovePasswordController.h"
#include "operations/WipeHeaderController.h"
#include "operations/EncryptHeaderController.h"
#include "operations/DecryptHeaderController.h"
#include "operations/BackupHeaderController.h"
#include "operations/RestoreHeaderController.h"
#include "operations/ChangeDescriptionController.h"
#include "dialogs/ContainerCreateDialog.h"
#include "dialogs/ChangeDescriptionDialog.h"
#include "entity/containerdescriptor.h"
#include "entity/settings.h"
#include "src/bctool.h"
#include "containerinfowidget.h"
#include "settingsdialog.h"
#include "reencryptdialog.h"
#include "formatdialog.h"
#include "trayiconmanager.h"
#include "aboutdialog.h"
#include "logdialog.h"
#include "container/exception.h"
#include "rawlinkdialog.h"
#include "welcomewidget.h"

#define SHOW_EXCEPTION(err_msg, e) BCOperationShowErrorString(QString(err_msg).arg(e.what()));


MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent), creationDialog(0),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    try
    {
        if (!BCOperationCheckKernelInitialization())
        {
            ::exit(EXIT_FAILURE);
        }
    }
    catch(BCException e)
    {
        BCOperationShowMessage(kBCTool_Error, e.what());
        ::exit(EXIT_FAILURE);
    }

    this->welcomeWidget = new WelcomeWidget();
    this->ui->stackedWidget->addWidget(this->welcomeWidget);
    connect(welcomeWidget, SIGNAL(create()), ui->createAction, SLOT(trigger()));
    connect(welcomeWidget, SIGNAL(locate()), ui->locateAction, SLOT(trigger()));

    this->hotkeyListener_ = new HotkeyListener();
    connect(this->hotkeyListener_, SIGNAL(OnHotkeyPressed()), this, SLOT(on_ejectAllAction_triggered()));

    // Set left side to minimum, not expandable size.
    this->ui->splitter->setStretchFactor(0, 0);
    this->ui->splitter->setStretchFactor(1, 1);

    this->containerModel_ = new ContainerListModel(this);
    ui->containerTree->InitModel(this->containerModel_);

    this->trayIcon_ = new TrayIconManager(this->containerModel_);

    this->ProcessDescriptorChanges();

    // containerTree_
    connect(ui->containerTree, SIGNAL(OnSelectedDescriptorChanged(ContainerDescriptor*)),
            this, SLOT(SelectedDescriptorChangedSlot(ContainerDescriptor*)));
    connect(ui->containerTree, SIGNAL(OnEject()), this, SLOT(on_ejectAction_triggered()));
    connect(ui->containerTree, SIGNAL(OnMount()), this, SLOT(on_mountAction_triggered()));
    connect(ui->containerTree, SIGNAL(OnDelete()), this, SLOT(on_removeAction_triggered()));
    // connect(this->ui->containerTree, SIGNAL(aboutToRemove(ContainerDescriptor*)), this, SLOT(ContainerIsAboutToRemove(ContainerDescriptor*)));

    // mainToolbar_
    connect(ui->mainToolBar, SIGNAL(OnCreate()), this, SLOT(on_createAction_triggered()));
    connect(ui->mainToolBar, SIGNAL(OnBrowse()), this, SLOT(on_locateAction_triggered()));
    connect(ui->mainToolBar, SIGNAL(OnMount()), this, SLOT(on_mountAction_triggered()));
    connect(ui->mainToolBar, SIGNAL(OnEject()), this, SLOT(on_ejectAction_triggered()));
    connect(ui->mainToolBar, SIGNAL(OnEjectForce()), this, SLOT(on_forceEjectAction_triggered()));
    connect(ui->mainToolBar, SIGNAL(OnEjectAll()), this, SLOT(on_ejectAllAction_triggered()));

    // trayIcon_
    connect(this->trayIcon_, SIGNAL(showControlPanel()), this, SLOT(TrayIconShowControlPanelSlot()));
    connect(this->trayIcon_, SIGNAL(quit()), this, SLOT(on_exitAction_triggered()));
    connect(this->trayIcon_, SIGNAL(ejectAll()), this, SLOT(on_ejectAllAction_triggered()));
    connect(this->trayIcon_, SIGNAL(selectDescriptor(ContainerDescriptor*)),
            this, SLOT(TrayIconDescriptorSelectedSlot(ContainerDescriptor*)));

    this->LoadPreviousViewSettings();
    this->UpdateHotkeyListener();
    this->trayIcon_->ShowIcon();

#ifndef BC_DEBUG
    ui->showLogAction->setVisible(false);
#endif
}

MainWindow::~MainWindow()
{
    this->SaveViewSettings();
    containerModel_->SaveToConfig();
    delete this->hotkeyListener_;
    delete ui;
}


/////////////////////
//      SLOTS       /
/////////////////////

void MainWindow::onContainerOperationFinished(bool /* success */)
{
    qDebug() << "MainWindow: background operation is finished on a container";


    BackgroundJob* job = qobject_cast<BackgroundJob*>(sender());

    if (job) {
        QString path = job->container()->GetPath();
        qDebug() << "Put " << path << " on top of recently used";
        BCSettings s;
        containerModel_->putOnTopOfRecent(path);
    }

    this->ProcessDescriptorChanges();
}

void MainWindow::SelectedDescriptorChangedSlot(ContainerDescriptor *)
{
    this->ProcessDescriptorChanges();
}

void MainWindow::TrayIconDescriptorSelectedSlot(ContainerDescriptor *descriptor)
{
    if (ui->containerTree->SelectDescriptor(descriptor))
    {
        if (descriptor->IsLinked()) // If linked or mounted.
        {
            ui->ejectAction->trigger();
        }
        else
        {
            ui->mountAction->trigger();
        }
    }
    else
    {
        // tray icon: show an error as a message box
        BCOperationShowErrorString(tr("Couldn't find a descriptor."));
    }
}


void MainWindow::showError(QString str)
{
    // TODO: use stack of errors in main window (somewhat like Firefox uses instead of popups)
    BCOperationShowErrorString(str);
}


void MainWindow::ContainerIsAboutToRemove(ContainerDescriptor * descriptor)
{
    ContainerInfoWidget *ciw = getContainerInfoWidgetFor(descriptor);
    ui->stackedWidget->removeWidget(ciw);
}


void MainWindow::TrayIconShowControlPanelSlot()
{
    this->show();
    this->activateWindow();
    this->raise();
}


void MainWindow::on_locateAction_triggered()
{
    try
    {
        QStringList selectedFiles = BCHelperLoadFiles(tr("Container Files Selection"),
                                        BC_FILETYPES_AVAILABLE, true);

        QStringList alreadyAddedFiles;

        for (int i = 0; i < selectedFiles.count(); i++)
        {
            ContainerDescriptor *descriptor = new ContainerDescriptor(0, selectedFiles.at(i));
            descriptor->update();
            bool alreadyAdded = this->containerModel_->IndexOfContainer(descriptor->GetPath()) != -1;

            if (!alreadyAdded) {
                this->containerModel_->Add(descriptor);
            }
            else {
                alreadyAddedFiles.append(selectedFiles.at(i));
            }
        }

        int alreadyAddedCount = alreadyAddedFiles.count();

        // Show list of files that haven't been added.
        if (alreadyAddedCount)
        {
            // TODO: Select the first of the already added files.

        }
    }
    catch(BCException ex)
    {
        BCOperationShowMessage(kBCTool_Error, ex.what());
    }

    containerModel_->SaveToConfig();
}

void MainWindow::on_mountAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();
    Q_ASSERT(descriptor);

    if (descriptor->isBusy())
        return;

    if (descriptor->IsMounted() || descriptor->IsLinked())
        return;

    // TODO: disable all the actions for a busy container

    BackgroundJob *j = new MountController(this, descriptor);

    // TODO: probably we do not need it anymore as 'changed()' signal will be emited when finished
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));

    descriptor->runJob(j);
}

void MainWindow::on_linkAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();
    Q_ASSERT(descriptor);

    if (descriptor->isBusy())
        return;

    if (descriptor->IsMounted() || descriptor->IsLinked())
        return;

    BackgroundJob *j = new LinkController(this, descriptor);

    // TODO: probably we do not need it anymore as 'changed()' signal will be emited by 'descriptor' object
    // to update action enabled states.
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));

    descriptor->runJob(j);
}

void MainWindow::on_rawFileLinkAction_triggered()
{
    // Select file to link.
    QStringList fileToRawLink = BCHelperLoadFiles(tr("Select file to raw link:"),
                                   "All files (*)", false);

    if (!fileToRawLink.count())
        return;

    RawLinkDialog dlg;
    RawLinkParameters p;
    p.file = fileToRawLink.at(0);

    dlg.SetParameters(&p);
    dlg.exec();
    if (dlg.result() == QDialog::Accepted)
    {
        // TODO: Linking read-only with random password is unuseful.
        BCOperationRawLinkFile(p.file, p.algName, p.modeName, false);
    }
}

void MainWindow::on_createAction_triggered()
{
    if (creationDialog)
        return;

    creationDialog = new ContainerCreateDialog(this);
    connect(creationDialog, SIGNAL(accepted()), this, SLOT(onCreationDlgOK()));
    connect(creationDialog, SIGNAL(rejected()), this, SLOT(onCreationDlgCanceled()));
    creationDialog->show();
}

void MainWindow::onCreationDlgCanceled()
{
    creationDialog->deleteLater();
    creationDialog = 0;
}

void MainWindow::onCreationDlgOK()
{
    disconnect(creationDialog, SIGNAL(accepted()), this, SLOT(onCreationDlgOK()));

    ContainerDescriptor *descriptor = creationDialog->getData();

    // check possibility of creation a container with this path and size
    QStringList errors;
    if (!BCOperationCheckNewDescriptor(descriptor, &errors))
    {
        QString errorString = errors.join("\n");
        BCOperationShowErrorString(tr("Please correct errors below and try again: \n%0").arg(errorString));
        delete descriptor;
        return;
    }

    CreationController *cc = new CreationController(this, descriptor, creationDialog);

    // add container creation-in-progress to list
    containerModel_->Add(descriptor);

    // create ConstainerInfoWidget for it (to display creation progress)
    ui->containerTree->SelectDescriptor(descriptor);

    descriptor->runJob(cc);

    // Now container creation is continuing in background and main UI can create another container
    creationDialog = 0;
}


void MainWindow::on_settingsAction_triggered()
{
    SettingsDialog dlg;
    dlg.exec();
    this->UpdateHotkeyListener();
}

void MainWindow::on_forceEjectAction_triggered()
{
    try
    {
        ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();
        if (descriptor->IsMounted())
            BCOperationUmountContainter(descriptor, true);
        else if (descriptor->IsLinked())
            BCOperationUnlinkContainer(descriptor, true);

        this->ProcessDescriptorChanges();
    }
    catch(BCException e)
    {
        SHOW_EXCEPTION("Error ejecting: %0", e);
    }
}

void MainWindow::on_ejectAction_triggered()
{
    try
    {
        ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();
        if (!descriptor)
            return;

        if (descriptor->IsMounted())
        {
            BCOperationUmountContainter(descriptor, false);
            this->ProcessDescriptorChanges();
        }
        else if (descriptor->IsLinked())
        {
            BCOperationUnlinkContainer(descriptor, false);
            this->ProcessDescriptorChanges();
        }
    }
    catch (BCException e)
    {
        SHOW_EXCEPTION("Error ejecting: %0", e);
    }
}

void MainWindow::on_ejectAllAction_triggered()
{
    BCOperationUmountAll(this->settings_.IsEjectAllForced());
    this->containerModel_->Update();
}

void MainWindow::on_exitAction_triggered()
{
    this->close();
}

void MainWindow::on_removeAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();
    if (!descriptor)
        return;
    if (BCOperationDeleteContainer(descriptor))
    {
        this->containerModel_->Remove(descriptor);
    }

    containerModel_->SaveToConfig();
}

void MainWindow::on_headerEncryptAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    BackgroundJob *j = new EncryptHeaderController(this, descriptor);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}

void MainWindow::on_headerDecryptAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    BackgroundJob *j = new DecryptHeaderController(this, descriptor);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}


void MainWindow::on_wipeHeaderAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    BackgroundJob *j = new WipeHeaderController(this, descriptor);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}


void MainWindow::on_changeDescriptionAction_triggered()
{
    ChangeDescriptionDialog dlg(this);

    if (dlg.exec() != QDialog::Accepted)
        return;

    QString new_description = dlg.getNewDescription();

    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    BackgroundJob *j = new ChangeDescriptionController(this, descriptor, new_description);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}


void MainWindow::on_headerBackupAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    BackgroundJob *j = new BackupHeaderController(this, descriptor);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}

void MainWindow::on_headerRestoreAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    BackgroundJob *j = new RestoreHeaderController(this, descriptor);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}

void MainWindow::on_aboutAction_triggered()
{
    AboutDialog about;
    about.exec();
}

void MainWindow::on_aboutQtAction_triggered()
{
    QMessageBox::aboutQt(this);
}

void MainWindow::on_reencryptAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();
    if (!descriptor)
        return;

    ReencryptDialog dlg(this);
    dlg.setModal(true);
    dlg.show();

    while (true)
    {
        while (!dlg.HaveResult())
        {
            BCHelperSleep(10);
            QApplication::processEvents();
        }

        if (dlg.GetResult() == QDialog::Accepted)
        {
            // We're here, if user clicked Ok.
            QString algName = dlg.GetAlgorythmName();
            QString modeName = dlg.GetModeName();
            if (!BCOperationReencryptContainer(descriptor, algName, modeName))
            {
                dlg.ResetResult();
                continue;
            }

            return;
        }

        // We're here, if user clicked Cancel.
        return;
    }
}

void MainWindow::on_createHiddenAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();
    Q_ASSERT(descriptor);

    if (descriptor->isBusy())
        return;

    if (descriptor->IsMounted()) {
        BCOperationShowErrorString("Container should be not mounted");
        return;
    }


    // this condition and comment copied from Mac version (mac/hidden_part/BCCreateHiddenVolumeController.m:15)
    //
    // Minimum container volume size can be 10 MB so for hidden part creation at least 20 MB volume is required.
    // However even on new empty 20 MB volume free space is only 18 MB so to suite rerequirement of 10 MB volume
    // we need a larger container, that is why 25 MB requirement.
    if (descriptor->size < BCTool::minimalContainerSize * 2.5) {
        BCOperationShowErrorString("Container size must be at least 25 MB to create hidden part.");
        return;
    }

    BackgroundJob *j = new HiddenCreateController(this, descriptor);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}

void MainWindow::on_passwordAddAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    BackgroundJob *j = new AddPasswordController(this, descriptor);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}

void MainWindow::on_passwordRemoveAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    BackgroundJob *j = new RemovePasswordController(this, descriptor);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}

void MainWindow::on_passwordChangeAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    BackgroundJob *j = new ChangePasswordController(this, descriptor);
    connect(j, SIGNAL(finished(bool)), this, SLOT(onContainerOperationFinished(bool)));
    descriptor->runJob(j);
}

void MainWindow::on_formatAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    if (!descriptor)
        return;

    FormatDialog dlg;
    dlg.setModal(true);
    dlg.show();

    while (true)
    {
        while(!dlg.HaveResult())
        {
            QApplication::processEvents();
            BCHelperSleep(10);
        }

        if (dlg.GetResult() == QDialog::Accepted)
        {
            QString fsType = dlg.GetFileSystem();
            BCOperationFormat(descriptor, fsType);
        }

        // We're here in case the user pressed Cancel.
        return;
    }
}

void MainWindow::on_checkFilesystemAction_triggered()
{
    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    if (!descriptor)
        return;

    BCOperationFsck(descriptor, QString());
}

void MainWindow::on_showLogAction_triggered()
{
    LogDialog log;
    log.exec();
}

void MainWindow::UpdateHotkeyListener()
{
    bool result = true;

#define CHECK_RESULT(msg) if (!result) { BCOperationShowMessage(kBCTool_Error, msg); return; }

    if (this->settings_.IsEjectAllHotKeyEnabled())
    {
        if (this->hotkeyListener_->IsListening())
            result = this->hotkeyListener_->StopListening();

        CHECK_RESULT("Error while stopping global hotkey listener.");

        BCHotkey hotkey = this->settings_.EjectAllHotKey();
        result = this->hotkeyListener_->SetHotkey(hotkey);

        CHECK_RESULT("Error while setting global hotkey.");

        result = this->hotkeyListener_->StartListening();

        CHECK_RESULT("Error while starting global hotkey listening.");
    }

#undef CHECK_RESULT
}

/////////////////////////////////////////////
//      Window Events and View Settings     /
/////////////////////////////////////////////

void MainWindow::closeEvent(QCloseEvent * /*event*/)
{
    this->SaveViewSettings();
    qApp->exit(0);
}

void MainWindow::changeEvent(QEvent *e)
{
    if (e->type() == QEvent::WindowStateChange)
    {
        if (this->isMinimized() && this->settings_.IsMinimizingToTray())
        {
            this->hide();
            e->ignore();
        }
    }
    QMainWindow::changeEvent(e);
}


ContainerInfoWidget * MainWindow::getContainerInfoWidgetFor(ContainerDescriptor *descriptor)
{
    ContainerInfoWidget *ret = descriptor->container_info_widget;

    if (! ret) {
        ret = new ContainerInfoWidget(descriptor);
        descriptor->container_info_widget = ret;

        // please note: ownership of the widget is passed to stackedWidget
        ui->stackedWidget->addWidget(descriptor->container_info_widget);
    }

    return ret;
}

void MainWindow::ProcessDescriptorChanges()
{
    bool enable_file_present_actions = false;
    bool enable_parsed_actions = false;
    bool enable_unparsed_actions = false;
    bool enable_mount_actions = false;
    bool enable_eject_actions = false;

    ContainerDescriptor *descriptor = ui->containerTree->GetSelectedDescriptor();

    if (descriptor) {
        ContainerInfoWidget *ciw = getContainerInfoWidgetFor(descriptor);
        ui->stackedWidget->setCurrentWidget(ciw);

        // TODO:  if background process is active only 'Cancel' is available
        // descriptor->isBusy() - still 'true' when called from MainWindow::onBackgroundJobFinished() because 'Controller' still exists

        if (descriptor->IsExists()) {
            enable_file_present_actions = true;

            if (descriptor->IsParsed())
                enable_parsed_actions = true;
            else
                enable_unparsed_actions = true;
        }

        // file may be absent but container is stil mounted/linked
        if (descriptor->IsLinked())
            enable_eject_actions = true;
        else if (descriptor->IsExists())
            enable_mount_actions = true;
    }
    else {
        ui->stackedWidget->setCurrentWidget(this->welcomeWidget);
    }

    // always enabled if any container selected
    ui->headerRestoreAction->setEnabled(enable_file_present_actions);
    ui->wipeHeaderAction->setEnabled(enable_file_present_actions);
    ui->passwordAddAction->setEnabled(enable_file_present_actions);
    ui->passwordChangeAction->setEnabled(enable_file_present_actions);
    ui->passwordRemoveAction->setEnabled(enable_file_present_actions);
    ui->headerBackupAction->setEnabled(enable_file_present_actions);
    ui->formatAction->setEnabled(enable_file_present_actions);
    ui->checkFilesystemAction->setEnabled(enable_file_present_actions);
    ui->changeDescriptionAction->setEnabled(enable_file_present_actions);
    ui->createHiddenAction->setEnabled(enable_file_present_actions);
    ui->reencryptAction->setEnabled(enable_file_present_actions);

    ui->headerDecryptAction->setEnabled(enable_unparsed_actions);
    ui->headerEncryptAction->setEnabled(enable_parsed_actions);

    ui->linkAction->setEnabled(enable_mount_actions);
    ui->mountAction->setEnabled(enable_mount_actions);
    ui->mainToolBar->SetMountEnabled(enable_mount_actions);

    ui->ejectAction->setEnabled(enable_eject_actions);
    ui->forceEjectAction->setEnabled(enable_eject_actions);
    ui->mainToolBar->SetEjectEnabled(enable_eject_actions);

    this->trayIcon_->UpdateMenu();
}


void MainWindow::LoadPreviousViewSettings()
{
    bool isMaximized;
    isMaximized = this->settings_.MainFormIsMaximized();
    if (isMaximized)
    {
        this->showMaximized();
        return; // Do not load size in this case,
                // because it will be size of maximized window.
    }
    else
        this->showNormal();

    QSize size = this->size();
    QPoint point = this->pos();

    size = this->settings_.MainFormSize(size);
    point = this->settings_.MainFormLocation(point);

    // Set minimum position, so window border can't be upper than zero point.
    if (point.x() < 70)
        point.setX(70);
    if (point.y() < 70)
        point.setY(70);

    this->setGeometry(point.x(), point.y(), size.width(), size.height());
}

void MainWindow::SaveViewSettings()
{
    this->settings_.SetMainFormLocation(this->pos());
    this->settings_.SetMainFormMaximized(this->isMaximized());
    this->settings_.SetMainFormSize(this->size());
}


void MainWindow::on_helpAction_triggered()
{
    QString p = "../help/index.html";
    if (QFileInfo(p).exists()) {
        QDesktopServices::openUrl(p);
        return;
    }

    p = "/usr/share/doc/bestcrypt/help/index.html";
    if (QFileInfo(p).exists()) {
        QDesktopServices::openUrl(p);
        return;
    }

    QDesktopServices::openUrl(QString("http://www.jetico.com/web_help/bc2_linux/"));
}

