#include <stdio.h>
#include <QDebug>

#include "containerlistmodel.h"
#include "entity/containerdescriptor.h"
#include "container/exception.h"
#include "entity/settings.h"
#include "operations/bcoperation.h"
#include "devicemonitor.h"

#include "version.h"

// Key names in QSetting
#define BC_CFG_CONTAINERS_RECENT        "RecentList"
#define BC_CFG_CONTAINERS_ARRAY         "Containers"
#define BC_CFG_CONTAINER_PATH           "Path"
#define BC_CFG_MOUNT_POINT              "MountPoint"
#define BC_CFG_MOUNT_READ_ONLY          "AlwaysMountReadOnly"

namespace {
    const int RECENT_MENU_SIZE = 10;
}


ContainerListModel::ContainerListModel(QObject *parent)
    :QObject(parent)
{
    ignoreEvents = false;

    LoadFromConfig();
    LoadFromDevices();

    // Subscribe to device events in system.
    connect(DeviceMonitor::GetInstance(), SIGNAL(diskEvent()), this, SLOT(onDiskEvent()));
    DeviceMonitor::GetInstance()->StartListening();
}

ContainerListModel::~ContainerListModel()
{
}

void ContainerListModel::Add(ContainerDescriptor *descriptor)
{
    this->InternalAdd(descriptor);
    putOnTopOfRecent(descriptor->GetPath());
    EmitDataChanged();
}

void ContainerListModel::Remove(ContainerDescriptor *descriptor)
{
    int index = this->IndexOfContainer(descriptor->GetPath());
    if (-1 != index)
    {
        recents.removeOne(descriptor->GetPath());
        descriptors.removeAt(index);
        EmitDataChanged();
    }
}

int ContainerListModel::IndexOfContainer(const QString& fullPath) const
{
    ContainerDescriptor *cur = 0;
    for (int i=0; i < descriptors.count(); i++)
    {
        cur = descriptors.at(i);
        if (cur->GetPath() == fullPath)
        {
            return i;
        }
    }

    return -1;
}

void ContainerListModel::Update()
{
    for (int i = 0; i < descriptors.count(); i++)
    {
        try
        {
            descriptors.at(i)->update();
        }
        catch (BCException ex)
        {
            BCOperationShowErrorString(QString("Error while updating descriptor for %0: %1").arg(descriptors.at(i)->GetPath()).arg(ex.what()));
        }
    }
    EmitDataChanged();
}

void ContainerListModel::Clear()
{
    descriptors.clear();
    recents.clear();
    EmitDataChanged();
}


void ContainerListModel::LoadFromConfig()
{
    QSettings settings;

    int containerCount = settings.beginReadArray(BC_CFG_CONTAINERS_ARRAY);

    for (int i = 0; i < containerCount; i++)
    {
        settings.setArrayIndex(i);

        ContainerDescriptor *descriptor = new ContainerDescriptor(this, settings.value(BC_CFG_CONTAINER_PATH).toString());
        descriptor->lastMountPoint = settings.value(BC_CFG_MOUNT_POINT).toString();
        descriptor->mountReadOnly = settings.value(BC_CFG_MOUNT_READ_ONLY).toBool();

        if (! descriptor->IsExists()) {
            delete descriptor;
            continue;
        }

        InternalAdd(descriptor);
    }

    loadRecentList();

    Update();
}


void ContainerListModel::SaveToConfig()
{
    BCSettings bs;
    if (! bs.IsContainerListSaving())
        return;

    // we should keep in recent list only containers from main list
    QStringList filtered_recent_list;

    QSettings &settings = bs.settings();

    // Remove old container list and add new.
    settings.remove(BC_CFG_CONTAINERS_ARRAY);
    settings.beginWriteArray(BC_CFG_CONTAINERS_ARRAY);

    for (int i = 0; i < descriptors.count(); i++)
    {
        ContainerDescriptor *cur = descriptors.at(i);
        if (cur) {
            settings.setArrayIndex(i);
            settings.setValue(BC_CFG_CONTAINER_PATH, cur->GetPath());
            settings.setValue(BC_CFG_MOUNT_READ_ONLY, cur->mountReadOnly);
            settings.setValue(BC_CFG_MOUNT_POINT, cur->lastMountPoint);

            if (recents.contains(cur->GetPath()))
                filtered_recent_list.append(cur->GetPath());
        }
    }
    settings.endArray();

    recents = filtered_recent_list;
    storeRecentList();
}

void ContainerListModel::InternalAdd(ContainerDescriptor *descriptor)
{
    // Add only containers that are not in model yet.
    int index = this->IndexOfContainer(descriptor->GetPath());
    // qDebug() << "ContainerListModel::InternalAdd(): " << descriptor->GetPath() << index;

    if (-1 == index)
    {
        descriptor->setParent(this);
        descriptors.append(descriptor);
    }
}

void ContainerListModel::EmitDataChanged()
{
    if (!this->ignoreEvents)
        emit this->changed();
}

void ContainerListModel::onDiskEvent()
{
    fprintf(stderr, "Disk event came, reloading list of devices...\n");
    bool isEventsIgnored = this->ignoreEvents;
    this->ignoreEvents = false;

    // Load mounted descriptors.
    this->LoadFromDevices();

    // Update present descriptors.
    this->Update();

    this->ignoreEvents = isEventsIgnored;
    EmitDataChanged();
}


void ContainerListModel::LoadFromDevices()
{
    std::vector<bc_mount_entry> mountInfo;
    if (! BCTool::GetMountInfo(&mountInfo)) {
        qWarning() << "Unable to get mount points";
        return;
    }

    for (std::vector<bc_mount_entry>::iterator i = mountInfo.begin()
         ; i != mountInfo.end()
         ; ++i)
    {
        bc_mount_entry &entry = *i;
        // qDebug() << "LoadFromDevices(): bc_mount_entry: " << entry.deviceName <<  entry.containerPath << entry.mountPoint;
        ContainerDescriptor *descriptor = new ContainerDescriptor(this, entry.containerPath);
        descriptor->update();
        InternalAdd(descriptor);
    }

    EmitDataChanged();
}

void ContainerListModel::loadRecentList()
{
    QSettings settings;

    int containerCount = settings.beginReadArray(BC_CFG_CONTAINERS_RECENT);

    for (int i = 0 ; i < containerCount ; i++) {
        settings.setArrayIndex(i);
        QString p = settings.value("path").toString();

        if (IndexOfContainer(p) >= 0)
            recents.append(p);
    }

    settings.endArray();
}

void ContainerListModel::storeRecentList()
{
    QSettings settings;

    // Remove old container list and add new.
    settings.remove(BC_CFG_CONTAINERS_RECENT);
    settings.beginWriteArray(BC_CFG_CONTAINERS_RECENT);

    for (int i = 0; i < recents.count(); i++) {
        settings.setArrayIndex(i);
        settings.setValue("path", recents[i]);
    }
    settings.endArray();
}

void ContainerListModel::putOnTopOfRecent(QString path)
{
    if (path.isEmpty()) {
        qDebug() << "Skip empty path from adding to recent list";
        return;
    }

    QStringList new_recents;

    // put this element on top
    new_recents.append(path);

    // followed by others
    for (int i = 0 ; i < recents.count() && new_recents.size() <= RECENT_MENU_SIZE; i++) {
        if (! new_recents.contains(recents[i]))
            new_recents.append(recents[i]);
    }

    recents = new_recents;
}

ContainerListModel::list_t ContainerListModel::getRecentDescriptors() const
{
    list_t ret;

    for (int i = 0 ; i < recents.size() ; i++) {
        int ind = IndexOfContainer(recents[i]);
        if (ind >= 0) {
            ret.append(descriptors[ind]);
        }
    }
    return ret;
}
