#include <stdio.h>
#include <glob.h>
#include <errno.h>
#include <execinfo.h>
#include <unistd.h>

#include <QFileDialog>
#include <QMutex>
#include <QWaitCondition>
#include <QDir>
#include <QFileInfo>

#include "qt_helpers.h"
#include "src/bctool.h"


void BCHelperPrintStackTrace(const QString& msg)
{
    void *array[100];
    size_t size;

    size = backtrace(array, 100);
    printf("%s\n", msg.toLatin1().constData());
    backtrace_symbols_fd(array, size, STDOUT_FILENO);
}

int BCHelperGetIndexInList(QStringList list, QString word)
{
    return list.indexOf(QRegExp("\\b" + word + "\\b", Qt::CaseInsensitive));
}

std::vector<std::string> BCHelperQStringListToStdVectorOfStrings(const QStringList &ll)
{
    std::vector<std::string> ret;

    for (int i = 0 ; i < ll.size() ; i++)
        ret.push_back(ll[i].toStdString());
    return ret;
}


static QString GetFileNameWithExtension(QString fileName, const QString& nameFilter)
{
    // * Get an extension from the selected name filter.
    int extStart = nameFilter.indexOf("(") + 2; // Also ignore the '*' sign.
    int extEnd = nameFilter.indexOf(")") - 1;

    if ((extStart == -1) || (extEnd == -1))
        return fileName;

    QString ext = nameFilter.mid(extStart, extEnd - extStart + 1);

    // * Add the extension to the file name, if it is needed.
    if (!fileName.endsWith(ext, Qt::CaseInsensitive))
        return fileName + ext;
    return fileName;
}

QStringList BCHelperLoadFiles(const QString& dialogCaption,
                              const QString& extensions, bool allowMultiple)
{
    QFileDialog openDialog;
    openDialog.setWindowTitle(dialogCaption);

    openDialog.setAcceptMode(QFileDialog::AcceptOpen);
    openDialog.setNameFilter(extensions);

    if (!allowMultiple)
        openDialog.setFileMode(QFileDialog::ExistingFile);
    else
        openDialog.setFileMode(QFileDialog::ExistingFiles);

    if (openDialog.exec())
    {
        QStringList selectedFiles = openDialog.selectedFiles();

        if (!allowMultiple)
        {
            QStringList result;
            if (selectedFiles.count())
                result.append(selectedFiles.at(0));
            return result;
        }
        // Else, if multiple selection is allowed.
        return selectedFiles;
    }

    return QStringList();
}


QString BCHelperSaveFile(const QString& defaultName, const QString& dialogCaption,
                         const QString& extensions)
{
    QFileDialog saveDialog;
    saveDialog.setWindowTitle(dialogCaption);
    saveDialog.setFileMode(QFileDialog::AnyFile);
    saveDialog.setAcceptMode(QFileDialog::AcceptSave);
    saveDialog.setNameFilter(extensions);

    if (defaultName.length())
        saveDialog.selectFile(defaultName);

    if (saveDialog.exec())
    {
        QStringList selectedFiles = saveDialog.selectedFiles();
        if (selectedFiles.count())
            return GetFileNameWithExtension(selectedFiles.at(0), saveDialog.selectedNameFilter());
    }

    return QString();
}

void BCHelperDeleteFileIfExists(const QString &fileName)
{
    QFile f(fileName);
    if (f.exists())
        f.remove();
}

bool BCHelperGetSupportedFS(QStringList *out_supportedFS)
{
    Q_ASSERT(out_supportedFS);

    glob_t searchResults;
    int result = glob("/sbin/mkfs.*", 0, NULL, &searchResults);
    if (result)
    {
        fprintf(stderr, "Glob error: %s\n", strerror(errno));
        return false;
    }

    for (size_t i = 0; i < searchResults.gl_pathc; ++i)
    {
        const char *path = searchResults.gl_pathv[i];
        if (path)
        {
            QFileInfo f(path);
            QString mkfsName = f.fileName();
            Q_ASSERT(mkfsName.startsWith("mkfs."));
            QString fsName = mkfsName.right(mkfsName.length() - 5);
            out_supportedFS->append(fsName);
        }
    }
    return true;
}


QString BCHelperGetLocation(const QString& dialogCaption, const QString& basePath)
{
    QFileDialog folderDialog;
    folderDialog.setFileMode(QFileDialog::Directory);
    folderDialog.setOptions(QFileDialog::ShowDirsOnly);
    folderDialog.setWindowTitle(dialogCaption);

    if (basePath.count())
    {
        folderDialog.setDirectory(basePath);
    }

    if (folderDialog.exec())
    {
        QStringList selectedFolders = folderDialog.selectedFiles();
        if (selectedFolders.count())
        {
            return QString(selectedFolders.at(0));
        }
    }

    return QString();
}


void BCHelperSleep(unsigned long msec)
{
    QMutex mutex;
    QWaitCondition waitCondition;

    // No matter, when we wait for msecs needed.
    if (!mutex.tryLock(msec))
        return;
    waitCondition.wait(&mutex, msec);
    mutex.unlock();
}

bool BCHelperLoadIcon(const QString& iconName, QIcon *out_icon)
{
    QImage img(iconName);
    if (img.isNull())
        return false;

    QPixmap pixmap = QPixmap::fromImage(img);
    QIcon icon(pixmap);
    (*out_icon) = icon;
    return true;
}



void BCHelperFixSize(QWidget *widget)
{
    QSize fixedSize;

    // Unfix size.
    widget->setMinimumSize(0, 0);
    widget->setMaximumSize(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX);

    // First, set widget size to minimal, that fits it's contents.
    widget->adjustSize();

    // If minimum and maximum size is set and equal, then use it.
    if ((!widget->minimumSize().isEmpty()) &&
            (widget->maximumSize() == widget->minimumSize()))
    {
        fixedSize = widget->minimumSize();
    }
    else // Else use size hint.
    {
        fixedSize = widget->sizeHint();
    }

    // Set fixed size.
    widget->setFixedSize(fixedSize.width(), fixedSize.height());
}
