#include <QDebug>
#include <QTimer>
#include <QDir>

#include "CreationController.h"
#include "CreationWorker.h"
#include "QtBCTool.h"
#include "model/devicemonitor.h"
#include "operations/bcoperation.h"
#include "entity/containerdescriptor.h"
#include "model/containerlistmodel.h"
#include "dialogs/ContainerCreateDialog.h"
#include "operations/QtPwdGet.h"


/**
 * @brief New container creation controller (UI-thread side)
 * @param descriptor of the container
 * @param createDialog - we take ownership of the object
 */
CreationController::CreationController(QWidget *parentWidget, ContainerDescriptor *d, ContainerCreateDialog *dlg)
    : BackgroundJob(parentWidget, d), creationDialog(dlg)
{
    dmIsListening = DeviceMonitor::GetInstance()->IsListening();
    if (dmIsListening)
        DeviceMonitor::GetInstance()->StopListening();

    worker = new CreationWorker();

    // Aware that OK is already clicked in Creation dialog
    creationDialog->dumpParamsTo(&worker->createParams, worker->createPwdGet.getContainerContext(), &worker->mountParams);

    // QtPwdGetGlue is not needed here:
    creationDialog->setQtPwdGet(&worker->createPwdGet);
    connect(&worker->createPwdGet, SIGNAL(needPassword(bool)), creationDialog, SLOT(onNeedPassword(bool)));

    progressUpdateTimer = new QTimer(this);
    connect(progressUpdateTimer, SIGNAL(timeout()), this, SLOT(onNeedToUpdateProgress()));
    progressUpdateTimer->start(500);

    // optionally we can mount a container
    if (worker->createParams.mount_after_creation) {
        worker->mountParams.filename = descriptor->GetPath().toStdString();
        if (worker->mountParams.mountPoint.empty()) {
            QDir mountLocation = QDir::homePath() + QDir::separator() + descriptor->GetFileBaseName();
            worker->mountParams.mountPoint = mountLocation.path().toStdString();
        }
        QDir mountLocation(worker->mountParams.mountPoint.c_str());
        if (! mountLocation.exists())
            mountLocation.mkpath(".");
    }
}


CreationController::~CreationController()
{
    creationDialog->deleteLater();
    worker->deleteLater();
}


void CreationController::onCancelation()
{
    qDebug() << "PROGRESS: " << Q_FUNC_INFO;
    worker->progress.cancel();
}


void CreationController::on_finished(bool success)
{
    progressUpdateTimer->stop();

    if (dmIsListening)
        DeviceMonitor::GetInstance()->StartListening();

    descriptor->update();

    if (! success)
        descriptor->error = "Error during creation: " + descriptor->error;

    emit finished(success);
}


void CreationController::onNeedToUpdateProgress()
{
    static QString text("Randomizing container");

    emit progress(text, worker->progress.percent, 100);
}


Worker* CreationController::getWorker()
{
    return worker;
}

