#include <QStringList>
#include <QDebug>

#include "QtPwdGet.h"
#include "model/qt_helpers.h"
#include "bctool.h"


QtPwdGet::QtPwdGet(QObject *parent) :
    QObject(parent),
    opened_key_files(0),
    external_header_channel(0)
{
}

QtPwdGet::~QtPwdGet()
{
    if (opened_key_files)
        delete opened_key_files;

    if (external_header_channel)
        delete external_header_channel;
}


/**
 * @brief implementation of getPassword() callback for multi-lib
 * @param prevFailed
 * @return user interaction result (OK or Cancel)
 */
BCContainerResult QtPwdGet::getPassword(bool prevFailed)
{
    qDebug() << "QtPwdGet::getPassword(): emited needPassword() signal (prevFailed=" << prevFailed << ")";

    // signal UI thread to display the dialog
    emit needPassword(prevFailed);

    // pause worker thread until a password is get
    password_semaphore.acquire();

    qDebug() << "QtPwdGet::getPassword(): password is received";

    // return appropriate code to back to multi-lib
    if (isEmpty())
        return kBCContainer_Cancelled;

    return kBCContainer_Success;
}

QString QtPwdGet::getLastError() const
{
    return last_error;
}


/**
 * @brief Set cPwdGet fields: m_password, keyfiles, etc.
 * and unpause a Worker thread.
 * @param pass
 * @param keyfiles
 */
bool QtPwdGet::setPasswordBundle(QString pass, QStringList keyfiles, QString external_header)
{
    if (password_semaphore.available()) {
        qDebug() << "setPasswordBundle(): do nothing because no worker waiting in getPassword()";
        return false;
    }
    qDebug() << "setPasswordBundle(): storing password";

    if (pass.isEmpty()) {
        qDebug() << "Worker thread: empty password was received. this is Cancel";
        cPwdGet::clear();
        password_semaphore.release();
        return false;
    }

    qDebug() << "Worker thread: password and keyfiles was received";
    
    QByteArray pp = pass.toUtf8();

    if (pp.size() > (kBCPasswordMax-1)) {
        last_error = "Password is too long";
        cPwdGet::clear();
        // continue to hold Worker, do not release semaphore here
        return false;
    }
    
    cPwdGet::setPasswordString(pp.constData());
    
    // set key files
    if (opened_key_files)
        delete opened_key_files;
    opened_key_files = new OpenedKeyfiles( BCHelperQStringListToStdVectorOfStrings(keyfiles) );
    opened_key_files->assingKeyfilesTo(this);
    
    // set header backup channel
    if (! external_header.isEmpty()) {
        if (external_header_channel) {
            delete external_header_channel;
            external_header_channel = 0;
        }
        external_header_channel = new BCFileChannel(external_header.toStdString());
        external_header_channel->open(kBCFileMode_Read);
        cPwdGet::setKeyBlockChannel(external_header_channel);
    }

    // TODO: set public keys DB

    // TODO: set private keys DB

    // unpause worker thread
    qDebug() << "setPasswordBundle(): calling semaphore release...";
    password_semaphore.release();
    qDebug() << "setPasswordBundle(): semaphore released";

    return true;
}


