#include <signal.h>
#include <stdio.h>
#include <mntent.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <semaphore.h>
#include <fcntl.h>
#include <execinfo.h>
#include <glob.h>

#include <QDir>
#include <QMessageBox>
#include <QStyle>
#include <QUrl>
#include <QMutex>
#include <QProgressDialog>
#include <QDesktopServices>
#include <QApplication>
#include <QDebug>

#include "bcoperation.h"
#include "container/container.h"
#include "container/BCFileChannel.h"
#include "container/keygen.h"
#include "container/exception.h"
#include "kernel/BCKernel.h"
#include "entity/containerdescriptor.h"
#include "dialogs/ContainerAccessDialog.h"
#include "src/version.h"
#include "disk/disk.h"
#include "model/qt_helpers.h"
#include "QtPwdGet.h"


static void HandleSignal(int signal)
{
    qDebug("HandleSignal: signal #%i.", signal);
}


static int ReportProgress(const char *message, bc_u64 byteCount, bc_u64 bytesTotal, void */*arg*/)
{
    static QProgressDialog * progressDialog = 0;

    int percent = 0, percentMax = 100;
    if (bytesTotal != 0)
        percent = (int)(byteCount * 100LL / bytesTotal);

    // If progress dialog is already exists, change it's parameters.
    if (!progressDialog)
    {
        // Create a new progress dialog.
        QWidget *activeWindow = QApplication::activeWindow();
        progressDialog = new QProgressDialog(message, "Cancel", 0,
                                             percentMax, activeWindow);
        progressDialog->setWindowModality(Qt::WindowModal);
        progressDialog->setWindowTitle("BestCrypt");
    }

    progressDialog->setLabelText(message);
    progressDialog->setValue(percent);

    progressDialog->update();
    QApplication::processEvents();

    if (progressDialog->wasCanceled())
    {
        delete progressDialog;
        progressDialog = 0;
        return false;
    }
    if (byteCount != bytesTotal)
        return true;

    delete progressDialog;
    progressDialog = 0;

    return true;
}

static int volume_progress_report(bc_u64 nbytes, bc_u64 total, void* arg) {
    return ReportProgress("Processing volume", nbytes, total, arg);
}


//! Temporary support for non-responsivness version of BCTool
class BCTool2: public BCTool   // TODO: use QtBCTool instead
{
    bool AskUser(const char *msg, bool defaultAnswer)
    {
        return BCOperationAskUser(msg, defaultAnswer);
    }

    void messageCallback_(bctool_msg_level level, const char *msg)
    {
        char c = '?';
        if (level >= 0 && level <= 3) {
            const char mm[] = "DIWE";
            c = mm[level];
        }

        qDebug() << "QtBCTool: message:" << c << ":" << msg;

        if (level > kBCTool_Debug) {
            BCOperationShowMessage(level, msg);
        }
    }

    bool generateRandom(bool strong, bc_u8 *out, size_t size)
    {
        // copy/paste from QtBCTool
        QFile file("/dev/urandom");
        QByteArray bytes = file.read(size);
        memcpy(out, bytes.data(), size);
        return true;
    }
};

//! Temporary workaround for old operations during a switch to controller-worker scheme
class PwdAsker: public QtPwdGet
{
public:
    BCContainerResult getPassword(bool prevFailed)
    {
        ContainerAccessDialog dlg;
        dlg.setTitle(m_query, prevFailed);

        if (dlg.exec() ==  QDialog::Rejected)
            return kBCContainer_Cancelled;

        dlg.savePasswordBundleTo(this);
        return kBCContainer_Success;
    }

    void on_entry_point() {}
};




// Create 'Unified Interaction' object (common part for CLI and GUI)
static BCTool *CreateUI()
{
    BCTool *ui = new BCTool2();
    return ui;
}


bool BCOperationAskUser(QString msg, bool defaultAnswer)
{
    QMessageBox msgBox;
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setText(msg);
    msgBox.setIcon(QMessageBox::Question);
    msgBox.setDefaultButton(defaultAnswer? QMessageBox::Yes : QMessageBox::No);
    msgBox.setWindowTitle("BestCrypt Control Panel");
    return msgBox.exec() == QMessageBox::Yes;
}


const char *BCOperationCreateMessage(const char *msg, ...);



/*
  ***********************************************
                Interface
  ***********************************************
*/

bool BCOperationCheckKernelInitialization()
{
    BCTool *ui = CreateUI();
    bool result = ui->CheckKernelInitialization();
    delete ui;
    return result;
}

const char *BCOperationCreateMessage(const char *msg, ...)
{
    const int MaxLen = 4096;
    static char message[MaxLen];
    va_list list;
    va_start(list, msg);
    vsnprintf(message, MaxLen - 1, msg, list);
    va_end(list);
    message[MaxLen - 1] = '\0';
    return message;
}


void BCOperationShowErrorString(const QString &str)
{
    BCOperationShowMessageString(kBCTool_Error, str);
}


void BCOperationShowWarningString(const QString &str)
{
    BCOperationShowMessageString(kBCTool_Warning, str);
}


void BCOperationShowMessage(bctool_msg_level level, const char* msg)
{
    BCOperationShowMessageString(level, msg);
}


void BCOperationShowMessageString(bctool_msg_level level, const QString &msg)
{
    QString caption;
    QMessageBox::Icon icon;

    switch (level)
    {
        case kBCTool_Info:
            icon = QMessageBox::Information;
            caption = "Information";
            break;

        case kBCTool_Warning:
            icon = QMessageBox::Warning;
            caption = "Warning";
            break;

        case kBCTool_Error:
            icon = QMessageBox::Critical;
            caption = "Error";
            break;

        case kBCTool_Debug:
            // TODO: Ignore debug messages here.
            break;

        default:
            Q_ASSERT(0);
    }

    QMessageBox msgBox;
    msgBox.setText(msg);
    msgBox.setWindowTitle(caption);
    msgBox.setIcon(icon);
    msgBox.exec();
}


bool BCOperationCheckNewDescriptor(ContainerDescriptor *descriptor, QStringList *out_errors)
{
    Q_ASSERT(0 != out_errors);

    QFileInfo fileInfo(descriptor->GetPath());
    //QFile file(descriptor->GetPath());
    std::string containerPath = descriptor->GetPath().toStdString();
    bool result = true;

    if (fileInfo.exists())
    {
        QMessageBox msgBox(QMessageBox::Warning, "File already exists", "File with same name already exists. Are you sure you want to replace this file?", QMessageBox::Ok | QMessageBox::Cancel);
        if (msgBox.exec() != QMessageBox::Ok)
        {
            out_errors->append(QObject::tr("* File with same name is already exists in directory."));
            return false;
        }
        // else clean file.
    }

    // Try to create file in specified location.
    FILE *f = fopen(containerPath.c_str(), "w");
    if (!f)
    {
        int error = errno;
        out_errors->append(QObject::tr("* Cannot create file '%0': %1")
                           .arg(descriptor->GetPath()).arg(strerror(error)));
        result = false;
    }
    else
    {
        fclose(f);
        remove(containerPath.c_str());
    }

    if (descriptor->description.count() >= BC_DESCRIPTION_SIZE)
    {
        out_errors->append(QObject::tr("* Description is too long (%0 symbols). Maximum length is %1.")
                           .arg(descriptor->description.count()).arg(BC_DESCRIPTION_SIZE));
        result = false;
    }

    return result;
}


bool BCOperationRawLinkFile(const QString& file, const QString& algName,
                            const QString& modeName, bool readOnly)
{
    std::string fileStr = file.toStdString();
    std::string algStr = algName.toStdString();
    std::string modeNameStr = modeName.toStdString();
    BCTool *ui = CreateUI();
    bool succeeded = ui->RawLink(fileStr.c_str(), algStr.c_str(), modeNameStr.c_str(),
                                 readOnly);
    delete ui;
    return succeeded;
}

bool BCOperationUmountContainter(ContainerDescriptor *descriptor, bool force)
{
    descriptor->update();
    if (!descriptor->IsMounted())
        return true;
    std::string mountPointStr = descriptor->GetMountPoint().toStdString();

    Q_ASSERT(mountPointStr.length());

    BCTool *ui = CreateUI();
    bool result = ui->Umount(mountPointStr.c_str(), force); // Already showed all errors inside.
    delete ui;

    descriptor->UpdateMountState();

    return result;
}

void BCOperationUmountAll(bool force)
{
    BCTool *ui = CreateUI();
    int totalMounted, unmounted;
    (void) ui->UmountAll(force, &totalMounted, &unmounted);
    delete ui;
    // All detailed error messages are already shown inside.
    if (!totalMounted)
    {
        BCOperationShowMessage(kBCTool_Info, "There are no containers to process.");
    }
    else if (totalMounted != unmounted)
    {
        BCOperationShowErrorString("Some of your containers are still unmounted.\nIf you really need to remove them right now, please reboot your computer.");
    }
    else
    {
        BCOperationShowMessage(kBCTool_Info, "All containers are unmounted successfully.");
    }
}

bool BCOperationUnlinkContainer(ContainerDescriptor *descriptor, bool force)
{
    descriptor->update();
    if (!descriptor->IsLinked())
        return true;

    std::string deviceName = descriptor->GetDeviceName().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->Unlink(deviceName.c_str(), force);
    delete ui;
    descriptor->UpdateMountState();
    return result;
}


bool BCOperationGetAlgList(QList<BCAlgInfo> *out_algList)
{
    Q_ASSERT(out_algList);

    BCAlgInfo *algs = NULL;
    size_t algsTotal;
    bc_cipher_result result = kBCCipher_Success;

    result = BCKernelListAlgs(&algs, &algsTotal);
    if (kBCCipher_Success != result)
    {
        if (kBCCipher_InternalError != result)
        {
            BCOperationShowErrorString(QObject::tr("Couldn't load list of supported ciphers: %0").arg(BCKernelGetErrorString(result)));
        }
        else
        {
            int error = errno;
            BCOperationShowErrorString(QObject::tr("Couldn't load list of supported ciphers: %0").arg(strerror(error)));
        }

        return false;
    }

    for (unsigned i = 0; i < algsTotal; i++)
    {
        out_algList->append(algs[i]);
    }

    BCKernelFreeAlgList(algs);
    return true;
}

bool BCOperationIsModeSupported(const BCAlgInfo& alg, bc_mode_id modeId)
{
    return (bool)BCKernelIsModeSupported(&alg, modeId);
}

QString BCOperationGetModeName(bc_mode_id modeId)
{
    return QString(BCKernelModeName(modeId));
}

bc_mode_id BCOperationGetModeId(QString modeName)
{
    std::string modeStr = modeName.toStdString();
    return BCKernelModeId(modeStr.c_str());
}


QString getDefaultMountPointFor(const QString &containerFilename)
{
    QFileInfo fi(containerFilename);
    QString mountLocation = QDir::homePath() + QDir::separator() + ".local/BestCrypt" + QDir::separator() + fi.baseName();
    return mountLocation;
}

void createMountPointDirectory(QString mountLocation)
{
    QDir d(mountLocation);
    if (! d.exists()) {
        d.mkpath(".");
        //QFile f1(d.filePath(".created_by_BestCrypt"));
        //f1.open( IO_Raw | IO_ReadWrite );
    }
}


bool BCOperationReencryptContainer(ContainerDescriptor *descriptor, QString algName, QString modeName)
{
    BCTool *ui = CreateUI();
    std::string algNameStr = algName.toStdString();
    std::string modeNameStr = modeName.toStdString();
    std::string path = descriptor->GetPath().toStdString();

    bool result = ui->Reencrypt(path.c_str(), algNameStr.c_str(), modeNameStr.c_str());

    delete ui;
    descriptor->update();
    return result;
}


bool BCOperationDeleteContainer(ContainerDescriptor *descriptor)
{
    std::string path = descriptor->GetPath().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->DeleteContainer(path.c_str());
    delete ui;
    descriptor->update();
    return result;
}


bool BCOperationFormat(ContainerDescriptor *descriptor, const QString& fsType)
{
    std::string path = descriptor->GetPath().toStdString();
    std::string fsTypeStr = fsType.toStdString();

    BCTool *ui = CreateUI();
    bool result = false; // ui->Format(path.c_str(), fsTypeStr.c_str(), true);
    delete ui;
    return result;
}


bool BCOperationFsck(ContainerDescriptor *descriptor, const QString& fsType_)
{
    std::string path = descriptor->GetPath().toStdString();
    std::string fsType = fsType_.toStdString();
    BCTool *ui = CreateUI();
    bool result = false; // ui->CheckFS(path.c_str(), fsType.c_str());
    delete ui;
    return result;
}
