#include <cstring>
#include <cerrno>
#include <unistd.h>

#include "bctool.h"


bool BCTool::WipeHeader(BCHelperContext &mainContext)
{
    if (! isOKForContainerFile(mainContext))
        return false;

    bool agree = AskUser("The key blok of the container will be permanently wiped. "
                         "Please make double sure that you have a backup copy of the key block.\n"
                         "Please note that container with encrypted or wiped out key blocks can't be reliably identified as BestCrypt containers.\n\n"
                         "Do you still want to continue the key block wiping?", false);

    if (! agree)
        return false;

    mainContext.pwdGet->m_query = "Request password for main part";

    BCContainerResult cr = BCWipeHeader(mainContext);

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Unable to wipe container's header: %s (code %d)", ContainerResultString(cr), cr);
        return false;
    }

    ShowInfo("Container header has been successfuly wiped!");

    volumeInfoUpdatedHook();

    return true;
}


bool BCTool::EncryptHeader(BCHelperContext &mainContext)
{
    mainContext.pwdGet->m_query = "Request password for main part";

    if (! prepareForOperation(mainContext)) {
        return false;
    }

    BCContainerResult cr = BCEncryptHeader(mainContext);

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Unable to encrypt container header: %s (code %d)", ContainerResultString(cr), cr);
        return false;
    }

    ShowInfo("Header was successfuly encrypted!");

    volumeInfoUpdatedHook();

    return true;
}

bool BCTool::DecryptHeader(BCHelperContext &mainContext)
{
    mainContext.pwdGet->m_query = "Request password for main part";

    if (! prepareForOperation(mainContext)) {
        return false;
    }

    BCContainerResult cr = BCDecryptHeader(mainContext);

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Unable to decrypt container header: %s (code %d)", ContainerResultString(cr), cr);
        return false;
    }

    ShowInfo("Header was successfuly decrypted!");

    volumeInfoUpdatedHook();

    return true;
}


bool BCTool::BackupHeader(BCHelperContext &mainContext, BCToolHeaderParams *headerParams)
{
    if(BCTool::PathExists(headerParams->headerFilepath.c_str())) {
        ShowError("File '%s' already exists\n", headerParams->headerFilepath.c_str());
        return false;
    }

    if (! prepareForOperation(mainContext)) {
        return false;
    }

    BCFileChannel fch_back(headerParams->headerFilepath.c_str());
    BCContainerResult cr = BCBackupHeader(mainContext, &fch_back);

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Unable to store key block backup: %s (code %d)", ContainerResultString(cr), cr);
        return false;
    }

    ShowInfo("Container key block backup was successfuly stored!");

    return true;
}


bool BCTool::RestoreHeader(BCHelperContext &mainContext, BCToolHeaderParams *headerParams)
{
    if (! isOKForContainerFile(mainContext))
        return false;

    // we must show Restore header dialog before enter this method to ask kbb file path
    if(! BCTool::PathExists(headerParams->headerFilepath.c_str())) {
        ShowError("Key block backup file '%s' does not exist\n", headerParams->headerFilepath.c_str());
        return false;
    }

    BCFileChannel fch_back(headerParams->headerFilepath.c_str());
    BCContainerResult cr = BCRestoreHeader(mainContext, &fch_back, true /* prepare */);
    mainContext.outVolume->should_free_key = 1;

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Unable to restore key block backup: %s (code %d)", ContainerResultString(cr), cr);
        return false;
    }

    BCToolMountParams mountParams;
    mountParams.filename = headerParams->containerFilename;
    mountParams.mountReadOnly = true;

    char mpBuf[4096];
    strcpy(mpBuf, "/tmp/testmount_XXXXXX");

    if (mkdtemp(mpBuf) == 0) {
        ShowError("Unable to create temporary directory for mount point: %s", strerror(errno));
        return false;
    }
    mountParams.mountPoint = mpBuf;


    if (! Mount(mainContext, &mountParams)) {
        ShowError("Unable to restore the header: Test mount attempt failed");
        return false;
    }

    if (! Umount(mountParams.mountPoint.c_str(), false)) {
        ShowWarning("The container with specified header file was mounted for test purpose"
        " but unmount wasn't successfull. Container's key block will be restored anyway.");
    }

    if (rmdir(mountParams.mountPoint.c_str()) != 0) {
        ShowWarning("Unable to delete temporary directory which was created for test mount.");
    }

    cr = BCRestoreHeader(mainContext, &fch_back, false /* not preparation, doing real job now */);

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Unable to restore key block backup: %s (code %d)", ContainerResultString(cr), cr);
        return false;
    }

    ShowInfo("Container key block backup was successfuly restored!");

    return true;
}


/**
 * @brief Set new container description
 * @param mainContext
 * @param new_description is utf-8 string with a description
 * @return true on success
 */
bool BCTool::ChangeDescription(BCHelperContext &mainContext, std::string new_description)
{
    mainContext.pwdGet->m_query = "Request password for main part";

    if (! prepareForOperation(mainContext)) {
        return false;
    }

    BCContainerResult cr = BCUpdateDescription(mainContext, new_description.c_str());

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Unable to change container description: %s (code %d)", ContainerResultString(cr), cr);
        return false;
    }

    volumeInfoUpdatedHook();

    return true;
}

