#include "bctool.h"
#include "fslib/fslib.h"
#include "spacemap/spacemap.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include "container/BCFileChannel.h"


/**
 * @brief Parses filesystem image and get free space map from it
 * @param linked container device to reading freespace map from
 * @return extents chain
 */
SpaceMap *BCTool::GetFreeSpaceFromMainPart(std::string deviceName)
{
    SpaceMap *smap = 0;
    fs_extent_t *freeSpace = 0;
    fs_volume_t vol;

    int error = -1;

    char deviceToProcess[BC_MAX_PATH] = {0};

    // Search the first partition, if any. If not, use whole device.
    if (GetFirstDevicePartition(deviceName.c_str(), deviceToProcess, BC_MAX_PATH))
    {
        this->ShowError("Error getting device name: %s.\n", strerror(errno));
        return NULL;
    }

    // Try to parse volume.
    error = fslib_open_volume(deviceToProcess, O_RDWR, &vol);
    if (error)
    {
        this->ShowError("Couldn't parse container volume: %s. Possible reason is that the filesystem isn't supported.\n",
                        strerror(error));
        return NULL;
    }

    if (!vol.get_free_space)
    {
        this->ShowError("Getting of free space is not supported for this file system.\n");
        fslib_close_volume(&vol);
        return NULL;
    }

    // Get free space from volume.
    freeSpace = vol.get_free_space(&vol);

    fslib_close_volume(&vol);

    if (!freeSpace)
    {
        this->ShowError("Couldn't retrieve free space information from the container device.\n");
        return NULL;
    }

    // convert free_space_library extents to SpaceMap
    smap = SpaceMapCreate();
    assert(smap);
    fs_extent_t* p = freeSpace;
    while (p) {
        SpaceMapAddNew(smap, p->start, p->size);
        p = p->next;
    }

    // never use the first 10 MB of a container
    SpaceMap * beginArea = SpaceMapCreate();
    sm_result smres = SpaceMapAddNew(beginArea, 0, BCTool::minimalContainerSize);
    assert(smres == kSpaceMap_Success);
    smres = SpaceMapSubstract(smap, beginArea);
    assert(smres == kSpaceMap_Success);
    SpaceMapFree(beginArea);

exit:
    if (freeSpace)
        fs_put_extent_list(freeSpace);

    return smap;
}



/**
 * @brief Get free space map from main part of container
 * @param helperContext
 * @param mountParams
 * @return SpaceMap or 0 in case of error
 */
SpaceMap * BCTool::GetFreeSpaceFromMainPart(BCHelperContext &helperContext, BCToolMountParams *mountParams)
{
    if (! prepareForOperation(helperContext))
        return NULL;

    // TODO: can we get a free space map without actual 'linking' as a block device?

    // Link the main part of container.

    // TODO: set flag in mountParams like  'main_part_only = true'
    if (! this->Link(helperContext, mountParams)) {
        return 0; // All errors are already displayed by callbacks
    }

    // Get free space map.
    SpaceMap *map = this->GetFreeSpaceFromMainPart(mountParams->device_name);

    // Do not check return value, all errors are already displayed inside.
    this->Unlink(mountParams->device_name, false);

    return map;
}


/**
 * Calling this method is tricky.
 * Before a call you must assing to helperContext.header a pointer
 * to the header of container's "main part".
 * After returning from this call you must zero it to avoid freeing.
 * @return true if existing hidden part protected and spacemap was modified
 */
bool BCTool::ProtectExistingHiddenPart(BCHelperContext &mainContext, cPwdGet *hiddenPwdGet, SpaceMap *freespace)
{
    BCHelperContext helperContext(NULL, kBCFileMode_ReadWrite, hiddenPwdGet);

    // pass main part header to procedure of decoding hidden part
    helperContext.header = mainContext.header;

   BCContainerResult cr;
    do {
        cr = BCDecodeVolume(helperContext);
        helperContext.outVolume->should_free_key = 1;

        if (cr == kBCContainer_Success) {
            if (hiddenPwdGet->isMainVolumePassword()) {
                this->ShowError("The password you've entered is not from the existing hidden part of the main volume. Please try again.");
                helperContext.pwdGet->clear();
            } else {
                break;
            }
        } else {
            break;
        }
    } while (cr != kBCContainer_Cancelled);

    // prolong 'header' life beyond helperContext.
    // TODO: use 'm_adoptedHeader' property of BCHelperContext to control header's life time
    helperContext.header = NULL;

    // delete helperContext.progressActor;
    // helperContext.progressActor = NULL;

    if (cr == kBCContainer_Cancelled) {
        return false;
    }

    if (cr != kBCContainer_Success) {
        this->ShowError("Unable to check hidden volume space map: code %d", cr);
        return false;
    }

    if (! helperContext.outVolume->info.map) {
        this->ShowError("Unable to check hidden volume space map");
        return false;
    }

    if (helperContext.outVolume->info.map->count > 1) {
        SpaceMapSort(helperContext.outVolume->info.map);
    }

    // Check if we have anough space to shift first extent to take into account space map of hidden part
    if (helperContext.outVolume->info.map->start->extent.start <= 4096) {
        this->ShowError("Internal error: 1525");
        // TODO:  [self _clearBeforeFinish];
        return false;
    }

    // Shift extent to the size of kBCV8_HiddenPartSpaceMapSize
    helperContext.outVolume->info.map->start->extent.start -= 4096;
    helperContext.outVolume->info.map->start->extent.length += 4096;

    if (kSpaceMap_Success != SpaceMapSubstract(freespace, helperContext.outVolume->info.map)) {
        this->ShowError("Unable to calculate hidden volume size. You can try to run hidden part creation without protecting exisitng hidden part");
        // [self _clearBeforeFinish];
        return false;
    }

    return true;
}

/**
 * @brief Create new hidden part inside existing container
 * @param helperContext with CPwd_Get for creation a container
 * @param hiddenPartExtents Space map for new hidden part.\
 * It must have length and structure that is exactly describes hidden part to create.
 * Important! Space map must be alligned to the volume sector size.
 * Before calling this function user must set seed buffer with BCSetSeedBuffer.
 */
bool BCTool::CreateHiddenPart(BCHelperContext &mainVolumeHelper, cPwdGet *hiddenNewPwdGet, SpaceMap *spaceMap, BCToolCreateParams *createParams, BCToolMountParams *mountParams)
{
    if (! isOKForContainerFile(mainVolumeHelper))
        return false;

    SpaceMap * allignedMap = BCSpaceMapByTrimming(spaceMap, createParams->hiddenPartSize);
    if (! allignedMap) {
        this->ShowError("Not enought free space left for one more hidden part");
    }

#if 0
    // TODO
    if (bytesAvailable < BCTool::minimalContainerSize) {
        this->ShowError("Not enought free space left for one more hidden part");
        // [self _clearBeforeFinish];
        return false;
    }
#endif

    bc_u8 seed[128];  // keygens/kgghost/kgghost.h  kgghost::kSeedLength
    // seed buffer should be filled with strong random numbers
    if (! generateRandom(true, seed, sizeof(seed))) {
        this->ShowError("Unable to generate random data");
        return false;
    }
    BCSetSeedBuffer(seed, sizeof(seed));

    BCContainerResult cr = BCCreateHiddenPart(mainVolumeHelper, hiddenNewPwdGet, allignedMap);
    // clean-up
    memset(seed, 0, sizeof(seed));
    SpaceMapFree(allignedMap);

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Hidden volume cannot be created: code %d", cr);
        return false;
    }

    // set a password to newly created hidden part
    BCHelperContext helperContext(mainVolumeHelper.channel, kBCFileMode_ReadWrite, hiddenNewPwdGet);

    cr = BCDecodeVolume(helperContext);
    helperContext.outVolume->should_free_key = 1;

    if (cr == kBCContainer_Cancelled) {
        this->ShowInfo("Canceled on setting new password");
        return false;
    }

    if (cr != kBCContainer_Success) {
        ShowError("Unable to decode new hidden: code %d", cr);
        return false;
    }

    // TODO: show status (progress message)
    ShowDebug("Formatting new volume");

    if (! DoLink(helperContext, mountParams)) {
        ShowError("Failed to link the created hidden part.");
        return false;
    }

    ShowDebug("Formatting linked hidden part %s", mountParams->device_name.c_str());

    int formatResult = this->FormatVolume(mountParams->device_name, createParams->fsType.c_str(), createParams->volumeLabel.c_str());
    if (formatResult != 0)
    {
        this->ShowWarning("mkfs returned error %i. Hidden part may remain unformatted. Please check the log.", formatResult);
        return false;
    }
    ShowDebug("Formatting OK");

    if (createParams->mount_after_creation) {
        Mount(helperContext, mountParams);
    }
    else {
        Unlink(mountParams->device_name, false);
    }

    return true;
}
