#include "bctool.h"
#include "container/file.h"
#include "container/helpers.h"
#include "spacemap/spacemap_extra.h"

#include <unistd.h>

bool BCTool::DoLink(BCHelperContext &hctx, BCToolMountParams *mountParams)
{
    bc_cipher_result kernelResult;

    assert(! mountParams->filename.empty());
    const char *filename = mountParams->filename.c_str();

    BCVolume *volume = hctx.outVolume;

    int options = mountParams->mountReadOnly? kBCDisk_ReadOnly : 0;

    char device_name_buf[4096];
    if (volume->info.map) {
        kernelResult = BCKernelSetupDisk(filename,
                              volume->info.alg, volume->info.mode, volume->key_handle,
                              volume->info.map, options, device_name_buf);
    }
    else {
        // workaround for sha keygen
        SpaceMap *map = SpaceMapCreate();
        SpaceMapAddNew(map, volume->info.offset, volume->info.size);
        kernelResult = BCKernelSetupDisk(filename,
                              volume->info.alg, volume->info.mode, volume->key_handle,
                              map, options, device_name_buf);
        SpaceMapFree(map);
    }
    mountParams->device_name = device_name_buf;

    if (kBCCipher_Success != kernelResult)
    {
        this->ShowError("Could not link container file %s to device. Error while creating block device: %s\n"
                        ,filename, BCKernelGetErrorString(kernelResult));
        return false;
    }

    // Send BLKRRPART ioctl to our device for kernel to rescan it's partition table.
    if (this->RunProcessL("/sbin/bcrrpart_helper", mountParams->device_name.c_str(), ARG_TERM))
    {
        this->ShowWarning("Could not rescan partitions on device. BLKRRPART ioctl failed.\n");
    }

    // I/O error during reading freespacemap from main part. Hope it helps
    sleep(1);

    return true;
}


bool BCTool::Link(BCHelperContext &helperContext, BCToolMountParams *mountParams)
{
    assert(! mountParams->filename.empty());

    if (! mountParams->device_name.empty())
        return true; // already mounted

    const char *file = mountParams->filename.c_str();

    if (! isOKForContainerFile(helperContext))
        return false;

    BCContainerResult res = BCDecodeVolume(helperContext);
    helperContext.outVolume->should_free_key = 1;

    if (res != kBCContainer_Success)
    {
        if (res != kBCContainer_Cancelled && res != kBCContainer_InvalidPassword)
        {
            this->ShowError("Could not parse container file %s: %s\n", file,
                            BCContainerErrorDecription(res));
        }
        return false;
    }

    // Check container file size to prevent backup mounting.
    if ((!helperContext.outVolume->info.hidden)
            && (!BCTool::IsBlockSpecial(file)) // ignore condition for block devices.
            && (
                (BCFileSize(file) != helperContext.outVolume->info.size) // for containers created with bcrypt 1.8-0, the container size is equal to the file size.
                && (BCFileSize(file) < (helperContext.outVolume->info.offset + helperContext.outVolume->info.size))
               ))
    {
        this->ShowError("File size is less than the size specified in a container header. Possibly you are trying to mount a container header backup.\n");
        return false;
    }
    else if (BCTool::IsBlockSpecial(file)
             && (helperContext.outVolume->info.size == BCFileSize(file))
             && (!helperContext.outVolume->info.hidden))
    {
        // Deal with bug in previous versions of BestCrypt here.
        // Version 1.9-2 stored a whole size of the block device in header file,
        // without excluding of the header size, so we should modify the value
        // before sending it to a driver.

        // Bug results in IO error when dealing with last 8 blocks.
        //
        // The exact solution should:
        // 1) Write new, right size to the container header.
        // 2) Explain how to deal with block device that changed it's size.
        //
        // The solution below simply fixes the size of the container before
        // sending it to the kernel.

        // Old versions had kgSHA as default keygen, so we can remove only 4096KB.
        helperContext.outVolume->info.size -= 4096;

        sm_result smr;
        SpaceMap *modifiedMap;

        smr = SpaceMapGetMapOfSize(helperContext.outVolume->info.map, helperContext.outVolume->info.size, &modifiedMap);
        if (kSpaceMap_Success != smr)
        {
            this->ShowError("Could not reallocate space map while linking: %s. Please contact our technical support.\n", SpaceMapGetErrorString(smr));
            return false;
        }

        SpaceMapFree(helperContext.outVolume->info.map);
        helperContext.outVolume->info.map = modifiedMap;
    }

    bool result = this->DoLink(helperContext, mountParams);

    return result;
}

bool BCTool::Unlink(std::string deviceName, bool force)
{
    if (!BCTool::IsBlockSpecial(deviceName.c_str()))
    {
        this->ShowError("%s is not a block device.\n", deviceName.c_str());
        return false;
    }

    int error = BCKernelRemoveDisk(deviceName.c_str(), force? 1 : 0);

    if (error)
    {
        this->ShowError("Could not remove device %s: %s\n", deviceName.c_str(), BCKernelGetErrorString(error));
        return false;
    }

    return true;
}
