#include "bctool.h"

#include <unistd.h>
#include <sys/types.h>
#include <glob.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "container/helpers.h"

/// @brief If mounted container has partitions, gets path to device representing first device partition,
///        otherwise gets main device name.
/// @param maxNameSize Maximum length of both mainDeviceName and out_mountDeviceName strings.
int BCTool::GetFirstDevicePartition(const char *mainDeviceName, char *out_partitionDeviceName, int maxNameSize)
{
    assert(out_partitionDeviceName);

    char searchPattern[BC_MAX_PATH];
    // Create search pattern.
    snprintf(searchPattern, maxNameSize, "%s?", mainDeviceName);

    glob_t searchResults;
    int globResult = glob(searchPattern, 0,  NULL, &searchResults);
    int result = 0;

    do
    {
        // If nothing was found, use main device name.
        if (globResult)
        {
            if (globResult == GLOB_NOMATCH)
            {
                strncpy(out_partitionDeviceName, mainDeviceName, maxNameSize);
                break;
            }

            // Errno will be set in this case.
            result = -1;
            break;
        }

        assert(searchResults.gl_pathc);

        // Glob results are sorted, so simply get the first result.
        strncpy(out_partitionDeviceName, searchResults.gl_pathv[0], maxNameSize);
    }
    while (0);

    globfree(&searchResults);
    return result;
}

bool BCTool::Mount(BCHelperContext &mainContext, BCToolMountParams *mountParams)
{
    // TODO: license check here
    // if ([licence isDemo] && [licence isExpired]) {
    //     mountParams->readOnly = true;
    // }

    if (! isOKForContainerFile(mainContext))
        return false;

    // Show mount dialog. User selects mount point there, and 
    // we try to 'Link' the encrypted device first
    bool result = this->Link(mainContext, mountParams);
    if (!result)
        return false;

    result = DoMount(mountParams);

    // Do device unlink in case of an error.
    if (!result)
    {
        this->Unlink(mountParams->device_name, false);
    }

    return result;
}


//! Do mount. Precondition: device is already linked
bool BCTool::DoMount(BCToolMountParams *mountParams)
{
    assert(! mountParams->device_name.empty());

    // now extract parameters from mount dialog
    std::string mountPoint = mountParams->mountPoint;

    if (! PathExists(mountPoint.c_str()))
    {
        // TODO: create it here???
        this->ShowError("Mount point %s does not exist.\n", mountPoint.c_str());
        return false;
    }
    else if (! IsDirectory(mountPoint.c_str())) {
        this->ShowError("Mount point %s is not a directory.\n", mountPoint.c_str());
        return false;
    }

    if (geteuid() /* root owns everything */ && !IsOwner(mountPoint.c_str()))
    {
        this->ShowError("You are not the owner of %s.\n", mountPoint.c_str());
        return false;
    }

    // Get the name of device representing first partition.
    char partitionDeviceName[BC_MAX_PATH];
    int error = GetFirstDevicePartition(mountParams->device_name.c_str(), partitionDeviceName, BC_MAX_PATH);
    if (error)
    {
        this->ShowError("Error while getting device for mounting: %s", strerror(errno));
        return false;
    }

    if (! callMountCmd(partitionDeviceName, mountParams)) {
        return false;
    }

    return true;
}


bool BCTool::callMountCmd(const char *partitionDeviceName, BCToolMountParams *mountParams)
{
    // Run our suid helper to mount device.
    const char *args[0xff] = { NULL };
    const char** parg = args;
    *parg++ = "/sbin/bcmount_helper";
    *parg++ = partitionDeviceName;
    *parg++ = mountParams->mountPoint.c_str();
    
    if (mountParams->mountReadOnly)
    {
        *parg++ = "-r";
    }
    
    if (! mountParams->fsType.empty())
    {
        *parg++ = "-t";
        *parg++ = mountParams->fsType.c_str();
    }
    
    if (! mountParams->mountOptions.empty())
    {
        *parg++ = "-o";
        *parg++ = mountParams->mountOptions.c_str();
    }
    
    char uidString[10], gidString[10];
    
    if (mountParams->uid || mountParams->gid)
    {
        if (geteuid())
        {
            this->ShowError("Only root can specify mount point's owner.\n");
            return false;
        }
        
        if (mountParams->uid)
        {
            snprintf(uidString, 10, "%i", mountParams->uid);
            *parg++ = "-u";
            *parg++ = uidString;
        }
        
        if (mountParams->gid)
        {
            snprintf(gidString, 10, "%i", mountParams->gid);
            *parg++ = "-g";
            *parg++ = gidString;
        }
    }

    int exit_code = RunProcessV("/sbin/bcmount_helper", args);
    
    if (exit_code != 0)
    {
        if (32 == exit_code)
            this->ShowError("File system is not recognized.\n");
        else
            this->ShowError("bcmount_helper failed with exit code %d", exit_code);

        return false;
    }

    return true;
}

