#include "bctool.h"

#include <signal.h>
#include <limits.h>
#include <unistd.h>

#include "container/file.h"
#include "container/container.h"
#include "container/BCFileChannel.h"
#include "keygens/kgghost/kgghost.h"
#include "FilesystemUtils.h"

#define ROUND_UP(value, blocksize) ( ((value) + (blocksize) - 1) & ~((blocksize) - 1) )


//! Temporary set signal handlers
class BCTool::SignalHandlerSentry {
public:
    SignalHandlerSentry(SignalCallback cb) {
        // Install interrupt handlers and create container.
        // TODO: review and change to sigaction().
        intHandler = signal(SIGINT, cb);
        quitHandler = signal(SIGQUIT, cb);
        termHandler = signal(SIGTERM, cb);
        abortHandler = signal(SIGABRT, cb);
    }
    ~SignalHandlerSentry() {
        // Restore signal handlers before return.
        signal(SIGINT, intHandler);
        signal(SIGQUIT, quitHandler);
        signal(SIGTERM, termHandler);
        signal(SIGABRT, abortHandler);
    }
private:
    sighandler_t intHandler;
    sighandler_t quitHandler;
    sighandler_t termHandler;
    sighandler_t abortHandler;
};


static int progress_caller(bc_u64 nbytes, bc_u64 total, void *arg)
{
    BCToolCreationProgress *p = reinterpret_cast<BCToolCreationProgress*>(arg);
    return (p)? p->func(nbytes, total) : 0;
}


bool BCTool::Create(BCHelperContext *mainContext, BCToolCreateParams *createParams, BCToolMountParams *mountParams, BCToolCreationProgress *progress)
{
    BCContainerContext &cctx = mainContext->pwdGet->getContainerContext();

    if (this->IsBlockSpecial(cctx.path))
    {
        if (cctx.size != 0)
        {
            this->ShowWarning("File %s is a block device, specified size will be ignored.", cctx.path);
        }

        cctx.size = BCFileSize(cctx.path);
        if (! cctx.size)
        {
            this->ShowError("Error getting size of the block device.\n");
            return false;
        }
    }
    else if (this->IsFile(cctx.path)) {
        this->ShowError("Unable to create container: File %s already exists.\n", cctx.path);
        return false;
    }

    if (cctx.size == 0)
    {
        this->ShowError("Please specify container size.\n");
        return false;
    }
    else if (cctx.size % 512) // Check that size is a multiple to block size.
    {
        ROUND_UP(cctx.size, 512);
        this->ShowWarning("Container size is not even to block size, will round up to %llu bytes.\n", (long long unsigned)cctx.size);
    }

    if (cctx.size < minimalContainerSize) {
        this->ShowError("Container size is too small. Minimum value is 10 Mb"); //  StringUtil::humanSizeString(minimalContainerSize)
        return false;
    }

    uint64_t freespace_bytes = FilesystemUtils::getFreeSpace(cctx.path);

    if (freespace_bytes > 0 && cctx.size > freespace_bytes) {
        this->ShowError("Container size is greater than available free space.");
        return false;
    }

    // TODO: use it for console application on caller's side (main.cc probably)
//    SignalHandlerSentry sig_sentry(this->HandleSignal);

    bool result = true;

    bc_u8 _seed[128];  // keygens/kgghost/kgghost.h  kgghost::kSeedLength

    // seed buffer should be filled with strong random numbers
    if (! generateRandom(true, _seed, sizeof(_seed))) {
        this->ShowError("Unable to generate random data");
        return false;
    }

    BCSetSeedBuffer(_seed, sizeof(_seed));

    try {
        BCContainerResult res = BCCreateContainer(*mainContext);
        if (res != kBCContainer_Success)
        {
            if (res == kBCContainer_IOError)
            {
                this->ShowError("Error while creating container: %s\n"
                                "Hint: check there is enought disk space.\n",
                                BCContainerErrorDecription(res));
            }
            else
            {
                this->ShowError("Error while creating container: %s\n", BCContainerErrorDecription(res));
            }
            return false;
        }
    }
    catch(CipherException &e) {
        ShowError("BCCreateContainer have thrown an exception: %s\n", e.what());
        if (e.code() == kBCCipher_NoMemory) {
            ShowInfo("Key slots in kernel module are exhausted");
        }
        return kBCContainer_InternalError;
    }

    if (cctx.encryptHeader) {
        mainContext->encryptHeader();
    }

    mainContext->outVolume->channel = mainContext->channel;
    mainContext->outVolume->should_free_key = 1;

    if (kBCContainer_Success != mainContext->header->store(mainContext->channel)) {
        BCFileDelete(cctx.path);
        return kBCContainer_InternalError;
    }

    if (kBCContainer_Success != BCRandomizeVolumeBody(mainContext->outVolume, &progress_caller, (void*)progress)) {
        BCFileDelete(cctx.path);
        return false;
    }

    // Format new volume if user requested.
    if (! createParams->fsType.empty())
    {
        char deviceName[PATH_MAX] = { 0 };
        bc_cipher_result error = BCKernelSetupDisk(cctx.path, mainContext->outVolume->info.alg, mainContext->outVolume->info.mode,
                                   mainContext->outVolume->key_handle, mainContext->outVolume->info.map, kBCDisk_Default, deviceName);
        if (kBCCipher_Success != error)
        {
            this->ShowError("Could not create device with a new container: %s\n", BCKernelGetErrorString(error));
            return false;
        }

        int exit_code = this->FormatVolume(deviceName, createParams->fsType.c_str(), createParams->volumeLabel.c_str());
        if (exit_code)
        {
            this->ShowError("Could not format volume, mkfs returned %d.\n", exit_code);
            return false;
        }
        sync();

        // optionally, mount a container
        if (createParams->mount_after_creation) {
            mountParams->device_name = deviceName;
            DoMount(mountParams);
            return true;
        }

        error = BCKernelRemoveDisk(deviceName, 0);
        if (error)
        {
            // Show error but continue, we have already created the container.
            this->ShowError("Could not remove block device: %s. Container remains linked to device.\n",
                            BCKernelGetErrorString(error));
        }
    }
    else
    {
        this->ShowWarning("Please do not forget to format the container.\n");
    }

    return true;
}
