#include <string>
#include <vector>
#include <cstdio>
#include <cstdlib>
#include <climits>
#include <cassert>
#include <cerrno>

#include <stdarg.h>
#include <unistd.h>
#include <stdint.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <mntent.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <linux/fs.h>

#include "bctool.h"
#include "container/volume.h"
#include "container/container.h"
#include "container/BCFileChannel.h"
#include "container/file.h"
#include "container/cipher_kernel.h"
#include "container/error.h"
#include "kernel/BCKernel.h"
#include "disk/disk.h"
#include "container/isaac.h"


BCTool::BCTool()
{
    this->minimumMsgLevel = kBCTool_Info;

    static CipherKernelModule globalCipher;
    CipherFactory::m_gCipher = &globalCipher;
}

BCTool::~BCTool()
{
}


bool BCTool::prepareForOperation(BCHelperContext &mainContext, cPwdGet *additionalPwdGet)
{
    if (! isOKForContainerFile(mainContext))
        return false;

    BCContainerResult result = BCPrepareContainerForOperation(mainContext, additionalPwdGet);

    if (result != kBCContainer_Success) {
        if (result != kBCContainer_Cancelled) {
            ShowError("Unable to open main volume: BCContainerResult is %d", result);
        }
        return false;
    }

    volumeInfoUpdatedHook();
    return true;
}


bool BCTool::WantChange(const char *file)
{
    if(!BCTool::PathExists(file)) {
        this->ShowError("File %s does not exist\n", file);
        return false;
    }

    if(!BCTool::CanWrite(file)) {
        this->ShowError("You do not have write permissions to file %s\n", file);
        return false;
    }

    return true;
}


bool BCTool::WantRead(const char *file)
{
    if(!BCTool::PathExists(file)) {
        this->ShowError("File %s does not exist\n", file);
        return false;
    }

    if(!BCTool::CanRead(file)) {
        this->ShowError("You do not have read permissions to file %s\n", file);
        return false;
    }

    return true;
}


bool BCTool::DeleteContainer(const char *container)
{
    if (!this->AskUser("After this operation container data will be unrecoverable. Are you sure you want to continue?", false))
        return false;

    if (!this->WantChange(container))
        return false;

    bc_file_handle handle;
    bc_file_result fileResult = BCFileOpen(container, kBCFileMode_ReadWrite, &handle);
    if (kBCFileResult_Success != fileResult)
    {
        this->ShowError("Error opening file %s: %s", container, strerror(errno));
        return false;
    }

    bc_u64 fileSize = BCFileSize(container);
    bc_u64 processSize = fileSize > 4096 ? 4096 : fileSize;

    char buf[1024] = { 0 };
    bc_u64 i = 0;
    bc_u64 bytesWritten = 0;
    bc_u64 bytesTotal = 0;
    unsigned t = (unsigned)time(NULL);
    const void *seed = (const void*)&t;
    unsigned int seedSize = sizeof(unsigned);
    void *isaacContext = IsaacInit(seed, seedSize);

    for (; i < processSize; i += bytesWritten)
    {
        IsaacGenerate(isaacContext, buf, sizeof(buf));
        bytesWritten = BCFileWrite(&handle, buf, sizeof(buf));
        bytesTotal += bytesWritten;

        if ((sizeof(buf) != bytesWritten) && (bytesTotal < fileSize))
        {
            this->ShowError("Error writing to a container file %s: %s", container, strerror(errno));
            BCFileClose(&handle);
            return false;
        }
    }

    IsaacRelease(isaacContext);
    BCFileClose(&handle);
    BCFileDelete(container);
    return true;
}


bool BCTool::AddPassword(BCHelperContext &mainContext, cPwdGet *addPwdGet, cPwdGet *newPwdGet)
{
    mainContext.pwdGet->m_query = "Request password for main part";
    addPwdGet->m_query = "Please enter existing hidden part password";
    newPwdGet->m_query = "Please enter new password";

    if (! prepareForOperation(mainContext, addPwdGet)) {
        return false;
    }

    cPwdGet * volumePasswordGetter = NULL;
    if (addPwdGet->isEmpty()) {
        // we will add a password to main part
        volumePasswordGetter = mainContext.pwdGet;
    } else {
        /* we will add a password to this hidden part.
         * Password for another hidden part may be overwritten (lost) during this operation. */
        ShowWarning("You are going to add a new password to an existing hidden part.\n"
                    "This operation will likely destroy one another hidden parts' password so you may lose "
                    "an alternative password of this hidden part or (much worse) "
                    "you will not be able to open another existing hidden part anymore.\n");
        volumePasswordGetter = addPwdGet;
    }

    /* TODO: to add a password to existing hidden part
     * with preserving other passwords for this or another hidden parts
     * it should be done like on 'creation new hidden part': with
     * another helperContext object with same 'header' pointer.
     */

    BCContainerResult cr = BCAddPassword(mainContext, volumePasswordGetter, newPwdGet);

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Password cannot be added: code %d", cr);
        return false;
    }

    ShowInfo("New password has been successfuly added!");

    return true;
}


bool BCTool::ChangePassword(BCHelperContext &mainContext, cPwdGet *addPwdGet, cPwdGet *newPwdGet)
{
    mainContext.pwdGet->m_query = "Request password for main part";
    addPwdGet->m_query = "Please enter existing volume password to change.";

    if (! prepareForOperation(mainContext, addPwdGet)) {
        return false;
    }

    cPwdGet * volumePasswordGetter = NULL;
    if (addPwdGet->isEmpty()) {
        newPwdGet->m_query = "Please enter new password for the main part";
        volumePasswordGetter = mainContext.pwdGet;
    } else {
        newPwdGet->m_query = "Please enter new password for the hidden part";
        volumePasswordGetter = addPwdGet;
    }

    BCContainerResult cr = BCChangePassword(mainContext, volumePasswordGetter, newPwdGet);

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr != kBCContainer_Success) {
        ShowError("Password cannot be changed: code %d", cr);
        return false;
    }

    ShowInfo("Password has been successfuly changed!");

    return true;
}


bool BCTool::RemovePassword(BCHelperContext &mainContext, cPwdGet *addPwdGet)
{
    mainContext.pwdGet->m_query = "Request password for main part";
    addPwdGet->m_query = "Please enter existing volume password to remove.";

    if (! prepareForOperation(mainContext, addPwdGet)) {
        return false;
    }

    BCContainerResult cr = BCRemovePassword(mainContext, addPwdGet);

    if (cr == kBCContainer_Cancelled)
        return false;

    if (cr == kBCContainer_LastKeyBlock) {
        bool agree = AskUser("Only one password left for volume. "
                "You are trying to remove last password for volume. "
                "It will only be possible to access the container data using backup keyblock."
                "Do you want to continue?", false);
        if (agree) {
            cr = BCRemovePassword(mainContext, addPwdGet, true);
        }
    }

    if (cr != kBCContainer_Success) {
        ShowError("Password cannot be removed: code %d", cr);
        return false;
    }

    ShowInfo("Password has been successfuly removed!");

    return true;
}


int BCTool::FormatVolume(std::string deviceName, const char* fsType, const char *fsLabel)
{
    // Run mkfs on linked device.
    const char* mkfs = "/sbin/mkfs";
    int error;

    if (! fsLabel) {
        fsLabel = "New volume";
    }

    //printf("Formatting device %s to %s\n", deviceName, fsType);

    if (0 == strcasecmp(fsType, "ext2") || 0 == strcasecmp(fsType, "ext3") || 0 == strcasecmp(fsType, "ext4"))
    {
        error = this->RunProcessL(mkfs, "-t", fsType, "-L", fsLabel, "-m", "0", deviceName.c_str(), ARG_TERM);
    }
    else if (0 == strcasecmp(fsType, "exfat"))
    {
        error = this->RunProcessL(mkfs, "-t", fsType, "-n", fsLabel, deviceName.c_str(), ARG_TERM);
    }
    else if (0 == strcasecmp(fsType, "vfat"))
    {
        error = this->RunProcessL(mkfs, "-t", fsType, "-n", fsLabel, "-v", deviceName.c_str(), ARG_TERM);
    }
    else if (0 == strcasecmp(fsType, "ntfs"))
    {
        error = this->RunProcessL(mkfs, "-t", fsType, "-L", fsLabel, "-v", deviceName.c_str(), ARG_TERM);
    }
    else if (0 == strcasecmp(fsType, "hfs"))
    {
        error = this->RunProcessL(mkfs, "-t", fsType, "-v", fsLabel, deviceName.c_str(), ARG_TERM);
    }
    else
    {
        error = this->RunProcessL(mkfs, "-t", fsType, deviceName.c_str(), ARG_TERM);
    }

    return error;
}

int BCTool::RunProcessL(const char* name, ...)
{
    char* args[1024] = { (char *)NULL };
    char** pargs = args;
    char** plast = &args[1023];

    *pargs++ = (char*)name;

    va_list list;
    va_start(list, name);

    while (NULL != (*pargs++ = va_arg(list, char*)))
    {
        assert(pargs != plast);
    }

    va_end(list);

    return this->RunProcessV(name, (const char**) args);
}

int BCTool::RunProcessV(const char *name, const char **args)
{
    {
        char buf[4096];
        memset(buf, 0, sizeof(buf));
        for (const char **p = args ; *p ; p++) {
            if (p != args)
                strncat(buf, " ", sizeof(buf));
            strncat(buf, *p, sizeof(buf));
        }
        ShowDebug("Running %s...", buf);
    }

    pid_t pid = fork();

    if (pid < 0)
    {
        this->ShowError("Fork failed: %s\n", strerror(errno));
        return -1;
    }

    if (0 == pid)
    {
        // child
        const char* env[] = { "PATH=/sbin:/usr/sbin:/bin:/usr/bin", ARG_TERM };

        execve(name, (char* const*)args, (char* const*)env);

        // not reached if ok
        this->ShowError("execve(%s) failed: %s\n", name, strerror(errno));
        return -1;
    }

    // parent
    int status;
    waitpid( pid, &status, 0);
    return WEXITSTATUS(status);
}

bool BCTool::CheckKernelInitialization()
{
    bc_cipher_result result = BCKernelIsInitialized();
    if (kBCCipher_Success != result)
    {
        this->ShowError("Kernel modules are not loaded."
                        "\nIt might occur if you have a new kernel version installed."
                        "\nTo be able to use BestCrypt with a new kernel version, please go to the root of BestCrypt sources and execute the following commands:"
                        "\n\tmake module-clean"
                        "\n\tmake module"
                        "\n\tmake module-install");
        return false;
    }

    return true;
}


bool BCTool::CheckContainerInUse(const char *containerPath)
{
    // If we can't check that, consider container free, because we've tried.

    char fullPath[PATH_MAX], tmpFullPath[PATH_MAX];

    if (!realpath(containerPath, fullPath))
    {
        this->ShowError("Container is in use. Please eject it before this action.");
        return false;
    }

    std::vector<bc_mount_entry> mounts;
    if (!this->GetMountInfo(&mounts))
    {
        this->ShowError("Container is in use. Please eject it before this action.");
        return false;
    }

    std::vector<bc_mount_entry>::iterator i = mounts.begin();
    for (; i != mounts.end(); i++)
    {
        if (realpath(i->mountPoint, tmpFullPath))
        {
            if (0 == strncmp(fullPath, tmpFullPath, PATH_MAX))
                return true;
        }
    }

    return false;
}


// TODO: arguments: helperContext, creationParams
bool BCTool::Format(const char *file, const char *fsType, cPwdGet *pwdGetObj, bool showWarning)
{
    if (!this->WantChange(file))
        return false;

    if (this->CheckContainerInUse(file))
    {
        return false;
    }




    // TODO:  just Link() ?

    BCVolume vol;
    BCFileChannel fch(file);
    bc_file_mode mode = kBCFileMode_ReadWrite | kBCFileMode_Exclude;
    BCHelperContext ctx(&fch, mode, pwdGetObj, &vol);

    BCContainerResult res = BCDecodeVolume(ctx);
    ctx.outVolume->should_free_key = 1;

    if (kBCContainer_Success != res)
    {
        if ((kBCContainer_Cancelled != res) && (kBCContainer_InvalidPassword != res))
        {
            this->ShowError("Could not decode container file.\n");
        }

        return false;
    }

    char deviceName[BC_MAX_PATH] = { 0 };
    bc_cipher_result error = BCKernelSetupDisk(file, vol.info.alg, vol.info.mode, vol.key_handle,
                              vol.info.map, kBCDisk_Default, deviceName);
    BCPutVolume(&vol);

    if (kBCCipher_Success != error)
    {
        this->ShowError("Could not setup new device: %s\n", BCKernelGetErrorString(error));
        return false;
    }







    if (showWarning)
    {
        if (!this->AskUser("Warning! After formatting container data will be destroyed! Are you sure you want to continue?", false))
        {
            return false;
        }
    }

    bool result = true;
    int err = this->FormatVolume(deviceName, fsType, "New volume");
    if (err)
    {
        this->ShowError("Could not format volume, mkfs returned %d\n", err);
        result = false;
    }

    error = BCKernelRemoveDisk(deviceName, 0);
    if (error)
    {
        this->ShowError("Could not remove block device: %s\n", BCKernelGetErrorString(error));
        result = false;
    }

    if (result)
    {
        this->ShowInfo("Format successful.\n");
    }

    return result;
}

bool BCTool::CheckFS(BCHelperContext &hctx)
{
#if 0
    if (!this->WantChange(file))
        return false;

    if (this->CheckContainerInUse(file))
    {
        return false;
    }


    // TODO: just Link() ?
    BCContainerResult res = BCDecodeVolume(hctx);
    hctx.outVolume->should_free_key = 1;

    if (kBCContainer_Success != res)
    {
        if ((kBCContainer_Cancelled != res) && (kBCContainer_InvalidPassword != res))
        {
            this->ShowError("Could not decode container file.\n");
        }
        return false;
    }

    char deviceName[PATH_MAX] = { 0 };
    bc_cipher_result error = BCKernelSetupDisk(file, vol.info.alg, vol.info.mode, vol.key_handle,
                              vol.info.map, kBCDisk_Default, deviceName);

    if (kBCCipher_Success != error)
    {
        BCPutVolume(&vol);
        this->ShowError("Could not setup new disk.\n");
        return false;
    }

    int exit_code;
    // If user did not specify fstype try to use implicit.
    if (fsType == (const char*)NULL)
    {
        exit_code = this->RunProcessL("/sbin/fsck", deviceName, ARG_TERM);
    }
    else
    {
        exit_code = this->RunProcessL("/sbin/fsck", "-t", fsType, deviceName, ARG_TERM);
    }

    bool succeeded = true;
    if (exit_code)
    {
        if (exit_code == 1)
            this->ShowWarning("Filesystem errors are corrected successfully.");
        else
        {
            this->ShowError("fsck failed with result %d. Check `man fsck` for more detail.\n", exit_code);
            succeeded = false;
        }
    }

    // Wait one second for reference to be closed.
    sleep(1);

    if (kBCCipher_Success != BCKernelRemoveDisk(deviceName, 0))
    {
        this->ShowError("Could not remove device. Your container remains linked to %s.\n", deviceName);
        succeeded = false;
    }

#endif
    return false; // succeeded;
}



void BCTool::ShowMessage(bctool_msg_level level, const char* msg, va_list args)
{
    const int MaxLen = 4096;
    char str[MaxLen] = { '\0' };
    vsnprintf(str, MaxLen, msg, args);
    this->messageCallback_(level, str);
}

void BCTool::ShowError(const char *msg, ...)
{
    va_list args;
    va_start(args, msg);
    this->ShowMessage(kBCTool_Error, msg, args);
    va_end(args);
}

void BCTool::ShowWarning(const char *msg, ...)
{
    va_list args;
    va_start(args, msg);
    this->ShowMessage(kBCTool_Warning, msg, args);
    //va_end(args);
}

void BCTool::ShowInfo(const char *msg, ...)
{
    va_list args;
    va_start(args, msg);
    this->ShowMessage(kBCTool_Info, msg, args);
    va_end(args);
}

void BCTool::ShowDebug(const char *msg, ...)
{
    va_list args;
    va_start(args, msg);
    this->ShowMessage(kBCTool_Debug, msg, args);
    va_end(args);
}


/**
 * @brief Get active crypted device list from kernel driver
 * @param out_entriesPtr - list of active crypted block devices 
 * @return true on success
 */
bool BCTool::GetMountInfo(std::vector<bc_mount_entry> *out_entriesPtr)
{
    assert(out_entriesPtr != NULL);
    std::vector<BCDiskRef> disks;
    BCDiskRef::listDisks(&disks);

    std::vector<BCDiskRef>::iterator i = disks.begin();
    std::vector<BCDiskRef>::iterator end = disks.end();

    while (i != end)
    {
        const char *deviceName = i->nodeName();

        // Look for device mount point.
        // Configured devices may not be actually mounted.
        FILE *mtab = setmntent("/etc/mtab", "r");
        if (NULL == mtab)
        {
            fprintf(stderr, "Error: Could not open mtab: %s\n", strerror(errno));
            return false;
        }

        const char *mountPoint = NULL;
        struct mntent *mountEntry = NULL;

        while ((mountEntry = getmntent(mtab)) && !feof(mtab))
        {
            if (0 == strncmp(deviceName, mountEntry->mnt_fsname, strlen(deviceName)))
            {
                mountPoint = mountEntry->mnt_dir;
                break;
            }
        }

        bc_mount_entry entry;
        memset(&entry, 0, sizeof(entry));

        strncpy(entry.deviceName, deviceName, BC_MAX_PATH);
        if (i->backingStore())
        {
            strncpy(entry.containerPath, i->backingStore(), BC_MAX_PATH);

            std::string containerPathTmp(entry.containerPath);

            if (std::string::npos == containerPathTmp.find(" (deleted)"))
                entry.isFileExists = true;
//            else
//                entry.isFileExists = false;

            if (mountPoint)
            {
                strncpy(entry.mountPoint, mountPoint, BC_MAX_PATH);
            }
//            else
//                entry.mountPoint[0] = '\0';
        }
//        else
//        {
//            entry.containerPath[0] = '\0';
//            entry.isFileExists = false;
//            entry.mountPoint[0] = '\0';
//        }
        out_entriesPtr->push_back(entry);
        endmntent(mtab);
        ++i;
    }

    return true;
}


// TODO: does multi-lib/container.h has this function?
bool BCTool::GetContainerInfo(const char *path, bc_container_info *out_info)
{
    assert(path);
    assert(out_info);

    // Get information about container.
    BCVolumeInfo volumeInfo;
    BCFileChannel fch(path);

    if (!BCParseContainerFile(&fch, &volumeInfo))
        return false;

    BCAlgInfo algInfo;
    if (kBCCipher_Success != BCKernelLookupAlgById(volumeInfo.alg, &algInfo))
    {
        this->ShowError("Could not find algorithm with id 0x%x.\n", volumeInfo.alg);
        return false;
    }

    // Store information in output buffer.
    strncpy(out_info->containerPath, path, BC_MAX_PATH);
    out_info->size = volumeInfo.size;
    out_info->algId = volumeInfo.alg;
    strncpy(out_info->algName, algInfo.name, BC_ALG_NAME_SIZE);
    out_info->modeId = volumeInfo.mode;
    strncpy(out_info->modeName, BCKernelModeName(volumeInfo.mode), BC_ALG_MODE_NAME_SIZE);
    strncpy(out_info->description, volumeInfo.description, BC_DESCRIPTION_SIZE);

    return true;
}

// static
bool BCTool::PathExists(const char *path)
{
    int error = access(path, F_OK);
    return !error;
}

// static
bool BCTool::CanRead(const char *path)
{
    int res = access(path, R_OK);
    return res == 0;
}

// static
bool BCTool::CanWrite(const char* path)
{
    int error = access(path, W_OK);
    return !error;
}

// static
bool BCTool::IsFile(const char* path)
{
    struct stat st;
    if(0 != stat(path, &st))
    {
        return false;
    }

    return S_ISREG(st.st_mode);
}

// static
bool BCTool::IsDirectory(const char* path)
{
    struct stat st;
    if(0 != stat(path, &st))
    {
        return false;
    }

    return S_ISDIR(st.st_mode);
}

// static
bool BCTool::IsBlockSpecial(const char* path)
{
    struct stat st;
    if(0 != stat(path, &st))
    {
        return false;
    }

    return S_ISBLK(st.st_mode);
}

// static
bool BCTool::IsOwner(const char* path)
{
    struct stat st;
    if(0 != stat(path, &st))
    {
        return false;
    }

    return (st.st_uid == getuid());
}


bool BCTool::isOKForContainerFile(BCHelperContext &helperContext)
{
    BCFileChannel *fc = dynamic_cast<BCFileChannel*>(helperContext.channel);
    if (!fc) {
        this->ShowError("Internal error: unable to cast 'channel' to 'BCFileChannel'");
        return false;
    }

    const char *fileName = fc->getFullPath().c_str();
    bc_file_mode wish_mode = helperContext.mode;

    // Check file name.
    if (strlen(fileName) == 0)
    {
        this->ShowError("File name cannot be empty");
        return false;
    }

    // Check file.
    if (PathExists(fileName))
    {
        if (IsBlockSpecial(fileName))
        {
            if (!this->AskUser("Specified file is block device. All data on it will be lost. "
                               "Are you sure you want it? (y/[n]):", false))
            {
                return false;
            }
        }
    }

    if (IsDirectory(fileName)) {
        this->ShowError("Specified path is a directory");
        return false;
    }

    if (! CanRead(fileName)) {
        this->ShowError("Specified file is not readable");
        return false;
    }

    // Check write permissions if requested.
    if(wish_mode & kBCFileMode_ReadWrite && ! CanWrite(fileName))
    {
        this->ShowError("You do not have write permissions to this file");
        return false;
    }

    if (wish_mode | kBCFileMode_Exclude && this->CheckContainerInUse(fileName))
    {
        this->ShowWarning("This container is already in use");
        return false;
    }

    return true;
}


// static
bc_u64 BCTool::GetBlockDeviceSize(const char* bdev)
{
    int fd = open(bdev, O_RDONLY);

    if (fd < 0)
    {
        perror("open");
        return 0;
    }

    unsigned long long size;
    if (0 != ioctl(fd, BLKGETSIZE, &size))
    {
        perror("ioctl");
        return 0;
    }

    size *= S_BLKSIZE;
    return size;
}

// static
int BCTool::GetUidByName(const char* name)
{
    struct passwd *pass;
    if ( NULL == name ) return -1;

    uid_t uid = atol(name);

    if( (pass = getpwnam(name)) || (pass = getpwuid(uid)) )
    return pass->pw_uid;

    return -1;
}

// static
int BCTool::GetGidByName(const char* name)
{
    struct group *grp;
    if ( NULL == name ) return -1;

    gid_t gid = atol(name);

    if( (grp = getgrnam(name)) || (grp = getgrgid(gid)) )
    return grp->gr_gid;

    return -1;
}

// static
bool BCTool::IsBCMount(const char* mountPoint)
{
    struct stat st;
    if (0 != stat(mountPoint, &st))
    {
        return false;
    }

    int blockMajor = BCDiskRef::blockMajor();
    return (blockMajor == (int)major(st.st_dev));
}


// TODO: to multi-lib
const char * BCTool::ContainerResultString(BCContainerResult rc)
{
    switch(rc) {
    case kBCContainer_Success:          return "Operation successfuly completed";
    case kBCContainer_InvalidArgument:  return "Invalid function argument";
    case kBCContainer_InvalidAlgorithm: return "Invalid encryption algorithm";
    case kBCContainer_InvalidPassword:  return "Incorrect password";
    case kBCContainer_Damaged:          return "Container is damaged";
    case kBCContainer_NoKeySpace:       return "No space left for new password";
    case kBCContainer_IOError:          return "Container file read/write error. Please check file permissions.";
    case kBCContainer_NoMemory:         return "Memory allocation error";
    case kBCContainer_Unsupported:      return "Operation not supported";
    case kBCContainer_LastKeyBlock:     return "Only one password left for volume";
    case kBCContainer_InternalError:    return "Internal error occured";
    case kBCContainer_UnknownFormat:    return "Unknown container format";
    default:                            return "Unknown container error";
    };
    // XXX
}

// list should be in sync with bcrypt-mac-git/mac/create/BCCreateContainerWizardController.m
bool BCTool::isDeprecatedEncryptionAlgorithm(const char *alg_name)
{
    if (strcmp(alg_name, "IDEA") == 0
     || strcmp(alg_name, "DES") == 0
     || strcmp(alg_name, "CAST") == 0
     || strcmp(alg_name, "BLOWFISH") == 0
     || strcmp(alg_name, "BLOWFISH-128") == 0
        )
            return true;

    return false;
}
