#ifndef BCTOOL_H
#define BCTOOL_H

#include <vector>
#include <string>
#include <stdint.h>
#include <stdarg.h>

#include "container/user.h"
#include "container/volume.h"
#include "container/keygen.h"
#include "container/BCFileChannel.h"
#include "container/cPwdGet.h"
#include "kernel/BCKernel.h"


#define BC_ALG_MODE_NAME_SIZE   4
#define BC_ALG_NAME_SIZE        16

#define ARG_TERM                ((char*)0)

// Enable printf-like parameter check for function in class (first parameter is 'this' pointer).
#define PRINTF_LIKE __attribute__((format(printf, 2, 3)))

#define _1MB    uint64_t((1024 * 1024))

extern const char *PROG_NAME;

typedef enum
{
    kBCTool_Debug = 0,
    kBCTool_Info,
    kBCTool_Warning,
    kBCTool_Error
} bctool_msg_level;

typedef void    (*SignalCallback)(int signal);


typedef std::vector<std::string> strings_t;

/// @brief Information about mounted containers from kernel driver
typedef struct
{
    char deviceName[BC_MAX_PATH];           ///< BestCrypt device name, ex. /dev/bcrypta
    char containerPath[BC_MAX_PATH];        ///< Container file linked to device.
    char mountPoint[BC_MAX_PATH];           ///< Container mount point. Contains empty string for containers linked without mounting.
    bool isFileExists;                      ///< True, if container file is still exists on disk, otherwise false.
} bc_mount_entry;


/// @brief Container mount params
struct BCToolMountParams
{
    std::string filename;
    std::string device_name;
    std::string mountPoint;         ///< User-specified mount point.
    bool mountReadOnly;             ///< Will container need to be mounted as read-only?
    bool useDefaultMountPoint;      ///< Should we use user-specified mount point or default (ignored on Linux).
    std::string fsType;             ///< Explicitly specify '-t' argument for 'mount' command
    std::string mountOptions;       ///< Additional arguments for 'mount' command
    int uid;                        ///< Specify 'uid' option for mount
    int gid;                        ///< Specify 'gid' option for mount

    BCToolMountParams()
        : mountReadOnly(false), useDefaultMountPoint(true), uid(0), gid(0) {}
};


//! Params needed for creation a container or hidden part
struct BCToolCreateParams
{
    std::string fsType;
    std::string volumeLabel;
    uint64_t  hiddenPartSize;
    bool confirm_before_format;
    bool mount_after_creation;

    BCToolCreateParams()
     : hiddenPartSize(0), confirm_before_format(false), mount_after_creation(false) {}
};


//! Params for header backup/restore
struct BCToolHeaderParams
{
    std::string headerFilepath;
    std::string containerFilename;
};


/// @brief Container file information.
/// This information is available for containers with unencrypted header only.
typedef struct
{
    bc_alg_id   algId;
    bc_mode_id  modeId;
    bc_u64      size;
    char        containerPath[BC_MAX_PATH];
    char        description[BC_DESCRIPTION_SIZE];
    char        algName[BC_ALG_NAME_SIZE];
    char        modeName[BC_ALG_MODE_NAME_SIZE];
} bc_container_info;

/// @brief Benchmark information item.
typedef struct
{
    char            algName[BC_ALG_NAME_SIZE];          ///< Name of the algorithm being benchmarked.
    char            modeName[BC_ALG_MODE_NAME_SIZE];    ///< Name of the algorithm mode used.
    unsigned long   averageTime;                        ///< Average time in nanoseconds to encrypt one megabyte of garbage.
} bc_bench_info;



/*!
 * \brief The OpenedKeyfiles class is a list of opened keyfiles.
 * It opens every 'keyfile' passed to its constructor
 * and maintains array of pointers to opened iChannel objects.
 */
class OpenedKeyfiles
{
public:
    OpenedKeyfiles(strings_t keyfiles)
    {
        for(unsigned i = 0 ; i < keyfiles.size() ; i++) {
            iChannel *c = new BCFileChannel(keyfiles[i].c_str());
            c->open(kBCFileMode_Read);
            _ptrs.push_back(c);
        }
    }

    ~OpenedKeyfiles()
    {
        for(unsigned i = 0 ; i < _ptrs.size() ; i++)
            delete _ptrs[i];
    }

    void assingKeyfilesTo(cPwdGet *pwObj) const
    {
        pwObj->setKeyFilesTable((iChannel**)_ptrs.data(), _ptrs.size());
    }

private:
    OpenedKeyfiles();
    OpenedKeyfiles(const OpenedKeyfiles &);
    std::vector< iChannel*> _ptrs;
};


class BCToolCreationProgress
{
public:
    BCToolCreationProgress()
    : m_canceled(false) {}

    // should return 0 if 'm_canceled' == true, 1 if we continue
    virtual int func(bc_u64 nbytes, bc_u64 total) = 0;

    virtual void cancel() { m_canceled = true; }

protected:
    bool m_canceled;
};


class BCTool
{
public:

    class SignalHandlerSentry;

    bctool_msg_level minimumMsgLevel;

    static const uint64_t minimalContainerSize = 10 * _1MB;

    BCTool();
    virtual ~BCTool();

    bool CheckKernelInitialization();

    bool prepareForOperation(BCHelperContext &mainContext, cPwdGet *additionalPwdGet = 0);

    virtual bool generateRandom(bool strong, bc_u8 *out, size_t size) = 0;

    /**
      @brief Tries to find a BestCrypt device that uses container with specified path.
            If yes, displays a message.
      @return True, if a device was found, otherwise false.
    */
    bool CheckContainerInUse(const char *containerPath);

    bool Mount(BCHelperContext &mainContext, BCToolMountParams *mountParams);

    bool Umount(const char *mountPoint, bool force);

    bool UmountAll(bool force, int *out_mountedCount, int *out_unmountedCount);

    bool Link(BCHelperContext &helperContext, BCToolMountParams *mountParams);

    bool RawLink(const char *file, const char *algName, const char *modeName, bool readOnly);

    bool Unlink(std::string deviceName, bool force);

    bool Create(BCHelperContext *mainContext, BCToolCreateParams *createParams, BCToolMountParams *mountParams, BCToolCreationProgress *progress = 0);

    SpaceMap * GetFreeSpaceFromMainPart(BCHelperContext &helperContext, BCToolMountParams *mountParams);

    bool ProtectExistingHiddenPart(BCHelperContext &helperContext, cPwdGet *hiddenPwdGet, SpaceMap *freespace);

    bool CreateHiddenPart(BCHelperContext &helperContext, cPwdGet *hiddenPwdGet, SpaceMap *spaceMap, BCToolCreateParams *createParams, BCToolMountParams *mountParams);

    bool Format(const char *file, const char *fsType, cPwdGet *pwdGetObj, bool showWarning);

    bool CheckFS(BCHelperContext &hctx);

    bool AddPassword(BCHelperContext &mainContext, cPwdGet *addPwdGet, cPwdGet *newPwdGet);

    bool ChangePassword(BCHelperContext &mainContext, cPwdGet *addPwdGet, cPwdGet *newPwdGet);

    bool RemovePassword(BCHelperContext &mainContext, cPwdGet *addPwdGet);

    bool Reencrypt(const char *file, const char *algName, const char *modeName);

    bool WipeHeader(BCHelperContext &mainContext);

    bool BackupHeader(BCHelperContext &mainContext, BCToolHeaderParams *headerParams);

    bool RestoreHeader(BCHelperContext &mainContext, BCToolHeaderParams *headerParams);

    bool ChangeDescription(BCHelperContext &mainContext, std::string new_description);

    bool EncryptHeader(BCHelperContext &mainContext);

    bool DecryptHeader(BCHelperContext &mainContext);

    bool DeleteContainer(const char *container);

    static bool GetMountInfo(std::vector<bc_mount_entry> *out_entriesPtr);

    bool GetContainerInfo(const char *path, bc_container_info *out_info);

    bool Benchmark(const char *algName, const char *modeName, std::vector<bc_bench_info> *out_result);

    bool isOKForContainerFile(BCHelperContext &helperContext);

    static bool PathExists(const char *path);
    static bool CanRead(const char *path);
    static bool CanWrite(const char* path);
    static bool IsFile(const char* path);
    static bool IsDirectory(const char* path);
    static bool IsBlockSpecial(const char* path);
    static bool IsOwner(const char* path);
    static bc_u64 GetBlockDeviceSize(const char* bdev);
    static int GetUidByName(const char* name);
    static int GetGidByName(const char* name);
    static bool IsBCMount(const char* mountPoint);

    // TODO: candidates for moving to multi-lib
    static bool isDeprecatedEncryptionAlgorithm(const char *alg_name);
    static const char * ContainerResultString(BCContainerResult rc);

protected:
    virtual bool AskUser(const char *question, bool defaultAnswer) = 0;
    virtual void messageCallback_(bctool_msg_level level, const char *msg) = 0;

    //! called if BCPrepareContainerForOperation succed
    virtual void volumeInfoUpdatedHook() {}

private:
    bool DoLink(BCHelperContext &hctx, BCToolMountParams *mountParams);
    bool callMountCmd(const char *partitionDeviceName, BCToolMountParams *mountParams);
    int RunProcessL(const char *name, ...);
    int RunProcessV(const char *name, const char **args);
    int FormatVolume(std::string deviceName, const char *fsType, const char *fsLabel);
    int GetFirstDevicePartition(const char *mainDeviceName, char *out_partitionDeviceName, int maxNameSize);
    SpaceMap *GetFreeSpaceFromMainPart(std::string deviceName);
    bool DoMount(BCToolMountParams *mountParams);
    int DoUmount(const char* mountPoint, const char* deviceName, bool force);

#if 0
    /**
      @brief Cuts end of the map, if map's size is bigger than specified in 'size' parameter.
    */
    SpaceMap *TransformToSubmapOfSize(SpaceMap *map, bc_u64 size);
#endif


public:
    void ShowError(const char *msg, ...) PRINTF_LIKE;
    void ShowWarning(const char *msg, ...) PRINTF_LIKE;
    void ShowInfo(const char *msg, ...) PRINTF_LIKE;
    void ShowDebug(const char *msg, ...) PRINTF_LIKE;
    void ShowMessage(bctool_msg_level level, const char* msg, va_list args);

    bool WantChange(const char *file);
    bool WantRead(const char *file);
};

#endif // BCTOOL_H
