#include <termios.h>
#include <errno.h>

#include "cli_user_interaction.h"

#define BC_INVALID_ICONV_RESULT ((iconv_t)-1)

extern bool gInterrupted;  // from main.cc


static void* SetEchoOff()
{
    struct termios *term_flags = (struct termios*)malloc(sizeof(struct termios));
    if (tcgetattr(fileno(stdin), term_flags)) {
        fprintf(stderr, "tcgetattr failed: %s\n", strerror(errno));
        free(term_flags);
        return NULL;
    }

    tcflag_t old_flags = term_flags->c_lflag;
    term_flags->c_lflag &= ~ECHO;

    if (tcsetattr(fileno(stdin), TCSAFLUSH, term_flags)) {
        fprintf(stderr, "tcsetattr failed: %s\n", strerror(errno));
        free(term_flags);
        return NULL;
    }

    term_flags->c_lflag = old_flags;

    return term_flags;
}

static void SetEchoOn(void *state)
{
    struct termios *term_flags = (struct termios*)state;
    if (tcsetattr(fileno(stdin), TCSAFLUSH, term_flags)) {
        fprintf(stderr, "tcsetattr failed: %s\n", strerror(errno));
    }

    free(term_flags);
}

int getPasswordString(char *buffer, size_t bufferSize)
{
    void *state = SetEchoOff();

    memset(buffer, '\0', bufferSize);
    // Get the password.
    unsigned i = 0;
    char c = '\0';

    for (; (i < bufferSize-1) && (c != '\n') && (c != EOF); i++) {
        c = fgetc(stdin);
        if ((c != '\n') && (c != EOF)) {
            buffer[i] = c;
        }
    }

    printf("\n");

    // Restore an old state.
    if (state)
        SetEchoOn(state);

    return 0;
}

// Will get password converted to UTF8 on success
static bool askPassword(const char* query, BCPassword* out_pwd)
{
    out_pwd->clear();

    printf("Empty password will cancel the operation.\n");
    printf("%s", query);
    char* pwd = (char*)malloc(kBCPasswordMax);

    if (!pwd) {
        return false;
    }

    if (getPasswordString(pwd, kBCPasswordMax)) {
        free(pwd);
        return false;
    }
    
    strncpy(out_pwd->buf, pwd, strlen(pwd)+1);
    out_pwd->length = strlen(pwd);

    // on Linux input is always UTF8
    // out_pwd->length = convertToUTF8(pwd, strlen(pwd), out_pwd->buf, sizeof(out_pwd->buf));
    memset(pwd, 0, strlen(pwd));
    free(pwd);
    return (out_pwd->length != 0);
}

// Ask a yes/no question.
static bool ask_user(const char* message, bool default_answer) {
	char c;
	fprintf(stderr, "%s", message);
	c = tolower(fgetc(stdin));

	if ('\n' == c) return default_answer;

	while ('\n' != fgetc(stdin));

	return ('y' == c)?  true : false;
}

static void show_message(bctool_msg_level level, const char *msg) {
    switch (level) {
        case kBCTool_Debug:
            printf("Debug: ");
            break;

        case kBCTool_Info:
            //printf("Info: ");
            break;

        case kBCTool_Warning:
            fprintf(stderr, "Warning: ");
            break;

        case kBCTool_Error:
            fprintf(stderr, "Error: ");
            break;

        default:
            assert(0);
    }

    printf("%s\n", msg);
}


// catch termination signals when writing to container file to delete it.
static void sighandler(int sig) {
    if (ask_user("Do you really want to terminate process? (y/[n]):", false)) {
        gInterrupted = true;
    }
}

static int progress_report(const char *message, bc_u64 nbytes, bc_u64 total, void* arg) {
    if(gInterrupted) {
        return false;
    }

    printf("%s: %llu\\%llu bytes (%u%%)  \r", message, (long long unsigned)(nbytes),
           (long long unsigned)(total), (unsigned)(((nbytes + 1) * 100) / total));

    if(nbytes == total) {
        printf("\n");
    }

    return true;
}

static int volume_progress_report(bc_u64 nbytes, bc_u64 total, void* arg) {
    return progress_report("Processing volume", nbytes, total, arg);
}



void PwdGet_Common_CLI::setKeyFiles(strings_t keyfiles)
{
    if (opened_key_files)
        delete opened_key_files;

    opened_key_files = new OpenedKeyfiles(keyfiles);
}



BCContainerResult PwdGet_Common_CLI::getPassword ( bool prevFailed )
{
    static bool env_pass_has_tried = false;
    
    const char *pp = getenv("BESTCRYPT_PASSWORD");
    if (!env_pass_has_tried && pp) {
        env_pass_has_tried = true;
        if (strlen(pp) > kBCPasswordMax-1) {
            return kBCContainer_InvalidPassword; // password is too long
        }
        m_password.set(pp);
        return kBCContainer_Success;
    }

    if (! m_query || strlen(m_query) < 1)
        m_query = BC_DEFAULT_PASSWORD_QUERY;
    printf("%s ", m_query);

    char passbuf[4096];
    ::getPasswordString(passbuf, sizeof(passbuf));

    if (strlen(passbuf) > kBCPasswordMax-1) {
        return kBCContainer_InvalidPassword; // password is too long
    }

    if (strlen(passbuf) > 0) {
        m_password.set(passbuf);
        memset(passbuf, 0, sizeof(passbuf));

        opened_key_files->assingKeyfilesTo(this);

        return kBCContainer_Success;
    }

    return kBCContainer_Cancelled;
}


bool CLI_BCTool::generateRandom(bool strong, bc_u8 *out_buf, size_t total)
{
#ifdef BC_DEBUG
    if (strong) {
        printf("\n\nWARNING: Debug mode: fast pseudo-random will be used instead strong random numbers\n");
        printf("WARNING: Created container in this way is not safe!\n\n");
        strong = false;
    }
#endif

    const char *filename = (strong)? "/dev/random" : "/dev/urandom";

    FILE *fp = fopen(filename, "rb");
    if(! fp) {
        fprintf(stderr, "Error: open %s failed: %s\n", filename, strerror(errno));
        return false;
    }

    if (strong) {
        printf("Randomizing: Getting random numbers from the system...\n");
        printf("Please press random keys and move your mouse (if any) to speed up the process...\n");
    }

    void *state = SetEchoOff();

    unsigned nread = 0;
    do {
        out_buf[nread++] = fgetc(fp);
    } while(nread != total);

    if (state)
        SetEchoOn(state);

    if (strong) {
        printf("Randomizing: Done\n");
    }

    fclose(fp);
    return true;
}


bool CLI_BCTool::AskUser(const char *msg, bool defaultAnswer)
{
    return ask_user(msg, defaultAnswer);
}

void CLI_BCTool::messageCallback_(bctool_msg_level level, const char *msg)
{
    show_message(level, msg);
}



CLI_CertlibProgressActor::CLI_CertlibProgressActor() :
      m_mainBlockLimit(0)
    , m_extBlockLimit(0)
    , m_progress(0)
{
}

void CLI_CertlibProgressActor::init(int value0, int value1)
{
    // qDebug() << "PROGRESS: Entry of " << Q_FUNC_INFO;

    m_mainBlockLimit = value0;
    m_extBlockLimit = value1;
    // emit num_of_steps(m_mainBlockLimit);
}

int CLI_CertlibProgressActor::func(const char * _message, int value0, int value1)
{
    if (m_progress == 0) {
        // emit started();
    }

    if (value0 == 1) {
        // emit stopped();
        m_progress = 0;
        // qDebug() << "PROGRESS: func(): return 0 (stopped)";
        return 0;
    }

    ++m_progress;

    if (m_progress == m_mainBlockLimit) {
        if (! ask_user("Do you want to try beyond main keyblock slots?", false)) {
            return 1; // it is enought
        }
        // continue beyond main slots
        // emit trying_extra_steps();
        // emit num_of_steps(m_extBlockLimit);
    }

    if (m_extBlockLimit > m_mainBlockLimit && m_progress > m_extBlockLimit) {
        // qDebug() << "PROGRESS: func(): return 1";
        return 1;
    }

    int max = (m_progress > m_mainBlockLimit)? m_extBlockLimit : m_mainBlockLimit;

    if (_message)
        printf("%s          (%d/%d)    \n",_message, m_progress, max);

    int ret = (gInterrupted)? 1 : 0;
    return ret;
}

