/*
 * Copyright 2012-2016 Jetico Inc. Oy   All rights reserved.
 * BestCrypt command line interface implementation and main.
 */

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include <assert.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <mntent.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/vfs.h>

#include <iostream>
#include <vector>
#include <iomanip>

#include "kernel/BCKernel.h"
#include "container/container.h"
#include "container/user.h"
#include "container/BCFileChannel.h"
#include "disk/disk.h"
#include "bctool.h"
#include "cli_user_interaction.h"
#include "certificate/CertificateLib/PGPPkts/pktdefs.h"
#include "version.h"
#include "StringUtils.h"

const char *PROG_NAME = "bestcrypt";

#define DIE(message, args...) { fprintf(stderr, message, ## args); exit(EXIT_FAILURE); }
#define ROUND_UP(value, blocksize) ( ((value) + (blocksize) - 1) & ~((blocksize) - 1) )

// Options
static char*            o_description = 0;      /* -d description of container */
static const char*      o_fstype = 0;           /* -t type of used file system */
static const char*      o_alg = 0;              /* -a algorithm name */
static const char*      o_algmode = 0;          /* -M algorithm mode */
static char*            o_options = 0;          /* -o additional mount(2) options */
static uint64_t         o_cont_size = 0;        /* -s container size in bytes */
static int              o_verbose = 0;          /* -v verbose */
static int              o_ro = 0;               /* -r read only */
static int              o_mode = 0;             /* -m mountpoint mode */
static int              o_uid = 0;              /* -u user id for mount point (root only)*/
static int              o_gid = 0;              /* -g group id for mount point (root only)*/
static int              o_force = 0;            /* -f force operation (like umount) */
//static int              o_all = 0;              /* -A umnount all containers for umount command*/
static strings_t        o_keyfiles;             /* keyfiles for open/create a container */

bool gInterrupted = false;

// multi-lib/container/linux/user.cc
extern int getPasswordString(char *buffer, size_t bufferSize);




static int get_uid_by_name(const char* name)
{
	struct passwd *pass;
   	if ( NULL == name ) return -1;

	uid_t uid = atol(name);

    if( (pass = getpwnam(name)) || (pass = getpwuid(uid)) )
	return pass->pw_uid;

	return -1;
}


static int get_gid_by_name(const char* name)
{
	struct group *grp;
   	if ( NULL == name ) return -1;

	gid_t gid = atol(name);

   	if( (grp = getgrnam(name)) || (grp = getgrgid(gid)) )
	return grp->gr_gid;

	return -1;
}

/******************************************************************************************
 * bctool info file
 *
 * Display information about container file.
 * Should never ask any passwords and thus ignore encrypted header possibility
 ******************************************************************************************/

static int bc_info(int argc, char** argv) {
    const char* path = argv[1];

	if(!path) {
        fprintf(stderr, "Missing container path.\n\nPlease specify container path to see information about container,"
               " or use 'bctool benchmark' to see all available algorithms and encryption modes"
               " and their speed.\n");
		return EXIT_FAILURE;
	}

    bc_container_info info;

    int result = EXIT_FAILURE;

    CLI_BCTool bctool;

    if (bctool.GetContainerInfo(path, &info)) {
        printf("Container information: \n");
        printf("  file:         %s\n", info.containerPath);
        printf("  size:         %lu bytes\n", (unsigned long)info.size);
        printf("  algorithm:    %s-%s\n", info.algName, info.modeName);
        printf("  description:  %s\n", info.description);

        result = EXIT_SUCCESS;
    }
    else {
        fprintf(stderr, "Unable to get information about %s: not a container file or has encrypted header.\n", path);
        result = EXIT_FAILURE;
    }
    return result;
}



/******************************************************************************************
 * bctool list
 *
 * Lists all currently configured devices and their mount points if mounted.
 ******************************************************************************************/

static int bc_list(int argc, char** argv) {
    BCTool *ui = new CLI_BCTool;
    std::vector<bc_mount_entry> entries;
    int result = EXIT_FAILURE;
    if (ui->GetMountInfo(&entries)) {
        std::vector<bc_mount_entry>::iterator i = entries.begin();
        std::vector<bc_mount_entry>::iterator end = entries.end();

        while (i != end) {
            printf("%s\t\t", i->deviceName);
            if (strlen(i->mountPoint)) {
                printf("%s", i->mountPoint);
            }
            else {
                printf("Not mounted.");
            }
            printf("\t\t%s\n", i->containerPath);
            ++i;
        }

        result = EXIT_SUCCESS;
    }
    delete ui;

    return result;
}


/******************************************************************************************
 * bctool new -a algid -M modeid -s size [-d description] [-v] file
 *
 * Display information about container file.
 * Should never ask any passwords and thus ignore encrypted header possibility
 ******************************************************************************************/

static int bc_new(int argc, char** argv, const char* algname, const char* modename, uint64_t size, const char* description, const char* fstype, strings_t keyfiles)
{
    if (argc < 2) {
        DIE("Please provide container file name.\n");
    } else if (argc > 2) {
        DIE("Too many arguments.\n");
    }

    const char* fileName = argv[1];

    BCTool *bctool = new CLI_BCTool;

    // Get specified algorithm and mode.
    if (! algname)
        algname = "rijndael";
    BCAlgInfo alg;
    if (kBCCipher_Success != BCKernelLookupAlgByName(algname, &alg)) {
        bctool->ShowError("Unknown algorithm: %s.\n", algname);
        return EXIT_FAILURE;
    }

    if (BCTool::isDeprecatedEncryptionAlgorithm(algname)) {
        bctool->ShowError("Deprecated algorithm: %s.\n", algname);
        return EXIT_FAILURE;
    }

    // Use CBC mode by default.
    if (! modename)
        modename = "CBC";

    bc_mode_id modeId = BCKernelModeId(modename);

    if (! BCKernelIsModeSupported(&alg, modeId)) {
        bctool->ShowError("Algorithm %s does not support mode %s.\n", algname, modename);
        return EXIT_FAILURE;
    }

    PwdGet_Common_CLI pwdGetObj(BC_DEFAULT_NEW_PASSWORD_QUERY);

    BCToolCreateParams createParams;

    if (fstype)
        createParams.fsType = fstype;

    strncpy(pwdGetObj.getContainerContext().path, fileName, BC_MAX_PATH);
    pwdGetObj.getContainerContext().size = size;

    pwdGetObj.getContainerContext().algID = alg.id;
    pwdGetObj.getContainerContext().modeID = modeId;
    pwdGetObj.getContainerContext().hashAlgID = pgphaSHA256;

    // TODO: check it. Was it like that in BestCrypt for Linux 2.0 ?
    pwdGetObj.getContainerContext().iterations = 256;

    const size_t salt_size = sizeof(pwdGetObj.getContainerContext().salt);


    // random numbers for salt may be pseudo-random
    bctool->generateRandom(false, pwdGetObj.getContainerContext().salt, salt_size);
    pwdGetObj.getContainerContext().saltSize = salt_size;
    pwdGetObj.setKeyFiles(keyfiles);

    if (description) {
        strncpy(pwdGetObj.getContainerContext().description,
            description,
            sizeof(pwdGetObj.getContainerContext().description));
    }

    BCFileChannel fch(fileName);
    bc_file_mode file_mode = 0;
    BCHelperContext ctx(&fch, file_mode, &pwdGetObj);

    BCToolMountParams mountParams;

    bool succeeded = bctool->Create(&ctx, &createParams, &mountParams);
    delete bctool;

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool link file [-r]
 *
 * Parse container file and create a new crypto loop binded to it
 ******************************************************************************************/

static int bc_link(int argc, char** argv, bool ro, strings_t keyfiles)
{
  // TODO: make use of 'ro' param

    if (argc < 2) {
        DIE("Please specify container name");
    } else if (argc > 3) {
        DIE("Too many arguments");
    }

    char device_name[BC_MAX_PATH];

    const char* filename = argv[1];
    bc_file_mode file_mode = kBCFileMode_Exclude;
    file_mode |= (ro)? kBCFileMode_Read : kBCFileMode_ReadWrite;

    BCTool *ui = new CLI_BCTool;

    PwdGet_Common_CLI pwdGetObj;
    pwdGetObj.setKeyFiles(keyfiles);

    BCFileChannel fch(filename);
    BCHelperContext ctx(&fch, file_mode, &pwdGetObj);

    BCToolMountParams mountParams;

    bool succeeded = ui->Link(ctx, &mountParams);

    delete ui;

    if (succeeded)
        printf("File is successfully linked to %s\n", device_name);

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool link_raw [-r] -a algname -m modename file
 *
 * Generate random key and create a new crypto loop device binded to arbitrary file
 ******************************************************************************************/

static int bc_link_raw(int argc, char** argv, const char* algname, const char* modename, bool ro)
{
	if (argc < 2) {
        DIE("Please specify file name to link.\n");
	} else if (argc > 2) {
        DIE("Too many arguments.\n");
	}

    if (!algname) {
        DIE("Please specify algorithm name.\n");
    }

    const char* fname = argv[1];

    BCTool *ui = new CLI_BCTool;

    bool succeeded = ui->RawLink(fname, algname, modename, ro);
    delete ui;
    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool unlink device
 *
 * Unlink crypto loop device previously linked manually.
 ******************************************************************************************/

static int bc_unlink(int argc, char** argv)
{
	if( argc < 2 ) {
        DIE("Please specify device name\n");
	} else if (argc > 2) {
        DIE("Too many arguments\n");
	}

    const char* devname = argv[1];

    BCTool *ui = new CLI_BCTool;
    bool succeeded = ui->Unlink(devname, (bool)o_force);
    delete ui;
    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/********************************************************************************************************
 * bctool mount [-r] [-t fstype] [-u owner] [-g group] [-m mode] [-o options] file mount_point
 *
 * Mount container at the specified mount point.
 *
 * -u, -g and -m options allow the user to set mount point owner, group and mode respectively.
 * This is to support the case when root is mounting the container for a different user
 *
 * -o option passes its value directly to mount(1) -o
 *
 * -t option allows user to specify file system type so we dont have to guess
 *
 * -r is for reading only
 ********************************************************************************************************/

static int bc_mount(int argc, char** argv, const char* fstype, int uid, int gid, int mode, bool ro, const char* opts, strings_t keyfiles)
{
	if (argc < 3) {
        DIE("Please specify container file and mount point.\n");
	} else if (argc > 3) {
        DIE("Too many arguments.\n");
	}

    const char* filename = argv[1];

    BCTool *ui = new CLI_BCTool;

    bc_file_mode file_mode = (ro)? kBCFileMode_Read : kBCFileMode_ReadWrite;
    file_mode |= kBCFileMode_Exclude;

    BCToolMountParams mountParams;
    mountParams.filename = filename;
    mountParams.mountReadOnly = ro;
    mountParams.mountPoint = argv[2];
    mountParams.useDefaultMountPoint = false;
    mountParams.uid = uid;
    mountParams.gid = gid;
    if (fstype)
        mountParams.fsType = fstype;
    if (opts)
        mountParams.mountOptions = opts;

    PwdGet_Common_CLI pwdGetObj;
    pwdGetObj.setKeyFiles(keyfiles);

    BCFileChannel fch(filename);
    BCHelperContext ctx(&fch, mode, &pwdGetObj);
    CLI_CertlibProgressActor progress;
    ctx.progressActor = &progress;

    bool succeeded = ui->Mount(ctx, &mountParams);
    delete ui;

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}


/********************************************************************************************************
 * bctool umount [-f] mount_point
 *
 * Unmount container at specified mount point
 * -f flag forces unmount even if container is busy.
 ********************************************************************************************************/

bool bc_umount(int argc, char** argv, bool force)
{
	if (argc < 2) {
        DIE("Please specify mount point.\n");
	} else if (argc > 2) {
        DIE("Too many arguments.\n");
	}

    const char* mpoint = argv[1];

    BCTool *ui = new CLI_BCTool;
    bool succeeded = ui->Umount(mpoint, force);
    delete ui;
    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/********************************************************************************************************
 * bctool umountall [-f]
 *
 * Unmount all mounted containers
 * -f flag forces unmount even if container is busy.
 ********************************************************************************************************/

bool bc_umount_all(int argc, char** argv, bool force)
{
	if (argc > 1) {
        DIE("Too many arguments\n");
	}
    int total_mounted, umounted;
    BCTool *ui = new CLI_BCTool;
    bool succeeded = ui->UmountAll(force, &total_mounted, &umounted);
    delete ui;
    if (total_mounted)
    {
        printf("%d of %d containers successfully unmounted.\n", umounted, total_mounted);
        if ((umounted != total_mounted) && force)
        {
            printf("Warning! Some containers aren't unmounted!\nIf you really need to unmount them right now, please reboot your computer.");
        }
    }
    else
    {
        printf("There are no containers to unmount.\n");
    }

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool format -t fstype file
 *
 * Link and create a file system on container.
 ******************************************************************************************/

static int bc_format(int argc, char** argv, const char* fstype)
{
	if (argc < 2) {
        DIE("Please specify container name\n");
	} else if (argc > 2) {
        DIE("Too many arguments\n");
	} else if (!fstype) {
        DIE("Please specify file system\n");
    }

	const char* file = argv[1];
    BCTool *ui = new CLI_BCTool;
    PwdGet_Common_CLI pwdGetObj;
    bool succeeded = ui->Format(file, fstype, &pwdGetObj, false);
    delete ui;
    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}


/******************************************************************************************
 * bctool fsck [-t fstype] file
 *
 * Check formatted container file system
 ******************************************************************************************/

static int bc_fsck(int argc, char** argv, const char* fstype)
{
	if (argc < 2) {
        DIE("Please specify container name\n");
	} else if (argc > 2) {
        DIE("Too many arguments\n");
    }

    const char* file = argv[1];
    BCTool *ui = new CLI_BCTool;

    BCFileChannel fch(file);
    bc_file_mode mode = kBCFileMode_ReadWrite | kBCFileMode_Exclude;
    PwdGet_Common_CLI pwdGetObj;
    BCHelperContext hctx(&fch, mode, &pwdGetObj);

    bool succeeded = ui->CheckFS(hctx);
    delete ui;

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool passwd|add_passwd|del_passwd file
 *
 * Change, remove or add container password
 ******************************************************************************************/

static int bc_passwd(int argc, char** argv, BCKeyOp op)
{
    // check container file write access
    if (argc < 2) {
        DIE("Please specify container file\n");
    } else if (argc > 2) {
        DIE("Too many arguments\n");
    }

    BCFileChannel fch(argv[1]);
    bc_file_mode mode = kBCFileMode_ReadWrite | kBCFileMode_Exclude;
    PwdGet_Common_CLI pwdGetObj;
    BCHelperContext mainContext(&fch, mode, &pwdGetObj);

    PwdGet_Common_CLI addPwdGet;
    PwdGet_Common_CLI newPwdGet;

    CLI_BCTool bctool;

    bool success = false;
    switch(op) {
    case kBCKey_Recode:
        success = bctool.ChangePassword(mainContext, &addPwdGet, &newPwdGet);
        break;
    case kBCKey_Duplicate:
        success = bctool.AddPassword(mainContext, &addPwdGet, &newPwdGet);
        break;
    case kBCKey_Delete:
        success = bctool.RemovePassword(mainContext, &addPwdGet);
        break;
    default:
        assert(0);
    }

    return success? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool encrypt_header|decrypt_header file
 *
 * Encrypt or decrypt container header.
 ******************************************************************************************/

static int bc_header(int argc, char** argv, bool encrypt)
{
    if (argc < 2) {
        DIE("Please specify container file\n");
    } else if (argc > 2) {
        DIE("Too many arguments\n");
    }

    BCFileChannel fch(argv[1]);
    bc_file_mode mode = kBCFileMode_ReadWrite | kBCFileMode_Exclude;
    PwdGet_Common_CLI pwdGetObj;
    BCHelperContext mainContext(&fch, mode, &pwdGetObj);

    CLI_BCTool bctool;

    bool success = false;

    if (encrypt) {
        success = bctool.EncryptHeader(mainContext);
    }
    else {
        success = bctool.DecryptHeader(mainContext);
    }

    return success? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool backup container file
 *
 * Backup container header into backup file.
 ******************************************************************************************/

static int bc_backup(int argc, char** argv)
{
    if (argc < 3) {
        DIE("Please specify container and backup files\n");
    } else if (argc > 3) {
        DIE("Too many arguments\n");
    }

    BCToolHeaderParams headerParams;
    headerParams.containerFilename = argv[1];
    headerParams.headerFilepath = argv[2];

    BCFileChannel fch(argv[1]);
    bc_file_mode mode = kBCFileMode_ReadWrite | kBCFileMode_Exclude;
    PwdGet_Common_CLI pwdGetObj;
    BCHelperContext mainContext(&fch, mode, &pwdGetObj);

    CLI_BCTool bctool;

    bool result = bctool.BackupHeader(mainContext, &headerParams);

    return result? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool restore container file
 *
 * Restore container header from backup file.
 ******************************************************************************************/

static int bc_restore(int argc, char** argv)
{
    if (argc < 3) {
        DIE("Please specify container and backup files\n");
    } else if (argc > 3) {
        DIE("Too many arguments\n");
    }


    BCToolHeaderParams headerParams;
    headerParams.containerFilename = argv[1];
    headerParams.headerFilepath = argv[2];

    BCFileChannel fch(argv[1]);
    bc_file_mode mode = kBCFileMode_ReadWrite | kBCFileMode_Exclude;
    PwdGet_Common_CLI pwdGetObj;
    BCHelperContext mainContext(&fch, mode, &pwdGetObj);

    CLI_BCTool bctool;

    bool result = bctool.RestoreHeader(mainContext, &headerParams);

    return result? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool make_hidden file -s size -t fstype
 *
 * Create container hidden part of given size.
 ******************************************************************************************/

static int bc_hidden(int argc, char** argv, unsigned long long size, const char* fstype, strings_t keyfiles)
{
    if (argc < 1) {
        DIE("Please specify the hidden part size.");
    }
    if (argc > 2) {
        DIE("Too many arguments.");
    }

    const char* filename = argv[1];
    BCFileChannel fch(filename);

    bc_file_mode file_mode = kBCFileMode_ReadWrite | kBCFileMode_Exclude;

    CLI_BCTool bctool;

    PwdGet_Common_CLI pwdGetObj("Enter password for main part");
    BCHelperContext mainContext(&fch, file_mode, &pwdGetObj);



    BCToolMountParams mountParams;
    mountParams.filename = filename;



    // currently it does Link, get free space map, Unlink
    SpaceMap *hiddenExtents = bctool.GetFreeSpaceFromMainPart(mainContext, &mountParams);

    if (!hiddenExtents) {
        return EXIT_FAILURE;
    }


    bctool.ShowInfo("Protection of existing hidden part is not implemented yet in command line interface");

    // TODO: protect existing hidden parts




    PwdGet_Common_CLI hiddenPwdGet("Set password for hidden part");
    // pwdGetObj2.setKeyFiles(keyfiles);

    BCToolCreateParams createParams;
    createParams.fsType = fstype;
    createParams.hiddenPartSize = size;
    createParams.volumeLabel = "HIDDEN";

    // TODO: it should be another helperContext but with the same 'header' pointer

    // mount after creation is not supported in CLI
    bool result = bctool.CreateHiddenPart(mainContext, &hiddenPwdGet, hiddenExtents, &createParams, &mountParams);

    return result? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool reencrypt file -a algorithm -M mode
 *
 * Reencrypt container with new algorithm, mode and key.
 ******************************************************************************************/

static int bc_reencrypt(int argc, char** argv, const char* algname, const char* modename)
{
    if (argc < 2) {
        DIE("Please specify container file\n");
    } else if (argc > 2) {
        DIE("Too many arguments\n");
    }

    const char* file = argv[1];

    BCTool *ui = new CLI_BCTool;
    bool result = ui->Reencrypt(file, algname, modename);
    delete ui;

    return (result)? EXIT_SUCCESS : EXIT_FAILURE;
}


static int bc_benchmark(int argc, char **argv, const char *algname, const char *modename)
{
    std::vector<bc_bench_info> bench_result;

    printf("Benchmarking...\n");
    BCTool *ui = new CLI_BCTool;
    bool succeeded = ui->Benchmark(algname, modename, &bench_result);
    delete ui;

    if (!succeeded)
        return EXIT_FAILURE;

    printf("Benchmark results: \n");
    std::vector<bc_bench_info>::iterator i = bench_result.begin();
    std::cout.precision(2);

    for (; i != bench_result.end(); i++)
    {
        bc_bench_info *info = &(*i);
        std::cout << std::setw(15) << info->algName << std::setw(5) << info->modeName << std::setw(10) << std::fixed << ((double)1e9 / (double)info->averageTime) << " MB/s" << std::endl;
    }

    return EXIT_SUCCESS;
}

static int bc_is_kernel_initialized()
{
    BCTool *ui = new CLI_BCTool;
    bool result = ui->CheckKernelInitialization();
    delete ui;

    if (!result)
        return 0;
    return 1;
}


static void usage(void) {
    const char *usage_string ="Usage: bctool command [options] arguments\n\
\n\
Commands available:\n\
 * addpasswd file_name - add additional password for container\n\
 * backup container_file backup_file - backup container header to file\n\
 * benchmark [-a alg_name] [-M alg_mode] - benchmark algorithms\n\
 * decryptheader file_name - decrypt container header\n\
 * delpasswd file_name - delete additional password(s) from container\n\
 * encryptheader file_name - encrypt container header\n\
 * format file_name -t filesystem_type - format container\n\
 * fsck file_name [-t fs_type] - check container file system\n\
 * info file_name - print information about container\n\
 * link file_name [-r] [-f] - link container to device\n\
 * list - list all linked devices, their container files and mount points\n\
 * makehidden file_name -s size [-t fs_type] [-v] - (temporary disabled) make hidden part in container\n\
 * mount file_name mount_point [-t fstype] [-m mode] [-u user] [-g group] [-r] - mount container\n\
 * new file_name -a alg_name -s size -M alg_mode [-d description] [-t fstype] [-v] - create container\n\
 * passwd file_name - change container's password\n\
 * rawlink file_name device_name -a alg_name [-r] - raw link file to BestCrypt device\n\
 * restore container_file backup_file - restore container header from backup\n\
 * reencrypt file_name [-a alg_name] [-M alg_mode] [-v] - reencrypt container\n\
 * umount mount_point [-f] - unmount container\n\
 * umountall [-v] [-f] - unmount all mounted containers\n\
 * unlink device_name - unlink container from device\n\
\n\
options:\n\
 -t fstype       specify filesystem type (default: msdos)\n\
 -a alg_name     specify algorithm name\n\
 -M alg_mode     specify algorithm mode\n\
                 alg_mode can be CBC, LRW or XTS (default is CBC).\n\
                 Not all modes are supported by all algorithms.\n\
 -s size         specify container size\n\
                 example:\n\
                  -s 32768 - 32768 bytes (32 kilobytes);\n\
                  -s 10K   - 10 Kb (10240 bytes);\n\
                  -s 10M   - 10 Mb (10240 kilobytes);\n\
                  -s 10G   - 10 Gb (10240 megabytes)\n\
                  -s 2T    -  2 Tb (2048  gigabytes)\n\
 --keyfile <path> Specify a keyfile during open or creation a container.\n\
                 This option may be used several times.\n\
 -d text         specify description of container (up to 64 symbols)\n\
 -r              mount read only\n\
 -m mode         specify mountpoint mode in octal (default 0700)\n\
 -u user         specify mountpoint's owner (only root can do that)\n\
 -g group        specify mountpoint's group\n\
 -v,--verbose    print more details during operation\n\
 -V,--version    print version and exit\n\
 -h,--help,--usage   print this help\n\
\n";

    fprintf(stderr, "%s\n", usage_string);
}


static int run_command(const char* cmd, int argc, char** argv)
{
    if (0 == strcmp(cmd, "addpasswd"))      return bc_passwd(argc, argv, kBCKey_Duplicate);
    if (0 == strcmp(cmd, "backup"))         return bc_backup(argc, argv);
    if (0 == strcmp(cmd, "benchmark"))      return bc_benchmark(argc, argv, o_alg, o_algmode);
    if (0 == strcmp(cmd, "decryptheader"))  return bc_header(argc, argv, false);
    if (0 == strcmp(cmd, "delpasswd"))      return bc_passwd(argc, argv, kBCKey_Delete);
    if (0 == strcmp(cmd, "encryptheader"))  return bc_header(argc, argv, true);
    if (0 == strcmp(cmd, "format"))         return bc_format(argc, argv, o_fstype);
    if (0 == strcmp(cmd, "fsck"))           return bc_fsck(argc, argv, o_fstype);
    if (0 == strcmp(cmd, "info"))           return bc_info(argc, argv);
    if (0 == strcmp(cmd, "link"))           return bc_link(argc, argv, o_ro, o_keyfiles);
    if (0 == strcmp(cmd, "list"))           return bc_list(argc, argv);
    if (0 == strcmp(cmd, "makehidden"))     return bc_hidden(argc, argv, o_cont_size, o_fstype, o_keyfiles);
    if (0 == strcmp(cmd, "mount"))          return bc_mount(argc, argv, o_fstype, o_uid, o_gid, o_mode, o_ro, o_options, o_keyfiles);
    if (0 == strcmp(cmd, "new"))            return bc_new(argc, argv, o_alg, o_algmode, o_cont_size, o_description, o_fstype, o_keyfiles);
    if (0 == strcmp(cmd, "passwd"))         return bc_passwd(argc, argv, kBCKey_Recode);
    if (0 == strcmp(cmd, "rawlink"))        return bc_link_raw(argc, argv, o_alg, o_algmode, o_ro);
    if (0 == strcmp(cmd, "restore"))        return bc_restore(argc, argv);
    if (0 == strcmp(cmd, "reencrypt"))      return bc_reencrypt(argc, argv, o_alg, o_algmode);
    if (0 == strcmp(cmd, "umount"))         return bc_umount(argc, argv, o_force);
    if (0 == strcmp(cmd, "umountall"))      return bc_umount_all(argc, argv, o_force);
    if (0 == strcmp(cmd, "unlink"))         return bc_unlink(argc, argv);



    fprintf(stderr, "Unknown command %s\n", cmd);
    usage();

    return EXIT_FAILURE;
}


int main(int argc, char** argv)
{
    setvbuf(stderr, (char*)NULL, _IONBF, 0);
    setvbuf(stdout, (char*)NULL, _IONBF, 0);

    int c;
    while (1)
	{
        static struct option long_options[] =
        {
          {"keyfile", required_argument, 0, 0x81},
          /* These options set a flag. */
          {"verbose", no_argument,       0, 'v'},
          {"version", no_argument,       0, 'V'},
          {"usage", no_argument,         0, 'h'},
          {"help", no_argument,          0, 'h'},
          {0, 0, 0, 0}
        };

        /* getopt_long stores the option index here. */
        int option_index = 0;

        c = getopt_long(argc, argv, "Vvfhprt:a:o:s:d:m:M:u:g:",
                        long_options, &option_index);

        /* Detect the end of the options. */
        if (c == -1)
            break;

        switch (c) {
        case 'V':
            fprintf(stdout, "BestCrypt version %d.%d.%d rev%d\nCopyright Jetico Inc. Oy See LICENSE for details.\n",
                    VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH, VERSION_REVISION);
            exit(EXIT_SUCCESS);

        case 'v':
            o_verbose = 1;
            break;

//        case 'A':
//            o_all = 1;
//            break;

        case 'r':
            o_ro = 1;
	        break;

        case 'f':
            o_force = 1;
	        break;

        case 't':
            o_fstype = optarg;
	        break;

        case 'a':
            o_alg = optarg;
            break;

        case 'M':
            o_algmode = optarg;
            break;

        case 'o':
            o_options = optarg;
            break;

        case 'd':
            o_description = optarg;
 	        break;

        case 's':
            o_cont_size = StringUtils::sizeStringToBytes(optarg);
          	if (o_cont_size == 0) {
                fprintf(stderr, "Invalid container size\n");
                exit(EXIT_FAILURE);
            }
            else {
#ifdef BC_DEBUG
                printf("Entered container size is %" PRId64 " bytes\n", o_cont_size);
#endif
            }
            break;

        case 'm':
            o_mode = strtol(optarg, NULL, 8);
            o_mode = o_mode & 07777;
            if (0 == o_mode) {
                fprintf(stderr, "Invalid mode, using 0700\n");
                o_mode = 0700;
            }
            break;

        case 'u':
          	if (0 != getuid()) {
                fprintf(stderr, "Only root can specify mountpoint's owner\n");
                exit(EXIT_FAILURE);
            }

            o_uid = get_uid_by_name(optarg);

            if (-1 == o_uid) {
                fprintf(stderr, "Invalid user name %s\n", optarg);
                exit(EXIT_FAILURE);
            }

            break;

        case 'g':
          	o_gid = get_gid_by_name(optarg);
          	if (-1 == o_gid) {
                fprintf(stderr, "Invalid group name %s\n", optarg);
                exit(EXIT_FAILURE);
            }

            break;

        case 0x81:
            if (!BCTool::CanRead(optarg)) {
                fprintf(stderr, "Error: specified key file doesn't exists: %s\n", optarg);
                exit(EXIT_FAILURE);
            }
            o_keyfiles.push_back(optarg);
            break;

        case 'h':
           	usage();
          	exit(EXIT_SUCCESS);

        default:
           	usage();
          	exit(EXIT_FAILURE);

        }
    }

    // get rid of used by getopt_long arguments
    argc -= optind;
    argv += optind;

    const char* cmd = argv[0];

    if (NULL == cmd) {
        fprintf(stderr, "\nError: Missing command.\n\n");
        usage();
        exit(EXIT_FAILURE);
    }

    if (!bc_is_kernel_initialized())
        return EXIT_FAILURE;

    // pass the rest of arguments to specific command
	return run_command(cmd, argc, argv);
}

