/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * Keygen module interface.
 */

#ifndef BC_KEYGEN_H
#define BC_KEYGEN_H

#include "intsize.h"
#include "error.h"
#include "password.h"
#include "memory.h"
#include "../spacemap/spacemap.h"
#include "cipherbase.h"
#include "iChannel.h"
#include "../certificate/CertificateLib/progressActor.h"
#include "../certificate/CertificateLib/SMemory/SBlocks.h"

#define BC_DESCRIPTION_SIZE 128

// Encoded volume info
struct BCVolumeInfo {
    bc_alg_id   alg;                                // Container encryption algorithm id.
    bc_mode_id  mode;                               // Container encryption mode id.
    bc_u32      hash;                               // Keygen hash algorithm id.
    bc_u32      hidden;                             // Is this is a hidden volume.
    char        description[BC_DESCRIPTION_SIZE];   // User description utf8 string. Not set for hidden volumes.
    bc_u64      size;                               // Size of the main part of the container.
    bc_u64      offset;                             // Offset of the main part of the container.
    SpaceMap*   map;                                // Extent map, used for hidden keys.
	Extent* 	extent_array;						// Extent array generated from map, used to calculate physical extents by SpaceMapGetPhysicalExtents. 
	bc_u64		extent_count;
	bc_u8		use_new_synchro;					// Container file was created with new synchro based on 64bit value.
	
    BCVolumeInfo()
    {
        this->map = NULL;
		this->extent_array = NULL;
		this->extent_count = 0;
        clear();
    }

    void clear()
    {
        alg = 0;
        mode = kBCMode_UDF;
        hash = 0;
        hidden = 0;
        memset(description, 0, BC_DESCRIPTION_SIZE);
        size = 0;
        offset = 0;
        if (map) SpaceMapFree(map);
        if (extent_array) SpaceMapFreeArray( &extent_array );
        extent_count = 0;
        use_new_synchro = 0;
    }
    
    BCVolumeInfo & operator=(const BCVolumeInfo & _other) {
        alg = _other.alg;
        mode = _other.mode;
        hash = _other.hash;
        hidden = _other.hidden;
        memcpy(description, _other.description, BC_DESCRIPTION_SIZE);
        size = _other.size;
        offset = _other.offset;
        if (_other.map) map = SpaceMapCreateCopy(_other.map);
        else map = NULL;
        
        if (_other.extent_array && _other.extent_count > 0) {
			size_t size = (size_t)(_other.extent_count * sizeof(Extent));
			assert( size == _other.extent_count * sizeof(Extent) );
            extent_array = (Extent*)malloc(size);
            memcpy(extent_array, _other.extent_array, size);
            extent_count = _other.extent_count;
        } else {
            extent_array = NULL;
            extent_count = 0;
        }
        use_new_synchro = _other.use_new_synchro;
        return *this;
    }
    
    ~BCVolumeInfo()
    {
        clear();
    }

	bc_mode_id getMode()
	{
		return ( bc_mode_id )mode;
	}
};


// Key decoding operation type.
typedef enum {
	kBCKey_Recode,			        // Decode key and recode it inplace with new user secret.
	kBCKey_Duplicate,		        // Decode key and encode its copy with new secret.
    kBCKey_Delete			        // Decode key and delete it.
} BCKeyOp;


struct BCKeygen;

/**
 * Container header interface.
 * Header should contain information about main encoded volume and passwords
 * Each keygen specifies its own header format and implements its interface. 
 */
class IBCHeader {
public:
    
    // owner
    const struct BCKeygen* keygen;

    /**
     * Get main volume info.
     * Behaviour is undefined if header is encrypted.
     * @param info          Initialized main volume info.
     */
    virtual void mainVolumeInfo(struct BCVolumeInfo* info) = 0;

    /**
     * Initialize keygen.
     * @param algid         Encryption algorithm id
     * @param modeid        Encryption mode id
     * @param size          Encrypted data size in bytes
     * @param description   Optional container description
     * @param password      Initial container password
     */
    virtual BCContainerResult init(bc_alg_id algid, bc_mode_id modeid, bc_u64 size, const char* description, BCPassword & password) = 0;
	virtual BCContainerResult initEx(
			bc_alg_id algid, bc_mode_id modeid, bc_u64 size, const char* description, BCPassword & password,
			bc_u16 kgID = 0, int hash = 0, bc_u32 iterations = 0, const bc_u8 *salt = NULL, int saltSize = 0 )
	{
		return init( algid, modeid, size, description, password );
	}

    /**
     * Generate and store a new hidden key.
     * @param spaceMap          Hidden part space map.
     * @param mainPassword      Main part password.
     * @param hiddenPassword    New hidden part password.
     */
    //virtual BCContainerResult createHiddenKey(SpaceMap *spaceMap, BCPassword *mainPassword, BCPassword* hiddenPassword) = 0;
    virtual BCContainerResult createHiddenKey(SpaceMap *spaceMap, cPTB *mainKey, BCPassword* hiddenPassword) = 0;

	/**
	 * Ask user secret (password, etc) and decode volume key.
	 *	According to bc drive style pool is saved at key handle, so it is not required to cary it thru sources
	 *	*** _param out_pool	    Decrypted random data pool on success.
     * @param password      UTF8 encoded initial user password.
     * @param out_info      Initialized decoded volume info on success.
     * @param out_key       Decoded key handle on success.
	 * @param cipher_out	Initialised cipher.
	 * @param keyOut		Optional plain text key for future using, f.e. for encrypt header.
	 * @return	            kBCKeygen_Success if user provided correct password, error otherwise.
	 */
    virtual BCContainerResult decodeKey(BCPassword & password, struct BCVolumeInfo* out_info, bc_key_handle* out_key, CipherBase **cipher_out, cPTB *keyOut = NULL ) = 0;

	/**
	 * Derives cipher and key handle from key. Fills volume info from the key.
	 * @param keyIn			Plain text key generated by decodeKey.
	 * @param out_info      Initialized decoded volume info on success.
     * @param out_key       Decoded key handle on success.
	 * @param cipher_out	Initialised cipher.
	 * @return	            kBCKeygen_Success if user provided correct password, error otherwise.
	 */
	virtual BCContainerResult generateCipherInfo( const cPTB &keyIn, struct BCVolumeInfo* out_info, bc_key_handle* out_handle, CipherBase **cipher_out ) = 0;

    /**
     * Verify a single password without decoding anything.
     * @param password      Password to verify.
     * @param out_info      Volume info for correct password.
     */
    virtual BCContainerResult verifyPassword(BCPassword & password, struct BCVolumeInfo* out_info, cProgressActor *progressActor, cPTB *keyOut ) = 0;

    /**
     * Add new container password
     * @param password      Existing password.
     * @param newPassword   New password to add.
     */
    virtual BCContainerResult addPassword(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & newPassword) = 0;

    /**
     * Change existing password
     * @param password      Existing password.
     * @param newPassword   New password.
     */
    virtual BCContainerResult changePassword(BCPassword & mainPassword, BCPassword & volumePassword, BCPassword & newPassword) = 0;

    /**
     * Remove existing password
     * @param password      Existing password to remove
     */
    virtual BCContainerResult removePassword(BCPassword & mainPassword, BCPassword & oldPassword) = 0;

    /**
     * Tell if container header is encrypted.
     */
    virtual bool isEncrypted() = 0;

    /**
     * Tell if container header is valid.
     */
    virtual bool isValid() = 0;
    virtual int setBusiedName( const char *compName, int len ) = 0;
    virtual int getBusiedName( char *compName, int len ) = 0;

    /**
     * Encrypt container header.
     */
    virtual BCContainerResult encrypt(BCPassword & password) = 0;

    /**
     * Decrypt container header.
     */
    virtual BCContainerResult decrypt(BCPassword & password) = 0;

    /**
     * Update container description.
     * Each format has a maximum description length.
     */
    virtual BCContainerResult setDescription(const char* description, const cPTB & mainKey) = 0;

    /**
     * Mark container as undergoing reencryption.
     * @param mark          Set or remove mark.
     */
    virtual void markForReencryption(bool mark) = 0;

    /**
     * Tell if container is marked for reencryption.
     */
    virtual bool isMarkedForReencryption() = 0;
    
    /**
     * Store keygen data into a file.
     * File might be a container file or a blank file in case of a backup.
     * @param file          File name to write to.
     */
    //virtual BCContainerResult store(const char* file) = 0;
    virtual BCContainerResult store( iChannel *channel ) = 0;

	virtual BCContainerResult encrypt( const cPTB &  _mainKey )
	{ return kBCContainer_Unsupported;}

	virtual BCContainerResult decrypt( const cPTB &  _mainKey )
	{ return kBCContainer_Unsupported;}
    
    /**
     * Get name of keygen used
     */
    virtual const char * getKeygenName() { return NULL; }
    
    /**
     * Returns true if header encryption ot decryption is allowed
     */
    virtual bool isHeaderEncryptionAllowed() { return true; }
    /**
     * Returns true if container can be reencrypted
     */
    virtual bool isReencryptionAllowed() { return false; }
    
    /**
     * Return number of key blocks of the main volume
     */
    virtual bc_u16 getMainKeyblocksCount() { return 1; }

    /**
     * Store key block backup.
     * @param channel   Channel object of the new key block file.
     */
    virtual BCContainerResult storeKeyblockBackup(iChannel *channel) { return kBCContainer_Unsupported; }
    
    virtual BCContainerResult wipeDatablock(const cPTB &  _mainKey) { return kBCContainer_Unsupported; }
    
    /**
     * Destroy keygen.
     */
    virtual ~IBCHeader() {}

protected:

    IBCHeader() {}
    IBCHeader(const IBCHeader& /*rhv*/) {}
    IBCHeader& operator = (const IBCHeader& /*rhv*/) { return *this; }

};


// Scoped header pointer for resource management
typedef BCScopedPtr<IBCHeader> BCHeaderHandle;


enum {
    kBCKeygen_EncryptedHeaders = 0x01 // keygen supports encrypted headers
};


/**
 * Known keygen module description.
 */
struct BCKeygen {
	int		    id;             // Unique id
	const char*	name;           // Name string
	const char*	description;    // Detailed description
	int		    version;        // Supported container version
    int         flags;


    /**
     * Create blank keygen.
     */
    IBCHeader*      (*kgcreate) (void);

    
    /**
     * Load keygen from file.
     * @param channel			contaner channel interface.
     * @param keyBlockChannel	channel for key block, key block may be saved separately and in KBB format. 
     * @param keyFilesT			table of channels with key files.
     */
    IBCHeader*      (*kgload) ( iChannel *channel, iChannel *keyBlockChannel );


	/**
	 * Parse main volume info from container file.
	 * @param file	        Container file path
	 * @param info	        Initialized keygen structure on success.
	 */
	//BCContainerResult	(*kgparse) (const char* file, struct BCVolumeInfo* info);
	BCContainerResult	(*kgparse) ( iChannel *channel, struct BCVolumeInfo* info);
    
};


#endif

